/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const HttpMethod_cxx_Version =
    "$Id: HttpMethod.cxx,v 1.7 2002/03/19 20:17:22 jason Exp $";


#include "HttpMethod.hxx"
#include "Verify.hxx"
#include <cassert>


using Vocal::Http::Method;
using Vocal::Http::Type;
using Vocal::Http::Token;
using Vocal::Http::Exception;
using Vocal::Http::EXTENSION;


const Token     Method::TokenOPTIONS("OPTIONS");
const Token     Method::TokenGET("GET");
const Token     Method::TokenHEAD("HEAD");
const Token     Method::TokenPOST("POST");
const Token     Method::TokenPUT("PUT");
const Token     Method::TokenDELETE("DELETE");
const Token     Method::TokenTRACE("TRACE");
const Token     Method::TokenCONNECT("CONNECT");


Method::Table       Method::myTable;
Method::DecodeTable Method::myDecodeTable;


Method::Method(Type pType)
{
    init();
    
    type(pType);
    
    test_covered("0");
}


Method::Method(const Token & pToken)
    :   myType(EXTENSION)
{
    init();
    
    if ( pToken.size() )
    {
        type(pToken);
    }
    
    test_covered("1");
}        


Method::Method(const Method & src)
    :   myType(src.myType),
        myToken(src.myToken)
{
    test_covered("2");
}


Method::~Method()
{
    test_covered("3");
}


const Method &  
Method::operator=(const Method & src)
{
    if ( this != &src )
    {
        myType = src.myType;
        myToken = src.myToken;
        
        test_covered("4");
    }

    test_covered("5");
    
    return ( *this );
}


bool            
Method::operator==(const Method & src) const
{
    test_covered("6");
    
    return ( !isExtension() ? myType == src.myType : myToken == src.myToken );
}


bool            
Method::operator<(const Method & src) const
{
    test_covered("7");
    
    if ( isExtension() && src.isExtension() )
    {
        test_covered("8");
        
        return ( myToken < src.myToken );
    }
    
    return ( myType < src.myType );
}


Type
Method::type() const
{
    test_covered("9");

    return ( myType );
}


void
Method::type(Type pType)
{
    myToken.erase();
    
    if ( pType >= END )
    {
        pType = EXTENSION;

        test_covered("10");
    }
    myType = pType;
    
    test_covered("11");
}


void
Method::type(const Token & pToken)
throw ( Vocal::Http::Exception )
{
    myToken.erase();
    
    DecodeTable::const_iterator i = myDecodeTable.find(pToken);
    if ( i != myDecodeTable.end() )
    {
        myType = i->second;

        test_covered("12");

        return;
    }

    // This may be an extension
    //
    if ( pToken.size() )
    {
        myType = EXTENSION;
        myToken = pToken;

        test_covered("13");
        
        return;
    }

    test_covered("14");

    throw ( Vocal::Http::Exception(Exception::BadMethod) );
}
        

bool
Method::isExtension() const
{
    test_covered("15");

    return ( myType == EXTENSION );
}


const Token &  
Method::token() const
{
    if ( isExtension() )
    {
        test_covered("16");
        
        return ( myToken );
    }
    test_covered("17");

    return ( *myTable[myType] );
}


size_t   
Method::totalTypes()
{
    test_covered("18");
    
    return ( END );
}


ostream &       
Method::writeTo(ostream & out) const
{
    test_covered("19");

    return ( out << token() );
}


void    
Method::init()
{
    if ( myTable.size() == 0 )
    {
        initTable();
        initDecodeTable();
        
        test_covered("20");
    }
}

void    
Method::initTable()
{
    myTable[OPTIONS]    = &TokenOPTIONS;
    myTable[GET]        = &TokenGET;
    myTable[HEAD]       = &TokenHEAD;
    myTable[POST]       = &TokenPOST;
    myTable[PUT]        = &TokenPUT;
    myTable[DELETE]     = &TokenDELETE;
    myTable[TRACE]      = &TokenTRACE;
    myTable[CONNECT]    = &TokenCONNECT;

    assert( totalTypes() == myTable.size() );

    test_covered("21");
}
        

void    
Method::initDecodeTable()
{
    myDecodeTable[TokenOPTIONS] = OPTIONS;
    myDecodeTable[TokenGET]     = GET;
    myDecodeTable[TokenHEAD]    = HEAD;
    myDecodeTable[TokenPOST]    = POST;
    myDecodeTable[TokenPUT]     = PUT;
    myDecodeTable[TokenDELETE]  = DELETE;
    myDecodeTable[TokenTRACE]   = TRACE;
    myDecodeTable[TokenCONNECT] = CONNECT;

    assert( totalTypes() == myDecodeTable.size() );

    test_covered("22");
}
        
