
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const KeepAliveAgent_cxx_Version =
    "$Id: KeepAliveAgent.cxx,v 1.8 2001/01/12 01:35:50 wbreu Exp $";
//Standard Include
#include "global.h"
#include <strstream.h>
#include <unistd.h>

#if defined(__linux__)
#include <algo.h>
#else
#include <algorithm>
#endif

//User defined
#include "VNetworkException.hxx"
#include "Connection.hxx"
#include "cpLog.h"
#include "COPSKAMsg.hxx"
#include "KeepAliveAgent.hxx"
#include "COPSTransceiver.hxx"

static const int FIVE_SECS = 5;

KeepAliveAgent* KeepAliveAgent::_instance = 0;
KeepAliveAgent::ConnectionList KeepAliveAgent::_connectionList;

KeepAliveAgent&
KeepAliveAgent::instance(int kaTimer)
{
    if (!_instance)
    {
        _instance = new KeepAliveAgent(kaTimer);
        cpLog(LOG_DEBUG, "Created instance for %s\n", _instance->className().c_str());
    }
    return *_instance;
}

KeepAliveAgent::KeepAliveAgent(int kaTimer)
    : _kaTimer(kaTimer)
{
#if defined(__linux__)
    if (atexit(KeepAliveAgent::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };
#endif
    //create a separate thread for Agent
    _aThread = new VThread;
    _aThread->spawn(KeepAliveAgent::run, this);
}

void*
KeepAliveAgent::run(void* args)
{
    KeepAliveAgent* self = static_cast < KeepAliveAgent* > (args);

    //Loop through all registered connection and send keepAlive
    //Messages.The limitation is that every connection has the same
    //Keep Alive time
    for (; ; )
    {
        for (ConnectionList::iterator itr = self->_connectionList.begin();
                itr != self->_connectionList.end(); itr++)
        {
            if ((*itr).isLive())
                self->sendKeepAlive(*itr);
        }
        sleep(self->_kaTimer);
    }
    return (0);
}


KeepAliveAgent::~KeepAliveAgent()
{}


void
KeepAliveAgent::destroy()
{
    delete KeepAliveAgent::_instance;
    KeepAliveAgent::_instance = 0;
}

void
KeepAliveAgent::sendKeepAlive(Connection& conn)
{
    cpLog(LOG_DEBUG, "Sending keep-alive:%s", conn.getDescription().c_str());
    COPSHeader hdr(COPS_OP_KA);
    COPSKAMsg msg;
    msg << hdr ;
    COPSTransceiver tr;
    try
    {
        tr.sendMsg(msg , conn.getConnId());
        ///KeepAlive message should be echoed back from PDP within 5 seconds
        if(conn.isReadReady(FIVE_SECS))
        {
            Sptr<COPSMsg> kMsg  = tr.receiveMsg(conn.getConnId());
            if((kMsg.getPtr() == 0) ||  !(kMsg->getHeader().isAKeepAlive()))
            {
                //close the connection
                conn.close();
            }
        }
        else
        {
            cpLog(LOG_INFO, "Nothing received");
        }
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_ALERT, "Failed to write to connection:%s", conn.getDescription().c_str());
        conn.close();
    }
}

void
KeepAliveAgent::registerConn(const Connection& conn)
{
    cpLog(LOG_DEBUG, "registering connection:%s", conn.getDescription().c_str());
    _connectionList.push_back(conn);
}

void
KeepAliveAgent::unregisterConn(const Connection& conn)
{
    ConnectionList::iterator itr = find(_connectionList.begin(), _connectionList.end(), conn);
    if (itr != _connectionList.end()) _connectionList.erase(itr);
}
