
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSClientOpenMsg_cxx_Version =
    "$Id: COPSClientOpenMsg.cxx,v 1.10 2000/12/18 23:42:53 bko Exp $";
#include <assert.h>

#include "COPSPepId.hxx"
#include "COPSIpv4LastPdpAddr.hxx"
#include "COPSIpv6LastPdpAddr.hxx"
#include "COPSClientSI.hxx"
#include "COPSIntegrity.hxx"
#include "COPSClientOpenMsg.hxx"
#include "cpLog.h"

COPSClientOpenMsg::COPSClientOpenMsg()
        : _pepId(0),
        _clientSI(0),
        _pdpAddress(0),
        _integrity(0)
{}

COPSClientOpenMsg::COPSClientOpenMsg(const COPSClientOpenMsg& other)
{
    copyRhs(other);
}


COPSClientOpenMsg::COPSClientOpenMsg(const char* data)
        : _pepId(0),
        _clientSI(0),
        _pdpAddress(0),
        _integrity(0)
{
    parse(data);
}

void
COPSClientOpenMsg::parse(const COPSHeader& hdr, const char* data) throw (COPSBadDataException&)
{
    assert(hdr.getOpCode() == COPS_OP_OPN);
    _hdr = new COPSHeader(hdr);
    parse(data);
    setMsgLength();
}

COPSClientOpenMsg::~COPSClientOpenMsg()
{}


void
COPSClientOpenMsg::parse(const char* data) throw (COPSBadDataException&)
{
    COPSMsg::parseHeader(data);

    while (_dataStart < _dataLength)
    {
        COPSObjHeader objHdr;
        objHdr.parse(data + _dataStart);
        switch (objHdr.getCNum())
        {
            case COPS_PEPID:
            {
                cpLog(LOG_DEBUG, "Creating COPSPepId");
                _pepId = new COPSPepId(data + _dataStart);
                _dataStart += _pepId->getDataLength();
            }
            break;
            case COPS_LAST_PDP_ADDR:
            {
                cpLog(LOG_DEBUG, "Creating COPSLastPdpAddress");
                if (objHdr.getCType() == 1)
                {
                    _pdpAddress = new COPSIpv4LastPdpAddr(data + _dataStart);
                }
                else if (objHdr.getCType() == 2)
                {
                    _pdpAddress = new COPSIpv6LastPdpAddr(data + _dataStart);
                }
                _dataStart += _pdpAddress->getDataLength();
            }
            break;
            case COPS_CSI:
            {
                cpLog(LOG_DEBUG, "Creating COPSClientSI");
                _clientSI = new COPSClientSI(data + _dataStart);
                _dataStart += _clientSI->getDataLength();
            }
            break;
            case COPS_MSG_INTEGRITY:
            {
                cpLog(LOG_DEBUG, "Creating COPSIntegrity");
                _integrity = new COPSIntegrity(data + _dataStart);
                _dataStart += _integrity->getDataLength();
            }
            break;
            default:
            {
                cpLog(LOG_ALERT, "Bad Message format, unknown object type");
                throw COPSBadDataException("Bad Message format", __FILE__, __LINE__, 0);
            }
            break;
        }
    }
    checkSanity();
}

void
COPSClientOpenMsg::checkSanity() const throw (COPSBadDataException&)
{
    if (!_hdr || !_pepId )
    {
        cpLog(LOG_ALERT, "Bad Message format");
        throw COPSBadDataException("Bad message format",
                                   __FILE__, __LINE__, 0);
    }
}

void
COPSClientOpenMsg::dump(ostream& os)const
{
    if (_hdr.getPtr()) os << *_hdr << endl;
    if (_pepId.getPtr()) os << *_pepId << endl;
    if (_clientSI.getPtr()) os << *_clientSI << endl;
    if (_pdpAddress.getPtr()) os << *_pdpAddress << endl;
    if (_integrity.getPtr()) os << *_integrity << endl;
}

COPSClientOpenMsg&
COPSClientOpenMsg::operator<<(const COPSPdpAddress& obj)
{
    assert(!_pdpAddress);
    assert(obj.isLastPdpAddress());
    //Message integrity object should be the very last one
    //If it is already added
    assert(_integrity == 0);

    if (obj.isIpv4PdpAddress())
    {
        const COPSIpv4LastPdpAddr* tPtr = dynamic_cast < const COPSIpv4LastPdpAddr* > (&obj);
        assert(tPtr);
        _pdpAddress = new COPSIpv4LastPdpAddr(*tPtr);
    }
    else
    {
        const COPSIpv6LastPdpAddr* tPtr = dynamic_cast < const COPSIpv6LastPdpAddr* > (&obj);
        assert(tPtr);
        _pdpAddress = new COPSIpv6LastPdpAddr(*tPtr);
    }
    setMsgLength();
    return (*this);
}

COPSClientOpenMsg&
COPSClientOpenMsg::operator<<(const COPSClientSI& obj)
{
    assert(!_clientSI);
    assert(obj.isClientSI());
    _clientSI = new COPSClientSI(obj);
    setMsgLength();
    return (*this);
}

COPSClientOpenMsg&
COPSClientOpenMsg::operator<<(const COPSPepId& obj)
{
    assert(!_pepId);
    assert(obj.isPepId());
    _pepId = new COPSPepId(obj);
    setMsgLength();
    return (*this);
}

COPSClientOpenMsg&
COPSClientOpenMsg::operator<<(const COPSIntegrity& obj)
{
    assert(!_integrity);
    assert(obj.isMessageIntegrity());
    _integrity = new COPSIntegrity(obj);
    setMsgLength();
    return (*this);
}

COPSClientOpenMsg&
COPSClientOpenMsg::operator<<(const COPSHeader& obj)
{
    assert(!_hdr);
    assert(obj.getOpCode() == COPS_OP_OPN);
    _hdr = new COPSHeader(obj);
    setMsgLength();
    return (*this);
}

void
COPSClientOpenMsg::setMsgLength()
{
    u_int16_t len = 0;
    if (_pepId.getPtr()) len += _pepId->getDataLength();
    if (_clientSI.getPtr()) len += _clientSI->getDataLength();
    if (_pdpAddress.getPtr()) len += _pdpAddress->getDataLength();
    if (_integrity.getPtr()) len += _integrity->getDataLength();
    _hdr->setMsgLength(len);
}

void
COPSClientOpenMsg::writeData(int id) const throw(VNetworkException&)
{
    checkSanity();
    if (_hdr.getPtr()) _hdr->writeData(id);
    if (_pepId.getPtr()) _pepId->writeData(id);
    if (_clientSI.getPtr()) _clientSI->writeData(id);
    if (_pdpAddress.getPtr()) _pdpAddress->writeData(id);
    if (_integrity.getPtr()) _integrity->writeData(id);
}

COPSClientOpenMsg& 
COPSClientOpenMsg::operator=(const COPSClientOpenMsg& other)
{
    if(this != &other)
    {
        copyRhs(other);
    }
    return *this;
}

void
COPSClientOpenMsg::copyRhs(const COPSClientOpenMsg& other)
{
    _hdr = other._hdr;
    _dataLength = other._dataLength;
    _dataStart = other._dataStart;
    _pepId = other._pepId;
    _clientSI = other._clientSI;
    _pdpAddress = other._pdpAddress;
    _integrity = other._integrity;
}
