
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const COPSHeader_cxx_Version =
    "$Id: COPSHeader.cxx,v 1.10 2001/05/13 07:26:01 icahoon Exp $";

#include "VocalCommon.hxx" 
#include <netinet/in.h>
#include <cstdio>

#include "COPSHeader.hxx"
#include "COPSUtil.hxx"
#include "cpLog.h"

#define LOG(x) cpLog(LOG_DEBUG, (x))

COPSHeader::COPSHeader()
{
    _dataRep._versionNflg = 0x10;
    _dataRep._opCode = 0;
    _dataRep._cType = 0;
    _dataRep._msgLength = 0;
}

COPSHeader::COPSHeader(u_int8_t opCode, u_int16_t clientType)
{
    _dataRep._versionNflg = 0x10;
    _dataRep._opCode = opCode;
    _dataRep._cType = clientType;
    _dataRep._msgLength = 0;
    if (isAKeepAlive()) _dataRep._cType = 0;
}

COPSHeader::COPSHeader(const char* buf)
{
    u_int16_t a;
    memcpy(&a, buf, 2);
    memcpy(&_dataRep, &a, 2);

    memcpy(&a, buf + 2, 2);
    a = ntohs(a);
    memcpy(&_dataRep._cType, &a, 2);

    u_int32_t b;
    memcpy(&b, buf + 4, 4);
    b = ntohl(b);
    memcpy(&_dataRep._msgLength, &b, 4);
}

//Set the solicitation flag
void
COPSHeader::setFlag(u_int8_t flg)
{
    u_int8_t tFlg = _dataRep._versionNflg;
    tFlg &= 0x01;
    tFlg &= flg;
    _dataRep._versionNflg &= 0x10;
    _dataRep._versionNflg |= tFlg;
};

///
COPSHeader::COPSHeader(const COPSHeader& other)
{
    _dataRep._versionNflg = other._dataRep._versionNflg;
    _dataRep._opCode = other._dataRep._opCode;
    _dataRep._cType = other._dataRep._cType;
    _dataRep._msgLength = other._dataRep._msgLength;
}

bool
COPSHeader::isARequest() const
{
    LOG("COPSHeader::isARequest");
    return (_dataRep._opCode == COPS_OP_REQ);
}

bool
COPSHeader::isADecision() const
{
    LOG("COPSHeader::isADecision");
    return (_dataRep._opCode == COPS_OP_DEC);
}

bool
COPSHeader::isAReport() const
{
    LOG("COPSHeader::isAReport");
    return (_dataRep._opCode == COPS_OP_RPT);
}

bool
COPSHeader::isADeleteReq() const
{
    LOG("COPSHeader::isADeleteReq");
    return (_dataRep._opCode == COPS_OP_DRQ);
}

bool
COPSHeader::isASyncStateReq() const
{
    LOG("COPSHeader::isASyncStateReq");
    return (_dataRep._opCode == COPS_OP_SSQ);
}

bool
COPSHeader::isAClientOpen() const
{
    LOG("COPSHeader::isAClientOpen");
    return (_dataRep._opCode == COPS_OP_OPN);
}

bool
COPSHeader::isAClientAccept() const
{
    LOG("COPSHeader::isAClientAccept");
    return (_dataRep._opCode == COPS_OP_CAT);
}

bool
COPSHeader::isAClientClose() const
{
    LOG("COPSHeader::isAClientClose");
    return (_dataRep._opCode == COPS_OP_CC);
}

bool
COPSHeader::isAKeepAlive() const
{
    LOG("COPSHeader::isAKeepAlive");
    return (_dataRep._opCode == COPS_OP_KA);
}

bool
COPSHeader::isASyncComplete() const
{
    LOG("COPSHeader::isASyncComplete");
    return (_dataRep._opCode == COPS_OP_SSC);
}

void
COPSHeader::setMsgLength(u_int32_t len)
{
    if (len % 4)
        cpLog(LOG_ALERT, "Message is not aligned on 32 bit intervals, len(bytes):%d" , len );
    assert((len % 4) == 0);
    _dataRep._msgLength = len + 8;
}

u_int32_t
COPSHeader::getMsgLength() const
{
    return _dataRep._msgLength;
}

///
void
COPSHeader::writeData(int id) const throw(VNetworkException&)
{
    HeaderData sentData;

    u_int16_t a;  //version, flags and opcode
    u_int16_t b;  //clientType
    memcpy(&a, &_dataRep, 2);
    memcpy(&b, &_dataRep._cType, 2);
    b = htons(b);
    u_int32_t c32 = htonl(_dataRep._msgLength);

    memcpy(&sentData, &a, 2);
    sentData._cType = b;
    sentData._msgLength = c32;

    COPSUtil::writeData(id, &sentData, 8);
}
