
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
#ifndef OSP_CFG_HXX
#define OSP_CFG_HXX



#include "global.h"
#include <iostream>
#include <string>
#include <cassert>

#include "ospdatatypes.h"

#include "Sptr.hxx"
#include "VNetworkException.hxx"
#include "VMissingDataException.hxx"
#include "OSPServerData.hxx"

static const char* const OSPCfgVersion = "$Id: OSPCfg.hxx,v 1.14 2001/10/30 21:48:45 icahoon Exp $";

#define _Open   open
#define _Read   read
#define _Close  close
#define _Lseek  lseek
#define _Strdup strdup
#define  O_BINARY  (0)

/// max size for call id
static const int CALL_ID_SZ = 256;   /* matching w/ OSP SDK v2.5.2 */
/// size of destination field
static const int DESTINATION_SZ = 64;
/// max size of auth. token
static const int TOKEN_SZ = 2096*8;
/// max number of call ids
static const int NUM_CALL_IDS = 3;
/// size of timestamp
static const int TIMESTAMP_SZ = 32;

/// OSP service point url size
static const int SVC_PT_SZ = 262;
/// size of local certificate
static const int CERT_SZ = 1024;
/// size of private key
static const int PKEY_SZ = 1024;

/// max service points supported
static const int MAX_SERVICE_POINTS = 10;

/** C++ wrapper class containing OSP Provider information -
 * OSP Server connection characteristics:
 * shadow data from PolicyCfg but type friendly for use with
 * the SDK:
 */
class OSPCfg
{
    public:
        /// method to access the Singleton OSPCfg instance
        static OSPCfg& Instance();

        /// destructor
        virtual ~OSPCfg( );

        /// clean up on exit()
        static void destroy( void );

        /**
         * read the Xml file from provision server.
         * use PolicyCfg object to populate data needed for the OSP Server
         * provider
	 */
        void readCfgFile(void);

        /// get config data for the OSP Server provider:
        void getServerConfiguration( void ) throw (VMissingDataException & );

        /// print out object content - for debug only
        void display( void );

        /// OSP audit url - not used but need to be provided
        string auditurl;
	/// array for OSP Server url's
        char *servicepoints[MAX_SERVICE_POINTS];

	/// TransNexus customer id
        string custid;
	/// TransNexus device id
        string devid;
	/// number of OSP Serice Points configured
        int num_serv_points;
	/** whether local validation or remote validation is done.
	 *  1 = local validation (for now, always do local val
	 *  0 = remote validation
	 */
        bool localValidation;
	/// life time for SSL connection in seconds
        int sslLifeTime;          // in seconds
	/// max HTTP simultaneous connections allowed
        int httpMaxConnections;
	/** # of seconds an HTTP connection should be
	 *  maintained after the completion of a communication
	 *  exchange
	 */
        int httpPersistence;      // in unit of second
	/// # seconds of delay before reconnection retries
        int httpRetryDelay;       // ditto
	/// max number of reconnection retries
        int httpRetryLimit;       // max # of connection retries
	/** max milliseconds to wait for a response from the OSP
	 *  Server
	 */
        int httpTimeout;          // in milliseconds

        /// number of CA certificates (default to 1 for now)
        size_t NUM_CA_CERTS;
	/// num of local certificaates (default to 1 for now)
        size_t NUM_LOCAL_CERTS;
	/// num of public/private key pairs (default to 1 for now)
        size_t NUM_PKEYS;
	/// pointer to pointer to CA certificates
        unsigned char** certs;
	/// pointer to sizes of certificates
        int* certssizes;
	/// pointer to pointer to local certificates
        unsigned char** localcerts;
	/// pointer to sizes of local certs.
        int* localcertssizes;
	/// pointer to pointer to private keys
        unsigned char** pkeys;
	/// pointer to sizes of private keys
        int* pkeyssizes;
	/// private key structure
        OSPTPRIVATEKEY privatekey;
        ///

    protected:
        /// constructor
        OSPCfg();

    private:

	/// OSP Configurator singleton object
        static OSPCfg* _instance;

        /// Is object of this class well formed?
        bool isValid();

        /// clean up object
        void clear();

        /// copy constructor - not for use
        OSPCfg( const OSPCfg& rhs );
         
        /// assignment operator - not for use
        OSPCfg& operator=( const OSPCfg& rhs );

        ///
        static const char* const osptag;
        ///
        static const char* const servertag;
        ///
        static const char* const customeridtag;

}
;

#endif
