
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const PdpAgent_cxx_Version =
    "$Id: PdpAgent.cxx,v 1.25 2002/12/12 02:34:48 bko Exp $";
//Standard Include
#include <errno.h>
#include <sys/wait.h>
#include <stdio.h>
#include <strstream.h>
#include <signal.h>
#include <unistd.h>

#include <algorithm>

//User defined
#include "VNetworkException.hxx"
#include "VMissingDataException.hxx"
#include "NetworkAddress.h"
#include "PdpAgent.hxx"
#include "Connection.hxx"
#include "COPSHeader.hxx"
#include "COPSMsgParser.hxx"
#include "COPSProcessor.hxx"
#include "LockHelper.hxx"
#include "PolicyCfg.hxx"
#include "COPSTransceiver.hxx"
#include "VPolicyServerData.hxx"
#include "ProvisionInterface.hxx"
#include "cpLog.h"
#include "CommandLine.hxx"
#include "Cops_def.hxx"

static const int WELL_KNOWN_PDP_PORT = 3288;
static const int ATHOUSAND = 1000;
static const int BUFSIZE = 256;
static const int BIG_BUFSIZE = 1026;

PdpAgent* PdpAgent::_instance = 0;
PdpAgent::ConnectionMap PdpAgent::_connectionMap;
#if defined(__SUNPRO_CC)
extern "C" {

void PdpAgent_sigChld(int sigNo)
{
  PdpAgent::sigChld(sigNo);
    
}

void PdpAgent_sigTerm(int sigNo)
{
  PdpAgent::sigTerm(sigNo);
}

}

#else
#endif
void
PdpAgent::sigTerm(int signo)
{
    delete PdpAgent::_instance;
    PdpAgent::_instance = 0;
}

void
PdpAgent::sigChld(int signo)
{
    pid_t pid;
    int status;
    while (( pid = waitpid( -1, &status, WNOHANG)) > 0)
    {
        // don't call cpLog()
    }
}


PdpAgent&
PdpAgent::instance()
{
    if (!_instance)
    {
        _instance = new PdpAgent();
        cpLog(LOG_DEBUG, "Created instance for %s\n", _instance->className().c_str());
    }
    return *_instance;
}

PdpAgent::PdpAgent()
{
    int port = WELL_KNOWN_PDP_PORT;
#if defined (_linux_)    
 if (atexit(PdpAgent::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    };
#endif
    _serverSocket = new TcpServerSocket(port);
#if defined (__SUNPRO_CC)
    {
      signal(SIGCHLD, PdpAgent_sigChld);
      signal(SIGTERM, PdpAgent_sigTerm);
    }
#else
    signal(SIGCHLD, PdpAgent::sigChld);
    signal(SIGTERM, PdpAgent::sigTerm);
#endif
    char hName[BUFSIZE];
    int retVal = gethostname(hName, BUFSIZE);
    if (retVal != 0)
    {
        cpLog(LOG_ALERT, "Failed to get the PDP host name, reason %s",
              strerror(errno));
        exit( -1);
    }
    NetworkAddress na(hName);
    _pdpHostIp = na.getIpName().getString();

#ifdef HAS_OSP
    VPolicyServerData pData;
    ProvisionInterface::instance().getPolicyData(_pdpHostIp + ":3288", pData);
    int nThreads = pData.getMaxThreads();
    //Get Keep Alive timer value ans set it in seconds
    _kaTimerValue = (pData.getKeepAlive()) ? pData.getKeepAlive() / ATHOUSAND : 0;

#else
    int nThreads = 8;
    _kaTimerValue = 30;
#endif

    cpLog(LOG_DEBUG, "Creating %d threads for PDP", nThreads);
    cpLog(LOG_DEBUG, "Setting keepallive timer value to %d seconds", _kaTimerValue);

    _threadPool = new VThreadPool(nThreads);
    if(CommandLine::instance()->getInt("HEARTBEAT"))
    {
        // NB. make sure provisioningInterface is instantiated before
        //     heartbeat can begin.
        cpLog(LOG_DEBUG, "Creating heartbeat thread, port=%d", port);
        itsHeartbeatTxThread = new HeartbeatTxThread( port);
    }

}

void
PdpAgent::run()
{
    //Loop through for Incoming messages

    if (itsHeartbeatTxThread != 0) itsHeartbeatTxThread->run();

    for (; ; )
    {
        if (_serverSocket->getServerConn().isReadReady(1))
        {
            cpLog(LOG_DEBUG, "Accepting connection");
            Connection conn;
            _serverSocket->accept(conn);
            processRequest(conn);
        }
        for (ConnectionMap::iterator itr = _connectionMap.begin();
                itr != _connectionMap.end(); itr++)
        {
            Connection& conn = (*itr).second;
            if (conn.isLive() && conn.isReadReady(1))
            {
                cpLog(LOG_DEBUG , "Data ready on %s", conn.getDescription().c_str());
    
                VFunctor functor(PdpAgent::processRequestFunc, &conn);
                _threadPool->addFunctor(functor);
            }
        }
    }
}


PdpAgent::~PdpAgent()
{
    cpLog(LOG_DEBUG, "PdpAgent::~PdpAgent()");
    for (ConnectionMap::iterator itr = _connectionMap.begin();
            itr != _connectionMap.end(); itr++)
    {
        Connection& conn = (*itr).second;
        conn.close();
    }
    _serverSocket->close();
}

void
PdpAgent::processRequestFunc(void* con)
{
    Connection& conn = *(static_cast < Connection* > (con));
    cpLog(LOG_DEBUG, "Thread (%d) serving request on conn %s",
          pthread_self(), conn.getDescription().c_str());
    PdpAgent::instance().processRequest(conn);
}

void
PdpAgent::destroy()
{
    delete PdpAgent::_instance;
    PdpAgent::_instance = 0;
}

void
PdpAgent::processRequest(Connection& conn)
{
    try
    {
        cpLog(LOG_DEBUG, "Processing request");
        Sptr < COPSMsg > msg = COPSTransceiver::receiveMsg(conn.getConnId());
        if (msg->getHeader().isAClientOpen())
        {
            handleClientOpenMsg(conn, *msg);
        }
        else if (msg->getHeader().isADeleteReq())
        {
            handleDeleteRequestMsg(conn, *msg);
        }
        else if (msg->getHeader().isAClientClose())
        {
            handleClientCloseMsg(conn, *msg);
        }
        else if (msg->getHeader().isAKeepAlive())
        {
            handleKeepAliveMsg(conn, *msg);
        }
        else
        {
            cpLog(LOG_DEBUG, "Handling COPS message on %s." , conn.getDescription().c_str());
            COPSProcessor copsProcessor(conn.getConnId());
            copsProcessor.process(*msg);
        }
    }
    catch (VException& e)
    {
        cpLog(LOG_INFO, "CLOSING Connection (%s), reason:%s",
              conn.getDescription().c_str(),
              e.getDescription().c_str());
        conn.close();
        return ;
    }
}


void
PdpAgent::handleClientOpenMsg(const Connection& conn, const COPSMsg& msg)
{
    const COPSClientOpenMsg* cMsg = dynamic_cast < const COPSClientOpenMsg* > (&msg);
    const COPSPepId& pepId = cMsg->getPepId();
    string cntxt;
    try
    {
        cntxt = "(Client Close)";
        validateClientType(conn, msg);
        cntxt = "(Client Accept)";
        COPSHeader ahdr(COPS_OP_CAT, msg.getHeader().getClientType());
        COPSKATimer katimer(_kaTimerValue);
        COPSClientAcceptMsg acceptMsg;
        acceptMsg << ahdr ;
        acceptMsg << katimer ;
        acceptMsg.writeData(conn.getConnId());
        _connectionMap[pepId.getData().str()] = conn;
    }
    catch(VNetworkException& e)
    {
        cpLog(LOG_ALERT, "Failed to send %s message to connection %s",
                          cntxt.c_str(), conn.getDescription().c_str());
    }
}

void
PdpAgent::validateClientType(const Connection& conn, const COPSMsg& msg)
          throw (COPSBadDataException&)
{
    if((msg.getHeader().getClientType() != C_RSVP) &&
       (msg.getHeader().getClientType() != C_SIP))
    {
        //Unsupported client type
        COPSHeader cHdr (COPS_OP_CC, msg.getHeader().getClientType());
        COPSError err(COPS_ERR_UNSUPPORTED_CLIENT_TYPE, 0);
        COPSClientCloseMsg closeMsg;
        closeMsg << cHdr;
        closeMsg << err;
        closeMsg.writeData(conn.getConnId());
        throw COPSBadDataException("Unsupported client type",
                                   __FILE__, __LINE__, 0);
    }
}

void
PdpAgent::handleClientCloseMsg(Connection& conn, const COPSMsg& msg)
{
    cpLog(LOG_DEBUG, "Got close request, closing connection %s",
          conn.getDescription().c_str());
    conn.close();
}

void
PdpAgent::handleKeepAliveMsg(const Connection& conn, const COPSMsg& msg)
{
    cpLog(LOG_DEBUG, "Got keepAlive message");
    const COPSKAMsg* kaMsg = static_cast < const COPSKAMsg* > (&msg);
    char buf[BIG_BUFSIZE];
    ostrstream ostr(buf, BIG_BUFSIZE);
    ostr << *kaMsg << ends;
    cpLog(LOG_DEBUG, "Echoing KA Msg:%s", ostr.str());
    kaMsg->writeData(conn.getConnId());
}



void
PdpAgent::handleDeleteRequestMsg(const Connection& conn, const COPSMsg& msg)
{
    validateClientType(conn, msg);

    const COPSDeleteMsg* cMsg = dynamic_cast<const COPSDeleteMsg*>(&msg);

    COPSHandle handle ;
    char buf[BUFSIZE];
    ostrstream ostr(buf, BUFSIZE);
    ostr << cMsg->getClientHandle()->getId() << ends;
    cpLog(LOG_DEBUG, "Removing decision for handle %s", ostr.str());
    removeDecision(cMsg->getClientHandle()->getId());
}


void
PdpAgent::insertDecision(const COPSData& key, Sptr < COPSDecisionMsg > decMsg)
{
    _decisionMap[key.str()] = decMsg;
}

void
PdpAgent::removeDecision(const COPSData& key) throw (VMissingDataException&)
{
    if (_decisionMap.count(key.str()))
    {
        _decisionMap.erase(key.str());
    }
    else
    {
        char buf[BUFSIZE];
        sprintf(buf, "Missing decision for key %s", key.str().c_str());
        cpLog(LOG_WARNING, buf);
        throw VMissingDataException(buf, __FILE__, __LINE__, 0);
    }
}

const COPSContext&
PdpAgent::findDecisionContext(const COPSData& key) throw (VMissingDataException&)
{
    LockHelper guard(_lock);
    if (_decisionMap.count(key.str()))
    {
        COPSDecisionMsg& decMsg = *(_decisionMap[key.str()].getPtr());
        const COPSDecisionMsg::DecisionMap& dMap = decMsg.getDecisions();
        COPSDecisionMsg::DecisionMap::const_iterator itr = dMap.begin();
        return ((*itr).first);
    }
    else
    {
        char buf[BUFSIZE];
        sprintf(buf, "Missing decision for key %s", key.str().c_str());
        cpLog(LOG_WARNING, buf);
        throw VMissingDataException(buf, __FILE__, __LINE__, 0);
    }
}


///
bool
PdpAgent::isQoSAllowed(const string ip1, const string& ip2)
{
    for (LocalPolicyData::iterator itr = _localPolicyData.begin();
         itr != _localPolicyData.end(); itr++)
    {
        QoSData& qData = (*itr).second;
        if (qData.contains(ip1) && qData.contains(ip2))
        {
            cpLog(LOG_DEBUG, "QoS allowed for %s [%s -> %s]",
                  (*itr).first.c_str(), ip1.c_str(), ip2.c_str());
            return true;
        }
    }
    cpLog(LOG_DEBUG, "QoS not allowed for %s -> %s", ip1.c_str(), ip2.c_str());
    return false;
}


void
PdpAgent::insertLocalPolicy(const string& callId, const QoSData& qData)
{
    LockHelper guard(_lock);
    cpLog(LOG_DEBUG, "Inserting local policy for callId %s", callId.c_str());

    LocalPolicyData::iterator itr = _localPolicyData.find(callId);
    if (itr != _localPolicyData.end())
    {
        cpLog(LOG_DEBUG, "Inserting duplicate local policy for callId %s",
              callId.c_str());
    }
    else
    {
        _localPolicyData[callId] = qData;
    }
}

void
PdpAgent::removeLocalPolicy(const string& callId)
{
    LockHelper guard(_lock);
    cpLog(LOG_DEBUG, "Removing local policy for callId %s", callId.c_str());

    LocalPolicyData::iterator itr = _localPolicyData.find(callId);
    if (itr != _localPolicyData.end())
    {
        _localPolicyData.erase(itr);
    }
    else
    {
        cpLog(LOG_DEBUG, "Nonexistant QoSData to remove");
    }
}
