// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef Filter_hxx
#define Filter_hxx

#include "BaseIField.hxx"
#include "Sptr.hxx"

namespace Epoisses {

template <typename T> class Field;
class BaseColumnField;

/**
   Allows the user to specify Provisioning database filters with
   convenient syntax. 

   Filters are formed by operations on Fields and on Filters.
   E.g. 
   Filter a(Field<int>("foo") < 5);
   Filter b((Field<int>("foo") < 5) && 
   (Field<int>("bar") < Field<int>("baz")));

   The Filter is a wrapper to smart pointers so the
   arguments can be retained until the Filter and any parent Filters
   have gone out of scope. */
class Filter : public FilterableInterface
{
   public:
      class Operator
      {
            friend class Filter;
         protected:
            Operator(const string &_name,
                     bool _isUnary)
               : name(_name),
                 isUnaryFlag(_isUnary)
            {} 
            ~Operator() {} // not virtual -- derived classes do not add members
            bool isUnary() const
            {
               return isUnaryFlag;
            }
         private:
            const string name;
            const bool isUnaryFlag;
      };

      class UnaryOperator : Operator
      {
            friend class Filter;
         public:
            UnaryOperator(const string &_name)
               : Operator(_name, true)
            {}
      };

      class BinaryOperator : Operator
      {
            friend class Filter;
         public:
            BinaryOperator(const string &_name)
               : Operator(_name, false)
            {}
      };

#ifdef SHOULD_WORK
      // implicit conversion from Field<bool> to filter
      Filter(const Field<bool> &field);
#endif

      Filter(const Sptr<FilterableInterface>& _arg1,
             const UnaryOperator* _op)
         : op(_op),
           arg1(_arg1)
      {}

      Filter(const Sptr<FilterableInterface>& _arg1,
             const BinaryOperator* _op,
             const Sptr<FilterableInterface>& _arg2)
         : op(_op),
           arg1(_arg1),
           arg2(_arg2)
      {}

      //  empty Filter for initializing Filter var
      Filter()
         : op(Filter::NoOp)
      {}
      
      virtual ~Filter() {}

      bool empty() const
      {
         return op == NoOp;
      }

      Filter& operator=(const Filter &rhs)
      {
         if (this != &rhs)
         {
            op = rhs.op;
            arg1 = rhs.arg1;
            arg2 = rhs.arg2;
         }
         return *this;
      }
      
      bool
      operator==(const Filter &rhs) const
      {
         if (op != rhs.op)
         {
            return false;
         }
         
         if (op == NoOp)
         {
            return true;
         }
         
         if (arg1 == rhs.arg1)
         {
            return false;
         }
         
         if (op->isUnary())
         {
            return true;
         }
         
         return arg2 == rhs.arg2;
      }

      bool
      operator!=(const Filter &rhs) const
      {
         return !(*this == rhs);
      }
      
      static const UnaryOperator* NoOp; // empty filter
      static const UnaryOperator* IsNull; // comparison to NULL
      static const UnaryOperator* Self; // implied by Field<bool>
      static const UnaryOperator* Not;
      static const BinaryOperator* And;
      static const BinaryOperator* Or;

      static const BinaryOperator* Equal;
      static const BinaryOperator* NotEqual;
      static const BinaryOperator* Less;
      static const BinaryOperator* Greater;
      static const BinaryOperator* LessEqual;
      static const BinaryOperator* GreaterEqual;
      static const BinaryOperator* Like;
      // etc.

/*
      /// Create Negation Filter for this Filter
      Filter operator!()
      {
         // convert Filter by-value to heap Filter
         return Filter(Sptr<FilterableInterface>(new Filter(*this)), 
                       Not);
      }

      /// Create And-composing Filter from this and argument Filter
      Filter operator&&(const Filter &filter)
      {
         // convert Filter by-value to heap Filter
         return Filter(Sptr<FilterableInterface>(new Filter(*this)), 
                       And, 
                       Sptr<FilterableInterface>(new Filter(filter)));
      }

      /// Create Or-composing Filter from this and argument Filter
      Filter operator||(const Filter &filter)
      {
         // convert Filter by-value to heap Filter
         return Filter(Sptr<FilterableInterface>(new Filter(*this)), 
                       Or, 
                       Sptr<FilterableInterface>(new Filter(filter)));
      }
*/
      /// Copy constructor
      Filter(const Filter &rhs)
         : op(rhs.op),
           arg1(rhs.arg1),
           arg2(rhs.arg2)
      {}
   
      virtual void stringify(ostream &str) const;

      /// Ready made empty Filter for interface use.
      static const Filter None;

      /// For old test driver -- deprecated 
      static Filter parse(const string &str);

   private:
      const Operator *op;   
      Sptr<FilterableInterface> arg1;
      Sptr<FilterableInterface> arg2;
};


/// Global operator to logically (AND) combine Filters 
Filter
operator&&(const Filter &lhs,
           const Filter &rhs);

/// Global operator to logically (OR) combine Filters 
Filter
operator||(const Filter &lhs,
           const Filter &rhs);

/// Global operator to logically negate a Filter 
Filter
operator!(const Filter &arg);

Filter
isNull(const BaseColumnField& arg);

} // namespace Epoisses

#endif // Filter_hxx
