// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#include "Field.hxx"
#include <strstream.h>
#include <boost/tokenizer.hpp>
#include "Logger.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Epoisses;

IConstant::IConstant(const string &v)
   : BaseIField(convert(v))
{
   _name = convert(v); // !jf! revert string back to mixed case
}

IConstant::IConstant(const vector<string> &v)
   : BaseIField(convert(v))
{
   _name = convert(v); // !jf! revert string back to mixed case
}


void
IConstant::stringify(ostream &str) const
{
   str << _name;
}

string
IConstant::encode(const string &v)
{
   // .dlb. needs character by character encoding
   strstream s;
   s << "'" << v << "'" << ends;
   s.freeze(0);
   
   return s.str();
}

const string&
IConstant::convert(bool v)
{
   // does this depend on backend?
   static const string tr("true");
   static const string fa("false");
   return v ? tr : fa;
}

string
IConstant::convert(int v)
{
   strstream s;
   s << v << ends;
   s.freeze(0);
   return s.str();
}

string
IConstant::convert(long v)
{
   strstream s;
   s << v << ends;
   s.freeze(0);
   return s.str();
}


string
IConstant::convert(double v)
{
   // issues of precision here!
   strstream s;
   s << v << ends;
   s.freeze(0);
   return s.str();
}

string
IConstant::convert(const string &v)
{
   return encode(v);
}

string
IConstant::convert(const vector<string> &v)
{
   strstream s;
   for (vector<string>::const_iterator it = v.begin();
        it != v.end(); it++)
   {
      // for now, no embedded commas
      assert(it->find(',') == string::npos);
      s << *it;
      if (it+1 != v.end())
      {
         s << ",";
      }
   }
   
   s << ends;
   s.freeze(0);
   
   return encode(s.str());
}

void
IField<int>::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

// convert bound self to boolean
int IField<int>::value() const throw (NotBoundException&)
{
   if (!isBound) 
   {
      throw NotBoundException("Field<int>(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   } 
   return theValue;
}

void
IField<int>::bind(const string &s) throw (BadValueException&)
{
   theValue = atoi(s.c_str());
}

//======================================================================
   
void
IField<long>::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

// convert bound self to boolean
long
IField<long>::value() const throw (NotBoundException&)
{
   if (!isBound) 
   {
      throw NotBoundException("Field<long>(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   } 
   return theValue;
}

void
IField<long>::bind(const string &s) throw (BadValueException&)
{
   theValue = atol(s.c_str());
}

//======================================================================

void
IField<bool>::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

// convert bound self to boolean
bool IField<bool>::value() const throw (NotBoundException&)
{
   if (!isBound) 
   {
      throw NotBoundException("Field<bool>(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   } 
   return theValue;
}

void
IField<bool>::bind(const string &s) throw (BadValueException&)
{
   theValue = (s == "t");
}

//======================================================================

/// convert self to string
void
IField<double>::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

/// convert bound self to double
double
IField<double>::value() const throw (NotBoundException&)
{
   if (!isBound)
   {
      throw NotBoundException("Field<double>(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   }
   return theValue;
}

void
IField<double>::bind(const string &s) throw (BadValueException&)
{
   theValue = atof(s.c_str());
}

//======================================================================
void
IField<string>::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

const string &
IField<string>::value() const throw (NotBoundException&)
{
   if (!isBound)
   {
      throw NotBoundException("Field<string>(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   }
   return theValue;
}

void
IField<string>::bind(const string &s) throw (BadValueException&)
{
   // this one is easy
   theValue = s;
}

//======================================================================
void
IField< vector < string > >::stringify(ostream &str) const
{
   // XML sugar?
   str << _name;
}

const vector<string> &
IField< vector < string > >::value() const throw (NotBoundException&)
{
   if (!isBound)
   {
      throw NotBoundException("Field< vector < string > >(\"" + _name + 
                              "\") has not been bound to a query result",
                              __FILE__,
                              __LINE__);
   }
   return theValue;
}

void
IField< vector < string > >::bind(const string &s) throw (BadValueException&)
{
   theValue.clear();

   typedef boost::tokenizer<boost::char_delimiters_separator<char> > tokenizer;
   boost::char_delimiters_separator<char> sep(false, "", ",");
   tokenizer tok(s, sep);
   for(tokenizer::iterator it = tok.begin(); it != tok.end(); ++it)
   {
      theValue.push_back(*it);
   }
}

