/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <cstdio>
#include "cpLog.h"
#include <parse3tuple.h>
#include "PSConfiguration.hxx"
#include "NetworkAddress.h"
#include "Logger.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVS

using namespace Epoisses;

typedef enum
{
   DebugLevel = 0,
   LogFileName,
   RedundantServer,
   LocalPort,
   DaemonMode,
   NumThreads,
   MulticastHost,
   MulticastPort,
   DBConnection,
   database,
   host,
   port,
   user,
   password,
   dn,
   PSConfigurationTagMax
}  PSConfigurationTags;    


static const char* PSConfigurationTagString [] = 
{
   "DebugLevel",
   "LogFilename",
   "RedundantServer",
   "LocalPort",
   "DaemonMode",
   "NumThreads",
   "MulticastHost",
   "MulticastPort",
   "DBConnection",
   "database",
   "host",
   "port",
   "user",
   "password",
   "dn",
   "Unknown"
};


PSConfiguration* PSConfiguration::myInstance = 0;

void
PSConfiguration::parseConfigFile()
{
   FILE* fd = fopen (configFile.c_str(), "r");
   if (fd == NULL)
   {
      ErrLog (<< "No such provisioning config file exists: " << configFile);
      cerr << "No such provisioning config file exists: " << configFile << endl;
      exit(-1);
   }
   fclose(fd);

   assert ( parse3tuple( configFile.c_str(), parseCfgFileCallBack ) );
}

PSConfiguration*
PSConfiguration::instance()
{
   if ( myInstance == 0)
   {
      myInstance = new PSConfiguration();
   }
    
   return myInstance;
}


PSConfiguration::PSConfiguration()
   :debugLevel ("LOG_INFO"),
    logFileName ("psLogFile"),
    redundantServer ("NA"),
    redundantServerHostName (""),
    redundantServerHostPort ( -1 ),
    localPort ( 6005 ),
    daemonMode (true),
    numThreads ( 25 ),
    multicastHost ( "NA" ),
    multicastPort ( -1 ),
    configFile ("ps.cfg"),
    leftover ("")
{
   // !dlb! sadly, this is the only way to change the back end
   // needs serious refactoring

   // to serve the mascarpone team, by default connect to mySql
   // 
   dbInfo.dbBackend = "PGSQL";
   dbInfo.database = "cathay";
   dbInfo.host = "ottawa";
   dbInfo.port = 0; // used by LDAP only
   dbInfo.user = "david";
   dbInfo.password = "";
   dbInfo.dn = ""; // used by LDAP only

}

PSConfiguration::~PSConfiguration()
{
   if ( myInstance != 0)
   {
      delete myInstance;
   }
}

void
PSConfiguration::show()
{
   DebugLog( << "--- PS Configuration ---");
   if ( getDaemonMode() ) 
   {
      DebugLog( <<  "    Daemon Mode is on");
   }
   else
   {
      DebugLog( <<  "    Daemon Mode is off");
   }

   DebugLog( <<  "        Debug Level : \"" << getDebugLevel()  << "\"");
   DebugLog( <<  "      Log File Name : \"" << getLogFileName()  << "\"");
   DebugLog( <<  "         Local Port : \"" << getLocalPort() << "\"");
   DebugLog( <<  "  Number of Threads : \"" << getNumThreads() << "\"");
   DebugLog( <<  " ******* Database Info ********");
   if (dbInfo.dbBackend == "MYSQL")
   {
      DebugLog( <<  " DB backend : \"" << dbInfo.dbBackend << "\"");
      DebugLog( <<  "   Database : \"" << dbInfo.database << "\"");
      DebugLog( <<  "       Host : \"" << dbInfo.host << "\"");
      DebugLog( <<  "       User : \"" << dbInfo.user << "\"");
      DebugLog( <<  "   Password : \"" << dbInfo.password << "\"");
   }
   else if (dbInfo.dbBackend == "PGSQL")
   {
      DebugLog( <<  " DB backend : \"" << dbInfo.dbBackend << "\"");
      DebugLog( <<  "   Database : \"" << dbInfo.database << "\"");
      DebugLog( <<  "       Host : \"" << dbInfo.host << "\"");
      DebugLog( <<  "       User : \"" << dbInfo.user << "\"");
      DebugLog( <<  "   Password : \"" << dbInfo.password << "\"");
   }
   else if (dbInfo.dbBackend == "LDAP")
   {
      DebugLog( <<  " DB backend : \"" << dbInfo.dbBackend << "\"");
      DebugLog( <<  "    Base DN : \"" << dbInfo.database << "\"");
      DebugLog( <<  "       Host : \"" << dbInfo.host << "\"");
      DebugLog( <<  "       Port : \"" << port << "\"");
      DebugLog( <<  "    Bind DN : \"" << dbInfo.dn << "\"");
      DebugLog( <<  "   Password : \"" << dbInfo.password << "\"");
   }
   else
   {
      DebugLog (<< "DB backend not supported: " << dbInfo.dbBackend);
   }
   DebugLog( <<  " *******************************");
}

void
PSConfiguration::parseCfgFileCallBack( char* tag, char* type, char* value)
{

   switch( strConst2i( tag, 
                       PSConfigurationTagString, 
                       PSConfigurationTagMax ) )
   {
      case DebugLevel:
      {
         instance()->setDebugLevel( value );
      }
      break;

      case LogFileName:
      {
         instance()->setLogFileName( value );
      }
      break;

      case RedundantServer:
      {
         instance()->setRedundantServer( value );
      }
      break;

      case LocalPort:
      {
         int i;
         sscanf( value, "%d", &i );
         instance()->setLocalPort( i );
      }
      break;
 
      case DaemonMode:
      {
         instance()->setDaemonMode( (string( value ) == "True" ) );
      }
      break;

      case NumThreads:
      {
         int i;
         sscanf( value, "%d", &i );
         instance()->setNumThreads( i );
      }
      break;

      case MulticastHost:
      {
         instance()->setMulticastHost( value );
      }
      break;

      case MulticastPort:
      {
         int i;
         sscanf( value, "%d", &i );
         instance()->setMulticastPort( i );
      }
      break;

      case DBConnection:
      {
         instance()->setDBConnection( value );
      }
      break;
	
      case database:
      {
         instance()->setDatabase( value );
      }
      break;

      case host:
      {
         instance()->setDBHost( value );
      }
      break;

      case port:
      {
         int i;
         sscanf( value, "%d", &i );
         instance()->setDBPort( i );
      }
      break;

      case user:
      {
         instance()->setDBUser( value );
      }
      break;

      case password:
      {
         instance()->setDBPassword( value );
      }
      break;

      case dn:
      {
         instance()->setDN( value );
      }
      break;

      default:
      {
         InfoLog( << "Unknown Tag: " << tag );
      }
   }
}

void
PSConfiguration::setDebugLevel( string value )
{
   debugLevel = value;
}

void
PSConfiguration::setLogFileName( string value )
{
   logFileName = value;
}

void
PSConfiguration::setRedundantServer( string value )
{
   string::size_type start;
   string::size_type len;
   redundantServer = value;

   start = redundantServer.find(":");
   len = redundantServer.length();

   redundantServerHostName = (redundantServer.substr(0, start ) );

   if (start != string::npos)
   {
      string portStr = (redundantServer.substr(start + 1, len));
      redundantServerHostPort = atoi(portStr.c_str());
   }
   else
   {
      ErrLog (<< "redundant server argument requires host:port");
      exit( -1);
   }
   NetworkAddress tempAddr(redundantServerHostName, redundantServerHostPort);
   redundantServerHostName = tempAddr.getIpName();

}

void
PSConfiguration::setLocalPort( int value )
{
   localPort = value;
}

void
PSConfiguration::setDaemonMode( bool value )
{
   daemonMode = value;
}

void
PSConfiguration::setNumThreads( int value )
{
   numThreads = value;
}

void
PSConfiguration::setMulticastHost( string value )
{
   multicastHost = value;
}

void
PSConfiguration::setMulticastPort( int value )
{
   multicastPort = value;
}

void
PSConfiguration::setConfigFile( string value )
{
   configFile = value;
}


string
PSConfiguration::getDebugLevel()
{
   return debugLevel;
}

string
PSConfiguration::getLogFileName()
{
   return logFileName;
}

string
PSConfiguration::getRedundantServer()
{
   return redundantServer;
}

string
PSConfiguration::getRedundantServerHostName()
{
   return redundantServerHostName;
}

int
PSConfiguration::getRedundantServerHostPort()
{
   return redundantServerHostPort;
}


int
PSConfiguration::getLocalPort()
{
   return localPort;
}

bool
PSConfiguration::getDaemonMode()
{
   return daemonMode;
}

int
PSConfiguration::getNumThreads()
{
   return numThreads;
}

string
PSConfiguration::getMulticastHost()
{
   return multicastHost;
}

int
PSConfiguration::getMulticastPort()
{
   return multicastPort;
}

string
PSConfiguration::getConfigFile()
{
   return configFile;
}


/*
 * Database Information
 *
 */

///
const DBInfo
PSConfiguration::getDBInfo()
{
   return dbInfo;
}


void
PSConfiguration::setDBConnection(const string& value)
{
   dbInfo.dbBackend = value;
}

void
PSConfiguration::setDatabase(const string& value)
{
   dbInfo.database = value;
}

void
PSConfiguration::setDBHost(const string& value)
{
   dbInfo.host = value;
}

void
PSConfiguration::setDBUser(const string& value)
{
   dbInfo.user = value;
}

void
PSConfiguration::setDBPassword(const string& value)
{
   dbInfo.password = value;
}

void
PSConfiguration::setDBPort(int value)
{
   dbInfo.port = value;
}

void
PSConfiguration::setDN(const string& value)
{
   dbInfo.dn = value;
}

