/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



static const char* const PSProtocolService_cxx_Version = "$Id: PSProtocolService.cxx,v 1.1 2002/09/30 04:55:12 jason Exp $";

#include <memory>
#include <stdio.h>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/util/XMLException.hpp>
#include <xercesc/sax2/SAX2XMLReader.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <boost/lexical_cast.hpp>



#include "Logger.hxx"
#include "VNetworkException.hxx"
#include "VDbException.hxx"
#include "XmlContent.hxx"
#include "HttpHeader.hxx"
#include "PSProtocolService.hxx"
#include "PSReqXmlContentHandlers.hxx"
#include "PSRspXmlContentHandlers.hxx"

#include "XmlBlobContentHandler.hxx"
#include "PSBlobResult.hxx"

#include "Domain.hxx"
#include "BaseField.hxx"
#include "FieldList.hxx"
#include "AssignmentList.hxx"
#include "Filter.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROV

using namespace Epoisses;

static const char* saxMemBufId = "psResponse"; 

PSProtocolService::PSProtocolService() throw(VException&)
{
   try 
   {
      // init the XML4C2 system
      XMLPlatformUtils::Initialize();
   }
   catch (const XMLException& toCatch)
   {
      DebugLog (<< "Error during XMLPlatformUtils initialization caught: " 
                << StrX(toCatch.getMessage()));
      throw (VNetworkException("Error during XMLPlatformUtils init", __FILE__, __LINE__));
   }
}


/// Destructor
PSProtocolService::~PSProtocolService()
{
   DebugLog (<< "terminate XML4C2 system");
   XMLPlatformUtils::Terminate();
}

///
Sptr<PSDBRequest>
PSProtocolService::parseRequest( PSMessage::msgType type, 
                                  const string& xmlBody )
                                  throw( VException& )  
{
   //DebugLog (<< " parse request xml=" << xmlBody);

     Sptr<PSDBRequest> request = new PSDBRequest( type );

     if (xmlBody.size() == 0)
     {
         // DebugLog( <<  "0 length Xml content");
         return request;    /* empty result set */
     } 
     // now parse the xml content and fill in PSDBResult
     // create a SAX parser object.
     std::auto_ptr<SAX2XMLReader> parser(XMLReaderFactory::createXMLReader());
     try
     {
         PSReqXmlContentHandlers handler( request );
         parser->setContentHandler(&handler);
         parser->setErrorHandler(&handler);
         MemBufInputSource buf((const XMLByte*)xmlBody.c_str(), xmlBody.size(),
                               saxMemBufId, false);
         parser->parse(buf);
     }
     catch (const XMLException& toCatch)
     {
         DebugLog(<< "parse request error " << StrX(toCatch.getMessage()));
	 throw VNetworkException("parse request error", __FILE__, __LINE__);
     }
     
     return request;
}



///
Sptr<PSDBResult>
PSProtocolService::parseResponse( PSMessage::msgType type, 
                                  const string& rawHttpMsg )
//                                  throw( VException& )  
{
   //DebugLog(<< "PSProtocolService::parseResponse(" << rawHttpMsg << ")");
   
     Http::Message decodeMsg;

     try
     {
         decodeMsg.decode( rawHttpMsg );
     }
     catch ( Http::Exception & e )
     {
        DebugLog (<< "parseResponse Error: " << e.getDescription());
        throw VNetworkException("parse response error", __FILE__, __LINE__);
     }

     Sptr<PSDBResult> result = new PSDBResult( type );

     // get status
     const Http::Status& stsObj = decodeMsg.status();
     result->setStatus( stsObj.code());    
     /**  don't care for headers for now
     const list<Http::Header>& hdrs = decodeMsg.headers();
     if (hdrs.size() == 0)
     {
         DebugLog( "no Http headers decoded!");
         return result;
     }
     */
     const string& xmlBody = decodeMsg.body();
     if (xmlBody.size() == 0)
     {
         return result;    /* empty result set */
     } 
     // now parse the xml content and fill in PSDBResult
     // create a SAX parser object.
     auto_ptr<SAX2XMLReader> parser(XMLReaderFactory::createXMLReader());
     try
     {
         PSRspXmlContentHandlers handler( result );
         parser->setContentHandler(&handler);
         parser->setErrorHandler(&handler);
         MemBufInputSource memBufIS((const XMLByte*)xmlBody.c_str(),
                                    xmlBody.size(),
                                    saxMemBufId, 
                                    false);
         parser->parse(memBufIS);
     }
     catch (const XMLException& toCatch)
     {
        DebugLog (<< "caught: " << StrX(toCatch.getMessage()));
        DebugLog (<< "error parsing response: " << xmlBody);

        throw (VNetworkException("parseResponse error", __FILE__,__LINE__));
     }

     if (!result->getError().empty())
     {
        InfoLog (<< "Error in response from server: " << result->getError());
        throw VDbException(result->getError(), __FILE__,__LINE__);
     }
     
     return result;
}

///
Sptr<PSBlobResult>
PSProtocolService::parseXmlBlob(const string& xmlBlob)
  throw(VException&)  
{
   //DebugLog (<<  "parseXmlBlob: " << xmlBlob);

    Sptr<PSBlobResult> result = new PSBlobResult();

    if (xmlBlob.size() == 0)
      {
         DebugLog (<<  "No XmlBlob given");
	return result;
     } 

    // create a SAX parser object.
    std::auto_ptr<SAX2XMLReader> parser(XMLReaderFactory::createXMLReader());
    try
     {
       XmlBlobContentHandler handler(result);
       parser->setContentHandler(&handler);
       parser->setErrorHandler(&handler);
       MemBufInputSource memBufIS((const XMLByte*)xmlBlob.c_str(), 
                                  xmlBlob.size(),
                                  saxMemBufId,
                                  false);
       parser->parse(memBufIS);
     }
    catch (const XMLException& toCatch)
      {
        DebugLog(<< "parse blob error " << StrX(toCatch.getMessage()));
	throw VNetworkException ("parse xml blob error", __FILE__, __LINE__);
      }

    return result;
}


///
Sptr<Http::Message> 
PSProtocolService::buildHttpMessage( Http::Type mthod,
                       const string& psReqType,
                       const string& body )
{
    Http::Method method(mthod);   // GET/PUT/POST/DELETE
    Http::Url    url(psReqType);  // e.g. "GetRecordsInDomain"

    /// need other headers???
    list<Http::Header> headers;
    // Http::Header acceptCharset(Http::Header::AcceptCharset, 
    //                            "iso-8859-1,*,utf-8");
    // headers.push_back(acceptCharset);

    Http::Header contentLength(Http::Header::ContentLength, boost::lexical_cast<string>(body.size()).c_str());
    headers.push_back(contentLength);

    // Assemble a HttpMessage
    Sptr<Http::Message> request = new Http::Message(method, url, headers, body );

    return request;

}

// *****************************************************************
// Generic atomic data accessor functions
// *****************************************************************

/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildGetXmlDescriptionReq(
                                const string& xmlCategory, 
                                const string& xmlDescription)
  throw( VException& ) 
{
   //DebugLog(<<  "buildGetXMLDescription() category " << xmlCategory << ", description " << xmlDescription);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;

    const string body = builder.buildGetXmlDescriptionReq( xmlCategory,
                                                     xmlDescription);
  
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::GET_XML_DESC_STR,
                              body ) ); 
}

/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildListXmlCatagoriesReq()
  throw( VException& ) 
{
   //DebugLog(<<  "buildListXMLCatagories()");

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;

    const string body = builder.buildListXmlCatagoriesReq();
  
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::LIST_XML_CAT_STR,
                              body ) ); 
}


/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildListXmlDescriptionsReq(
    const string& xmlCategory)
  throw( VException& ) 
{
   //DebugLog(<<  "buildGetXMLDescription() category "<< xmlCategory);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;
  
    const string body = builder.buildListXmlDescriptionsReq( xmlCategory );  
    
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::LIST_XML_DESC_STR,
                              body ) ); 
}


/** retrive record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<Http::Message> PSProtocolService::buildGetRecordsInDomainReq(
                                const Domain& domain, 
                                const Filter& filterRules,
                                const FieldList& fieldList)
  throw( VException& ) 
{
   //DebugLog(<<  "buildGetRecordsInDomainReq() domain " << domain.getName() << ", rules " << filterRules.toString());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;
  
    const string body = builder.buildGetRecordsInDomainReq( domain,
                                                      filterRules, fieldList );
    
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::GET_RECORDS_IN_DOMAIN_STR,
                              body ) ); 
}

/** retrive record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<Http::Message> PSProtocolService::buildGetRecordsInDomainReq(
                                const Domain& domain, 
                                const Filter& filterRules,
                                const FieldList& fieldList,
                                const Field<void>& orderBy)
  throw( VException& ) 
{
   //DebugLog(<<  "buildGetRecordsInDomainReq() domain " << domain.getName() << ", rules " << filterRules.toString() << " order " << orderBy.toString());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;
  
    const string body =  builder.buildGetRecordsInDomainReq(domain,
                                                      filterRules, fieldList,
                                                      orderBy);
    
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::GET_RECORDS_IN_DOMAIN_STR,
                              body ) ); 
}

///
Sptr<Http::Message> PSProtocolService::buildGetFieldsInDomainReq( 
                                const Domain& domain)
  throw( VException& ) 
{
   //DebugLog(<<  "buildGetFieldsInDomainReq() domain "<<  domain.getName());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */
    PSReqXmlContentHandlers builder;
  
    const string body = builder.buildGetFieldsInDomainReq( domain );
  
    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::GET_FIELDS_IN_DOMAIN_STR,
                              body ) ); 

}

/// update existing records with keyword/value pairs provided
Sptr<Http::Message> 
PSProtocolService::buildSetRecordsInDomainReq(
                                const Domain& domain, 
                                const Filter& filterRules,
				const AssignmentList& assignments)
  throw (VException& )
{
   //DebugLog(<<  "buildSetRecordsInDomainReq() domain=" <<  domain.getName() <<  " filter=" << filterRules.toString());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSetRecordsInDomainReq( domain,
                                                      filterRules, assignments );

    return( buildHttpMessage( Http::Method::PUT,
                              PSCommon::SET_RECORDS_IN_DOMAIN_STR,
                              body ) );

}


/// insert a new record
Sptr<Http::Message> 
PSProtocolService::buildAddRecordInDomainReq( const Domain& domain, 
                                              const AssignmentList& assignments)
   throw( VException& )
{
   //DebugLog(<<  "buildAddRecordInDomainReq() domain "<<  domain.getName());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildAddRecordInDomainReq( domain,
                                                     assignments );

    return( buildHttpMessage( Http::Method::PUT,
                              PSCommon::ADD_RECORD_IN_DOMAIN_STR,
                              body ) );

}


Sptr<Http::Message>
PSProtocolService::buildDeleteRecordsInDomainReq( const Domain& domain,
                                                  const Filter& filterRules )
   throw( VException& )
{
   //DebugLog(<<  "buildDeleteRecordsInDomainReq() domain: " << domain.getName() << " filterRules=" << filterRules.toString());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildDeleteRecordsInDomainReq( domain,
                                                         filterRules );

    return( buildHttpMessage( Http::Method::DELETE,
                              PSCommon::DELETE_RECORDS_IN_DOMAIN_STR,
                              body ) );

}

// *****************************************************************
// Session related functions
// *****************************************************************

Sptr<Http::Message> 
PSProtocolService::buildSessionOpenReq( int timeInSeconds ) 
                              throw(VException&)
{
   //DebugLog(<<  "buildSessionOpenReq() expiry time= " << timeInSeconds << " seconds");

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionOpenReq( timeInSeconds );

    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::SESSION_OPEN_STR,
                              body ) );

}


///
Sptr<Http::Message>
 PSProtocolService::buildSessionCloseReq( const string& sessionID ) throw( VException& )
{
   //DebugLog(<<  "buildSessionCloseReq() sessionID= " <<  sessionID);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionCloseReq( sessionID );

    return( buildHttpMessage( Http::Method::POST,
                              PSCommon::SESSION_CLOSE_STR,
                              body ) );

}


///
Sptr<Http::Message> PSProtocolService::buildSessionSearchDBReq( 
                       const string& sessionID, 
                       const Domain& domain, 
                       const Filter& filterRules,
                       const FieldList& fieldList,  
                       PSCommon::PSSortType sType,
                       const BaseField& sortField ) throw(VException&)
{
   //DebugLog(<<  "buildSessionSearchDBReq() sessionID  " << sessionID<< "domain= " << domain.getName()<< ", filterRules=" << filterRules.toString());


    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionSearchDBReq( sessionID, domain,
                                                   filterRules, fieldList, sType, sortField );

    return( buildHttpMessage( Http::Method::POST,
                              PSCommon::SESSION_SEARCH_DB_STR,
                              body ) );

}


///
Sptr<Http::Message>
 PSProtocolService::buildSessionGetDBResultReq( const string& sessionID, int startIdx,
                       int endIdx, const FieldList& fieldList)
                       throw (VException&)
{
   //DebugLog(<<  "buildSessionGetDBResultReq() sessionID= " << sessionId << " startIdx=" << startIdx << " endIdx=" << endIdx);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionGetDBResultReq( sessionID,
                                                      startIdx, endIdx, fieldList );

    return( buildHttpMessage( Http::Method::GET,
                              PSCommon::SESSION_GET_DB_RESULT_STR,   
                              body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionSetDBResultReq( const string& sessionID, 
                                               const map< string, string>& fieldValPairs)
                       throw (VException&)
{
   //DebugLog(<<  "buildSessionSetDBResultReq() sessionID= "<<  sessionID );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionSetDBResultReq( sessionID,
                                                      fieldValPairs );

    return( buildHttpMessage( Http::Method::PUT,
                              PSCommon::SESSION_SET_DB_RESULT_STR,
                              body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionSortDBResultReq( 
                       const string& sessionID, PSCommon::PSSortType sType,
                       const BaseField& sortField)
                       throw (VException&)
{
   //DebugLog(<<  "buildSessionSortDBResultReq() sessionID="<< sessionID );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionSortDBResultReq( sessionID,
                                                       sType, sortField );

    return( buildHttpMessage( Http::Method::POST,
                              PSCommon::SESSION_SORT_DB_RESULT_STR,
                              body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionDeleteAllReq( const string& sessionID ) 
                       throw (VException&)
{
   //DebugLog(<<  "buildSessionDeleteAllReq() sessionID="<<  sessionID );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildSessionDeleteAllReq( sessionID );

    return( buildHttpMessage( Http::Method::DELETE,
                              PSCommon::SESSION_DELETE_ALL_STR,
                              body ) );

}


Sptr<Http::Message> 
PSProtocolService::buildRegisterForUpdateReq( const string& type,
					      const string& hostport,
					      const Domain& domain,
					      const KeySet & keySet,
					      const string & uid)
  throw (VException&)
{
   //DebugLog (<<  "build reg for update req type= " << type << " hostport=" << hostport << " domain=" << domain.getName());
    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildRegisterForUpdateReq(hostport, domain, keySet, uid);

    return (buildHttpMessage(Http::Method::POST, type, body));
}


Sptr<Http::Message> 
PSProtocolService::buildUpdateReq( const string& type,
                                   const string& domain,
				   const FieldListtype& keyList,
                                   PSCommon::ChangeType changeType)
  throw (VException&)
{
   //DebugLog (<<  "build-" << type << "-Req() " << domain);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = builder.buildUpdateReq(domain, keyList, changeType);

    return (buildHttpMessage(Http::Method::PUT, type, body));
}


Sptr<Http::Message> 
PSProtocolService::buildUnRegisterForUpdateReq( const string& type,
						const string& hostport,
                                                const Domain& domain,
						const string& uid)
  throw (VException&)
{
    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = 
      builder.buildUnRegisterForUpdateReq(hostport, domain, uid);

    return (buildHttpMessage(Http::Method::POST, type, body));
}

Sptr<Http::Message>
PSProtocolService::buildRegisterForForeignUpdateReq(const string& type,
                                                    const string& hostport,
                                                    const Domain& foreignDomain,
                                                    const string& foreignKeyField,
                                                    const Domain& primaryDomain,
                                                    const string& uid) 
   throw (VException&)
{
    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    string body = builder.buildRegisterForForeignUpdateReq(hostport, 
                                                           foreignDomain, foreignKeyField, 
                                                           primaryDomain, uid);

    return (buildHttpMessage(Http::Method::POST, type, body));
}


Sptr<Http::Message>
PSProtocolService::buildForeignUpdateReq(const string& type,
                                         const string& uid,
                                         const FieldListtype &keyList,
                                         PSCommon::ChangeType changeType)
   throw (VException&)
{
    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = 
      builder.buildForeignUpdateReq(uid, keyList, changeType);

    return (buildHttpMessage(Http::Method::PUT, type, body));
}

Sptr<Http::Message> 
PSProtocolService::buildUnRegisterForForeignUpdateReq(const string& type,
                                                      const string& hostport,
                                                      const Domain& foreignDomain,
                                                      const string & uid) 
   throw (VException&)
{
    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    const string body = 
       builder.buildUnRegisterForForeignUpdateReq(hostport, foreignDomain, uid);

    return (buildHttpMessage(Http::Method::POST, type, body));
}

