// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#ifndef Cache_hxx
#define Cache_hxx

#include <boost/lexical_cast.hpp>
#include <ext/hash_map>
#include <ext/hash_set>
#include <iostream>
#include <string>

#include "CacheManager.hxx"
#include "Data.hxx"
#include "Fifo.h"
#include "Lock.hxx"
#include "Logger.hxx"
#include "PSDBCommonType.h"
#include "ProvisioningInterface.hxx"
#include "RWMutex.hxx"
#include "UpdateFunctor.hxx"
#include "VDbException.hxx"
#include "VNetworkException.hxx"
#include "global.h"
#include "ThreadIf.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Vocal;

namespace Epoisses {

/// Functor for registering to a Cache to be notified of changes to contents
class CacheUpdator
{
   public:
      CacheUpdator() {}
      virtual ~CacheUpdator() {}

      virtual void operator()(const KeySet &keySet) = 0;
};

// forward decl
template <class T>
class Cache;

/// Wrapper for all derived Cache implementations. This allows VectorCache
/// to determine the Cache implementation singleton from the SETDATA type.
/// Need not specialize if there is no additional functionality
/// (e.g. VectorCache typically does not require more than Cache<T>).
///
/// Specialize this template as the derived Cache<T>:
/// class CacheImpl<Line> : public Cache<Line>
/// {
///   public:
///      CacheImpl<Line>();
///   ....
///  };
///
/// Then typedef the singleton and always get the cache instance via the
/// singleton.
/// typedef SingletonHolder< CacheImpl<Line> > LineCache;

template <class T>
class CacheImpl : public Cache<T>
{
};

class NotFoundInProvisioningException : public VException
{
   public:
      NotFoundInProvisioningException(const string& msg,
                                      const string& file,
                                      const int line)
         : VException(msg, file, line)
      {}

      virtual string getName() const
      {
         return "NotFoundInProvisioningException";
      }
};

template <class D> 
class CacheUpdateDelegator;

// Forward declaration required to make friends
template<class SETDATA, class KEY, class SET> class VectorCache;

// Forward declaration required to make friends
template<class SETDATA, class KEY, class SET>
ostream& operator<<(ostream& stream, const VectorCache<SETDATA, KEY, SET>& cache);

/// Allows an application to maintain a reflection of persisted data.
/// The cache keeps itself up-to-date with the database by clearing on
/// modification. The set of keys in the database is also kept up to date, so
/// cache misses do not require hitting the database fruitlessly.
///
/// The cache recovers from pserver disconnection and reconnection by servicing
/// requests against the old cache while:
///   - the pserver is down
///   - during the time it takes to reload the new data (in case the pserver was
///     still running and updates were not getting through).
///
/// T::Db must be a Domain
template<class T>
class Cache : public CacheBase, private ThreadIf
{
   public:
      typedef typename T::Db PT;
      
      friend ostream& operator<<<>(ostream&, const Cache&);

      //template <class KEY, class SET>
      //friend ostream& operator<<(ostream&, const VectorCache<T, KEY, SET> &cache);

      typedef hash_set<typename PT::KeyType, hash<typename PT::KeyType> > ExistsSet;
      typedef hash_map<typename PT::KeyType, T*, hash<typename PT::KeyType> > CacheMap;
      typedef CacheUpdateDelegator<Cache> Delegator;
      friend class Delegator;

      explicit Cache(unsigned int reserve = 100)
         : _cache(new CacheMap(reserve)),
           _existsSet(new ExistsSet(reserve)),
           _alternateCache(new CacheMap(0)),
           _alternateExistsSet(new ExistsSet(0)),
           _mutex(),
           _alternateMutex(),

           // working and updating point to the same data until the pserver
           // comes back after being lost
           _workingCache(_cache),
           _workingExistsSet(_existsSet),
           _updatingCache(_cache),
           _updatingExistsSet(_existsSet),
           _workingMutex(&_mutex),
           _updatingMutex(&_mutex),

           _updator(Sptr<Delegator>(new Delegator(*this))),
           _pif(ProvisioningInterface::make())
      {
         registerForUpdates();
         run();

         try
         {
            load();
         }
         catch (VException& e)
         {
            DebugLog (<< "Caught: " << e);
            shutdown();
            throw;
         }
      }

      virtual ~Cache()
      {
         shutdown();

         {
            Threads::WriteLock lock(_mutex);
            Threads::WriteLock lock1(_alternateMutex);

            for (typename CacheMap::iterator i = _cache->begin();
                 i != _alternateCache->end(); i++)
            {
               delete i->second;
            }
            for (typename CacheMap::iterator i = _alternateCache->begin();
                 i != _alternateCache->end(); i++)
            {
               delete i->second;
            }
         }
         InfoLog(<< "You have used " << getClassName());
      }

      void registerCacheUpdate(CacheUpdator *updator)
      {
         Vocal::Threads::WriteLock lock(_applicationUpdatorsMutex);
         _applicationUpdators.push_back(updator);
      }

      void unregisterCacheUpdate(CacheUpdator *updator)
      {
         Vocal::Threads::WriteLock lock(_applicationUpdatorsMutex);
         _applicationUpdators.remove(updator);
      }
      
      virtual string
      getClassName() const
      {
         return "Cache<" + T::className() + ">";
      }

      vector<typename PT::KeyType>
      getKeys() const
      {
         vector<typename PT::KeyType> keys;
         keys.reserve(size());

         Threads::ReadLock lock(*_workingMutex);
         for (typename ExistsSet::const_iterator i = _workingExistsSet->begin();
              i != _workingExistsSet->end(); i++)
         {
            keys.push_back(*i);
         }
         return keys;
      }

      bool empty()
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingExistsSet->empty();
      }
      
      unsigned int size() const
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingExistsSet->size();
      }

      unsigned int cachedSize() const
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingCache->size();
      }

      /// Return true iff the object corresponding to the key exists in DB
      bool exists(const typename PT::KeyType& keyObj)
      {
         Threads::ReadLock lock(*_workingMutex);
         //DebugLog(<< getClassName() << "::exists " << _workingExistsSet);
         // look for cached
         return _workingExistsSet->count(keyObj);
      }
            
      /// Functor allowing either read or read/write invocations.
      class ApplyFunctor
      {
         public:
            ApplyFunctor() {}
            virtual ~ApplyFunctor() {}
            
            /// Read-only -- may return copy of t or component via trace
            virtual void operator()(const T&) {}
            /// Write-only -- no trace of call
            virtual void operator()(T& t) const {}
      };

      /// Use in derived classes to extract copy of stored
      class IdentityFunctor : public ApplyFunctor
      {
         public:
            IdentityFunctor() {}
            // read-only
            virtual void operator()(const T& data)
            {
               // keep a trace of the call
               _data = data;
            }
            T _data;

            T
            get() 
            {
               return _data;
            }
      };

      /// Invokes func.operator()(const T&) in a read lock context.
      // .dlb. could templatize and avoid the functor function call
      void applyRead(const typename PT::KeyType& keyObj,
                     ApplyFunctor& func) const 
      {
         {
            Threads::ReadLock lock(*_workingMutex);
            // look for cached
            typename CacheMap::const_iterator it = _workingCache->find(keyObj);
            if (it != _workingCache->end())
            {
               func(static_cast<const T&>(*(it->second)));
               return;
            }

            // check for pserver and existence
            if (!SingletonHolder<CacheManager>::Instance().isServiceAvailable() ||
                _workingExistsSet->find(keyObj) == _workingExistsSet->end())
            {
               //DebugLog(<< getClassName() << " missed " << keyObj);
               throw NotFoundInProvisioningException("Not in provisioning",
                                                     __FILE__,
                                                     __LINE__);
            }
         }

         T* nt;
         try
         {
            // check database
            nt = new T(retrieve(keyObj));
         }
         catch (VException& e)
         {
            // remove from exists set
            Threads::WriteLock lock(*_workingMutex);
            _updatingExistsSet->erase(keyObj);
            
            DebugLog(<< "In " << T::className() << " constructor; Caught: " << e);
            throw NotFoundInProvisioningException("Error in constructor",
                                                  __FILE__, __LINE__);
         }

         func(static_cast<const T&>(*nt));
         
         {
            // add to cache
            Threads::WriteLock lock(*_workingMutex);
            // look for resize and log?
            (*_workingCache)[keyObj] = nt;
         }
      }

      /// Notice that the functor cannot change state.
      void applyWrite(const typename PT::KeyType& keyObj,
                      const ApplyFunctor& func)
      {
         //DebugLog(<< getClassName() << "::applyWrite(" << keyObj << ")");
         {
             Threads::ReadLock lock(*_workingMutex);

             typename CacheMap::const_iterator it = _workingCache->find(keyObj);
             if (it != _workingCache->end())
             {
                func(*(it->second));
                return;
             }

             // check for pserver and existence
             if (!SingletonHolder<CacheManager>::Instance().isServiceAvailable() ||
                 _workingExistsSet->find(keyObj) == _workingExistsSet->end())
             {
                DebugLog(<< getClassName() << " missed " << keyObj);
                throw NotFoundInProvisioningException("Not in provisioning",
                                                      __FILE__,
                                                      __LINE__);
             }
         }

         T* nt = new T(retrieve(keyObj));
         // invoke operator()(*T)
         func(*nt);
         
         {
            // add to cache
            Threads::WriteLock lock(*_workingMutex);
            // look for resize and log?
            (*_workingCache)[keyObj] = nt;
         }
      }

      // generic accessor
      T get(const typename PT::KeyType& keyObj) const
      {
         IdentityFunctor f;
         applyRead(keyObj, f);
         return f.get();
      }

   private:
      void registerForUpdates()
      {
         try
         {
            PT pt;
            _pif.registerDomainForUpdate(pt, _updator);
            // start the onUpdate dequeuer task
            DebugLog(<< getClassName() << "starting update thread");
         }
         catch (VNetworkException &e)
         {
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
            return;
         }
      }

      /// Load the cache with all data from provisioning
      void load()
      {
         PT pt;
         FieldList all = pt.all();
         // Make sure the key field is included
         all.add(pt.getKeyField());

         Sptr<PSDBResult> result;
         try
         {
            result = _pif.getRecordsInDomain(pt,
                                             Filter::None,
                                             all);
         }
         catch (VNetworkException &e)
         {
            DebugLog (<< "Failed to retrieve from pserver (network error)");
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
            return;
         }

         Threads::WriteLock lock(*_updatingMutex);
         for (PSDBResult::BindIterator bind = result->begin(all);
              bind != result->end(); bind++) 
         {
            //DebugLog(<< "adding " << *pt.getKeyField());
            // construct local storage structure
            try
            {
               T* n = new T(pt);
               {
                  (*_updatingCache)[*pt.getKeyField()] = n;
               }
            }
            catch (VException& e)
            {
               DebugLog (<< "Error creating " << T::className() << ": "  << pt << " throwing it away");
            }
            
            // add to the exists set
            _updatingExistsSet->insert(*pt.getKeyField());
         } 
      }

      void unregisterForUpdates()
      {
         PT pt;
         try 
         {
            _pif.unregisterForUpdate(pt, _updator);
         } 
         catch (VException &e)
         {
            DebugLog(<< "Caught: " << e << " while shutting down " << getClassName());
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
         }
      }

      PT
      retrieve(const typename PT::KeyType& key) const
      {
         // get from provisioning
         PT pt;
         FieldList all = pt.all();
         Sptr<PSDBResult> result;
         try
         {
            result = _pif.getRecordsInDomain(pt,
                                             pt.getKeyField() == key,
                                             all);
         }
         catch (VNetworkException& e)
         {
            DebugLog (<< "Failed to retreive from pserver (network error)");
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
            throw NotFoundInProvisioningException("Not in provisioning",
                                                  __FILE__,
                                                  __LINE__);
         }
         
         assert(result->size() <= 1);
         
         if (!result->empty()) 
         {
            // bind to only match
            PSDBResult::BindIterator bind = result->begin(all);
            return pt;
         }
         else
         {
            DebugLog(<< getClassName() << " missed " << key);
            throw NotFoundInProvisioningException("Not in provisioning",
                                                  __FILE__,
                                                  __LINE__);
         }
      }
      
      /// Called on each update from provisioning -- in cache's thread
      void
      onUpdate(const KeySet& keySet)
      {
         //DebugLog(<< getClassName() << "::onUpdate " << keySet);

         PSCommon::ChangeType changeType = keySet._changeType;
         const vector<string> &keys = keySet._keys;
         
         {
            
            Threads::WriteLock lock(*_updatingMutex);

            for (vector<string>::const_iterator kIt = keys.begin();
                 kIt != keys.end(); kIt++)
            {
               typename PT::KeyType key = boost::lexical_cast<typename PT::KeyType>(*kIt);

               switch (changeType)
               {
                  case PSCommon::ADD_CHANGE :
                  {
                     //DebugLog(<< "noting new key " << key);
                     _updatingExistsSet->insert(key);
                     break;
                  }
                  case PSCommon::DELETE_CHANGE :
                  {
                     //DebugLog(<< "forgetting key " << key << " " << _updatingExistsSet);
                     _updatingExistsSet->erase(key);
                     // no break -- fall through to next case
                  }
                  case PSCommon::SET_CHANGE :
                  {
                     typename CacheMap::iterator it = _updatingCache->find(key);
                     if (it != _updatingCache->end())
                     {
                        //DebugLog(<< "forgetting contents " << key);
                        delete it->second;
                        _updatingCache->erase(it);
                     }
                     break;
                  }
                  default:
                     assert(false);
               }
            }
         }

         // call application updators -- may result in hits to the Cache
         {
            Vocal::Threads::ReadLock lock(_applicationUpdatorsMutex);
            for (list<CacheUpdator*>::const_iterator i = _applicationUpdators.begin();
                 i != _applicationUpdators.end(); i++)
            {
               (**i)(keySet);
            }
         }
      }

      /// Called on each update from provisioning -- in general client thread
      void onUpdateEnqueue(const KeySet& keySet)
      {
         fifo.add(keySet);
      }

      /// final, from ThreadIf -- execute updates as soon as possible
      virtual void thread()
      {
         while (true)
         {
            try 
            {
                KeySet keySet = fifo.getNext();
               // kind of a hack -- an empty keySet means finish
               if (keySet.empty())
               {
                  break;
               }
               onUpdate(keySet);
            }
            catch (VException& e)
            {
               DebugLog(<< "Caught: " << e << "... continuing");
            }
            catch (...)
            {
               DebugLog(<< "Caught unknown exception... continuing");
            }
         }
      }

   protected:
      virtual void serviceAvailable()
      {
         // unregister in case the psever remained up
         try 
         {
            PT pt;
            try 
            {
               _pif.unregisterForUpdate(pt, _updator);
            } 
            catch (VException &e)
            {
            }
         }
         catch (VDbException &e)
         {
            DebugLog(<< "unregister failed -- new pserver");
            // OK, already deregistered -- maybe bounced pserver
         }

         DebugLog(<< "enabling cached data");
         // swap in alternate cache/exists for updates
         // but continue using cache/exists for requests
         {
            Threads::WriteLock lock(_mutex);
            Threads::WriteLock lock1(_alternateMutex);

            _updatingCache = _alternateCache;
            _updatingExistsSet = _alternateExistsSet;
            _updatingMutex = &_alternateMutex;
         }
         
         // updates will affect updating cache/exists
         registerForUpdates();
         load();

         // swap newly updated cache/exists into request service
         {
            Threads::WriteLock lock(_mutex);
            Threads::WriteLock lock1(_alternateMutex);
            _workingCache = _updatingCache;
            _workingExistsSet = _updatingExistsSet;
            _updatingMutex = &_mutex;

            swap(_cache, _alternateCache);
            swap(_existsSet, _alternateExistsSet);

            // clear out the alternate cache/exists
            for (typename CacheMap::iterator i = _alternateCache->begin();
                 i != _alternateCache->end(); i++)
            {
               delete i->second;
            }
            _alternateCache->clear();
            _alternateExistsSet->clear();
         }
      }

      virtual void
      shutdown()
      {
         // tell the fifo its done
         KeySet n;
         fifo.add(n);
         join();

         unregisterForUpdates();
      }

      // datastructures and their alternates
      CacheMap* _cache;
      ExistsSet* _existsSet;
      CacheMap* _alternateCache;
      ExistsSet* _alternateExistsSet;
      Vocal::Threads::RWMutex _mutex;
      Vocal::Threads::RWMutex _alternateMutex;

      // !dlb! problems with friendship -- not really public!!
   public:
      // indirections on the datastructures -- usually all point to the non-alternates
      CacheMap* _workingCache;
      ExistsSet* _workingExistsSet;
      CacheMap* _updatingCache;
      ExistsSet* _updatingExistsSet;
      Vocal::Threads::RWMutex* _workingMutex;
      Vocal::Threads::RWMutex* _updatingMutex;
   protected:

      Sptr<Delegator> _updator;
      ProvisioningInterface& _pif;

   private:
      Fifo<KeySet> fifo;

      Vocal::Threads::RWMutex _applicationUpdatorsMutex;
      list<CacheUpdator*> _applicationUpdators;
      
      // no value semantics
      Cache(const Cache&);
      Cache& operator=(const Cache&);
};

template<class T>
ostream&
operator<<(ostream& stream, const Cache<T>& cache)
{
   stream << cache.getClassName() << "[";
   Threads::ReadLock lock(*cache._workingMutex);
   for (typename Cache<T>::CacheMap::const_iterator cIt = cache._workingCache->begin();
        cIt != cache._workingCache->end(); cIt++)
   {
      stream << cIt->first << " -> " << *cIt->second << ", " << endl;
   }
   stream << "]";
   return stream;
}      

/// Generic delegator. Use to communicate updates to, say, a cache.
template <class D> 
class CacheUpdateDelegator : public UpdateFunctor
{
   public:
      CacheUpdateDelegator(D& delegatee)
         : _delegatee(delegatee)
      {}

      virtual string
      getClassName() const
      {
         return "[CacheUpdateDelegator]" + _delegatee.getClassName();
      }

      // final -- don't override
      virtual void
      operator()(const KeySet& keySet)
      {
         //DebugLog(<< getClassName() << " update queue " << keySet);
         _delegatee.onUpdateEnqueue(keySet);
      }

   private:
      D& _delegatee;
};

} // namespace Epoisses

#undef VOCAL_SUBSYSTEM

#endif // Cache_hxx
