/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const PSHttpXMLInterface_cxx_Version = "$Id: PSHttpXMLInterface.cxx,v 1.1 2002/09/30 04:55:14 jason Exp $";

#include "Logger.hxx"
#include "VNetworkException.hxx"
#include "PSCommon.hxx"
#include "PSHttpXMLInterface.hxx"
#include "PSProtocolService.hxx"
#include "TimeAccumulate.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Epoisses;

ProvisioningInterface&
PSHttpXMLInterface::getInstance(const char* primHost, const int primPort)
   throw (VException&)
{
   if (!myInstance)
   {

      if (primHost == 0 || primPort == 0)
      {
         string err = "PServer hostname and port not provided";
         ErrLog(<< err);
         VNetworkException exception(err, __FILE__, __LINE__);
         throw exception;
      }
      myInstance  =  new PSHttpXMLInterface(primHost, primPort);
      myInstance->setInterfaceType( PSCommon::HTTPXML );
   }
   return *myInstance;
}

///  A concrete implementation of ProvisioningInterface
PSHttpXMLInterface::PSHttpXMLInterface(const char* primHostname, const int primPort)
   : ProvisioningInterface(primHostname, primPort)
{
}

/**
   Frees the memory associated the Instance.Gets register with
   atexit(), so that when program exits, it will be called automatically.
*/
void PSHttpXMLInterface::destroy()
{
   ProvisioningInterface::destroy();
}

/// Destructor
PSHttpXMLInterface::~PSHttpXMLInterface(){}

/** retrieve record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<PSDBResult> 
PSHttpXMLInterface::getRecordsInDomain(const Domain& domain, 
                                       const Filter& filter,
                                       const FieldList& fieldList)
   throw( VException& ) 
{
   DebugLog(<< "getRecordsInDomain domain " 
            << domain.toString() 
            << " rules "<< filter.toString());

   TimeAccumulate t("PSHttpXMLInterface::getRecordsInDomain");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildGetRecordsInDomainReq(domain, filter, fieldList);

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::GET_RECORDS_IN_DOMAIN_RSP, resp ));
}

/** retrieve record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<PSDBResult> 
PSHttpXMLInterface::getRecordsInDomain(const Domain& domain, 
                                       const Filter& filter,
                                       const FieldList& fieldList,
                                       const Field<void>& orderBy)
   throw( VException& ) 
{
   DebugLog(<< "getRecordsInDomain domain " 
            << domain.toString() << ", rules "
            << filter.toString() << ", orderBy " 
            << orderBy.toString());

   TimeAccumulate t("PSHttpXMLInterface::getRecordsInDomain");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildGetRecordsInDomainReq(domain, filter, fieldList, orderBy);
   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::GET_RECORDS_IN_DOMAIN_RSP, resp ));
}

///
Sptr<PSDBResult>
PSHttpXMLInterface::getFieldsInDomain(const Domain& domain)
   throw( VException& ) 
{
   DebugLog(<< "getFieldsInDomain domain " << domain.toString());

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = proto.buildGetFieldsInDomainReq(domain);

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );

   return(proto.parseResponse( PSMessage::GET_FIELDS_IN_DOMAIN_RSP, resp ));
}


/// update existing records with keyword/value pairs provided
Sptr<PSDBResult>
PSHttpXMLInterface::setRecordsInDomain(const Domain& domain, 
                                       const Filter& filter,
                                       const AssignmentList& assignments)
   throw (VException& )
{
   DebugLog(<< "setRecordsInDomain domain "
            << domain.toString() << " rules " 
            << filter.toString() << " assignments "
            << assignments.toString());
   
   TimeAccumulate t("PSHttpXMLInterface::setRecordsInDomain");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildSetRecordsInDomainReq(domain, filter, assignments);

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );

   return(proto.parseResponse( PSMessage::SET_RECORDS_IN_DOMAIN_RSP, resp ));
}

/// insert a new record
Sptr<PSDBResult>
PSHttpXMLInterface::addRecordInDomain(const Domain& domain, 
                                      const AssignmentList& assignments)
   throw( VException& )
{
   DebugLog(<< "addRecordsInDomain domain " 
            << domain.toString() << ", assignments "
            << assignments.toString());

   TimeAccumulate t("PSHttpXMLInterface::addRecordsInDomain");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildAddRecordInDomainReq(domain, assignments)
      ;
   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::ADD_RECORD_IN_DOMAIN_RSP, resp ));
}

///
Sptr<PSDBResult>
PSHttpXMLInterface::deleteRecordsInDomain(const Domain& domain,
                                          const Filter& filter )
   throw( VException& ) 
{
   DebugLog(<< "deleteRecordsInDomain domain "
            << domain.toString() << "rules"
            <<filter.toString());

   TimeAccumulate t("PSHttpXMLInterface::deleteRecordsInDomain");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildDeleteRecordsInDomainReq(domain, filter);

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::DEL_RECORDS_IN_DOMAIN_RSP, resp ));
}


// *****************************************************************
// Session related functions
// *****************************************************************

// string sessionId sessionOpen( int timeInSeconds ) throw(VException&)
string PSHttpXMLInterface::sessionOpen( int timeInSeconds ) throw(VException&)
{
   DebugLog(<< "sessionOpen() expire time = "
            << timeInSeconds << " seconds");

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = proto.buildSessionOpenReq( timeInSeconds );

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse( PSMessage::SESSION_OPEN_RSP, resp );

   string sessionId = "";
   if (result != 0)
   {
      sessionId = result->getSessionID();
   }

   return sessionId;
}

///
bool PSHttpXMLInterface::sessionClose( const string& sessionID ) throw( VException& )
{
   DebugLog(<< "sessionClose sessionID = " << sessionID); 

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = proto.buildSessionCloseReq( sessionID );

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse( PSMessage::SESSION_CLOSE_RSP, resp );
   
   // if status code == 200 OK, return true
   if ((result !=0) && (result->getStatus() == 200))
   {
      return(true);
   }
   else
   {
      return(false);
   }
}

///
Sptr<PSDBResult>
PSHttpXMLInterface::sessionSearchDB(const string& sessionID, 
                                    const Domain& domain, 
                                    const Filter& filter,
                                    const FieldList& fieldList,  
                                    PSCommon::PSSortType sType,
                                    const BaseField& sortField ) throw(VException&)
{
   DebugLog(<< "sessionSearchDB session "
            << sessionID << ", domain "
            << domain.toString() << ", rules "
            << filter.toString());
    

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildSessionSearchDBReq(sessionID,
                                    domain,
                                    filter, fieldList, 
                                    sType, sortField );
   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::SESSION_SEARCH_DB_RSP, resp ));
}

// use the result iterator
#ifdef NOPE
///
Sptr<PSDBResult>
PSHttpXMLInterface::sessionGetDBResult(const string& sessionID, int startIdx,
                                       int endIdx, const FieldListtype& fieldList)
   throw (VException&) 
{
   DebugLog(<< "sessionGetDBResult sessionID "
            << sessionID << ", startIdx "
            << startIdx << ", endIdx "
            << endIdx);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildSessionGetDBResultReq( sessionID, startIdx, 
                                        endIdx, fieldList );

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::SESSION_GET_DB_RESULT_RSP, resp ));
}
#endif

///
Sptr<PSDBResult>
PSHttpXMLInterface::sessionSetDBResult(const string& sessionID, 
                                       const map< string, string>& fieldValPairs)
   throw (VException&)
{
   DebugLog(<< "sessionSetDBResult sessionID " << sessionID);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildSessionSetDBResultReq( sessionID, 
                                        fieldValPairs );

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::SESSION_SET_DB_RESULT_RSP, resp ));
}

///
Sptr<PSDBResult>
PSHttpXMLInterface::sessionSortDBResult(const string& sessionID, 
                                        PSCommon::PSSortType sType,
                                        const BaseField& sortField)
   throw (VException&) 
{
   DebugLog(<< "sessionSortDBResult sessionID " << sessionID);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = 
      proto.buildSessionSortDBResultReq(sessionID, sType, sortField);

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::SESSION_SORT_DB_RESULT_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionDeleteAll( const string& sessionID ) 
   throw (VException&)
{
   DebugLog(<< "sessionDeleteAll sessionID " << sessionID);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr< Http::Message > req = proto.buildSessionDeleteAllReq( sessionID );

   // exception will be thrown if bad request
   string resp = myInterface.sendRequest( req->encode() );
   
   return(proto.parseResponse( PSMessage::SESSION_DELETE_ALL_RSP, resp ));
}

// *****************************************************************
// Registration functions
// *****************************************************************

///
void
PSHttpXMLInterface::sendRegisterDomainForUpdate(const Domain& domain,
                                                const string& uid)
   throw (VException&)
{
   DebugLog (<<  "RegisterDomainForUpdate - domain = " << domain.toString());

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   static const KeySet emptyKeyList;

   Sptr<Http::Message> request = 
      proto.buildRegisterForUpdateReq(PSCommon::REGISTER_DOMAIN_FOR_UPDATE_STR,
                                      getHostPort(),
				      domain, 
				      emptyKeyList,
				      uid);

   // if thrown exception, client needs to catch it.
   string response = myInterface.sendRequest(request->encode());

   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse(PSMessage::REGISTER_DOMAIN_FOR_UPDATE_RSP, 
                                response);

   if ((result != 0) && (result->getError().size() != 0))
   {
      DebugLog (<< result->getError());
   }
}

///
void
PSHttpXMLInterface::sendRegisterRecordsForUpdate(const Domain& domain,
                                                 const KeySet& keySet, 
                                                 const string &uid)
   throw (VException&)
{

   DebugLog (<<  "RegisterRecordForUpdate - domain = " << domain.toString());

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   Sptr<Http::Message> request = 
      proto.buildRegisterForUpdateReq(PSCommon::REGISTER_RECORD_FOR_UPDATE_STR,
                                      getHostPort(),
				      domain, 
				      keySet, 
				      uid);

   // if thrown exception, client needs to catch it.
   string response = myInterface.sendRequest(request->encode());

   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse(PSMessage::REGISTER_RECORD_FOR_UPDATE_RSP, 
                                response);

   if ((result != 0) && (result->getError().size() != 0))
   {
      DebugLog (<< result->getError());
      return;
   }
}
 
///                       
void 
PSHttpXMLInterface::sendUnregisterForUpdate(const Domain &domain,
                                            const string &uid)
   throw (VException&)
{
   DebugLog (<<  "UnRegisterDomainForUpdate - domain: " << domain.getName() << ", uid: " << uid);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();

   Sptr<Http::Message> request = 
      proto.buildUnRegisterForUpdateReq(PSCommon::UN_REGISTER_DOMAIN_FOR_UPDATE_STR,
                                        getHostPort(),
                                        domain,
					uid);

   // if thrown exception, client needs to catch it.
   string response = myInterface.sendRequest(request->encode());

   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse(PSMessage::UN_REGISTER_DOMAIN_FOR_UPDATE_RSP, 
                                response);

   if ((result != 0) && (result->getError().size() != 0))
   {
      InfoLog (<< result->getError());
   }
   return;
}

void
PSHttpXMLInterface::sendRegisterDomainForForeignUpdate
(const Domain& foreignDomain,
 const string& foreignKeyField,
 const Domain& primaryDomain,
 const string& uid) throw (VException&)
{
   DebugLog (<<  "sendRegisterDomainForForeignUpdate - foreign domain = " <<  foreignDomain.toString());

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();
   static const set<string> emptyKeyList;

   Sptr<Http::Message> request = 
      proto.buildRegisterForForeignUpdateReq(PSCommon::REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_STR,
                                             getHostPort(),
                                             foreignDomain,
                                             foreignKeyField,
                                             primaryDomain,
                                             uid);

   // if thrown exception, client needs to catch it.
   string response = myInterface.sendRequest(request->encode());

   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse(PSMessage::REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_RSP, 
                                response);

   if ((result != 0) && (result->getError().size() != 0))
   {
      DebugLog (<<  result->getError());
   }
   return;
}

void
PSHttpXMLInterface::sendUnregisterForForeignUpdate
(const Domain& foreignDomain,
 const string& uid) throw (VException&)
{
   DebugLog (<<  "UnRegisterDomainForUpdate - domain: " << foreignDomain.getName() << ", uid: " << uid);

   /* let PSProtocolService helper class build the XML content
    * and the Http header(s)       
    */ 
   PSProtocolService& proto = PSProtocolServiceHolder::Instance();

   Sptr<Http::Message> request = 
      proto.buildUnRegisterForForeignUpdateReq(PSCommon::UN_REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_STR,
                                               getHostPort(),
                                               foreignDomain,
                                               uid);

   // if thrown exception, client needs to catch it.
   string response = myInterface.sendRequest(request->encode());

   Sptr<PSDBResult> result = 0;
   result = proto.parseResponse(PSMessage::UN_REGISTER_DOMAIN_FOR_FOREIGN_UPDATE_RSP, 
                                response);

   if ((result != 0) && (result->getError().size() != 0))
   {
      DebugLog (<<  result->getError());
   }
   return;
}

///
void
PSHttpXMLInterface::handleRequest(const string& request)
   throw (VException&)
{
   Http::Message httpMsg;

   // exception will be thrown if bad request
   httpMsg.decode(request);

   // For now, handle the PUT request only
   if (httpMsg.method().type() == Http::Method::PUT)
   {
      HandlePUTRequest(httpMsg);
   }
   else
   {
      DebugLog (<<  "Don't know this request: " <<  request);
   }
}

// does it really need a seperate PUT request ????
// what else it could be ??? GET/PUT/DELETE/POST
void
PSHttpXMLInterface::HandlePUTRequest(const Http::Message& httpMsg)
{
   //DebugLog (<< "PUT request: URL = " <<  httpMsg.url().url());

   PSProtocolService& protocol = PSProtocolServiceHolder::Instance();

   if (httpMsg.url().url() == PSCommon::UPDATE_DOMAIN_STR)
   {
      Sptr <PSDBRequest> request = protocol.parseRequest(PSMessage::UPDATE_DOMAIN_REQ, httpMsg.body());

      KeySet keySet(request->getChangeType(),
                    request->getFieldList());
       
      triggerDomainUpdates(request->getDomain(), keySet);
   }
   else if (httpMsg.url().url() == PSCommon::FOREIGN_UPDATE_DOMAIN_STR)
   {
      Sptr <PSDBRequest> request = protocol.parseRequest(PSMessage::FOREIGN_UPDATE_DOMAIN_REQ,
                                                         httpMsg.body());

      const string uid = request->getFunctorUId();

      KeySet keySet(request->getChangeType(),
                    request->getFieldList());

      triggerForeignDomainUpdates(uid, keySet);
   }
   else
   {
      DebugLog (<<  "Unrecognized PUT function");
   }
   return;
}
