// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#ifndef PROVISIONINGINTERFACE_HXX
#define PROVISIONINGINTERFACE_HXX

static const char* const ProvisioningInterface_hxx_Version = "$Id: ProvisioningInterface.hxx,v 1.1 2002/09/30 04:55:14 jason Exp $";

#include <memory>
#include <list>
#include <string>
#include <ext/hash_map>

#include "global.h"
#include "PSInterface.hxx"
#include "VMutex.h"

#include "Data.hxx"
#include "PSCommon.hxx"
#include "PSDBResult.hxx"
#include "PSDBCommonType.h"

#include "Domain.hxx"
#include "Field.hxx"
#include "FieldList.hxx"
#include "AssignmentList.hxx"
#include "Filter.hxx"
#include "UpdateFunctor.hxx"
#include "ForeignUpdateFunctor.hxx"
#include "StringHash.hxx"

class NetworkAddress;
class VException;

namespace Epoisses {

class PSListener;

/** Singelton Facade class - Only one instance is available per application.
    Initial access of the instance takes in a PServer Client ID to properly
    instantiate the correct  PServer Client Interface.  PRD 2.0 calls
    for a provisioning interface that sends HTTP encapsulated XML messages
    to the server.  In the future, if the requirements change to use
    SNMP or Java Spaces as the communication protocol, the API defined in 
    ProvisioningInterface will remain unchanged, but a different child
    class will be instantiated (see usage in PSClientCreator.hxx)

    Implements the communications to the provision server and notify content
    update as it happens on the Provision server.  In order to receive
    notification client has to register itself with the class with a callback
    functor with the domain and set of records (empty if all records in the
    domain are of interest) for which update notification is requested. Whenever
    the record(s) change(s) on the provision server, the callback function is
    called with the content of the item passed as the argument.

    Example of use:
    // assume a provision interface instance
    SPtr<ProvisioningInterface> pi = ...

    pi->getRecordsInDomain(Domain("user"),
    !Field<bool>("expired") &&  // filter with operators
    Field<int>("credits") > 0,
    (Field<int>("userid"),        // list of fields
    Field<string>("phoneNumber")));

    pi->setRecordsInDomain(Domain("user"),
    !Field<bool>("expired") &&
    Field<int>("credits") <= 0,
    (Field<bool>("updated") = true,   // list of assignments
    Field<int>("credits") = 10000));

    // Note: the like operator is *=
    // *= has lower precedence than && and ||, so parentheses are required:
    pi->getRecordsInDomain(Domain("user"),
    !Field<bool>("expired") &&
    (Field<string>("phoneNumber") *= "650*"),
    Field<string>("phoneNumber"));
*/

class ProvisioningInterface
{
   public:
      static ProvisioningInterface& make(const char* host=0, const int port=0);

      /**
         Frees the memory associated the Instance.Gets register with
         atexit(), so that when program exits, it will be called automatically.
      */
      static void destroy();

      /// Destructor
      virtual ~ProvisioningInterface();

      virtual const char* getClassName() const { 
         return "ProvisioningInterface"; 
      }

      ///
      PSCommon::PSIFType getInterfaceType() const { return myIFType; }

      ///
      void setInterfaceType(PSCommon::PSIFType typ) 
      { 
         myIFType = typ;
      }

      /// get the listening port
      int getListenPort() const;

      /// retrieve record(s) matching the domain, filter and fields of interest
      virtual Sptr<PSDBResult> 
      getRecordsInDomain(const Domain& domain, 
                         const Filter& filter,
                         const FieldList& fieldList) 
         throw( VException& ) =0;

      virtual Sptr<PSDBResult> 
      getRecordsInDomain(const Domain& domain, 
                         const Filter& filter,
                         const FieldList& fieldList,
                         const Field<void>& orderBy)
         throw( VException& ) =0;

      /// Return the list of field names for the domain
      virtual  Sptr<PSDBResult> 
      getFieldsInDomain( const Domain& domain )
         throw( VException& ) = 0;

      /// update existing records with keyword/value pairs provided
      virtual Sptr<PSDBResult>  
      setRecordsInDomain(const Domain& domain, 
                         const Filter& filter,
                         const AssignmentList& assignments)
         throw (VException& ) = 0;

      /// insert a new record
      virtual Sptr<PSDBResult>
      addRecordInDomain(const Domain& domain, 
                        const AssignmentList& assignments)
         throw( VException& ) = 0;

      /// delete the selected records
      virtual Sptr<PSDBResult> 
      deleteRecordsInDomain(const Domain& domain,
                            const Filter& filter )
         throw( VException& ) = 0;
      
      // *****************************************************************
      // Session related functions
      // *****************************************************************

      /** psServer returns -1 if PServer does not return
       *  otherwise, return value is the sessionID
       */
      virtual string sessionOpen( int timeInSeconds ) throw(VException&) = 0; 

      ///
      virtual bool sessionClose( const string& sessionID ) throw( VException& ) = 0;

      ///
      virtual Sptr<PSDBResult> sessionDeleteAll( const string& sessionID ) 
         throw (VException&) = 0;

      // dlb -- new interface
   public:
      /// Search and sort
      virtual Sptr<PSDBResult>  
      sessionSearchDB(const string& sessionID, 
                      const Domain& domain,
                      const Filter& filter,
                      const FieldList& fieldList, 
                      PSCommon::PSSortType sType,
                      const BaseField& sortField ) throw(VException&) = 0;

      /// Sort existing result
      virtual Sptr<PSDBResult> 
      sessionSortDBResult(const string& sessionID, 
                          PSCommon::PSSortType sType,
                          const BaseField& sortField)
         throw (VException&) = 0;

      // should this be public? !dlb! who calls this?
      virtual Sptr<PSDBResult>
      sessionSetDBResult(const string& sessionID, 
                         const map< string, string>& fieldValPairs)
         throw (VException&) = 0;

      // *****************************************************************
      // Registration functions
      // *****************************************************************

      /**
         Register client for notification for a given domain.
         update functor gets called whenever item is updated.
      */
      void
      registerDomainForUpdate(const Domain& domain,
                              Sptr<UpdateFunctor> func) throw (VException&);
      
      void
      unregisterForUpdate(const Domain& domain,
                          Sptr<UpdateFunctor> fun) throw (VException&);

      /**
         Register client for updates to primaryDomain when the corresponding rows
         of foreignDomain change. foreignKeyField is the name of the column
         mapping from foreign to primary domain. */
      void
      registerDomainForForeignUpdate(const Domain& foreignDomain,
                                     const Field<void>& foreignKeyField,
                                     const Domain& primaryDomain,
                                     Sptr<ForeignUpdateFunctor> func);

      void
      unregisterDomainForForeignUpdate(const Domain& foreignDomain,
                                       Sptr<ForeignUpdateFunctor> func);


      /// Get the network address of the currently used pServer
      const NetworkAddress getNetAddress()
      {
         return myInterface.getNetAddress();
      }

      /// return <host>:<port> !dlb! bleah
      string getHostPort();
      
   protected:
      
      virtual void
      sendRegisterDomainForUpdate(const Domain& domain,
                                  const string &uid) throw (VException&) = 0;

      virtual void
      sendUnregisterForUpdate(const Domain &domain,
                              const string &uid) throw (VException&) = 0;

      virtual void
      sendRegisterDomainForForeignUpdate(const Domain& foreignDomain,
                                         const string& foreignKeyField,
                                         const Domain& primaryDomain,
                                         const string& uid) throw (VException&) = 0;

      virtual void
      sendUnregisterForForeignUpdate(const Domain& foreignDomain,
                                     const string& uid) throw (VException&) = 0;

      // sets up local bookkeeping
      void
      registerUpdate(const Domain &domain,
                     const KeySet& keySet, 
                     Sptr<UpdateFunctor> func);

      void
      unregisterUpdate(const Domain &domain,
                       Sptr<UpdateFunctor> fun);
                                            
      // sets up local bookkeeping
      void
      registerForeignUpdate(Sptr<ForeignUpdateFunctor> func);

      void
      unregisterForeignUpdate(Sptr<ForeignUpdateFunctor> fun);

      friend class PSIFCreator;

      /// pointer to the singleton instance
      static ProvisioningInterface* myInstance;

      /// interface to the PS
      PSInterface myInterface;

      /**
         Constructor initializes with the name of provision server
         for redundancy passed as command line argument. Intentionally
         made private so that it can not be instantiated directly.
         It also creates an instance of listener in a separate thread
         to listen for data updates (through the interface)
      */
      ProvisioningInterface(const char* primHostname, const int primPort);

      ///
      virtual void handleRequest(const string& request) = 0;

      void
      triggerDomainUpdates(const string &domain,
                           const KeySet &keySet);

      void
      triggerForeignDomainUpdates(const string& uid,
                                  const KeySet &keySet);

   private:
      /// suppress equal operator
      bool operator==( const ProvisioningInterface& rhs );
      /// suppress copy contructor
      ProvisioningInterface( const ProvisioningInterface& rhs );
      /// suppress assignment operator
      ProvisioningInterface& operator=( const ProvisioningInterface& rhs );

      /// Provisioning Interface type
      PSCommon::PSIFType myIFType;

      /// Host name the library is running on
      string myHostName;

      /// make mySessionMap thread safe
      VMutex mySessionMutex;

      friend class PSListener;
      friend class ListenerClientHandler;

      /// listener for updates from server
      PSListener* myListener;
      /// host:port string used to setup call back functors
      string hostPort;

      struct KeySetUpdate
      {
         public:
            KeySetUpdate(const KeySet &_keySet,
                         Sptr<UpdateFunctor> _fun)
               : keySet(_keySet),
                 fun(_fun)
            {}
               
            bool
            operator==(const string &uid)
            {
               return fun->getUId() == uid;
            }

            KeySet keySet;
            Sptr<UpdateFunctor> fun;
      };

      typedef Sptr<ForeignUpdateFunctor> ForeignUpdate;

      // helper method for trying update functors
      void attemptUpdate(Sptr<UpdateFunctor> fun,
                         const KeySet &keys);

      // helper method for trying update functors
      void attemptForeignUpdate(Sptr<ForeignUpdateFunctor> fun,
                                const KeySet &keys);

      typedef hash_map <string, list<KeySetUpdate> > UpdatesByDomain; 
      typedef hash_map <string, ForeignUpdate > ForeignUpdatesByUId; 

      // Map from domain to keylists and update functor
      UpdatesByDomain updatesByDomain;
      // Map from domain to foreign update functor
      ForeignUpdatesByUId foreignUpdatesByUId;

      // mutex to keep above thread safe
      Vocal::Threads::Mutex updatesMutex;

      void dumpRegistration();
      void dumpForeignRegistration();
};

} // namespace Epoisses

#endif // PROVISIONINGINTERFACE_HXX

