// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#ifndef RelationCache_hxx
#define RelationCache_hxx

#include "VectorCache.hxx"
#include "Domain.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

namespace Epoisses {

/**
   This cache maintains a one-to-many relationship between two domains mediated
   by a relationship domain where the relationship itself contains data.

   key -- rel --> set
                   +
                 data

  SETDATA::Db must be a Domain
  KEY must be a Domain
  RELDATA::Db must be a Domain
 */
template<class SETDATA, class KEY, class RELDATA>
class RelationCache : public VectorCache<RELDATA, KEY>
{
      //friend ostream& operator<<<>(ostream&, const RelationCache&);
   public:
      RelationCache(int reserve = 100)
         : VectorCache<RELDATA, KEY>(reserve),
           _cache(SingletonHolder< CacheImpl<SETDATA> >::Instance())
      {}

      virtual string
      getClassName() const
      {
         return "RelationCache<" + SETDATA::className() + ", " + KEY::className() + "," + RELDATA::className() + ">";
      }

      bool empty()
      {
         return VectorCache<RELDATA, KEY>::empty();
      }
      
      unsigned int size() const
      {
         return VectorCache<RELDATA, KEY>::size();
      }

      bool exists(const typename KEY::KeyType& keyObj) throw (VDbException&)
      {
         return _cache.exists(keyObj);
      }

      /// Return a snapshot of the value set
      vector<SETDATA>
      operator[](const typename KEY::KeyType& keyObj)
         throw (NotFoundInProvisioningException&, VDbException&)
      {
         return lookup(keyObj);
      }
      
      vector<SETDATA>
      lookup(const typename KEY::KeyType& keyObj) 
         throw (NotFoundInProvisioningException&, VDbException&)      
      {
         vector<SETDATA> sv;

         vector<RELDATA> rels = VectorCache<RELDATA, KEY>::lookup(keyObj);
         sv.reserve(rels.size());

         // build up vector of pairs
         for (typename vector<RELDATA>::const_iterator i = rels.begin();
              i != rels.end(); i++)
         {
            try 
            {
               // requires typename KEY::KeyType RELDATA::getForeignKey() const
               sv.push_back(_cache.get(i->getForeignKey()));
            }
            catch (NotFoundInProvisioningException &e)
            {
               DebugLog(<< "Inconsistency in " << getClassName() << *i << " missing");
            }
         }

         static DataCompare<SETDATA> comparitor;
         sort(sv.begin(), sv.end(), comparitor);
         return sv;
      }

   private:
      CacheImpl<SETDATA>& _cache;
};

/// Output a RelationCache
template<class SETDATA, class KEY, class RELDATA>
ostream& operator<<(ostream& stream, const RelationCache<SETDATA, KEY, RELDATA>& cache)
{
   Threads::ReadLock lock(*cache._workingMutex);
   Threads::ReadLock lock1(*cache._dataCache._workingMutex);

   stream << cache.getClassName() << endl;
   for (typename VectorKeyCache<KEY, typename SETDATA::Db>::CacheMap::const_iterator cIt = cache._workingCache->begin();
        cIt != cache._workingCache->end(); cIt++)
   {
      stream << cIt->first << " -> {";
      for (typename vector<typename SETDATA::Db::KeyType>::const_iterator i = cIt->second.begin();
           i != cIt->second.end(); i++)
      {
         typename Cache<SETDATA>::CacheMap::const_iterator j;
         j = cache._dataCache._workingCache->find(*i);
         if (j != cache._dataCache._workingCache->end())
         {
            stream << "<";
            stream << cache._cache[j->second->getForeignKey()];
            stream << ", ";
            stream << *j->second;
            stream << ">";
         }
         if (i+1 != cIt->second.end())
         {
            stream << ", ";
         }
      }
      stream << "}" << endl;
   }
   return stream;
}

} // namespace Epoisses

#undef VOCAL_SUBSYSTEM

#endif // RelationCache_hxx
