// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#ifndef VectorCache_hxx
#define VectorCache_hxx

#include "VectorKeyCache.hxx"
#include "ForeignUpdateFunctor.hxx"
#include "Domain.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

namespace Epoisses 
{

/**
   The third template argument allows the singleton cache for SETDATA to be
   used over several relationships. The relationship is determined by the
   foreign key of SET. While SET is typically SETDATA::Db, it maye be a variant
   of that domain with another field specified as foreign.

   SETDATA::Db must be a Domain
   KEY must be a Domain
   SET must be a Domain
 */
template<class SETDATA, class KEY, class SET = typename SETDATA::Db>
class VectorCache : public VectorKeyCache<KEY, SET>
{
      // !dlb! compiler choked instantiating
      //friend ostream& operator<<<>(ostream&, const VectorCache&);
   public:
      VectorCache(int reserve = 100)
         :  VectorKeyCache<KEY, SET>(reserve),
            _dataCache(SingletonHolder< CacheImpl<SETDATA> >::Instance())
      {}

      virtual ~VectorCache()
      {
         InfoLog(<< "You have used " << getClassName());
      }
      
      void registerCacheUpdate(CacheUpdator *updator)
      {
         VectorKeyCache<KEY, SET>::registerCacheUpdate(updator);
      }

      void unregisterCacheUpdate(CacheUpdator *updator)
      {
         VectorKeyCache<KEY, SET>::unregisterCacheUpdate(updator);
      }
      
      virtual string
      getClassName() const
      {
         return "VectorCache<" + SETDATA::className() + ", " + KEY::className() + "," + SET::className() + ">";
      }

      bool empty()
      {
         return VectorKeyCache<KEY, SET>::empty();
      }
      
      unsigned int size() const
      {
         return VectorKeyCache<KEY, SET>::size();
      }

      bool exists(const typename KEY::KeyType& keyObj) throw (VDbException&)
      {
         return VectorKeyCache<KEY, SET>::exists(keyObj);
      }

      /// Return a snapshot of the value set
      vector<SETDATA> operator[](const typename KEY::KeyType& keyObj) throw (NotFoundInProvisioningException&, VDbException&)
      {
         return lookup(keyObj);
      }
      
      vector<SETDATA> lookup(const typename KEY::KeyType& keyObj) throw (NotFoundInProvisioningException&, VDbException&)      
      {
         // retrieve set of ids from VectorKeyCache
         vector<typename SET::KeyType> ids = 
            // call parent [] operator
            VectorKeyCache<KEY, SET>::operator[](keyObj);
               
         typename Cache<SETDATA>::IdentityFunctor func;
         vector<SETDATA> vdata;
         vdata.reserve(ids.size());
         for (typename vector<typename SET::KeyType>::const_iterator i = ids.begin();
              i != ids.end(); i++)
         {
            try
            {
               _dataCache.applyRead(*i, func);
               vdata.push_back(func._data);
            }
            catch (NotFoundInProvisioningException &e)
            {
               DebugLog(<< getClassName() << ": inconsistent cache state: " << *i << " missing; continuing");
            }
         }

         static DataCompare<SETDATA> comparitor;
         sort(vdata.begin(), vdata.end(), comparitor);
         return vdata;
      }
      
   public:
      // use the singleton
      CacheImpl<SETDATA>& _dataCache;
};

template <class SETDATA, class KEY, class SET>
std::ostream& operator<<(std::ostream& stream, const VectorCache<SETDATA, KEY, SET> &cache)
{
   Threads::ReadLock lock(*cache._workingMutex);
   Threads::ReadLock lock1(*cache._dataCache._workingMutex);

   stream << cache.getClassName() << endl;
   for (typename VectorKeyCache<KEY, SET>::CacheMap::const_iterator cIt = cache._workingCache->begin();
        cIt != cache._workingCache->end(); cIt++)
   {
      stream << cIt->first << " -> {";
      for (typename vector<typename SET::KeyType>::const_iterator i = cIt->second.begin();
           i != cIt->second.end(); i++)
      {
         typename Cache<SETDATA>::CacheMap::const_iterator j;
         j = cache._dataCache._workingCache->find(*i);
         if (j != cache._dataCache._workingCache->end())
         {
            stream << *j->second;
         }
         if (i+1 != cIt->second.end())
         {
            stream << ", ";
         }
      }
      stream << "}" << endl;
   }
   return stream;
}

} // namespace Epoisses

#undef VOCAL_SUBSYSTEM

#endif // VectorCache_hxx

