// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#ifndef VectorKeyCache_hxx
#define VectorKeyCache_hxx

#include "Cache.hxx"
#include "ForeignUpdateFunctor.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::PROVC

using namespace Vocal;

namespace Epoisses {

template <class D> 
class VectorKeyCacheUpdateDelegator;

template <class D> 
class VectorKeyCacheForeignUpdateDelegator;

/// specialized keyset to handle both kinds of updates
/// update type indicated in added boolean
class RelKeySet : public KeySet
{
   public:
      RelKeySet(bool isPrimary,
              PSCommon::ChangeType changeType,
              const vector<string> &keys)
         : KeySet(changeType, keys),
           _isPrimary(isPrimary)
      {}

      RelKeySet()
         : KeySet()
      {}

      const bool _isPrimary;
};

/**
   Generic cache to maintain a one (KEY) to many (SET) relationship. This cache
   maintains only the set of identifiers, not the values of the many domain of
   the relation.

   The sole accessor of this cache, operator[], returns a copy of the vector of
   the identifiers of the many domain.

   Typically, VectorKeyCache is used in composition with a Cache of the many
   domain. The cached keys are directed through the Cache to retrieve the latest
   values. See VectorCache.hxx

   KEY must be a Domain
   SET must be a Domain
*/
template<class KEY, class SET>
class VectorKeyCache : public CacheBase, private ThreadIf
{
      //friend ostream& operator<<<>(ostream&, const VectorKeyCache&);
   public:
      typedef hash_map<typename KEY::KeyType, vector<typename SET::KeyType>, hash<typename KEY::KeyType> > CacheMap;
      typedef hash_set<typename KEY::KeyType, hash<typename KEY::KeyType> > ExistsSet;
      typedef VectorKeyCacheUpdateDelegator<VectorKeyCache> Delegator;
      typedef VectorKeyCacheForeignUpdateDelegator<VectorKeyCache> ForeignDelegator;

      VectorKeyCache(unsigned int reserve)
         : _cache(new CacheMap(reserve)),
           _existsSet(new ExistsSet(reserve)),
           _alternateCache(new CacheMap(0)),
           _alternateExistsSet(new ExistsSet(0)),
           _mutex(),
           _alternateMutex(),

           // working and updating point to the same data until the pserver
           // comes back after being lost
           _workingCache(_cache),
           _workingExistsSet(_existsSet),
           _updatingCache(_cache),
           _updatingExistsSet(_existsSet),
           _workingMutex(&_mutex),
           _updatingMutex(&_mutex),

           updator(Sptr<UpdateFunctor>(new Delegator(*this))),
           foreignUpdator(Sptr<ForeignUpdateFunctor>(new ForeignDelegator(*this))),
           pif(ProvisioningInterface::make())
      {
         registerForUpdates();
         run();

         try
         {
            load();
         }
         catch (VDbException& e)
         {
            shutdown();
            throw;
         }
      }

      virtual ~VectorKeyCache()
      {
         shutdown();
      }

      void registerCacheUpdate(CacheUpdator *updator)
      {
         Vocal::Threads::WriteLock lock(_applicationUpdatorsMutex);
         _applicationUpdators.push_back(updator);
      }

      void unregisterCacheUpdate(CacheUpdator *updator)
      {
         Vocal::Threads::WriteLock lock(_applicationUpdatorsMutex);
         _applicationUpdators.remove(updator);
      }
      
      virtual string
      getClassName() const
      {
         return "VectorKeyCache<" + KEY::className() + ", " + SET::className() + ">";
      }

      vector<typename KEY::KeyType>
      getKeys() const
      {
         vector<typename KEY::KeyType> keys;
         keys.reserve(size());

         Threads::ReadLock lock(*_workingMutex);
         for (typename ExistsSet::const_iterator i = _workingExistsSet->begin();
              i != _workingExistsSet->end(); i++)
         {
            keys.push_back(*i);
         }
         return keys;
      }

      bool empty() const
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingExistsSet->empty();
      }
      
      unsigned int size() const
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingExistsSet->size();
      }

      unsigned int cachedSize() const
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingCache->size();
      }

      /// Does the primary key exist? (may not have any foreign keys)
      bool exists(const typename KEY::KeyType& keyObj)
      {
         Threads::ReadLock lock(*_workingMutex);
         return _workingExistsSet->find(keyObj) != _workingExistsSet->end();
      }

      vector<typename SET::KeyType>
      operator[](const typename KEY::KeyType& keyObj)
      {
         {
            Threads::ReadLock lock(*_workingMutex);
            // look for cached
            typename CacheMap::const_iterator it = _workingCache->find(keyObj);
            if (it != _workingCache->end())
            {
               // return a copy of the stored vector of keys
               return it->second;
            }

            // check for pserver and existence
            if (!SingletonHolder<CacheManager>::Instance().isServiceAvailable())
            {
               DebugLog (<< "Can't find " << keyObj << " in cache and pserver is down in " << getClassName());
               throw NotFoundInProvisioningException("pserver is down",
                                                     __FILE__,
                                                     __LINE__);
            }
            if ( _workingExistsSet->find(keyObj) == _workingExistsSet->end())
            {
               DebugLog (<< "Can't find " << keyObj << " in exists " << getClassName());
               throw NotFoundInProvisioningException("Not in provisioning (exists)",
                                                     __FILE__,
                                                     __LINE__);
            }
         }

         SET set;
         Sptr<PSDBResult> result;
         
         try
         {
            result = pif.getRecordsInDomain(set,
                                            set.getForeignKeyField() == keyObj,
                                            set.getKeyField());
         }
         catch (VNetworkException& e)
         {
            DebugLog (<< "Caught: " << e);
            DebugLog (<< getClassName() << " : " 
                      << " - pserver not available on getRecordsInDomain for " 
                      << keyObj);
            throw NotFoundInProvisioningException("Pserver not available on get",
                                                  __FILE__,
                                                  __LINE__);
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
         }
         
         Threads::WriteLock lock(*_workingMutex);
         // create empty
         vector<typename SET::KeyType>& storeVector = (*_workingCache)[keyObj];
         storeVector.reserve(result->size());
         // bind all matches
         for (PSDBResult::BindIterator bind = result->begin(set.getKeyField());
              bind != result->end(); bind++) 
         {
            // add to the storage vector
            storeVector.push_back(*set.getKeyField());
         } 
         storeVector.resize(result->size());

         // return a copy of the stored keys
         return storeVector;
      }

   protected:
      /// final, from ThreadIf -- execute updates as soon as possible
      virtual void thread()
      {
         while (true)
         {
            try 
            {
               
               RelKeySet setKeys = _fifo.getNext();
               // kind of a hack -- an empty keyset means finish
               if (setKeys.empty()) 
               {
                  break;
               }
               onUpdate(setKeys);
            }
            catch (VException& e)
            {
               DebugLog(<< "Caught: " << e << " " << getClassName() << "...continuing");
            }
            catch (...)
            {
               DebugLog(<< "Caught unknown exception..." << " " << getClassName() << "...continuing");
            }
         }
      }

   private:
      void registerForUpdates()
      {
         try
         {
            KEY key;
            pif.registerDomainForUpdate(key,
                                        updator);
         
            SET set;
            pif.registerDomainForForeignUpdate(set, 
                                               set.getForeignKeyField(),
                                               key,
                                               foreignUpdator);
         }
         catch (VNetworkException& e)
         {
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
         }
      }

      void unregisterForUpdates()
      {
         try
         {
            SET set;
            KEY key;
            try
            {
               pif.unregisterForUpdate(key, updator);
            }
            catch (VException& e)
            {
               // need to do the local part of the foreign unregister even if
               // the remote part of the above unregister failed
            }
            pif.unregisterDomainForForeignUpdate(set, foreignUpdator);
         }
         catch (VNetworkException& e)
         {
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
         }
      }

      /// Called on each update from provisioning -- in cache's thread
      /// Called on updates from the foreign keyed (SET) domain
      virtual void
      onUpdate(const RelKeySet& keySet)
      {
         const vector<string> &keys = keySet._keys;
         
         {
            Threads::WriteLock lock(*_updatingMutex);

            if (keySet._isPrimary)
            {
               for (vector<string>::const_iterator kIt = keys.begin();
                    kIt != keys.end(); kIt++)
               {
                  typename KEY::KeyType key = boost::lexical_cast<typename KEY::KeyType>(*kIt);
                  switch (keySet._changeType)
                  {
                     case PSCommon::ADD_CHANGE :
                        //DebugLog(<< "new primary: " << *kIt);
                        _updatingExistsSet->insert(key);
                        break;
                     case PSCommon::DELETE_CHANGE :
                        //DebugLog(<< "forget primary: " << *kIt);
                        _updatingExistsSet->erase(key);
                        _updatingCache->erase(key);
                        break;
                     case PSCommon::SET_CHANGE :
                        //DebugLog(<< "neglect primary mod: " << *kIt);
                        break;
                     default: 
                        assert(false);
                  }
               }
            }
            else
            {
               // changes to foreign reference rows don't matter (assuming the foreign
               // key is not changed!)
               assert(PSCommon::SET_CHANGE != keySet._changeType);

               for (vector<string>::const_iterator kIt = keys.begin();
                    kIt != keys.end(); kIt++)
               {
                  // The key of the foreign reference is not passed, so all we can do is
                  // clear the cache -- could do better with more info from pserver
                  typename CacheMap::iterator it = _updatingCache->find(boost::lexical_cast<typename KEY::KeyType>(*kIt));
                  if (it != _updatingCache->end())
                  {
                     //DebugLog(<< "foreign add/del -- forget primary: " << *kIt);
                     _updatingCache->erase(it);
                  }
               }
            }
         }
      
         {
            Vocal::Threads::ReadLock lock(_applicationUpdatorsMutex);
            for (list<CacheUpdator*>::const_iterator i = _applicationUpdators.begin();
                 i != _applicationUpdators.end(); i++)
            {
               (**i)(keySet);
            }
         }
      }

      friend class Delegator;
      friend class ForeignDelegator;
      
      /// Called on each update from provisioning -- in general client thread
      void onUpdateEnqueue(const RelKeySet& keySet)
      {
         _fifo.add(keySet);
      }

      /// Load the cache with all data from provisioning
      void load()
      {
         Sptr<PSDBResult> result;
         Sptr<PSDBResult> setResult;

         KEY primary;
         FieldList keys = (primary.getKeyField());

         SET set;
         FieldList setKeys = (set.getKeyField(),
                              set.getForeignKeyField());

         try
         {
            // initialize the exists set
            result = pif.getRecordsInDomain(primary,
                                            Filter::None,
                                            keys);

            setResult = pif.getRecordsInDomain(set,
                                               Filter::None,
                                               setKeys);
         }
         catch (VNetworkException& e)
         {
            SingletonHolder<CacheManager>::Instance().serviceUnavailable();
            return;
         }

         // Note: lock aggressively to queue updates until after the load
         Threads::WriteLock lock(*_updatingMutex);

         // intialize the relationship vectors
         for (PSDBResult::BindIterator bind = result->begin(keys);
              bind != result->end(); bind++) 
         {
            _updatingExistsSet->insert(*primary.getKeyField());
         } 

         // add to the vector by one's primary key
         for (PSDBResult::BindIterator setBind = setResult->begin(setKeys);
              setBind != setResult->end(); setBind++) 
         {
            vector<typename SET::KeyType>& storeVector = (*_updatingCache)[*set.getForeignKeyField()];
            storeVector.push_back(*set.getKeyField());
         } 
      }

   protected:
      virtual void serviceAvailable()
      {
         // unregister in case the psever remained up
         try 
         {
            SET set;
            KEY key;
            try
            {
               pif.unregisterForUpdate(key, updator);
            }
            catch (VException& e)
            {
               // need to do the local part of the foreign unregister even if
               // the remote part of the above unregister failed
            }
            pif.unregisterDomainForForeignUpdate(set, foreignUpdator);
         }
         catch (VDbException &e)
         {
            // OK, already deregistered -- maybe bounced pserver
         }

         // swap in alternate cache/exists for updates
         // but continue using cache/exists for requests
         {
            Threads::WriteLock lock(_mutex);
            Threads::WriteLock lock1(_alternateMutex);

            _updatingCache = _alternateCache;
            _updatingExistsSet = _alternateExistsSet;
            _updatingMutex = &_alternateMutex;
         }
         
         // updates will affect updating cache/exists
         registerForUpdates();
         load();

         // swap newly updated cache/exists into request service
         {
            Threads::WriteLock lock(_mutex);
            Threads::WriteLock lock1(_alternateMutex);
            _workingCache = _updatingCache;
            _workingExistsSet = _updatingExistsSet;
            _updatingMutex = &_mutex;

            swap(_cache, _alternateCache);
            swap(_existsSet, _alternateExistsSet);

            // clear out the alternate cache/exists
            _alternateCache->clear();
            _alternateExistsSet->clear();
         }
      }

      virtual void
      shutdown()
      {
         // tell the fifo its done
         RelKeySet n;
         _fifo.add(n);
         join();

         unregisterForUpdates();
      }

      // !dlb! not really public -- template friend hell
   public:
      CacheMap* _cache;
      ExistsSet* _existsSet;
      CacheMap* _alternateCache;
      ExistsSet* _alternateExistsSet;
      Vocal::Threads::RWMutex _mutex;
      Vocal::Threads::RWMutex _alternateMutex;

      // indirections on the datastructures -- usually point to the
      // non-alternates
      CacheMap* _workingCache;
      ExistsSet* _workingExistsSet;
      CacheMap* _updatingCache;
      ExistsSet* _updatingExistsSet;
      Vocal::Threads::RWMutex* _workingMutex;
      Vocal::Threads::RWMutex* _updatingMutex;

   private:
      Fifo<RelKeySet> _fifo;

      Vocal::Threads::RWMutex _applicationUpdatorsMutex;
      list<CacheUpdator*> _applicationUpdators;

      Sptr<UpdateFunctor> updator;
      Sptr<ForeignUpdateFunctor> foreignUpdator;
      ProvisioningInterface& pif;

      VectorKeyCache(const VectorKeyCache&);
      VectorKeyCache& operator=(const VectorKeyCache&);
};

template <class D> 
class VectorKeyCacheUpdateDelegator : public UpdateFunctor
{
   public:
      VectorKeyCacheUpdateDelegator(D& delegatee)
         : _delegatee(delegatee)
      {}

      virtual string
      getClassName() const
      {
         return "VectorKeyCacheUpdateDelegator<" + _delegatee.getClassName() + ">";
      }

      // final -- don't override
      virtual void
      operator()(const KeySet& keySet) 
      {
         //DebugLog(<< getClassName() << " update queue keySet");
         // indicate that this update is for the primary
         RelKeySet rkeySet(true,
                         keySet._changeType,
                         keySet._keys);
         _delegatee.onUpdateEnqueue(rkeySet);
      }

   private:
      D& _delegatee;
};

template <class D> 
class VectorKeyCacheForeignUpdateDelegator : public ForeignUpdateFunctor
{
   public:
      VectorKeyCacheForeignUpdateDelegator(D& delegatee)
         : _delegatee(delegatee)
      {}

      virtual string
      getClassName() const
      {
         return "VectorKeyCacheForeignUpdateDelegator<" + _delegatee.getClassName() + ">";
      }

      // final -- don't override
      virtual void
      operator()(const KeySet& keySet) 
      {
         //DebugLog(<< getClassName() << " update queue keySet");
         // indicate that this update is for the foreign
         RelKeySet rkeySet(false,
                           keySet._changeType,
                           keySet._keys);
         _delegatee.onUpdateEnqueue(rkeySet);
      }

   private:
      D& _delegatee;
};

template<class KEY, class SET>
ostream&
operator<<(ostream& stream, const VectorKeyCache<KEY, SET>& cache)
{
   stream << cache.getClassName() << "[";
   Threads::ReadLock lock(*cache._workingMutex);
   for (typename VectorKeyCache<KEY, SET>::CacheMap::const_iterator cIt = cache._cache->begin();
        cIt != cache._cache->end(); cIt++)
   {
      // key of the one domain instance
      stream << cIt->first << " -> {";
      for (typename vector<typename SET::KeyType>::const_iterator vfIt = cIt->second.begin();
           vfIt != cIt->second.end(); vfIt++)
      {
         // key of the many domain instance
         stream << *vfIt << ", ";
      }
      stream << "}" << endl;
   }
   stream << "]";
   return stream;
}      

} // namespace Epoisses

#undef VOCAL_SUBSYSTEM

#endif // VectorKeyCache_hxx
