// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#include <boost/lexical_cast.hpp>

#include "Domain.hxx"
#include "Cache.hxx"

using namespace Epoisses;

class User
{
   public:
      class Db : public Domain
      {
         public:
            /// Base provisioning interface class must have KeyType defined
            typedef string KeyType;

            /// Every provisioning interface class must define an explicit constructor
            /// of one optional string argument.
            explicit Db(string domainName = "testusers")
               : Domain(domainName),
                 userid("userid"),
                 firstname("firstname"),
                 lastname("lastname"),
                 expired("expired"),
                 count("count"),
                 credits("credits"),
                 speed("speed"),
                 aliases("aliases")
            {}

            /// Base provisioning interface class must define getKeyField method.
            Field<KeyType> getKeyField() const
            {
               return userid;
            }

            /// Every provisioning interface class must define all method.
            virtual FieldList all()
            {
               // don't use static -- could be threaded in binder
               FieldList list = ((userid,
                                  firstname,
                                  lastname,
                                  expired,
                                  count,
                                  credits,
                                  speed,
                                  aliases));
               return list;
            }

            /// Every provisioning interface should define className (for debug).
            static string
            className()
            {
               return "User::Db";
            }

            /// final -- virtual displatcher of className
            virtual string
            getClassName()
            {
               return className();
            }

            /// Base provisioning interface class must define a primary key Field.
            Field<KeyType> userid;
            // remaining Fileds correspond to domain columns
            Field<string> firstname;
            Field<string> lastname;
            Field<bool> expired;
            Field<int> count;
            Field<int> credits;
            Field<double> speed;
            Field< vector<string> > aliases;
      };

      explicit User(const Db& users)
         : userid(*users.userid),
           firstname(*users.firstname),
           lastname(*users.lastname),
           expired(*users.expired),
           count(*users.count),
           credits(*users.credits),
           speed(*users.speed),
           extra(0),
           extra1(0)
      {}

      /// Required for vector<T>
      User() {}

      /// Every provisioning interface should define className (for debug).
      static string
      className()
      {
         return "User";
      }

      string userid;
      string firstname;
      string lastname;
      bool expired;
      int count;
      int credits;
      double speed;
      
      int extra;
      int extra1;

      // easier to use than global << because of nested/forward
      string
      toString() const
      {
         return ("[" + userid + ", " +
                 firstname + ", " +
                 lastname + ", " +
                 boost::lexical_cast<string>(expired) + ", " +
                 boost::lexical_cast<string>(count) + ", " +
                 boost::lexical_cast<string>(credits) + ", " +
                 boost::lexical_cast<string>(speed) + ", " +
                 boost::lexical_cast<string>(extra) + ", " +
                 boost::lexical_cast<string>(extra1) + "]"
            );
      }
};

namespace Epoisses
{
class CacheImpl<User> : public Cache<User>
{
   public:
      User operator[](const string& userid) throw (NotFoundInProvisioningException&, VDbException&)
      {
         IdentityFunctor getUser;
         applyRead(userid, getUser);
         return getUser.get();
      }
};
}

typedef SingletonHolder< CacheImpl<User> > UserCache;

inline
ostream&
operator<<(ostream& stream, const User& user)
{
   stream << "user["
          << user.userid << ", "
          << user.userid << ", "
          << user.firstname << ", "
          << user.lastname << ", "
          << user.expired << ", "
          << user.count << ", "
          << user.credits << ", "
          << user.speed << ", "
          << user.extra
          << "]";
   return stream;
}

class SpecUser : public User
{
   public:
      class Db : public User::Db
      {
         public:
            explicit Db(string domainName = "testspecusers")
               : User::Db(domainName),
                 permission("permission")
            {}

            virtual FieldList all()
            {
               return User::Db::all().add(permission);
            }

            static string
            className()
            {
               return "SpecUser::Db";
            }

            // field in addition to inherited fields
            Field<string> permission;
      };

      explicit SpecUser(const Db& specUser)
         : User(specUser),
           permission(*specUser.permission)
      {}

      static string
      className()
      {
         return "SpecUser";
      }

      string permission;
};

// further domain inheritance
class ExtraSpecUser : public SpecUser
{
   public:
      class Db : public SpecUser::Db
      {
         public:
            explicit Db(string domainName = "testextraspecusers")
               : SpecUser::Db(domainName),
                 extras("extras")
            {}

            virtual FieldList all()
            {
               return SpecUser::Db::all().add(extras);
            }

            static string
            className()
            {
               return "ExtraSpecUser::Db";
            }
            
            Field<string> extras;
      };

      explicit ExtraSpecUser(const Db& extraSpecUser)
         : SpecUser(extraSpecUser),
           extras(*extraSpecUser.extras)
      {}

      static string
      className()
      {
         return "ExtraSpecUser";
      }

      string extras;
};

// other domain inheritance
class OtherSpecUser : public SpecUser
{
   public:
      class Db : public SpecUser::Db
      {
         public:
            explicit Db(string domainName = "testotherspecusers")
               : SpecUser::Db(domainName),
                 others("others")
            {}

            virtual FieldList all()
            {
               FieldList list = SpecUser::Db::all().add(others);
               return list;
            }

            static string
            className()
            {
               return "OtherSpecUser::Db";
            }

            Field<string> others;
      };

      explicit OtherSpecUser(const Db& otherSpecUser)
         : SpecUser(otherSpecUser),
           others(*otherSpecUser.others)
      {}

       static string
      className()
      {
         return "OtherSpecUser";
      }

     string others;
};
 
class Feature
{
   public:
      class Db : public Domain
      {
         public: 
            typedef string KeyType;
            typedef string ForeignKeyType;

            explicit Db(string domainName = "testfeatures")
               : Domain(domainName),
                 featureid("featureid"),
                 userid("userid"),
                 name("name")
            {}

            Field<KeyType> getKeyField() const
            {
               return featureid;
            }

            virtual FieldList all()
            {
               FieldList list = ((featureid, userid, name));
               return list;
            }

            static string 
            className()
            {
               return "Feature::Db";
            }

            Field<ForeignKeyType>
            getForeignKeyField()
            {
               return userid;
            }

            Field<KeyType> featureid; // primary key
            Field<string> userid; // foreign key
            Field<string> name;
      };

      explicit Feature(const Db& features)
         : featureid(*features.featureid),
           userid(*features.userid),
           name(*features.name)
      {}

      Feature()
      {}

      static string
      className()
      {
         return "Feature";
      }

      string featureid;
      string userid;
      string name;
      int extra;

      // easier to use than global << because of nested/forward
      string
      toString() const
      {
         return ("[" + featureid + ", " + userid + ", " + name + "]");
      }
};

ostream&
operator<<(ostream& s, const Feature& feature)
{
   s << "FeatureData[featureid=" << feature.featureid << " userid=" << feature.userid 
     << " name=" << feature.name << " extra=" << feature.extra << "]";
   return s;
}

namespace Epoisses
{

// sort VectorCache of Feature
class DataCompare<Feature>
{
  public:
   bool operator()(const Feature& d1, const Feature& d2) const
   {
      return d1.featureid < d2.featureid;
   }
};

class CacheImpl<Feature> : public Cache<Feature>
{
   public:
      Feature operator[](const string& userid) throw (NotFoundInProvisioningException&, VDbException&)
      {
         IdentityFunctor getFeature;
         applyRead(userid, getFeature);
         return getFeature.get();
      }
};
}

typedef SingletonHolder< CacheImpl<Feature> > FeatureCache;

class BadFeature
{
   public:
      class Db : public Domain
      {
         public: 
            typedef string KeyType;
            typedef string ForeignKeyType;

            explicit Db(string domainName = "testfeatures")
               : Domain(domainName),
                 featureid("featureid"),
                 userid("baduserid"),
                 name("name")
            {}

            Field<KeyType> getKeyField() const
            {
               return featureid;
            }

            virtual FieldList all()
            {
               FieldList list = ((featureid, userid, name));
               return list;
            }

            static string 
            className()
            {
               return "BadFeature::Db";
            }

            Field<ForeignKeyType>
            getForeignKeyField()
            {
               return userid;
            }

            Field<KeyType> featureid; // primary key
            Field<string> userid; // foreign key
            Field<string> name;
      };

      explicit BadFeature(const Db& features)
         : featureid(*features.featureid),
           userid(*features.userid),
           name(*features.name)
      {}

      BadFeature()
      {}

      static string
      className()
      {
         return "BadFeature";
      }

      string featureid;
      string userid;
      string name;
      int extra;

      // easier to use than global << because of nested/forward
      string
      toString() const
      {
         return ("[" + featureid + ", " + userid + ", " + name + "]");
      }
};

ostream&
operator<<(ostream& s, const BadFeature& feature)
{
   s << "BadFeatureData[featureid=" << feature.featureid << " userid=" << feature.userid 
     << " name=" << feature.name << " extra=" << feature.extra << "]";
   return s;
}

/// group storage class
class Group
{
   public:
      class Db : public Domain
      {
         public:
            typedef string KeyType;
            explicit Db(string domainName = "TestGroups")
               : Domain(domainName),
                 groupId("groupId"),
                 companyId("companyId")
            {}

            Field<KeyType> getKeyField() const
            {
               return groupId;
            }
      
            virtual FieldList all()
            {
               FieldList list = ((groupId, companyId));
               return list;
            }

            static string 
            className()
            {
               return "Group::Db";
            }

            Field<KeyType> groupId; // primary key
            Field<string> companyId;
      };

      Group(const Group::Db& group)
         : groupId(*group.groupId),
           companyId(*group.companyId)
      {}

      Group()
      {}

      static string
      className()
      {
         return "Group";
      }

      string groupId;
      string companyId;

      // easier to use than global << because of nested/forward
      string
      toString() const
      {
         return ("[" + groupId + ", " + companyId + "]");
      }
};

ostream&
operator<<(ostream& s, const Group& group)
{
   s << "Group[groupId=" << group.groupId << " companyId=" << group.companyId << "]";
   return s;
}

// multiple mappings; abstract domain
class GroupMap : public Domain
{
   public: 
      typedef int KeyType;
      typedef string ForeignKeyType;

      explicit GroupMap(string domainName = "TestGroupMaps")
         : Domain(domainName),
           groupMapId("groupMapId"),
           memberId("memberId"),
           groupId("groupId")
      {}

      Field<KeyType> getKeyField() const
      {
         return groupMapId;
      }
      
      virtual FieldList all()
      {
         FieldList list = ((groupMapId, memberId, groupId));
         return list;
      }

      static string 
      className()
      {
         return "GroupMap";
      }

      Field<KeyType> groupMapId; // primary key
      Field<string> memberId;
      Field<string> groupId;
};

// concrete domain; mapping from group to member
class GroupToMemberMap : public GroupMap
{
   public: 
      explicit GroupToMemberMap(string domainName = "TestGroupMaps") {}

      static string 
      className()
      {
         return "GroupToMemberMaps";
      }

      Field<ForeignKeyType>
      getForeignKeyField()
      {
         return memberId;
      }
};

// concrete domain; mapping from member to group
class MemberToGroupMap : public GroupMap
{
   public: 
      explicit MemberToGroupMap(string domainName = "TestGroupMaps") {}

      static string 
      className()
      {
         return "MemberToGroupMaps";
      }

      Field<ForeignKeyType>
      getForeignKeyField()
      {
         return groupId;
      }
};
