// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#include <boost/lexical_cast.hpp>
#include <iostream>
#include <strstream>
#include <stdio.h>
#include <cppunit/TestCase.h>
#include <cppunit/TestCaller.h>
#include <cppunit/TestSuite.h>
#include <cppunit/TextTestRunner.h>
#include <cppunit/TextTestResult.h>

#include "Log.hxx"
#include "Users.hxx"
#include "Logger.hxx"
#include "ProvisioningInterface.hxx"
#include "ThreadIf.hxx"
#include "TimeAccumulate.hxx"

#define VOCAL_SUBSYSTEM Subsystem::TEST

using namespace CppUnit;
using namespace Epoisses;

static char buff[3];
static const string cr = (sprintf(buff, "\n"), buff);

class MyCache : public Cache<User>
{
   public:
      MyCache()
         : Cache<User>(100)
      {}

      /// Return a snapshot of the value.
      User operator[](const string& keyObj);

      /// Set extra for the corresponding User
      void setExtra(const string& key,
                    int extra,
                    int extra1 = 0);

   protected:
      /// Update the extra field
      class SetExtraFunctor : public ApplyFunctor
      {
         public:
            SetExtraFunctor(int extra,
                            int extra1) 
               : _extra(extra),
                 _extra1(extra1)
            {}

            // write-only
            virtual void operator()(User& ud) const
            {
               ud.extra = _extra;
               ud.extra1 = _extra1;
               //DebugLog(<< "SetExtraFunctor(" << _extra << ", " << _extra1 << ") = " << ud.toString());
            }
         private:
            const int _extra;
            const int _extra1;
       };

      // exercise the hooks
      virtual void
      onCleared(User::Db::KeyType key,
                const User& oldValue)
      {
         DebugLog(<< "MyCache::onCleared(" << key << ", " << oldValue.toString() << ")");
      }
      
      virtual void
      onInsert(User::Db::KeyType key)
      {
         DebugLog(<< "MyCache::onInsert(" << key << ")");
      }
};

User
MyCache::operator[](const string& key)
{
   IdentityFunctor ifunc;
   applyRead(key, ifunc);
   return ifunc.get();
} 

void
MyCache::setExtra(const string& key,
                  int extra,
                  int extra1)
{
   applyWrite(key, SetExtraFunctor(extra, extra1));
}

class CacheTest : public TestCase
{
   public:
      CacheTest()
         : pif(ProvisioningInterface::make())
      {}

      virtual ~CacheTest()
      {}

      // sadly, cppunit makes an instance of the test on each call
      MyCache&
      getCache()
      {
         if (cache == NULL)
         {
            cache = new MyCache;
         }
         return *cache;
      }

      virtual void setUp()
      {
      }

      virtual void tearDown()
      {
      }

      void empty()
      {
         cerr << "CacheTest::empty " << this << endl;

         pif.deleteRecordsInDomain(users,Filter::None);
      }

      string getCacheAsString()
      {
         ostrstream contents;
         contents << getCache() << ends;
         contents.freeze(0);

         return contents.str();
      }

      void populate()
      {
         // create an assignment list from an array

         cerr << "CacheTest::populate " << this << endl;
         
         Assignment<string> setUserid(users.userid = "");
         Assignment<string> setFirstname(users.firstname = "");
         Assignment<string> setLastname(users.lastname = "");
         Assignment<bool> setExpired(users.expired = false);
         Assignment<int> setCount(users.count = 0);
         Assignment<int> setCredits(users.credits = 0);
         Assignment<double> setSpeed(users.speed = 0.0);

         AssignmentList usersList((setUserid, setFirstname, setLastname,
                                   setExpired, setCount, setCredits, setSpeed));

         int i = 1;
         for (; i < 10; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "Berry");
            setExpired.set((bool)(i%2));
            setCount.set(10*i);
            setCredits.set(10*i+1);
            setSpeed.set(10.0*i);

            pif.addRecordInDomain(users,
                                  usersList);
         }

         Assignment<string> setPermission(specusers.permission = "");
         AssignmentList specusersList(usersList.add(setPermission));
         
         for (; i < 20; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "SpecialBerry");
            setExpired.set((bool)(i%2));
            setCount.set(10*i);
            setCredits.set(10*i+1);
            setSpeed.set(10.0*i);
            setPermission.set("perm_" + boost::lexical_cast<string>(i));
            
            pif.addRecordInDomain(specusers,
                                  specusersList);
         }         
      }

      void testStart()
      {
         cerr << "CacheTest::testStart " << this << endl;

         vector<string> expectedKeys;
         for (int i = 1; i < 20; i++)
         {
            expectedKeys.push_back(boost::lexical_cast<string>(i));
         }
            
         vector<string> keys = getCache().getKeys();

         sort(expectedKeys.begin(), expectedKeys.end());
         sort(keys.begin(), keys.end());
         cerr << "!! " << Inserter(keys) << endl;
         CPPUNIT_ASSERT(keys == expectedKeys);

         CPPUNIT_ASSERT(getCacheAsString() == ("Cache<User>[1 -> user[1, 1, John1, 1Berry, 1, 10, 11, 10, 0], " + cr +
                                               "2 -> user[2, 2, John2, 2Berry, 0, 20, 21, 20, 0], " + cr +
                                               "3 -> user[3, 3, John3, 3Berry, 1, 30, 31, 30, 0], " + cr +
                                               "4 -> user[4, 4, John4, 4Berry, 0, 40, 41, 40, 0], " + cr +
                                               "5 -> user[5, 5, John5, 5Berry, 1, 50, 51, 50, 0], " + cr +
                                               "6 -> user[6, 6, John6, 6Berry, 0, 60, 61, 60, 0], " + cr +
                                               "7 -> user[7, 7, John7, 7Berry, 1, 70, 71, 70, 0], " + cr +
                                               "8 -> user[8, 8, John8, 8Berry, 0, 80, 81, 80, 0], " + cr +
                                               "9 -> user[9, 9, John9, 9Berry, 1, 90, 91, 90, 0], " + cr +
                                               "10 -> user[10, 10, John10, 10SpecialBerry, 0, 100, 101, 100, 0], " + cr +
                                               "11 -> user[11, 11, John11, 11SpecialBerry, 1, 110, 111, 110, 0], " + cr +
                                               "12 -> user[12, 12, John12, 12SpecialBerry, 0, 120, 121, 120, 0], " + cr +
                                               "13 -> user[13, 13, John13, 13SpecialBerry, 1, 130, 131, 130, 0], " + cr +
                                               "14 -> user[14, 14, John14, 14SpecialBerry, 0, 140, 141, 140, 0], " + cr +
                                               "15 -> user[15, 15, John15, 15SpecialBerry, 1, 150, 151, 150, 0], " + cr +
                                               "16 -> user[16, 16, John16, 16SpecialBerry, 0, 160, 161, 160, 0], " + cr +
                                               "17 -> user[17, 17, John17, 17SpecialBerry, 1, 170, 171, 170, 0], " + cr +
                                               "18 -> user[18, 18, John18, 18SpecialBerry, 0, 180, 181, 180, 0], " + cr +
                                               "19 -> user[19, 19, John19, 19SpecialBerry, 1, 190, 191, 190, 0], " + cr +
                                               "]"));
      }

      void testExistsFail()
      {
         cerr << "CacheTest::testExistsFail " << this << endl;         
         CPPUNIT_ASSERT(!getCache().exists("-1"));
      }

      void testFetch()
      {
         cerr << "CacheTest::testFetch " << this << endl;

         User d1 = getCache()["1"];
         User d11 = getCache()["11"];

         for (int i = 1; i < 11; i++)
         {
            getCache()[boost::lexical_cast<string>(i)];
         }

         string res = d1.toString() + ", " + d11.toString();

         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "[1, John1, 1Berry, 1, 10, 11, 10, 0, 0], [11, John11, 11SpecialBerry, 1, 110, 111, 110, 0, 0]");
         cerr << getCache().cachedSize() << endl;
         CPPUNIT_ASSERT(getCache().cachedSize() == 19);
      }

      void testExists()
      {
         cerr << "CacheTest::testExists " << this << endl;
         CPPUNIT_ASSERT(getCache().exists("1"));
      }

      void testExistsEmpty()
      {
         cerr << "CacheTest::testExistsEmpty " << this << endl;
         CPPUNIT_ASSERT(!getCache().exists(""));
      }

      void testFetchFail()
      {
         cerr << "CacheTest::testFetchFail " << this << endl;
         
         string res;
         
         try
         {
            User d1 = getCache()["-1"];
         }
         catch (NotFoundInProvisioningException& nfe)
         {
            res = nfe.getDescription();
         }
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "NotFoundInProvisioningException: Not in provisioning");
      }
      
      void testModifyInCache()
      {
         cerr << "CacheTest::testModifyInCache" << endl;
         
         // make sure a particular element is in the cache
         User d1 = getCache()["1"];
         getCache().setExtra("1", 17);
         d1 = getCache()["1"];
         
         string res = d1.toString();
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "[1, John1, 1Berry, 1, 10, 11, 10, 17, 0]");
      }

      void testModifyInCacheFetch()
      {
         cerr << "CacheTest::testModifyInCacheFetch" << endl;
         
         getCache().setExtra("18", 18);
         User d1 = getCache()["18"];
         
         string res = d1.toString();
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "[18, John18, 18SpecialBerry, 0, 180, 181, 180, 18, 0]");
      }

      void testModifyInCacheFail()
      {
         cerr << "CacheTest::testFetchFail " << this << endl;
         
         string res;
         
         try
         {
            getCache().setExtra("-1", 27);
         }
         catch (NotFoundInProvisioningException& nfe)
         {
            res = nfe.getDescription();
         }
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "NotFoundInProvisioningException: Not in provisioning");
      }

      
      void testWrite()
      {
         cerr << "CacheTest::testWrite " << this << endl;

         pif.setRecordsInDomain(users,
                                users.userid == "1",
                                users.count = -1);
         pif.setRecordsInDomain(users,
                                users.userid == "11",
                                users.count = -11);
         sleep(1);
         cerr << "!! " << getCache().cachedSize() << endl;
         // causes this record to be dropped from the cache
         CPPUNIT_ASSERT(getCache().cachedSize() == 17);
         // but doesn't drop the key
         CPPUNIT_ASSERT(getCache().exists("11"));
      }

      void testPostWrite()
      {
         cerr << "CacheTest::testPostWrite " << this << endl;

         User d1 = getCache()["1"];
         User d11 = getCache()["11"];

         string res = d1.toString() + ", " + d11.toString();

         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "[1, John1, 1Berry, 1, -1, 11, 10, 0, 0], [11, John11, 11SpecialBerry, 1, -11, 111, 110, 0, 0]");
         cerr << "!! " << getCache().cachedSize() << endl;
         CPPUNIT_ASSERT(getCache().cachedSize() == 19);
      }
      
      void testRemove()
      {
         cerr << "CacheTest::testRemove " << this << endl;
         InfoLog (<< "cache contents: " << getCache());

         // fake numeric comparison
         pif.deleteRecordsInDomain(users,
                                   (users.userid == "1" || users.userid == "2" ||
                                    users.userid == "3" || users.userid == "4" ||
                                    users.userid == "5" || users.userid == "15"));

         sleep(1);

         cerr << "!! " << getCacheAsString() << endl;
         CPPUNIT_ASSERT(getCacheAsString() == ("Cache<User>[6 -> user[6, 6, John6, 6Berry, 0, 60, 61, 60, 0], " + cr +
                                               "7 -> user[7, 7, John7, 7Berry, 1, 70, 71, 70, 0], " + cr +
                                               "8 -> user[8, 8, John8, 8Berry, 0, 80, 81, 80, 0], " + cr +
                                               "9 -> user[9, 9, John9, 9Berry, 1, 90, 91, 90, 0], " + cr +
                                               "10 -> user[10, 10, John10, 10SpecialBerry, 0, 100, 101, 100, 0], " + cr +
                                               "11 -> user[11, 11, John11, 11SpecialBerry, 1, -11, 111, 110, 0], " + cr +
                                               "12 -> user[12, 12, John12, 12SpecialBerry, 0, 120, 121, 120, 0], " + cr +
                                               "13 -> user[13, 13, John13, 13SpecialBerry, 1, 130, 131, 130, 0], " + cr +
                                               "14 -> user[14, 14, John14, 14SpecialBerry, 0, 140, 141, 140, 0], " + cr +
                                               "16 -> user[16, 16, John16, 16SpecialBerry, 0, 160, 161, 160, 0], " + cr +
                                               "17 -> user[17, 17, John17, 17SpecialBerry, 1, 170, 171, 170, 0], " + cr +
                                               "18 -> user[18, 18, John18, 18SpecialBerry, 0, 180, 181, 180, 18], " + cr +
                                               "19 -> user[19, 19, John19, 19SpecialBerry, 1, 190, 191, 190, 0], " + cr +
                                               "]"));
         CPPUNIT_ASSERT(!getCache().exists("1"));
         CPPUNIT_ASSERT(!getCache().exists("2"));
         CPPUNIT_ASSERT(!getCache().exists("15"));
      }

      void testPostRemove()
      {
         cerr << "CacheTest::testPostRemove " << this << endl;

         string err;
         try
         {
            User d1 = getCache()["1"];
         }
         catch (NotFoundInProvisioningException& cnfe)
         {
            err = cnfe.getDescription();
         }
         cerr << "!! " << err << endl;
         CPPUNIT_ASSERT(err == "NotFoundInProvisioningException: Not in provisioning");
         CPPUNIT_ASSERT(!getCache().exists("1"));
      }

      void testInsert()
      {
         cerr << "CacheTest::testInsert " << this << endl;

         // precondition
         CPPUNIT_ASSERT(!getCache().exists("1"));         
         pif.addRecordInDomain(users,
                               (users.userid = "1",
                                users.firstname = "newJohn",
                                users.lastname = "newBerry",
                                users.expired = false,
                                users.count = 111,
                                users.credits = 0,
                                users.speed = 10));
         sleep(1);
         
         CPPUNIT_ASSERT(getCache().exists("1"));
      }

      void testInsertDerived()
      {
         cerr << "CacheTest::testInsertDerived " << this << endl;

         // precondition
         CPPUNIT_ASSERT(!getCache().exists("2"));
         pif.addRecordInDomain(specusers,
                               (users.userid = "2",
                                users.firstname = "newSpecialJohn",
                                users.lastname = "newSpecialBerry",
                                users.expired = false,
                                users.count = 2111,
                                users.credits = 20,
                                users.speed = 210));
         sleep(1);
         
         CPPUNIT_ASSERT(getCache().exists("2"));
      }

      void testUnregister()
      {
         cerr << "CacheTest::testUnregister " << this << endl;
         
         delete cache;
         cache = NULL;
      }

      class InvariantException : public VException
      {
         public: 
            InvariantException(const string& msg,
                               const string& file,
                               int line)
               : VException(msg, file, line)
            {}

            virtual string getName() const
            {
               return "InvariantException";
            }
      };

      void checkInvariants(const User& ud) throw (VException&)
      {
         if (!(((ud.count+1 == ud.credits) ||
                (ud.count == -ud.credits)) &&
               ((ud.extra+1 == ud.extra1) ||
                (ud.extra == -ud.extra1))))
         {
            throw InvariantException("invariant failed",
                                     __FILE__, __LINE__);
         }
      }
      
      void readWriteModify1()
      {
         const int w = random() % 3;
         const string u = boost::lexical_cast<string>((random() % 15) + 1);

         switch (w)
         {
            case 0 : 
            {
               // read and check for invariants
               //cerr << "thread 1 read" << endl;
               checkInvariants(getCache()[u]);
               break;
            }
            case 1 : 
            {
               // modify extra data 
               //cerr << "thread 1 modify" << endl;
               getCache().setExtra(u, 5, 6);
               break;
            }
            case 2 : 
            {
               // write 
               //cerr << "thread 1 write" << endl;
               pif.setRecordsInDomain(users, 
                                      users.userid == u,
                                      (users.count = 17,
                                       users.credits = 18));
               break;
            }
         }
      }

      void readWriteModify2()
      {
         const int w = random() % 3;
         const string u = boost::lexical_cast<string>((random() % 15) + 1);

         switch (w)
         {
            case 0 : 
            {
               // read and check for invariants
               //cerr << "thread 2 read" << endl;
               checkInvariants(getCache()[u]);
               break;
            }
            case 1 : 
            {
               // modify extra data 
               //cerr << "thread 2 modify" << endl;
               getCache().setExtra(u, 11, -11);
               break;
            }
            case 2 : 
            {
               // write 
               //cerr << "thread 2 write" << endl;
               pif.setRecordsInDomain(users, 
                                      users.userid == u,
                                      (users.count = 19,
                                       users.credits = -19));
               break;
            }
         }
      }

      class OtherThread : public ThreadIf
      {
         public: 
            OtherThread(CacheTest& cacheTest)
               : _cacheTest(cacheTest)
            {}

            virtual void thread()
            {
               for (int i = 0; i < 50; i++)
               {
                  _cacheTest.readWriteModify2();
               }
            }
         private:
            CacheTest& _cacheTest;
      };
     
      void testConcurrency() 
      {
         cerr << "CacheTest::testConcurrency" << endl;

         testUnregister();
         
         // zero the database
         empty();

         // strategy -- two threads writing and modifying at random 
         //  each thread has its own invariant for writing and modifying 
         //  at each read, the prov state and the cache state should 
         //  independently be in one or the other invariant. 
          
         //                         (count, credits), (extra, extra1)
         // invariant 1: (x, x+1) 
         // invariant 2: (x, -x) 

         // initialize to the first invariant for prov data, second invariant
         // for cache data
         populate();

         cerr << "!! " << getCache() << endl;

         // start the other thread exercising its invariant
         OtherThread thr(*this);
         thr.run();

         // start this thread exercising its invariant
         for (int i = 0; i < 50; i++)
         {
            readWriteModify1();
         }
         thr.join();

         testUnregister();
      }

      void testBadCache()
      {
         try
         {
            Cache<BadFeature> badCache(10);
         }
         catch (VDbException& e)
         {
            return;
         }
         CPPUNIT_ASSERT(false);
      }
       
   private:
      ProvisioningInterface& pif;
      static MyCache* cache;
      User::Db users;
      SpecUser::Db specusers;
};

MyCache* CacheTest::cache = NULL;

// local macros
#define cacheTEST(_method) \
   suite->addTest(new TestCaller<CacheTest>(#_method, \
                                            &CacheTest::_method))

int main(int argc, char** argv)
{
   Log::initialize(Log::COUT, Log::DEBUG, argv[0]);
   ProvisioningInterface::make("localhost", 6005);

   TestSuite* suite = new TestSuite;
   assert(suite->countTestCases() == 0);
   
   TextTestResult result;
   cacheTEST(empty);
   cacheTEST(populate);
   cacheTEST(testStart);
   cacheTEST(testExistsFail);
   cacheTEST(testExistsEmpty);
   cacheTEST(testFetch);
   cacheTEST(testExists);
   cacheTEST(testFetchFail);
   cacheTEST(testModifyInCache);
   cacheTEST(testModifyInCacheFetch);
   cacheTEST(testModifyInCacheFail);
   cacheTEST(testWrite);
   cacheTEST(testPostWrite);

   cacheTEST(testRemove);
   cacheTEST(testPostRemove);
   cacheTEST(testInsert);
   cacheTEST(testInsertDerived);

   cacheTEST(testUnregister);

   cacheTEST(testConcurrency);
   
   cacheTEST(testBadCache);

   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();

   // ?dlb? register atExit?
   SingletonHolder<CacheManager>::Instance().shutdown();

   TimeAccumulate::dump();
   return 0;
}
