// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
#include <boost/lexical_cast.hpp>
#include <strstream>

#include "Logger.hxx"
#include "Users.hxx"
#include "cppunit/TestCase.h"
#include "cppunit/TestCaller.h"
#include "cppunit/TestSuite.h"
#include "cppunit/TextTestRunner.h"
#include "cppunit/TextTestResult.h"
#include "ProvisioningInterface.hxx"
#include "VDbException.hxx"
#include "VectorCache.hxx"
#include "ThreadIf.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::TEST

using namespace CppUnit;
using namespace Epoisses;

// test VectorCache template
class FeaturesByUserCacheImpl : public VectorCache<Feature, User::Db>
{
   public:
      FeaturesByUserCacheImpl() : VectorCache<Feature, User::Db>(100) {}
};

typedef SingletonHolder<FeaturesByUserCacheImpl> FeaturesByUserCache;

/// Add some data
class AddFeatures : public ThreadIf
{
   public:
      AddFeatures(int mod)
         : _mod(mod),
           _pif(ProvisioningInterface::make())
      {}

      virtual ~AddFeatures()
      {
          join();
      }

      // runs once
      virtual void thread()
      {
         Assignment<string> setFeatureid(features.featureid = "");
         Assignment<string> setUserid(features.userid = "");
         Assignment<string> setName(features.name = "");

         AssignmentList assignments((setFeatureid, setUserid, setName));

         // every user whose userid is divisible by _mod gets a feature
         for (int i = _mod; i < 20; i += _mod) {
            // retrieve first
            FeaturesByUserCache::Instance()[boost::lexical_cast<string>(i)];

            // slam values into setters
            setFeatureid.set("f" + boost::lexical_cast<string>(_mod) + "_" +
                             boost::lexical_cast<string>(i));
            setUserid.set(boost::lexical_cast<string>(i));
            setName.set("feature" + boost::lexical_cast<string>(_mod));

            // insert the feature
            _pif.addRecordInDomain(features,
                                   assignments);
         }
      }

   private:
      int _mod;
      Feature::Db features;
      ProvisioningInterface& _pif;
};

class VectorCacheTestParallel : public TestCase
{
   public:

      VectorCacheTestParallel()
         : pif(ProvisioningInterface::make())
      {}
        
      virtual ~VectorCacheTestParallel()
      {}

      virtual void setUp()
      {
      }

      virtual void tearDown()
      {
      }

      void empty()
      {
         cerr << "VectorCacheTestParallel::empty " << this << endl;

         pif.deleteRecordsInDomain(users,
                                   Filter::None);

         pif.deleteRecordsInDomain(features,
                                   Filter::None);

      }

      void populatePrimary()
      {
         cerr << "VectorCacheTestParallel::populatePrimary " << this << endl;
         
         // create an assignment list from an array
         Assignment<string> setUserid(users.userid = "");
         Assignment<string> setFirstname(users.firstname = "");
         Assignment<string> setLastname(users.lastname = "");
         Assignment<bool> setExpired(users.expired = false);
         Assignment<int> setCount(users.count = 0);
         Assignment<int> setCredits(users.credits = 0);
         Assignment<double> setSpeed(users.speed = 0.0);

         AssignmentList usersList((setUserid, setFirstname, setLastname,
                                   setExpired, setCount, setCredits, setSpeed));

         int i = 1;
         for (i = 0; i < 10; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "Berry");
            setExpired.set((bool)(i%2));
            setCount.set(i);
            setCredits.set(10*i);
            setSpeed.set(10.0*i);

            pif.addRecordInDomain(users,
                                  usersList);
         }

         Assignment<string> setPermission(specusers.permission = "");
         AssignmentList specusersList(usersList.add(setPermission));
         
         for (; i < 20; i++)
         {
            // slam values into assignments
            setUserid.set(boost::lexical_cast<string>(i));
            setFirstname.set("John" + boost::lexical_cast<string>(i));
            setLastname.set(boost::lexical_cast<string>(i) + "SpecialBerry");
            setExpired.set((bool)(i%2));
            setCount.set(i);
            setCredits.set(10*i);
            setSpeed.set(10.0*i);
            setPermission.set("perm_" + boost::lexical_cast<string>(i));
            
            pif.addRecordInDomain(specusers,
                                  specusersList);
         }         
      }

      void populateForeign()
      {
         cerr << "VectorCacheTestParallel::populateForeign " << this << endl;

         AddFeatures f1(1);
         AddFeatures f2(2);
         AddFeatures f3(3);
         AddFeatures f5(5);

         f1.run();
         f2.run();
         f3.run();
         f5.run();

         sleep(1);
      }

      // convert vector of Feature to string
      static string
      toString(const vector<Feature>& featureData)
      {
         string s("{");
         for (vector<Feature>::const_iterator frIt = featureData.begin();
              frIt != featureData.end(); frIt++)
         {
            s = s + frIt->toString() + ", ";
         }
         s = s + "}";
         return s;
      }

      void testFetch()
      {
         cerr << "VectorCacheTestParallel::testFetch " << this << endl;

         vector<Feature> d1 = FeaturesByUserCache::Instance()[string("1")];
         vector<Feature> d3 = FeaturesByUserCache::Instance()[string("3")];
         vector<Feature> d6 = FeaturesByUserCache::Instance()[string("6")];
         vector<Feature> d15 = FeaturesByUserCache::Instance()[string("15")];

         string res = toString(d1);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{[f1_1, 1, feature1], }");

         // should have 2 features
         res = toString(d3);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{[f1_3, 3, feature1], [f3_3, 3, feature3], }");

         // should have 3 features
         res = toString(d6);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{[f1_6, 6, feature1], [f2_6, 6, feature2], [f3_6, 6, feature3], }");

         // should have 3 features
         res = toString(d15);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{[f1_15, 15, feature1], [f3_15, 15, feature3], [f5_15, 15, feature5], }");
      }

      void testFetchNonExistent()
      {
         cerr << "VectorCacheTestParallel::testFetchNonExistent " << this << endl;

         try
         {
            vector<Feature> d1 = FeaturesByUserCache::Instance()[string("-1")];
            CPPUNIT_ASSERT(false);
         } 
         catch (NotFoundInProvisioningException& e)
         {
            return;
         }
      }
      
      void testModify()
      {
         cerr << "VectorCacheTestParallel::testModify " << this << endl;

         unsigned int s = FeaturesByUserCache::Instance().size();

         cerr << "!! " << toString(FeaturesByUserCache::Instance()["15"]) << endl;

         pif.setRecordsInDomain(features,
                                features.featureid == "f1_15",
                                features.name = "modified 1");
         pif.setRecordsInDomain(features,
                                features.featureid == "f5_15",
                                features.name = "modified 15");
         sleep(1);
         cerr << "!! " << FeaturesByUserCache::Instance() << endl;
         // should have no effect on cache
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance().size() == s);

         cerr << "!! " << toString(FeaturesByUserCache::Instance()["15"]) << endl;
      }

      void testRemoveForeign()
      {
         cerr << "VectorCacheTestParallel::testRemoveForeign " << this << endl;

         cerr << "!! " << FeaturesByUserCache::Instance() << endl;

         pif.deleteRecordsInDomain(features,
                                   features.featureid == "f5_15" ||
                                   features.featureid == "f3_9");
         sleep(1);
         cerr << "!! " << FeaturesByUserCache::Instance() << endl;
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance()["9"].size() == 1);
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance()["15"].size() == 2);
      }

      void testRemovePrimary()
      {
         cerr << "VectorCacheTestParallel::testRemovePrimary " << this << endl;

         cerr << "!! " << FeaturesByUserCache::Instance() << endl;

         pif.deleteRecordsInDomain(users,
                                   users.userid == "9");
         sleep(1);

         try
         {
            FeaturesByUserCache::Instance()["9"];
            CPPUNIT_ASSERT(false);
         }
         catch (NotFoundInProvisioningException &e)
         {
            return;
         }
      }

      void testAdd()
      {
         cerr << "VectorCacheTestParallel::testAdd " << this << endl;

         cerr << "!! " << toString(FeaturesByUserCache::Instance()["12"]) << endl;
         
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance()["12"].size() == 3);
         
         unsigned int s = FeaturesByUserCache::Instance().size();

         AddFeatures f4(4);
         f4.run();

         sleep(1);
         cerr << "!! " << FeaturesByUserCache::Instance() << endl;
         
         // causes 12 to drop
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance().size() == s-1);
      }

      void testPostAdd()
      {
         cerr << "VectorCacheTestParallel::testPostAdd " << this << endl;

         vector<Feature> d = FeaturesByUserCache::Instance()[string("12")];
         string res = toString(d);
         cerr << "!! " << res << endl;
         CPPUNIT_ASSERT(res == "{[f1_12, 12, feature1], [f2_12, 12, feature2], [f3_12, 12, feature3], [f4_12, 12, feature4], }");
      }

      void testRemoveAll()
      {
         cerr << "VectorCacheTestParallel::testRemoveAll " << this << endl;
         DebugLog( << "!! " << FeaturesByUserCache::Instance());

         pif.deleteRecordsInDomain(features,
                                   Filter::None);
         sleep(1);
         DebugLog( << "!! " << FeaturesByUserCache::Instance());
         CPPUNIT_ASSERT(FeaturesByUserCache::Instance().size() == 0);
      }

   private:
      ProvisioningInterface& pif;
      User::Db users;
      SpecUser::Db specusers;
      Feature::Db features;
};

// local macros
#define cacheTEST(_method) \
   suite->addTest(new TestCaller<VectorCacheTestParallel>(#_method, \
                                                          &VectorCacheTestParallel::_method))

int main(int argc, char** argv)
{
   //cpLogSetPriority(LOG_DEBUG);
   Log::initialize(Log::COUT, Log::DEBUG, "testVectorCache");
   ProvisioningInterface::make("localhost", 6005);

   TestSuite* suite = new TestSuite;
   assert(suite->countTestCases() == 0);
   
   TextTestResult result;
   cacheTEST(empty);
   cacheTEST(populatePrimary);
   cacheTEST(populateForeign);
   cacheTEST(testFetch);
   cacheTEST(testFetchNonExistent);
   cacheTEST(testModify);
   cacheTEST(testRemoveForeign);
   cacheTEST(testRemovePrimary);
   cacheTEST(testAdd);
   cacheTEST(testPostAdd);

   cacheTEST(testRemoveAll);

   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();
   return 0;
}

