#ifndef PGSQLCONNECTION_HXX
#define PGSQLCONNECTION_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <libpq-fe.h>
#include <ext/hash_map>
#include <string>

#include "global.h"
#include "StringHash.hxx"
#include "DBConnection.hxx"
#include "RWMutex.hxx"


namespace Epoisses {

// domain => parent
typedef hash_map<string, string> DomainToParentMap;
   
/// Forward Declare PGSQLResultset
class PGSQLResultset;

/// PGSQLConnection class
class PGSQLConnection : public DBConnection
{
      friend class ConnectionManagerImpl;
   private:
      /// Constructor
      PGSQLConnection(const string& dbname, 
                      const string& dbhost,
                      const string& dbuser, 
                      const string& dbpassword);

      /// Destructor
      virtual ~PGSQLConnection();
      
      /// Return a resultset object based on query
      virtual DBResultset* 
      DBSelect(const string& Domain, 
               const string& SearchFilter, 
               const FieldListtype& FieldList, 
               const Ordertype& Order, 
               const string &OrderField);

      /// Return a resultset object based on query
      virtual DBResultset* 
      DBSelect(const string& Domain,
               const string& SearchFilter, 
               const FieldListtype& FieldList);

      /// Return a resultset object based on query
      virtual DBResultset*
      DBSelect(const string& Domain, 
               const string& SearchFilter);

      /// Return a resultset object based on query
      virtual DBResultset* 
      DBSelect(const string& Domain,
               const string& SearchFilter, 
               const string& start, 
               const string& end, 
               const FieldListtype& FieldList);

      /// Return a resultset object based on query
      virtual DBResultset* 
      DBSelect(const string& Domain,
               const string& SearchFilter,
               const string& start,
               const string& end,
               const Ordertype& Order,
               const string &OrderField,
               const FieldListtype& FieldList);

      /// Validate the result, cleanup, and return it or throw
      DBResultset*
      DBSelectHelper(const char* sqlCommand);

      /// Return a string containing primary key of Domain
      virtual string DBPrimaryKey( const string& Domain);

      /// Return the parent of the domain
      virtual string DBParent(const string& Domain);

      /// Return the children of the domain
      virtual vector<string> DBChildren(const string& domain);

      /// Return the root of the domain
      virtual string DBRoot(const string& domain);

      /// Return a table containing fields/types/keys of Domain
      virtual ResultTabletype DBDescribe(const string& Domain);

      /// Return a list of domains in database
      virtual ResultRecordtype DBDescribe(void);

      /// Update Domain, based on searchfilter
      virtual void DBUpdate(const string& Domain, const string& SearchFilter,
                            const FieldValuetype& FieldValueSet);

      /// Update Domain, based on previous resultset
      virtual void DBUpdate(const string& Domain, DBResultset* Resultset,
                            const FieldValuetype& FieldValueSet);

      /// Insert record
      virtual int DBInsert(const string& Domain,
                           const FieldValuetype& FieldValueSet);

      /// Delete record, based on searchfilter
      virtual void DBDelete(const string& Domain, 
                            const string& SearchFilter);

      /// Delete record, based on previous resultset
      virtual void DBDelete(const string& Domain, 
                            DBResultset* Resultset);
 
      /// Create a new domain, use STRINGTYPE for all columns
      /// with or without checking current existence
      virtual void DBCreate(const string& Domain, 
                            const FieldListtype& FieldsList, 
                            const string& PrimaryKey, 
                            bool ExistenceCheck);

      /// Create a new domain, specify types for each column
      /// with or without checking current existence
      virtual void DBCreate(const string& Domain, 
                            const FieldListtype& FieldsList, 
                            const FieldListDatatype& Datatype, 
                            const string& PrimaryKey, 
                            bool ExistenceCheck);

      /// Create a new domain that inherits from an existing domain,
      /// with or without checking current existence
      // Note: primary key is inherited and may not be specified here
      virtual void DBCreate(const string& domain, 
                            const string& parentDomain,
                            const FieldListtype& fieldsList, 
                            const FieldListDatatype& datatypelist, 
                            bool existenceCheck);

      /// Alter domain, drop field
      virtual void DBAlter(const string& Domain, 
                           const string& Fieldname);

      /// Alter domain, change a type of a field
      virtual void DBAlter(const string& Domain, 
                           const string& Fieldname, 
                           const Datatype Newtype);

      /// Drop a Domain, with or without checking current existence
      virtual void DBDrop(const string& Domain, 
                          bool ExistenceCheck);

      /// Returns true if Domain exists, false otherwise
      virtual bool DBVerify(const string& Domain);

      /// Returns true if a sequence exists for the specified domain
      virtual bool DBVerifySequence(const string& Domain);      

      /// Create the SQL for the respective query
      void CreateSQLselect(char *Sqlcmd, 
                           const string& Domain,
                           const string& SearchFilter, 
                           const FieldListtype& FieldList,
                           const Ordertype& Order, 
                           const string& OrderField);

      /// Create the SQL for the respective query
      void CreateSQLselect(char *Sqlcmd, 
                           const string& Domain, 
                           const string& SearchFilter, 
                           const FieldListtype& FieldList);

      /// Create the SQL for the respective query
      void CreateSQLselect(char *Sqlcmd, 
                           const string& Domain,
                           const string& SearchFilter, 
                           const string& start, 
                           const string& end,
                           const FieldListtype& FieldList);

      /// Create the SQL for the respective query
      void CreateSQLselect( char* Sqlcmd,
                            const string& Domain,
                            const string& SearchFilter, 
                            const string& start,
                            const string& end,
                            const Ordertype& Order, 
                            const string& OrderField,
                            const FieldListtype& FieldList );

      /// Create the SQL for the respective query
      void CreateSQLselect(char *Sqlcmd, const string& Domain,
                           const string& SearchFilter );

      /// Create the SQL for the respective query
      void CreateSQLupdate(char *Sqlcmd, const string& Domain, 
                           const string& SearchFilter, 
                           const FieldValuetype& FieldValueSet);

      /// Create the SQL for the respective query
      void CreateSQLinsert(char *Sqlcmd, const string& Domain, 
                           const FieldValuetype& FieldValueSet);

      /// Create the SQL for the respective query
      void CreateSQLdelete(char *Sqlcmd, const string& Domain, 
                           const string& SearchFilter);

      /// Create the SQL for the respective query
      void CreateSQLcreate(char *Sqlcmd, const string& Domain, 
                           const FieldListtype& FieldList, 
                           const string& PrimaryKey);

      /// Create the sequence for default values of primary key
      void
      PGSQLConnection::CreateSQLsequence(char* Sqlcmd, 
                                         const string& Domain, 
                                         const string& PrimaryKey);

      /// Create the SQL for the respective query
      void CreateSQLcreate(char *Sqlcmd, const string& Domain, 
                           const FieldListtype& FieldList, 
                           const FieldListDatatype& DatatypeList, 
                           const string& PrimaryKey);

      /// Create the SQL for the respective query
      void CreateSQLcreate(char *Sqlcmd, 
                           const string& domain, 
                           const string& parentDomain,
                           const FieldListtype& fieldList, 
                           const FieldListDatatype& datatypeList);

      /// Return the PostgreSQL equivalent of DBModule Datatype
      void GetSQLtypestring(string& Sqltype, Datatype type);

      virtual DBResultset* DBRetrieveByOid(const string& Domain, 
                                           uint oid,
                                           const FieldListtype& FieldList);

      // !dlb! pool 
      /// PostgreSQL Connection handle
      PGconn* myPgSQLConnectionP;

      /// Read/Write mutex for the map.
      static Vocal::Threads::RWMutex parentMapMutex;
      // mapping from domain to parent domain
      static DomainToParentMap domainToParent;
};

} // namespace Epoisses

#endif
