/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */



#include <iostream>
#include <string>
#include "VDbException.hxx"
#include "PGSQLFilterparser.hxx"

using namespace std;
using namespace Epoisses;

// Constructor
PGSQLFilterparser::PGSQLFilterparser(const string& filterbuffer):
myScanner(filterbuffer),
mySQLclausestrm(mySQLclause, sizeof(mySQLclause))
{
    // Parse when constructed
    Parse();
}

// GetToken- Ask myScanner object for next token
void
PGSQLFilterparser::GetToken(void)
{
    myTokenP = myScanner.GetToken();
    myTokencode = myTokenP->GetCode();
    myTokenstring = myTokenP->GetTokenString();
}

// Parse-
void
PGSQLFilterparser::Parse(void)
{
    // Get First Token
    GetToken();
    
    // Parse Search Filter
    ParseFilter();

    // Check that the last token is the EndOfString
    if (myTokencode != tcEOS)
    {
        string err = "Error in Parse: End of String expected";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    mySQLclausestrm << ends;

}

// ParseFilter-
void
PGSQLFilterparser::ParseFilter(void)
{

    // Grammar specifies that filters must always begin w/ a Left Paren
    if ( myTokencode != tcLParen )
    {
        string err = "Error in ParseFilter: Filter must begin with Left Paren";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        mySQLclausestrm << myTokenstring << " " ;
    }

    GetToken();
    ParseFilterComposition();
    
    if ( myTokencode != tcRParen )
    {
        string err = "Error in ParseFilter: Filter must end with Right Paren";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        mySQLclausestrm << myTokenstring << " " ;
        GetToken();
    }

}

//
void
PGSQLFilterparser::ParseFilterComposition(void)
{

    // Indirect recursion to allow nested filters.
    // ParseFilter -> 
    //      ParseFilterComposition 
    //              -> ParseFilterList
    //                      -> ParseFilter
    switch ( myTokencode )
    {
        case tcAnd:
            // Push SQL 'AND' command on stack
            myParsestack.push("AND ");
            // Parse 1 or more filters
            ParseFilterList();
            break;

        case tcOr:
            // Push SQL 'OR' command on stack
            myParsestack.push("OR ");
            // Parse 1 or more filters
            ParseFilterList();
            break;

        case tcNot:
            // Add SQL 'NOT' to clause string
            mySQLclausestrm << "NOT " ;
            // Parse 1 filter
            GetToken();
            ParseFilter();
            break;

        case tcWord:
            ParseItem();
            break;

        default:
            string err = "Error in ParseFilterComposition: Token not recognized - " + myTokenstring;
            throw VDbException(err, __FILE__, __LINE__, 0);
    }
}

//
void
PGSQLFilterparser::ParseFilterList(void)
{
    string currentoperator;

    GetToken();

    do
    {
       ParseFilter();

       // For SQL 'WHERE' clause, check if there is currently an operator 
       // on the stack and check that the next token is not the end
       // of the current filter.  If both are true, add the operator 
       // to the mySQLclause stream.
       if ( (!myParsestack.empty()) && (myTokencode != tcRParen) )
       {
           currentoperator = myParsestack.top();
           mySQLclausestrm << currentoperator;
       }

    }
    while (myTokencode == tcLParen);

    // Pop the SQL operator of the stack since we're done parsing 
    // a filterlist
    myParsestack.pop();

}

// ParseItem- Items are ( Fieldname <op> "Value" )
void
PGSQLFilterparser::ParseItem(void)
{
    // Current Token is Field name
    mySQLclausestrm << myTokenstring << " ";
    
    // Next Token should be operator
    GetToken();
    switch ( myTokencode )
    {
        case tcLt:
        case tcGt:
        case tcLe:
        case tcGe:
            mySQLclausestrm << myTokenstring << " ";
            GetToken();
            // After parsing operator, next Token should be Value
            ParseValue();
            break;
        case tcEqual:
            // For SQL, replace '=' with 'LIKE' in case you're
            // dealing with a substring.
            mySQLclausestrm << "LIKE " ;
            GetToken();
            // Next Token should be StarValue
            ParseStarValue();
            break;
        default:
            string err = "Error in ParseItem: Token not recognized - " + 
               myTokenstring;
            throw VDbException(err, __FILE__, __LINE__, 0);
   }
}

// ParseValue- This value should not be a substring. This Token
// always comes after a >, <, <=, >=, operator and should be a literal
// value. Basically, the grammar states you can't say 
// (vocalphonenumber < "200*"). If you do this, the backed DB
// will probably give you incorrect results.
void
PGSQLFilterparser::ParseValue(void)
{
    if (( myTokencode != tcString) && (myTokencode != tcWord ))
    {
        string err = "Error in ParseValue: Token not recognized - " + 
           myTokenstring;
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        // PgSQL and MySql like single quoted string values
        mySQLclausestrm << "\'" << myTokenstring << "\'" << " " ;
    }
    GetToken();

}

// ParseStarValue- This value should be a substring. This Token
// always comes after an = operator.
void
PGSQLFilterparser::ParseStarValue(void)
{
    string::iterator p;
    int i;
    if ( myTokencode != tcString )
    {
        string err = "Error in ParseValue: Token must be string";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        // In order for SQL to handle the StarValue tokens correctly,
        // The stars (*), need to be replaced with percent symbols (%)
        for ( i = 0, p = myTokenstring.begin(); 
              p != myTokenstring.end(); p++, i++)
        {
            if (myTokenstring[i] == '*')
                myTokenstring[i] = '%';
        }

        // PgSQL and MySql like single quoted string values
        mySQLclausestrm << "\'" << myTokenstring << "\'" << " " ;
    }
    GetToken();
}

// GetParsedresult- Return SQL 'WHERE' clause
const char* 
PGSQLFilterparser::GetParsedresult(void)
{
    return mySQLclause;
}
