#include <boost/lexical_cast.hpp>
#include <boost/random.hpp>
#include <cppunit/TestCaller.h>
#include <cppunit/TestCase.h>
#include <cppunit/TestSuite.h>
#include <cppunit/TextTestResult.h>
#include <cppunit/TextTestRunner.h>
#include <strstream>
#include "ConnectionManager.hxx"
#include "CommandLineParser.hxx"
#include "Logger.hxx"
#include "PSConfiguration.hxx"
#include "VDbException.hxx"

#define VOCAL_SUBSYSTEM Vocal::Subsystem::TEST

using namespace CppUnit;
using namespace Vocal;
using namespace Epoisses;


class ConnectionTest : public TestCase
{
   public:

      ConnectionTest()
      {}
        
      virtual ~ConnectionTest()
      {}

      virtual void setUp()
      {
         DebugLog(<<"In Setup");         
         ConnectionManager::Connection dbconn;
         dbconn.DBDrop("testConnectionTestDomainDerived", false);
         dbconn.DBDrop("testConnectionTestDomain", false);
      }

      virtual void tearDown()
      {
/*
         ConnectionManager::Connection dbconn;
         dbconn.DBDrop("testConnectionTestDomainDerived", false);
         dbconn.DBDrop("testConnectionTestDomain", false);
*/
      }
      
      void testVerifyFail()
      {
         ConnectionManager::Connection dbconn;
         CPPUNIT_ASSERT(!dbconn.DBVerify("testConnectionTestDomain"));
      }

      void testVerify()
      {
         DebugLog(<<"**testVerify");
         
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);            
         CPPUNIT_ASSERT(dbconn.DBVerify("testConnectionTestDomain"));
      }
                  
      void testCreateExists()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);            
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);            
      }
      
      void testCreateCheckExists()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         try
         {
            dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", true);
            CPPUNIT_ASSERT(false);
         }
         catch (VException& e)
         {
         }
      }
      
      void testCreateInherit()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         
         FieldListtype moreFields;
         moreFields.push_back("bar");
         FieldListDatatype moreFieldTypes;
         moreFieldTypes.push_back(INTEGERTYPE);

         dbconn.DBCreate("testConnectionTestDomainDerived", "testConnectionTestDomain", 
                         moreFields, moreFieldTypes, false);

         dbconn.DBDrop("testConnectionTestDomainDerived", true);
         CPPUNIT_ASSERT(!dbconn.DBVerify("testConnectionTestDomainDerived"));
      }
      
      void testRemoveBase()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         
         FieldListtype moreFields;
         moreFields.push_back("bar");
         FieldListDatatype moreFieldTypes;
         moreFieldTypes.push_back(INTEGERTYPE);

         dbconn.DBCreate("testConnectionTestDomainDerived", "testConnectionTestDomain", 
                         moreFields, moreFieldTypes, false);
         try
         {
            dbconn.DBDrop("testConnectionTestDomain", false);
            CPPUNIT_ASSERT(false);
         }
         catch (VDbException &e)
         {
         }
      }
      
      void testRemoveCheckDoesNotExist()
      {
         ConnectionManager::Connection dbconn;
         
         try
         {
            dbconn.DBDrop("testConnectionTestDomain", true);
            CPPUNIT_ASSERT(false);
         }
         catch (VException &e)
         {
         }
      }      


      class RowComparitor
      {
         public:
            bool operator()(const vector<string>& r1,
                            const vector<string>& r2)
            {
               return r1[0] < r2[0];
            }
      };
      
      void testDescribeDomain()
      {
         cout << "**testDescribeDomain**" << endl;
         
         ConnectionManager::Connection dbconn;         
         FieldListtype fields;
         fields.push_back("primfoo");
         fields.push_back("bar");

         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         fieldTypes.push_back(FLOATTYPE);
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomain");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 2);
         CPPUNIT_ASSERT(res[0][0] == "bar");
         CPPUNIT_ASSERT(res[0][1] == "Float");

         CPPUNIT_ASSERT(res[1].size() == 3);
         CPPUNIT_ASSERT(res[1][0] == "primfoo");
         CPPUNIT_ASSERT(res[1][1] == "Integer");
         CPPUNIT_ASSERT(res[1][2] == "PRIMARYKEY");
      }

      void testDescribeDerivedDomain()
      {
         cout << "**testDescribeDerivedDomain**" << endl;
         
         ConnectionManager::Connection dbconn;         
         FieldListtype fields;
         fields.push_back("primfoo");

         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);
         
         FieldListtype derivedFields;
         derivedFields.push_back("bar");
         FieldListDatatype derivedFieldTypes;
         derivedFieldTypes.push_back(FLOATTYPE);
         dbconn.DBCreate("testConnectionTestDomainDerived", 
                         "testConnectionTestDomain",
                         fields, fieldTypes, false);

         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomainDerived");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 2);
         CPPUNIT_ASSERT(res[0][0] == "bar");
         CPPUNIT_ASSERT(res[0][1] == "Float");

         CPPUNIT_ASSERT(res[1].size() == 3);
         CPPUNIT_ASSERT(res[1][0] == "primfoo");
         CPPUNIT_ASSERT(res[1][1] == "Integer");
         CPPUNIT_ASSERT(res[1][2] == "PRIMARYKEY");
      }
    
      void testAlterAdd()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primfoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);
         dbconn.DBAlter("testConnectionTestDomain", "bar", INTEGERTYPE);
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomain");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 2);
         CPPUNIT_ASSERT(res[0][0] == "bar");
         CPPUNIT_ASSERT(res[0][1] == "Integer");

         CPPUNIT_ASSERT(res[1].size() == 3);
         CPPUNIT_ASSERT(res[1][0] == "primfoo");
         CPPUNIT_ASSERT(res[1][1] == "Integer");
         CPPUNIT_ASSERT(res[1][2] == "PRIMARYKEY");
      }

      void testAlterAddDerived()
      {
         ConnectionManager::Connection dbconn;         
         FieldListtype fields;
         fields.push_back("primfoo");

         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         fieldTypes.push_back(FLOATTYPE);
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);

         FieldListtype emptyFields;
         FieldListDatatype emptyFieldTypes;
         dbconn.DBCreate("testConnectionTestDomainDerived", 
                         "testConnectionTestDomain",
                         emptyFields, emptyFieldTypes, false);
         
         dbconn.DBAlter("testConnectionTestDomainDerived", "bar", INTEGERTYPE);
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomainDerived");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 2);
         CPPUNIT_ASSERT(res[0][0] == "bar");
         CPPUNIT_ASSERT(res[0][1] == "Integer");

         CPPUNIT_ASSERT(res[1].size() == 3);
         CPPUNIT_ASSERT(res[1][0] == "primfoo");
         CPPUNIT_ASSERT(res[1][1] == "Integer");
         CPPUNIT_ASSERT(res[1][2] == "PRIMARYKEY");
      }
      
      void testAlterAddBase()
      {
         cout << "**testAlterAddBase**" << endl;
         
         ConnectionManager::Connection dbconn;         
         FieldListtype fields;
         fields.push_back("primfoo");

         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);
         
         FieldListtype emptyFields;
         FieldListDatatype emptyFieldTypes;
         dbconn.DBCreate("testConnectionTestDomainDerived", 
                         "testConnectionTestDomain",
                         emptyFields, emptyFieldTypes, false);

         dbconn.DBAlter("testConnectionTestDomain", "bar", FLOATTYPE);         

         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomainDerived");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 2);
         CPPUNIT_ASSERT(res[0][0] == "bar");
         CPPUNIT_ASSERT(res[0][1] == "Float");

         CPPUNIT_ASSERT(res[1].size() == 3);
         CPPUNIT_ASSERT(res[1][0] == "primfoo");
         CPPUNIT_ASSERT(res[1][1] == "Integer");
         CPPUNIT_ASSERT(res[1][2] == "PRIMARYKEY");
      }

      void testAlterRemove()
      {
         cout << "**testAlterRemove**" << endl;
         
         ConnectionManager::Connection dbconn;         
         FieldListtype fields;
         fields.push_back("primfoo");
         fields.push_back("bar");

         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         fieldTypes.push_back(FLOATTYPE);
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primfoo", false);

         dbconn.DBAlter("testConnectionTestDomain", "bar");
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomain");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 3);
         CPPUNIT_ASSERT(res[0][0] == "primfoo");
         CPPUNIT_ASSERT(res[0][1] == "Integer");
         CPPUNIT_ASSERT(res[0][2] == "PRIMARYKEY");
      }
      
      void testAlterRemoveDerived()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         
         FieldListtype moreFields;
         moreFields.push_back("bar");
         FieldListDatatype moreFieldTypes;
         moreFieldTypes.push_back(INTEGERTYPE);

         dbconn.DBCreate("testConnectionTestDomainDerived", "testConnectionTestDomain", 
                         moreFields, moreFieldTypes, false);

         dbconn.DBAlter("testConnectionTestDomainDerived", "bar");
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomainDerived");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 3);
         CPPUNIT_ASSERT(res[0][0] == "primfoo");
         CPPUNIT_ASSERT(res[0][1] == "Integer");
         CPPUNIT_ASSERT(res[0][2] == "PRIMARYKEY");
      }
      
      void testAlterRemoveNoSuchField()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);         
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         try
         {
            dbconn.DBAlter("testConnectionTestDomain", "bar");
            CPPUNIT_ASSERT(false);            
         }
         catch (VDbException &e)
         {
         }
         
      }

      void testAlterRemovePrimaryKey()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);         
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);

         try
         {                     
            dbconn.DBAlter("testConnectionTestDomain", "primFoo");         
            CPPUNIT_ASSERT(false);
         }
         catch (VDbException &e)
         {
         }         

      }

      
      void testAlterRemoveBase()
      {
         ConnectionManager::Connection dbconn;
         
         FieldListtype fields;
         fields.push_back("primFoo");
         fields.push_back("bar");
         FieldListDatatype fieldTypes;
         fieldTypes.push_back(INTEGERTYPE);
         fieldTypes.push_back(STRINGTYPE);         
         
         dbconn.DBCreate("testConnectionTestDomain", fields, fieldTypes, "primFoo", false);
         
         FieldListtype moreFields;
         moreFields.push_back("baz");
         FieldListDatatype moreFieldTypes;
         moreFieldTypes.push_back(INTEGERTYPE);

         dbconn.DBCreate("testConnectionTestDomainDerived", "testConnectionTestDomain", 
                         moreFields, moreFieldTypes, false);

         dbconn.DBAlter("testConnectionTestDomain", "bar");
         
         ResultTabletype res = dbconn.DBDescribe("testConnectionTestDomain");
         cout << "# of rows in result:  " << res.size() << endl;
         static RowComparitor rowComp;
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 3);
         CPPUNIT_ASSERT(res[0][0] == "primfoo");
         CPPUNIT_ASSERT(res[0][1] == "Integer");
         CPPUNIT_ASSERT(res[0][2] == "PRIMARYKEY");

         res = dbconn.DBDescribe("testConnectionTestDomainDerived");
         sort(res.begin(), res.end(), rowComp);

         CPPUNIT_ASSERT(res[0].size() == 3);
         CPPUNIT_ASSERT(res[0][0] == "primfoo");
         CPPUNIT_ASSERT(res[0][1] == "Integer");
         CPPUNIT_ASSERT(res[0][2] == "PRIMARYKEY");

         CPPUNIT_ASSERT(res[1][0] == "baz");
         CPPUNIT_ASSERT(res[0][1] == "Integer");
      }
};

#define TEST(_method) \
   suite->addTest(new TestCaller<ConnectionTest>(#_method, \
                                                 &ConnectionTest::_method))


int main(int argc, char** argv)
{
   CommandLineParser args(argc, argv);
   Log::initialize(args.logType(), args.logLevel(), argv[0]);
   TestSuite* suite = new TestSuite;

   PSConfiguration::instance()->setConfigFile("dbcreate.cfg");
   PSConfiguration::instance()->parseConfigFile();

/*
   TEST(testVerify);
   TEST(testVerifyFail);

   TEST(testCreateExists);
   TEST(testCreateCheckExists);
   TEST(testCreateInherit);

   TEST(testRemoveBase);
   TEST(testRemoveCheckDoesNotExist);
   TEST(testDescribeDomain);

   TEST(testAlterAdd);
   TEST(testAlterAddDerived);
   TEST(testAlterAddBase);

   TEST(testAlterRemove);
   TEST(testAlterRemoveDerived);
*/
   TEST(testAlterRemoveBase);
/*
   TEST(testAlterRemoveNoSuchField);
   TEST(testAlterRemovePrimaryKey);   
*/ 
   TextTestResult result;
   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();
   DebugLog(<< "Finished");
   return 0;
}
