#include <boost/lexical_cast.hpp>
#include <boost/random.hpp>

#include <cppunit/TestCaller.h>
#include <cppunit/TestCase.h>
#include <cppunit/TestSuite.h>
#include <cppunit/TextTestResult.h>
#include <cppunit/TextTestRunner.h>
#include <strstream>
#include "CommandLineParser.hxx"
#include "Logger.hxx"
#include "VDbException.hxx"
#include "FastDBConnection.hxx"
#include "FastDBResultSet.hxx"


#define VOCAL_SUBSYSTEM Vocal::Subsystem::TEST

using namespace CppUnit;
using namespace Vocal;
using namespace Epoisses;


class FastDbTest : public TestCase
{
   public:

      FastDbTest()
      {}
        
      virtual ~FastDbTest()
      {}

      virtual void setUp()
      {
      }

      virtual void tearDown()
      {
      }
      
      class RowComparitor
      {
         public:
            bool operator()(const vector<string>& r1,
                            const vector<string>& r2)
            {
               return r1[0] < r2[0];
            }
      };


      void testInsertAndRetrieveAndDelete()
      {
         InfoLog( <<"*!testInsertAndRetrieve!*");
         FieldValuetype rec;
 
         rec["groupid"] = "'wombat'";
         rec["contact"] = "'wombat@foo.com'";
         rec["qvalue"] = "-1";
         rec["cseq"] = "1";
         rec["expires"] = "3600";
         rec["callid"] = "'9[u8q23nrqwaef'";
         
         int k = dbconn->DBInsert("registrationbindings", rec);
         DebugLog( << "Inserted Record with oid: " << k);
         

         string foo[] = { "bindingid", "groupid", "contact", "qvalue", "callid", "cseq", "expires" };
         FieldListtype flist(foo, foo+7);
         
         DBResultset* res = dbconn->DBSelect("registrationbindings", "qvalue=-1", flist);
         
         ResultTabletype rows = res->GetRecords();
         
         delete res;
         
         DebugLog(<< Inserter(rows));

         CPPUNIT_ASSERT((rows[0][1] == "wombat" )
                        && (rows[0][2] == "wombat@foo.com") 
                        && (rows[0][3] == "-1")
                        && (rows[0][4] == "9[u8q23nrqwaef")
                        && (rows[0][5] == "1")
                        && (rows[0][6] == "3600" ));
         
         dbconn->DBDelete("registrationbindings", "groupid='wombat'");
         
         DBResultset* empty = dbconn->DBSelect("registrationbindings", "groupid='wombat'", flist);
         
         CPPUNIT_ASSERT(empty->GetRecordCount() == 0);
      }

      void testUpdate()
      {
         string foo[] = { "bindingid", "groupid", "contact", "qvalue", "callid", "cseq", "expires" };

         FieldListtype flist(foo, foo+7);


         InfoLog( <<"*!testUpdate!*");
         FieldValuetype rec;
 
         rec["groupid"] = "'wombat'";
         rec["contact"] = "'wombat@foo.com'";
         rec["qvalue"] = "-1";
         rec["cseq"] = "1";
         rec["expires"] = "3600";
         rec["callid"] = "'9[u8q23nrqwaef'";

         int k = dbconn->DBInsert("registrationbindings", rec);
         DebugLog( << "Inserted Record with oid: " << k);
         
         FieldValuetype changes;

         changes["qvalue"] = "-8";
         
         dbconn->DBUpdate("registrationbindings", "groupid='wombat'", changes);

         DBResultset* res = dbconn->DBSelect("registrationbindings", "qvalue=-8", flist);
         ResultTabletype rows = res->GetRecords();         
         delete res;

         DebugLog(<< Inserter(rows));

         CPPUNIT_ASSERT((rows[0][1] == "wombat" )
                        && (rows[0][2] == "wombat@foo.com") 
                        && (rows[0][3] == "-8")
                        && (rows[0][4] == "9[u8q23nrqwaef")
                        && (rows[0][5] == "1")
                        && (rows[0][6] == "3600" ));
      }
      
      void testInsertAndRetrieveByOid()
      {
         InfoLog( <<"*!testInsertAndRetrieveByOid!*");
         FieldValuetype rec;
         
         rec["groupid"] = "'wombat'";
         rec["contact"] = "'wombat@foo.com'";
         rec["qvalue"] = "-1";
         rec["cseq"] = "1";
         rec["expires"] = "3600";
         rec["callid"] = "'9[u8q23nrqwaef'";
         
         uint k = dbconn->DBInsert("registrationbindings", rec);
         DebugLog( << "Inserted Record with oid: " << k);

         string foo[] = { "bindingid", "groupid", "contact", "qvalue", "callid", "cseq", "expires" };
         FieldListtype flist(foo, foo+7);
         
         DBResultset* res = dbconn->DBRetrieveByOid("registrationbindings", k, flist);
         
         ResultTabletype rows = res->GetRecords();
         
         delete res;
         
         DebugLog(<< Inserter(rows));

         CPPUNIT_ASSERT((rows[0][1] == "wombat" )
                        && (rows[0][2] == "wombat@foo.com") 
                        && (rows[0][3] == "-1")
                        && (rows[0][4] == "9[u8q23nrqwaef")
                        && (rows[0][5] == "1")
                        && (rows[0][6] == "3600" ));
         
         dbconn->DBDelete("registrationbindings", "groupid='wombat'");
         
         DBResultset* empty = dbconn->DBSelect("registrationbindings", "groupid='wombat'", flist);
         
         CPPUNIT_ASSERT(empty->GetRecordCount() == 0);
      }

      void testSelectFakeColumns()
      {
         InfoLog( <<"*!testSelectFakeColumns!*");
         FieldValuetype rec;
         
         rec["groupid"] = "'wombat'";
         rec["contact"] = "'wombat@foo.com'";
         rec["qvalue"] = "-1";
         rec["cseq"] = "1";
         rec["expires"] = "3600";
         rec["callid"] = "'9[u8q23nrqwaef'";
         
         uint k = dbconn->DBInsert("registrationbindings", rec);
         DebugLog( << "Inserted Record with oid: " << k);

         string foo[] = { "'foo'", "contact", "qvalue", "callid", "cseq", "expires" };
         FieldListtype flist(foo, foo+6);
         
         DebugLog(<< "Retrieving with fieldlist: " << Inserter(flist));
         DBResultset* res = dbconn->DBRetrieveByOid("registrationbindings", k, flist);
         
         ResultTabletype rows = res->GetRecords();
         
         delete res;
         
         DebugLog(<< Inserter(rows));

         CPPUNIT_ASSERT((rows[0][0] == "foo" )
                        && (rows[0][1] == "wombat@foo.com") 
                        && (rows[0][2] == "-1")
                        && (rows[0][3] == "9[u8q23nrqwaef")
                        && (rows[0][4] == "1")
                        && (rows[0][5] == "3600" ));
         
         dbconn->DBDelete("registrationbindings", "groupid='wombat'");
         
         DBResultset* empty = dbconn->DBSelect("registrationbindings", "groupid='wombat'", flist);
         
         CPPUNIT_ASSERT(empty->GetRecordCount() == 0);
      }
      
      static void CreateStatic()
      {
         DebugLog(<< "creating database" );
         
         dbconn = FastDBConnection::instance();
      }

      void testAndSelect()
      {
         InfoLog( <<"*!testAndSelect!*");
         FieldValuetype rec;
 
         rec["groupid"] = "'wombat'";
         rec["contact"] = "'wombat@foo.com'";
         rec["qvalue"] = "-1";
         rec["cseq"] = "1";
         rec["expires"] = "3600";
         rec["callid"] = "'9[u8q23nrqwaef'";
         
         int k = dbconn->DBInsert("registrationbindings", rec);
         DebugLog( << "Inserted Record with oid: " << k);
         

         string foo[] = { "bindingid", "groupid", "contact", "qvalue", "callid", "cseq", "expires" };
         FieldListtype flist(foo, foo+7);
         
         DBResultset* res = dbconn->DBSelect("registrationbindings", 
                                             "((groupid = 'wombat') and (contact = 'wombat@foo.com'))", flist);
         
         ResultTabletype rows = res->GetRecords();
         
         delete res;
         
         DebugLog(<< Inserter(rows));

         CPPUNIT_ASSERT((rows[0][1] == "wombat" )
                        && (rows[0][2] == "wombat@foo.com") 
                        && (rows[0][3] == "-1")
                        && (rows[0][4] == "9[u8q23nrqwaef")
                        && (rows[0][5] == "1")
                        && (rows[0][6] == "3600" ));
         
         dbconn->DBDelete("registrationbindings", "groupid='wombat'");
         
         DBResultset* empty = dbconn->DBSelect("registrationbindings", "groupid='wombat'", flist);
         
         CPPUNIT_ASSERT(empty->GetRecordCount() == 0);
      }
      
      static FastDBConnection* dbconn;
      
};


FastDBConnection* FastDbTest::dbconn = 0;
      

#define TEST(_method) \
   suite->addTest(new TestCaller<FastDbTest>(#_method, \
                                             &FastDbTest::_method))


int main(int argc, char** argv)
{
   CommandLineParser args(argc, argv);
   Log::initialize(args.logType(), args.logLevel(), argv[0]);
   TestSuite* suite = new TestSuite;

   FastDbTest::CreateStatic();

//   TEST(testInsertAndRetrieveAndDelete);
//   TEST(testUpdate);
//   TEST(testInsertAndRetrieveByOid);
//   TEST(testSelectFakeColumns);
   TEST(testAndSelect);

   TextTestResult result;
   TextTestRunner runner;
   runner.addTest(suite);
   runner.run();
   DebugLog(<< "Finished");

   return 0;
}
