// Copyright (c) 2002 Cathay Networks, Inc. All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
// 
// 3. The names "VOCAL", "Vovida Open Communication Application Library", and
// "Vovida Open Communication Application Library (VOCAL)" must not be used to
// endorse or promote products derived from this software without prior written
// permission. For written permission, please contact vocal@vovida.org.
// 
// 4. Products derived from this software may not be called "VOCAL", nor may
// "VOCAL" appear in their name, without prior written permission.
// 
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT ARE DISCLAIMED.
// IN NO EVENT SHALL CATHAY NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR
// ANY DAMAGES IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
// OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
// OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef RefPtr_hxx
#define RefPtr_hxx

namespace Epoisses
{

/**
   Smart poiner that behaves like a value and its references. Like references,
   copies of the original RefPtr instance may be used only within the scope
   of the original. When the original goes out of scope the wrapped pointer is
   deleted and all copies are implicitly invalidated. It is undefined to
   dereference an out-of-scope copy of a RefPtr instance.
*/
template <class T>
class RefPtr
{
   public:
      explicit RefPtr(T* p)
         : _original(true),
           _p(p)
      {}

      RefPtr()
         : _original(false),
           _p(0)
      {}

      RefPtr(RefPtr& sap)
         : _original(false),
           _p(sap._p)
      {}

      // allow pointer like automatic casts
      template<typename U>
      RefPtr(const RefPtr<U>& u)
         : _original(false),
           _p(u.get())
      {}

      RefPtr<T>
      operator=(const RefPtr<T>& rp)
      {
         if (_p != rp._p)
         {
            if (_original)
            {
               delete _p;
            }
            _original = false;
            _p = rp._p;
         }
         return *this;
      }

      virtual ~RefPtr()
      {
         if (_original)
         {
            delete _p;
         }
      }

      T& operator*() const
      {
         assert(_p);
         return *_p;
      }

      T* operator->() const
      {
         assert(_p);
         return _p;
      }

      operator bool() const
      {
         return _p != 0;
      }

      bool operator!() const
      {
         return _p == 0;
      }

      bool 
      operator==(const RefPtr& sap) const
      {
         return _p == sap._p;
      }

      template<typename Y, typename U>  friend
      bool operator==(const RefPtr<Y> & , const RefPtr<U>&);

      template<typename Y, typename U> friend
      bool operator!=(const RefPtr<Y> & , const RefPtr<U>&);

      template<typename Y, typename U> friend
      bool operator<(const RefPtr<Y> & , const RefPtr<U>&);

      T* get() const { return _p; }

   private:
      bool _original;
      T* _p;
};

template<typename T, typename U> 
inline bool 
operator==(const RefPtr<T>& a, const RefPtr<U>& b)
{
   return a._p == b._p;
}

template<typename T, typename U>
inline bool
operator!=(const RefPtr<T>& a, const RefPtr<U>& b)
{
   return a._p != b._p;
}

template<typename T, typename U>
inline bool
operator<(const RefPtr<T>& a, const RefPtr<U>& b)
{
   return a._p < b._p;
}

}

#endif
