#ifndef CPLOpProxy_Hxx
#define CPLOpProxy_Hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const CPLOpProxyVersion =
    "$Id: CPLOpProxy.hxx,v 1.23 2002/11/09 02:05:37 derekm Exp $";



#include <SipEvent.hxx>
#include <InviteMsg.hxx>
#include <StatusMsg.hxx>
#include <AckMsg.hxx>
#include <ByeMsg.hxx>
#include <CancelMsg.hxx>
#include <OptionsMsg.hxx>
#include <InfoMsg.hxx>
#include <TransferMsg.hxx>
#include <SipVia.hxx>
#include <SipReceive.hxx>
#include "CPLOperator.hxx"

namespace Vocal
{


/**
  *  Operator to do the default handling of most SIP events. 
  */
class CPLOpProxy: public CPLOperator
{
    public:
        /**
         * Defualt constructor
         */
        CPLOpProxy();

        /**
         * Destructor
         */
        ~CPLOpProxy();

        /**
         * Execute the operator for the state m/c
         *@param event Event from the Sip stack
         *@return Pointer to next state or 0 if there is no transition
         */
        const Sptr < State > process( const Sptr < SipProxyEvent > event );

        /**
         * Save CPL tag attributes
         *@param attrName Attribute's name
         *@param attrValue Value of the attribute
         *@return bool flag if successful in setting the attribute
         */
        bool setAttributes( const char* attrName, const char* attrValue );

        /**
         * Add special operators
         *@param op Pointer to the operator to add
         *@param prevNode ID of the last CPL tag
         *@return bool flag if successful
         */
        bool setNextOperator( Sptr < Operator > op, int prevNode );

        /**
         * Return the class name
         *@return Pointer to char string
         */
        const char* const name() const;

    protected:
        /**
         *Display information regarding the class's member variables
         */
        void logData() const;

    private:
        /**@name private methods */
        //@{
        /**
         * method to process a Sip Message
         * @param sipEvent A SIP event
         * @param msg reference to SipMsg to be processed
         * @return next state
         */
        Sptr < State > processMsg(Sptr < SipEvent > sipEvent, 
                                  Sptr < SipMsg > msg );

        /**
         * method to process an INVITE request
         * @param sipEvent A SIP event
         * @param inviteMsg reference to INVITE to be processed
         */
        void processInvite(Sptr < SipEvent > sipEvent, 
                           Sptr < InviteMsg > inviteMsg );

        /**
         * method to process a Sip Response message
         * @param sipEvent A SIP event
         * @param statusMsg reference to Sip Response to be processed
         */
        void processStatus(Sptr < SipEvent > sipEvent, 
                           Sptr < StatusMsg > statusMsg );

        /**
         * method to process an ACK request
         * @param sipEvent A SIP event
         * @param ackMsg reference to ACK request to be processed
         */
        void processAck(Sptr < SipEvent > sipEvent, Sptr < AckMsg > ackMsg );

        /**
         * method to process CANCEL request
         * @param sipEvent A SIP event
         * @param cancelMsg reference to CANCEL to be processed
         */
        void processCancel(Sptr < SipEvent > sipEvent, 
                           Sptr < CancelMsg > cancelMsg );

        /**
         * method to process BYE request, when recieved before a 200/OK
         * @param sipEvent A SIP event
         * @param byeMsg reference to BYE to be processed
         */
        void processByeAsCancel(Sptr < SipEvent > sipEvent, 
                                Sptr < ByeMsg > byeMsg );

        /**
         * method to proxy a request
         * @param sipEvent A SIP event
         * @param request reference to the SIP command to be proxied
         */
        void proxyRequest(Sptr < SipEvent > sipEvent, SipCommand& request );

        /**
         * method to proxy ACK request based on proxied INVITE
         * @param sipEvent A SIP event
         * @param statusMsg reference to Sip Response to be Acked
         */
        void ackBasedOnInvite(Sptr < SipEvent > sipEvent, 
                              Sptr < StatusMsg > statusMsg );

        /**
         * method to restore/remove FROM field for upstream/downstream
         * messages
         * @param sipEvent A SIP event
         */
        void restoreOrRemoveCallerId(Sptr < SipEvent > sipEvent);
        /**
         * method to restore/remove Prefix for upstream/downstream messages
         * @param sipEvent A SIP event
         * @prefix The feature code
         */
        void restoreOrRemovePrefix(Sptr < SipEvent > sipEvent, Data prefix);
        /**
         * method to restore Caller-ID for upstream responses
         * @param sipEvent A SIP event
         * @param statusMsg reference to the Response to be sent upstream
         */
        void restoreCallerIdForReply(Sptr < SipEvent > sipEvent, StatusMsg& statusMsg);
        /**
         * method to lookup for any pending Request for a given Request
         * @param sipEvent A SIP event
         * @param sipCommand reference to the Sip Request to use for lookup
         * @return reference to the pending sip Request
         */
        const Sptr < SipCommand > findPendingCommand(Sptr < SipEvent > sipEvent, Sptr < SipCommand > sipCommand ) const;
        //@}
};

 
} // namespace Vocal

#endif

//End of file

