#ifndef CPLOperator_Hxx
#define CPLOperator_Hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const CPLOperatorVersion =
    "$Id: CPLOperator.hxx,v 1.23 2002/11/09 02:05:38 derekm Exp $";



#include <list>
#include "Operator.hxx"
#include "CallProcessingLanguage.hxx"
#include "CPLState.hxx"
#include "SipTransceiver.hxx"
#include "InviteMsg.hxx"
#include "FsStateMachnData.hxx"

namespace Vocal
{

/**
  * Used as part of the CPL feature state machine.
  */
class CPLOperator: public Operator
{
    public:
        ///Constructor
        CPLOperator();

        ///Destructor
        ~CPLOperator();

        ///Clear All Operators
        void clearAllOperators();

        /** Save CPL tag attributes
          *@param attrName Attribute's name
          *@param attrValue Value of the attribute 
          *@return True if successful setting attributes */
        virtual bool setAttributes( const char* attrName,
                                    const char* attrValue ) = 0;

        /**Execute the operator for the state machine
          *@param event Event from the Sip stack 
          *@return Pointer to next state or 0 if the operator did not need to
          *process the event. 
          */
        virtual const Sptr < State > process( const Sptr < SipProxyEvent > event );

        /**Set-up timer to process time-outs if there's no response
          *@param event Event from the Sip stack
          *@param timeout the maximum time to wait (seconds)
          *@return True if timer set successfully */
        bool setTimer( Sptr < SipProxyEvent > event, unsigned int timeout ) const;

        /**Cancel timer before it times-out
          *@param event Event from the Sip stack
          *@return True if timer cancelled successfully */
        bool cancelTimer( Sptr < SipProxyEvent > event ) const;

        /**Push operator to end of class' list of operators
          *@param op Pointer to operator to be added */
        void addOperator( Sptr < Operator > op);

        /**Set the state this operator will go to after doing the process().
          *@param nextState Pointer to the next CPL state
          */
        void setState( Sptr < CPLState > nextState );

        /**Add special operators
           @param op Pointer to the operator to add
           @param prevNode ID of the last CPL tag.
           @return true if function did not have any errors setting next
            operator. 
        */
        virtual bool setNextOperator( Sptr < Operator > op,
                                      int prevNode = CPLNodeEndofNodes ) = 0;

        /**Return the class name.
          @return Pointer to character string */
        virtual const char* const name() const;

        /**Displays class data information including calling doCpLog of any of
          * its operators and states that it contains.  */
        virtual void doCpLog() const;
        //@}

        /**@name Iterators */
        //@{
        ///Iterator for the list of operators
        typedef list < Sptr < Operator > > ::iterator myOpIterator;

        ///Const iterator for the list of operators
        typedef list < Sptr < Operator > > ::const_iterator myOpConstIterator;
        //@}

    protected:
        /**@name Member Variables */
        //@{
        ///Traverse to this next state after processing is done in this operator.
        Sptr < CPLState > myNextState;

        ///Holds 0 or more operators
        list < Sptr < Operator > > myOperators;
        //@}

        /**@name Methods */
        //@{
        ///Display information regarding the class' member variables.
        virtual void logData() const = 0;
        //@}

        /**Send an ACK to a Status message.
         * @param sipEvent A SIP event (useful to retrieve SIP Stack)
         * @param statusMsg A SIP status message
         **/
        void ackStatusMsg( Sptr<SipProxyEvent> event );

        /**Add a Call Control Diversion header (cc-diversion) to the invite message
          * based upon the proxy Status stored in Call Info. 
          *@param event Event from the Sip stack 
          *@param sipMsg The message that will get new cc-diversion. 
          *@param baseUrl The Request URL before redirection.  
          *@return True if header is successfully added        
          */
        void addCallControlDiversion( Sptr < SipProxyEvent > event, 
                                      Sptr < SipMsg > sipMsg,
                                      Sptr < BaseUrl > baseUrl ) const;

        /**Retrieve the CallInfo from the event regardless of SipEvent or Timer
          * Event.
          *@param event Event from the Sip stack 
          *@return Pointer to the CallInfo        
          */
        Sptr < CallInfo > getCallInfo( Sptr < SipProxyEvent > event ) const;

        /**Retrieve the Contact URL from a Status message
          * @param statusMsg A Sip status message
          * @return The URL from the Status message's CONTACT field.
          **/
        Sptr<BaseUrl> getContactUrl( Sptr<SipProxyEvent> event ) const;

        /**Retrieve a InviteMsg from the event regardless of SipEvent or Timer
          * Event.
          *@param event Event from the Sip stack 
          *@return Pointer to an INVITE message
          */
        Sptr < InviteMsg > getInvite( Sptr < SipProxyEvent > event ) const;

        /**Retrieve a SipCommand from the event regardless of SipEvent or Timer
          * Event.
          *@param event Event from the Sip stack 
          *@return Pointer to a SIP command 
          */
        Sptr < SipCommand > getSipCommand( Sptr < SipProxyEvent > event ) const;

        /**Retrieve a SipMsg from the event regardless of SipEvent or Timer
          * Event.
          *@param event Event from the Sip stack 
          *@return Pointer to a SIP Msg
          */
        Sptr < SipMsg > getSipMsg( Sptr < SipProxyEvent > event ) const;

        /**Retrieve the SIP stack from event regardless of SipEvent or Timer
          * Event.
          *@param event Event from the Sip stack 
          *@return Pointer to the SIP stack 
          */
        Sptr < SipTransceiver > getStack( Sptr < SipProxyEvent > event ) const;

        /**Retrieve the FsStateMachnData from the CallInfo
          *@param event Event from the Sip stack 
          *@return Pointer to a FsStateMachnData 
          */
        Sptr < FsStateMachnData > getStateMachineData( Sptr < SipProxyEvent > event ) const;

        /** Send a reply
          *@param event An Event retrieved from the SIP stack.
          *@param sipMessage the Message to send
          */
        void sendReply( Sptr < SipProxyEvent > event, StatusMsg& sipMessage ) const;

        /** Send message Asynchronously
          *@param event An Event retrieved from the SIP stack.
          *@param sipCommand the Message to send
          */
        void sendAsync( Sptr < SipProxyEvent > event, SipCommand& sipCommand ) const;

    private:

};
 
}


#endif



