
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const CPLState_cxx_Version =
    "$Id: CPLState.cxx,v 1.24 2002/12/27 00:57:20 sprajpat Exp $";

#include <CallInfo.hxx>
#include "CPLState.hxx"
#include "CPLOperator.hxx"

using namespace Vocal;

///
CPLState::CPLState( const char *name )
{

    int bufSize = (strlen( name ) + 1) * sizeof(char);
    myStateName = (char*)malloc( bufSize );
    strcpy( myStateName, name );

#ifdef MEM_TEST
    cpLog( LOG_DEBUG, "new %s[%x]", myStateName, this );
#endif
} //CPLState


///
CPLState::CPLState()
{
    //This constructor should not be called.
    //Use the constructor that has a char* parameter.
    assert( 0 );

} //CPLState


///
CPLState::~CPLState()
{

#ifdef MEM_TEST
    cpLog( LOG_DEBUG, "~%s[%x]", myStateName, this );
#endif
    free( myStateName );
} //~CPLState


///
void
CPLState::addOperator( const Sptr < Operator > op )
{
    if ( op == 0 )
    {
        cpLog( LOG_ERR, "op equalled 0." );
        return;
    }  /*if*/

    if ( myOperators.size() == 0 )
    {
        State::addOperator( op );
        return;
    } /*if*/

    OperatorIter it = myOperators.begin();
    OperatorIter endOfList = myOperators.end();

    //There's an expected order of the operators in
    //the list.  CPLOpEnd is always the last operator
    //and CPLOpProxy is the penultimate operator.
    //All other operators are added before CPLOpProxy.
    if ( strcmp( op->name(), "CPLOpProxy" ) == 0 )
    {
        //Insert this operator before CPLOpEnd
        while ( it != endOfList )
        {
            if ( strcmp( (*it)->name(), "CPLOpEnd") == 0 )
            {
                myOperators.insert( it, op );
                return;
            } /*if*/
            ++it;
        } /*while*/
    } /*if*/
    else if ( strcmp( op->name(), "CPLOpEnd" ) == 0 )
    {
        //Insert this operator after CPLOpProxy
        while ( it != endOfList )
        {
            if ( strcmp( (*it)->name(), "CPLOpProxy") == 0 )
            {
                myOperators.insert( ++it, op );
                return;
            } /*if*/
            ++it;
        } /*while*/
    } /*if*/
    else
    {
        //Insert this operator before CPLOpProxy
        while ( it != endOfList )
        {
            if ( ( strcmp( (*it)->name(), "CPLOpProxy") == 0 ) || 
                 ( strcmp( (*it)->name(), "CPLOpEnd") == 0 ) ) 
            {
                myOperators.insert( it, op );
                return;
            } /*if*/
            ++it;
        } /*while*/
    } /*else*/

    //If we reach here, then just push op to the back.
    State::addOperator( op );
    return;
}  //addOperator


///
bool
CPLState::setNextOperator( Sptr < Operator > op, int prevNode )
{
    addOperator( op );

    return true;
}  //setNextOperator


///
void
CPLState::clearAllOperators()
{
    State::clearAll();
}  //clearAllOperators


///
Sptr < State >
CPLState::process( const Sptr < SipProxyEvent > event)
{
    Sptr < State > nextState;

    cpLog( LOG_DEBUG, "%s::process()", myStateName );

    doCpLog();

    //State must have operators
    if ( myOperators.size() == 0 )
    {
        cpLog( LOG_ERR, "No Ops to do in this state." );
        return 0;
    } /*if*/

    for ( OperatorIter iter = myOperators.begin();
            iter != myOperators.end(); iter++)
    {
        nextState = (*iter)->process( event );

        if ( nextState != 0 )
        {
            cpLog( LOG_DEBUG, "Next State=%s", nextState->name() );
            Sptr < CallInfo > call = event->getCallInfo();
            if ( call != 0 )
            {
                call->setState( nextState );
            }
            else assert( 0 );
            break;
        } /*if*/
        else
            cpLog( LOG_DEBUG, "%s returned State_0.", (*iter)->name() );

    } /*for*/

    assert (nextState != 0 );
    return nextState;

}  //process


///
const char*  const
CPLState::name() const
{
    if ( strlen( myStateName ) == 0 )
        return "CPLState";
    else
        return myStateName;
}  //name


///
void
CPLState::logData() const
{
    OperatorContainer::const_iterator it = myOperators.begin();
    OperatorContainer::const_iterator listEnd = myOperators.end();

    while (it != listEnd)
    {
        if ( myOperators.size() > 1 )
            cpLog( LOG_DEBUG, "%s - Operator", name() );
        ++it;
    }  /*while*/
} //logData


///
void
CPLState::doCpLog() const
{
    if ( myOperators.size() == 0 )
    {
        cpLog( LOG_DEBUG, "%s has 0 ops.", myStateName );
        return ;
    } /*if*/

    OperatorContainer::const_iterator it = myOperators.begin();
    OperatorContainer::const_iterator listEnd = myOperators.end();

    //Print out all operator names first
    string operatorsList;
    while (it != listEnd)
    {
        operatorsList += (*it)->name();
        operatorsList += " ";
        ++it;
    }  /*while*/
    cpLog ( LOG_DEBUG, "%s has ops( %s)", name(), operatorsList.c_str() );

    //Now call each operator's doCpLog() except for Proxy & End
    it = myOperators.begin();
    while (it != listEnd )
    {
        //Skip CPLOpProxy or there'll be an infinite loop
        if ((strcmp( (*it)->name(), "CPLOpEnd" ) != 0 ) &&
                ( strcmp( (*it)->name(), "CPLOpProxy" ) != 0) )
        {
            cpLog( LOG_DEBUG, "%s's operator follows:", name() );
            (*it)->doCpLog();
        } /*if*/
        else
        {
            cpLog( LOG_DEBUG, "%'s op-%s", name(), (*it)->name() );
        } /*else*/
        ++it;
    } /*for*/

} //doCpLog

//End of file

