/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package com.vovida.telephony;

import javax.telephony.*;
import javax.telephony.capabilities.*;
import com.vovida.telephony.*;
import com.vovida.telephony.events.*;
import com.vovida.telephony.capabilities.*;
import java.util.*;

/*******************************************************/
/* VCConnection is the Vocal Client Connection Class  */
/* which is an implementation of the connection        */
/* interface from the Jtpi spec.                       */
/*******************************************************/

/**
 * VCConnection is the Vocal Client Connection Class which is an
 * implementation of the connection interface from the Jtapi spec.
 */

public class VCConnection implements Connection
{
    protected String name;
    protected int state;
    protected Address address;
    protected Call call;
    protected CallEventNotifier callEvtNotifier;
    protected Vector terminalConnections;
    protected VCDynamicConnectionCapabilities connectionCapabilities;


    /**
     * private default constructor
     */
    private VCConnection() {}

    /**
     * Constructor
     */
    protected VCConnection (Address connAddr, Call connCall)
    {
	    try
	    {
	        name = connAddr.getName();
	        address = connAddr;
	        call = connCall;
	        connectionCapabilities = new VCDynamicConnectionCapabilities(this);
	        terminalConnections = new Vector();
	        ((VCAddress)connAddr).addConnection(this);
	        ((VCCall)connCall).addConnection(this);
            callEvtNotifier = ((VCCall)call).getCallEvtNotifier();
            setState(IDLE);
	    }
	    catch (InvalidStateException e)
	    {
	        e.printStackTrace();
 	    }
    }

/********************************************************/
/* JTapi Implementation                                 */
/********************************************************/

    // return the state of the connection
    /**
     * Refer To JTapi Connection Interface
     */
    public int getState()
    {
	    return state;
    }

    // return the call associated with this connection
    /**
     * Refer To JTapi Connection Interface
     */
    public Call getCall()
    {
	    return call;
    }

    // return the address associated with this connection
    /**
     * Refer To JTapi Connection Interface
     */
    public Address getAddress()
    {
	    return address;
    }

    // return the array of terminalconnections associated with this 
    // connection
    /**
     * Refer To JTapi Connection Interface
     */
    public TerminalConnection[] getTerminalConnections()
    {
	    TerminalConnection[] terminalConnectionsArray = null;
	    if (terminalConnections.size() > 0)
	    {
	        terminalConnectionsArray = new TerminalConnection[terminalConnections.size()];
	        terminalConnections.copyInto(terminalConnectionsArray);
	    }
	    return terminalConnectionsArray;
    }

    // disconnect this connection from a call
    // the connection must be in a certain state to be disconnected
    // the associated terminalconnections should be set to a DROPPED
    // state and if the call is no longer active (less than 2 connections
    // are connected) then the all connections are removed from the call

    /**
     * Refer To JTapi Connection Interface
     */
    public void disconnect() throws PrivilegeViolationException,
	ResourceUnavailableException, MethodNotSupportedException,
	InvalidStateException
    {
	    isProviderOK();
	    if ((state != Connection.CONNECTED) && (state != Connection.ALERTING)
	        && (state != Connection.INPROGRESS) && (state != Connection.FAILED))
	    {
	        throw (new InvalidStateException(this,
			    InvalidStateException.CONNECTION_OBJECT, state));
	    }

	    if (VCRequestUtil.callDisconnect(this))
	    {
	        setState(DISCONNECTED);
	    }
	    else
	    {
	        setState(FAILED);
	        return;
	    }
	
	    if (getTerminalConnections() != null)
	    {
	        TerminalConnection[] tcs = getTerminalConnections();
	        for (int i=0; i < tcs.length; i++)
	        {
		        ((VCTerminalConnection)tcs[i]).setState(TerminalConnection.DROPPED);
	        }
	    }	

	    if (call.getConnections() != null)
	    {
	        if (call.getConnections().length < 2)
	        {
	            // remove all connections from Call
	            ((VCCall)call).closeConnections();
	        }
	    }
    }

    // return the dynamic capabilities of the connection
    /**
     * Refer To JTapi Connection Interface
     */
    public ConnectionCapabilities getCapabilities()
    {
	    return connectionCapabilities;
    }

    // as above, however this function format is deprecated
    /**
     * 
     * Refer To JTapi Connection Interface
     * @deprecated
     */
    public ConnectionCapabilities getConnectionCapabilities(Terminal terminal, Address address)
	    throws InvalidArgumentException, PlatformException
    {
	    return null;
    }


/**************************************************************/
/* Additional Functionality                                   */
/**************************************************************/

    /**
     * add a terminalconnection to the list of terminalconnections 
     * associated with this call.
     */
    protected void addTerminalConnection(TerminalConnection termConn)
    {
	    terminalConnections.addElement(termConn);
    }

    /**
     * remove a terminalconnection from the list of terminalconnections 
     * associated with this call.
     */
    protected void removeTerminalConnection(TerminalConnection termConn)
    {
	    terminalConnections.removeElement(termConn);
    }


    /**
     * check that the provider is In Service
     */
    protected void isProviderOK() throws InvalidStateException
    {
        if (this.getCall().getProvider().getState() != Provider.IN_SERVICE)
        {
            throw (new InvalidStateException(this.getCall().getProvider(),
				InvalidStateException.PROVIDER_OBJECT,
				this.getCall().getProvider().getState()));
        }
        return;
    }


    /**
     * Set the state of the connection
     * if the new state is the same as the current state then there
     * is no need to create an event so return from this function
     * The new state is checked against the old one to ensure
     * that the transition is a valid one
     * Otherwise create the appropriate event and set the new state
     */

    public void setState(int newState) throws InvalidStateException
    {
	System.out.println("CONNECTION SET STATE " + newState);
	// do not want to set the same state again, causing more events
	if (newState != IDLE)
	{
	    if (state == newState)
		return;
	}
	// INCLUDE SOME EVT HANDLING
	switch (newState)
	{
	    case IDLE:
		{
		    callEvtNotifier.notifyObserversOfEvent(
					new VCConnCreatedEv(this));
		}
		break;
	    case INPROGRESS:
		{
		    if (state == IDLE)
		    {
		        callEvtNotifier.notifyObserversOfEvent(
					new VCConnInProgressEv(this));
		    }
		    else
		    {
			throw (new InvalidStateException(this,
			    InvalidStateException.CONNECTION_OBJECT,
			    state));
		    }
		}
		break;
	    case ALERTING:
		{
		    if ((state == IDLE) || (state == INPROGRESS))
		    {
		        callEvtNotifier.notifyObserversOfEvent(
					new VCConnAlertingEv(this));
		    }
		    else
		    {
			throw (new InvalidStateException(this,
                            InvalidStateException.CONNECTION_OBJECT,
                            state));
		    }
		}
		break;
	    case CONNECTED:
		{
		    if ((state == IDLE) || (state == INPROGRESS) || (state == ALERTING))
		    {
		        callEvtNotifier.notifyObserversOfEvent(
					new VCConnConnectedEv(this));
		    }
		    else
		    {
			throw (new InvalidStateException(this,
                            InvalidStateException.CONNECTION_OBJECT,
                            state));
		    }
		}
		break;
	    case DISCONNECTED:
		{
		    VCCall connCall = (VCCall)getCall();
		    ((VCAddress)getAddress()).removeConnection(this);
		    connCall.removeConnection(this);
		    Connection[] conns = connCall.getConnections();
		    if (conns != null)
		    {
		        if (conns.length == 1)
			{
		            connCall.closeConnections();
			}
		    }
		    
		    if ((state == IDLE) || (state == INPROGRESS) || 
			(state == ALERTING) || (state == FAILED) || 
			(state == CONNECTED))
                    {
		        callEvtNotifier.notifyObserversOfEvent(
					new VCConnDisconnectedEv(this));
		    }
		    else
		    {
			throw (new InvalidStateException(this,
                            InvalidStateException.CONNECTION_OBJECT,
                            state));
		    }
		}
		break;
	    case FAILED:
		{
		    if ((state == IDLE) || (state == INPROGRESS) || 
                        (state == ALERTING) || (state == FAILED))
		    {
		        callEvtNotifier.notifyObserversOfEvent(
					new VCConnFailedEv(this));
		    }
		    else
		    {
			throw (new InvalidStateException(this,
                            InvalidStateException.CONNECTION_OBJECT,
                            state));
		    }
		}
		break;
	    case UNKNOWN:
		{
		    callEvtNotifier.notifyObserversOfEvent(
					new VCConnUnknownEv(this));
		}
        default:
        {
        }
        break;
	}
	state = newState;

    }

}



