/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package com.vovida.telephony;

import java.util.*;

/**
 * The VCEvent class holds data which is received
 * in a message received at the JTapi client
 * The data is parsed from the incoming xml
 * format message and stored to the data members of
 * this class It is then possible to read each item
 * of data individually from an instance of this
 * class Instances are actually members of one of the
 * subclass RequestEvent, ResponseEvent, 
 * Notificationevent
 */

abstract class VCEvent
{
    final static protected int REQUESTEVENT = 1;
    final static protected int RESPONSEEVENT = 2;
    final static protected int NOTIFICATIONEVENT = 3;
    final static protected int TIMEEVENT = 5;

    protected String msg;
    protected int evtType;
    protected String msgId;
    protected int msgType;
    protected String callId;
    protected String[] addrs;
    protected String addr1;
    protected String addr2;
    protected String terminal;
    protected int value;


    /**
     * Constructor - create event from string received from jtapi server
     */
    public VCEvent(String msg)
    {
	// Set all event values to default values
	evtType = 0;
	msgId = null;
	msgType = 0;
	callId = null;
        addrs = null;
	addr1 = null;
	addr2 = null;
	terminal = null;
	value = 0;
	this.msg = msg;
	if (msg != null)
	{
	    parseMsg();
	}
    }

    /**
     * virtual function implemented in subclass
     */
    abstract public int getEvtType();

    /**
     * Parse msg from jtapi server and store parameters in Event object
     */
    protected void parseMsg()
    {
        msgId = tagValue("<id>", msg);
//	if (str != null)
//	    msgId = (Long.parseLong(str)).toString();
	String str = tagValue("<type>", msg);
	if (str != null)
	    msgType = Integer.parseInt(str);
	callId = tagValue("<call>", msg);
	addrs = tagValues("<addr>", msg);
	if (addrs != null)
	{
	    if (addrs.length > 0)
	        addr1 = addrs[0];
	    if (addrs.length > 1)
	        addr2 = addrs[1];
	}
	terminal = tagValue("<term>", msg);
	str = tagValue("<value>", msg);
	if (str != null)
	    value = Integer.parseInt(str);
    }

    /**
     * Get Msg Id
     */
    protected String getEvtMsgId()
    {
        return msgId;
    }

    /**
     * Get Call Id
     */
    protected String getEvtCallId()
    {
        return callId;
    }

    /**
     * Get Addresses
     */
    protected String[] getEvtAddresses()
    {
        return addrs;
    }

    /**
     * Get tarminal
     */
    protected String getEvtTerminal()
    {
        return terminal;
    }

    /**
     * Get Value - can have different meanings depending on Msg Type
     */
    protected int getEvtValue()
    {
        return value;
    }

    /**
     * Get message
     */
    protected String getEvtMessage()
    {
        return msg;
    }

    /**
     * Get Msg Type
     */
    protected int getEvtMsgType()
    {
        return msgType;
    }


    /**
     * Given a msg and tag, find the contents of the xml tag in the msg
     */
    static protected String tagValue(String tag, String msg)
    {
	String src = msg;
	String tagContents = null;
	String openTag = tag;
	String closeTag = "</" + tag.substring(1);
	int startpos = src.indexOf(openTag);
	int endpos = src.indexOf(closeTag);
	if ((startpos == -1) && (endpos == -1))
	{
	    return tagContents;
    	}
	if ((startpos != -1) && (endpos == -1))
	{
	    return tagContents;
	}
	if ((startpos == -1) && (endpos != -1))
	{
            return tagContents;
	}
	tagContents = src.substring(startpos + openTag.length(), endpos);
	return tagContents;
    }

  /**
   * Given a msg and tag, find the contents of the xml tag in the msg
   * where the xml tag may occur in the msg more than once
   */
    static protected String[] tagValues(String tag, String msg)
    {
	String src = msg;
        String tagContents;
	int newStartPos = 0;
	Vector v = new Vector();
	String[] arr;
        String openTag = tag;
        String closeTag = "</" + tag.substring(1);
	int count = 0;
	int startpos = src.indexOf(openTag);
	int endpos = src.indexOf(closeTag);
	while ((startpos != -1) && (endpos != -1))
	{
	    tagContents = src.substring(startpos + openTag.length(), endpos);
	    v.addElement(tagContents);
	    newStartPos = endpos + 1;
	    count++;
	    startpos = src.indexOf(openTag, newStartPos);
	    endpos = src.indexOf(closeTag, newStartPos);
	}
	if ((startpos == -1) && (endpos == -1))
        {
	    if (v.size() > 0)
	    {
		arr = new String[v.size()];
                v.copyInto(arr);
                return arr;
	    }
	    else
	    {
                return null;
	    }
        }
	if ((startpos != -1) && (endpos == -1))
        {
            return null;
        }
        if ((startpos == -1) && (endpos != -1))
        {
            return null;
        }
	return null;
    }

    /**
     * Parse the msg string to find the evt type required
     */
    static public int getEvtTypeForMsg(String str)
    {
	String typeStr = tagValue("<type>", str);
	if (typeStr != null)
	{
	    int typeValue = Integer.parseInt(tagValue("<type>", str));
	    return typeValue/100;
	}
	else
	{
	    // throw exception
	}
	return -1;
    }

}

