#ifndef _MARSHAL_HXX_
#define _MARSHAL_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const marshal_hxx_Version =
    "$Id: Marshal.hxx,v 1.28.2.1 2003/01/23 03:00:02 sprajpat Exp $";

#include <string>
#include <map>

#include "Sptr.hxx"
#include "SipCommand.hxx"
#include "SipVia.hxx"
#include "MrshlPolicy.hxx"
#include "MrshlConfMapEntry.hxx"
#include "VMrshlException.hxx"
#include "VBadDataException.hxx"


#if !defined(__linux__) || defined (__INTEL_COMPILER)
#include <set>
typedef set < string > ProxyServerList;
#else
#include "StringHash.hxx"
#if defined(__GNUC__) && (__GNUC__ < 3)
#include <hash_set>
#else
#include <ext/hash_set>
#endif

typedef hash_set< string , StringHash > ProxyServerList;
#endif

namespace Vocal
{
        
///
typedef enum
{                                  ///
    MRSHL_AUTH_BASIC,              ///
    MRSHL_AUTH_DIGEST,             ///
    MRSHL_AUTH_ACCESS,             ///
    MRSHL_AUTH_INVALID             /// Always the last entry
} MarshalAuthType;

/// Branches in Via
typedef enum
{                                  ///
    MRSHL_INCOMING_REDIRECT = 1,   ///
    MRSHL_INCOMING_CONTACT,        ///
    MRSHL_OUTGOING_REDIRECT,       ///
    MRSHL_OUTGOING_CONTACT         ///
} MarshalBranch;

///
typedef enum
{                                  ///
    MRSHL_STATUS_INCOMING = 1,     ///
    MRSHL_STATUS_OUTGOING,         ///
    MRSHL_STATUS_INVALID           /// Always the last entry
} MarshalStatus;

///
typedef enum
{                                  ///
    MRSHL_TYPE_GATEWAY = 1,        ///
    MRSHL_TYPE_USERAGENT,          ///
    MRSHL_TYPE_INTERNETWORK,       ///
    MRSHL_TYPE_CONFERENCE          ///
} MarshalType;

/*
 *  Conference call map type
 *  Used for storing active conference data for marshal
 *  Unique call id used as map key
 *  Conference Map Entry class used as map value
 */
typedef map < SipCallId, MrshlConfMapEntry > MrshlConfCallMap;

///
    
/**
 *  Marshal class is a singleton, encapsulates the global variables and
 *  functions
 */
class Marshal
{

    public:
        /**
         * Instance
         * @param int sipPort
         * @return Marshal&
         */
        static Marshal &instance( const int sipPort = -1 )
            throw ( VMrshlException& );

        /**
         * Destroy function deletes instance
         */
        static void destroy();
    
        ///
        ~Marshal() {}
    
        /// The control loop
        void run();

        /**
         * update the Marshal data should it change
         *
         */
        static void
        updateMarshal( const string& data,
                       const string& fileName,
                       const bool deletedFlag );

        /**
         * Called on solicitation from pserver for updating
         * the registrations
         */
        static void
        updateRegister( const string& data,
                       const string& fileName,
                       const bool deletedFlag );


        ///
        void doRegistrationForUpdate();

        /**
         * Takes a SIP command, returns true if it has a
	 * ProxyAuthorization header with a key that matches 
	 * that specified for the VOCAL system
         * @param Sptr<SipCommand> command
         * @return bool
         */
        bool hasMarshalKey( const Sptr < SipCommand > command ) const;

        /**
         * Insert the marshal's Via in the command;
	 * the branch parameter is appended to the branch
	 * value generated by the sip stack.
         * @param SipCommand& sipCommand
         * @param Data viaBranch
         * @return void
	 */
        void addViaWithBranch( SipCommand& sipCommand,
			       const Data& viaBranch );

        /**
         * Takes a SIP command, determines if it is incoming or outgoing from
         * the perspective of the VOCAL system, and marshals the command
         * accordingly. Marshalling a command includes the following:
         * check for loops based on appearance in Via;
         * set(incoming)/clear(outgoing) proxy authorization field;
         * return status of the marshaling process. This value may be
         * either a MarshalStatus enum or a sip error code.
         * @param Sptr < SipCommand > command
         * @return unsigned int either a MarshalStatus enum or a sip error code
         */
        unsigned int marshalIncomingCommand( Sptr < SipCommand > command );

        enum ChallengeType 
        {
            CHALLENGE_WWW_AUTH,
            CHALLENGE_PROXY_AUTH
        };


        /**
         * generate a Challenge status message to this command, and return it.
         *
         */
        Sptr<StatusMsg> generateChallenge(const Sptr < SipCommand > command, 
                                          int statusErr,
                                          ChallengeType type =
                                            CHALLENGE_PROXY_AUTH);

        /**
         * Get pending command to which this cancel applies
         * if none exist, return 481;
         * if one or more exists, return 200 to sender
         * and forward Cancel for each pending command
         * Note, only worries about first pending command
         * since marshals do not fork commands
         * trueCancel determines how requestLine should be marked
         * @param Sptr < SipEvent > sipEvent
         * @param bool trueCancel
         * @return void
         */
        void proxyCommandWithoutRoute( const Sptr < Vocal::SipEvent > sipEvent,
                                       const bool trueCancel = true );

        /**
         * Generate a nonce for the specifed user
         * @param string& user
         * @return string nonce
         */
        Data genMrshlNonce( const Data &user ) const;
    
        ///
        void showMrshlConfig() const;

        /// Accessor methods, set by provisioning
        MrshlPolicy &policy()
        {
            return myPolicy;
        }
    
        ///
        MarshalType serverType() const
        {
            return myServerType;
        }
    
        ///
        int sipPort() const
        {
            return mySipPort;
        }
        ///
        string proxyAuthKey() const
        {
            return myProxyAuthorizationKey;
        }
        ///
        bool requireCdr() const
        {
            return myRequireCdr;
        }

        ///
        const Data& transport() const { return myTransport; };

        ///
        string bridgeNumber() const
        {
            return myBridgeNumber;
        }
        ///
        list < string > &accessNumbers()
        {
            return myAccessNumbers;
        }
        ///
        MrshlConfCallMap &mrshlConfCallMap()
        {
            return myMrshlConfCallMap;
        }

        ///
        bool isNextHopAProxy(const SipCommand& sipMsg);

        /// get the gateway host
        Data gatewayHost() const
        {
            Lock lockable(myGatewayMutex);
            return myGatewayHost;
        }

        /// get the gateway port
        int gatewayPort() const
        {
            Lock lockable(myGatewayMutex);
            return myGatewayPort;
        }

    private:
        ///
        Marshal( const int sipPort );
    
        ///
        void loadPsData() throw ( VMrshlException&, VBadDataException& );

        /**
         * @param sipMsg the SipMsg that contains the via list
	 * @returns true if caller appears in Via list more than or equal to the limit
         */
        bool inViaList( const Sptr < SipMsg > sipMsg , int limit);

        /**
         * @param packetIp ipaddress to be checked agains gatewayIp
         * @return bool true if matches gatewayHost, false otherwise
         */
        bool checkGatewayIp( const Data &packetIp );

        /**
         * check user from packet against user list on provisioning
         *
	 * @param sipCommand contains the user to be check for access
         * @return true if found user in provisioning
         */
        bool AccessListCheck( const Sptr < SipCommand > sipCommand );

        /**
         * checks user for basic authentication
	 *
         * @param sipCommand contains user to be checked for basic auth
         * @return bool true if passed basic check
         */
        bool BasicCheck( const Sptr < SipCommand > sipCommand );

        /**
         * checks user for digest authentication
	 *
         * @param sipCommand contains user to be checked for digest auth
         * @return bool true if passed digest check
         */
        bool DigestCheck( const Sptr < SipCommand > sipCommand );

        /**
	 * checks for access, or digest authentication of user
	 *
         * @param sipCommand contains user to be checked
	 * @return bool result returns from access or digest check
         */
        bool UserAgentCheck( const Sptr < SipCommand > sipCommand );

        /**
	 * performs either UserAgentCheck or AccessListCheck based
         * on marshal server type
         *
         * @param sipCommand contains user to be checked
	 * @return bool result from UserAgentCheck or AccessListCheck
         */
        bool incomingCheck( const Sptr < SipCommand > sipCommand );
        
        ///
        /// Private Data Members
        ///
    
        ///
        static Marshal *myInstance;
        ///
        MrshlPolicy myPolicy;
        ///
        string myServerTypeStr;
        ///
        MarshalType myServerType;
        ///
        static string myGroupType;
        ///
        int mySipPort;
        ///
        string myProxyAuthorizationKey;
    
        /// Primary CDR server
        string myCdrServer;
    
        /// Alternate CDR server
        string myCdr2Server;
    
        ///
        int myCdrPort;
        ///
        int myCdr2Port;
    
        /// if true, reject calls if no CDR server
        bool myRequireCdr;
    

        /// protect myGatewayHost and myGatewayPort
        mutable VMutex myGatewayMutex;

        ///
        Data myGatewayHost;
        ///
        int myGatewayPort;
    
        /// Via header for marshal to insert in outgoing msgs
        SipVia myVia;
    
        /// unique number used by endpoints for all ad-hoc calls
        string myBridgeNumber;
    
        /// list of conference bridge numbers to map to
        list < string > myAccessNumbers;
    
        /// marshal conference call map
        MrshlConfCallMap myMrshlConfCallMap;

        ///
        ProxyServerList myListOfServers;

        //
        Data myTransport;
};
 
}

// Local Variables:
// mode:c++
// c-file-style:"bsd"
// c-basic-offset:4
// c-file-offsets:((inclass . ++))
// indent-tabs-mode:nil
// End:


#endif // _MARSHAL_HXX_
