
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const mrshlpolicy_cxx_Version =
    "$Id: MrshlPolicy.cxx,v 1.22 2002/06/07 20:53:51 veer Exp $";


#include <string>
#include <iostream.h>

#include "MrshlPolicy.hxx"
#include "SipOsp.hxx"
#include "SipOspAuthorization.hxx"
#include "SipTransceiver.hxx"

#ifndef NO_POLICY
#include "COPSData.hxx"
#endif

using namespace Vocal;

// static data declarations
//
PendingInviteMap MrshlPolicy::myPendingInviteMap;
VRWLock MrshlPolicy::myLock;
bool MrshlPolicy::myUseContentData( false );


MrshlPolicy::MrshlPolicy() :
    myResponseTime( 10000 )
{}


MrshlPolicy::~MrshlPolicy()
{
    myPendingInviteMap.clear();
}


void
MrshlPolicy::initialize( const int responseTime )
{
#ifndef NO_POLICY

    myResponseTime = responseTime;

    if ( getenv( "MARSHAL_CONTENT_DATA" ) )
    {
        myUseContentData = true;
    }

    try
    {
        PepAgent::instance().registerCallback( MrshlPolicy::policyCallback );

        // Time we get from provisioning in Milliseconds
        PepAgent::instance().setResponseTime( myResponseTime );
    }
    catch ( VMissingDataException &e )
    {
        // if there is no connection, continue anyways
        cpLog( LOG_ERR, "Error initializing policy server, reason: %s",
               e.getDescription().c_str() );
    }
    catch ( VException &ve )
    {
        cpLog( LOG_ERR, "%s", ve.getDescription().c_str() );
    }

#endif
}


bool
MrshlPolicy::inList( const Data &callId )
{
#ifndef NO_POLICY
    bool found = false;

    myLock.ReadLock();

    if (myPendingInviteMap.count( callId) )
    {
        found = true;
    }

    myLock.Unlock();

    return found;
#else
    return false;
#endif
}


void
MrshlPolicy::addInvite( const Data &callId, Sptr < SipEvent > sipEvt )
{
    myLock.WriteLock();
    myPendingInviteMap[callId] = sipEvt;
    myLock.Unlock();
}


bool
MrshlPolicy::removeInvite( const Data &callId )
{
    PendingInviteMap::iterator iter;
    bool found = false;

    myLock.WriteLock();

    iter = myPendingInviteMap.find( callId );
    if ( iter != myPendingInviteMap.end() )
    {
        myPendingInviteMap.erase( iter );
        found = true;
    }
    myLock.Unlock();

    return found;
}


/*
 * Callback function supplied to handle responses to request
 * made to the policy server 
 */

#ifndef NO_POLICY

void
MrshlPolicy::policyCallback( PolRequest& policyRequest )
{
    Data callId( policyRequest.getCallId() );

    cpLog( LOG_DEBUG, "Policy callback request for callId: %s", callId.logData() );


    Sptr < SipEvent > sipEvent = 0;
    Sptr < SipMsg > sipMsg = 0;
    Sptr < InviteMsg > inviteMsg = 0;

    PendingInviteMap::iterator iter;

    //
    // IMPORTANT: Before exiting method, you must release the lock
    //

    myLock.WriteLock();

    // locate associated Invite msg in pendingInviteMap
    //
    iter = myPendingInviteMap.find( callId );

    if ( iter != myPendingInviteMap.end() )
    {
        sipEvent = iter->second;

        if (sipEvent != 0)
        {
            // get Invite from SipEvent
            sipMsg = sipEvent->getSipReceive()->back();

            inviteMsg.dynamicCast( sipMsg );
        }
    }
    else
    {
        cpLog( LOG_ERR, "No match in pendingInviteMap for callId: %s",
               callId.logData() );
    }

    // If cannot find pending invite or sipEvent, unlock and return
    //
    if (inviteMsg == 0)
    {
        cpLog( LOG_ERR, "No INVITE in sipEvent, corrupted sipEvent" );

        myLock.Unlock();
        return ;
    }


    // check if authorization succeeded
    //
    if ( policyRequest.getStatus() != PolRequest::STATUS_OK )
    {
        // Abort the call
        cpLog( LOG_ALERT, "Policy server request rejected, send 401, check policy server" );

        StatusMsg statusMsg( *inviteMsg, 401 );

        sipEvent->getSipStack()-> sendReply( statusMsg );
    }

    // if call incoming, remove token
    //
    else if ( policyRequest.isIncoming() )
    {
        cpLog( LOG_INFO, "Found incoming policy request" );

        // token should have already been removed by policyValidate()

        sipEvent->getSipStack()->sendAsync( *inviteMsg );
    }

    // if call outgoing, add token to message
    //
    else
    {
        cpLog( LOG_INFO, "Found outgoing policy request" );

        // embed policy server token in Invite and send out
        // this INET marshal's gateway
        //
        cpLog( LOG_INFO, "Request granted, embed token: %s",
               policyRequest.getToken().c_str() );

        if ( MrshlPolicy::myUseContentData )
        {
            SipOsp osp;
            osp.setData( policyRequest.getToken() );

            SipContentType contype( "multipart", "mixed" );
	    contype.setTokenDetails("boundary", "next");

//            SipContentType ospContentType( "application", "OSP" );
//            osp.setCtype(ospContentType);

            // osp data will be copied
//            inviteMsg->setContentData( &osp );

#if 0	    
	    if ( inviteMsg->getNumContentData() > 1 )
	    {
		inviteMsg->setContentType(contype);
	    } 
	    else
	    {
		inviteMsg->setContentType(ospContentType);
	    }
#endif

        }
        else
        {
            // use the osp header method
            SipOspAuthorization osp( policyRequest.getToken() );
            inviteMsg->setSipOspAuthorization(osp);
        }

#if 0
        // This code is for using TCP. The sip stack will handle the
        // large token for LANs with the MTU set higher than 3000,
        // however WAN's have a lower MTU and the SIP messages will
        // not be transported
        //
        SipVia via(inviteMsg->getVia(0));
        via.setTransport("TCP");
        inviteMsg->removeVia( 0 );
        inviteMsg->setVia( via, 0 );
#endif
        sipEvent->getSipStack()->sendAsync( *inviteMsg );
    }


    // remove pending invite from map and unlock
    //
    myPendingInviteMap.erase(iter);
    myLock.Unlock();
}
#endif


void
MrshlPolicy::policyAbort( const Data &callId )
{
#ifndef NO_POLICY

    // abort policy
    cpLog( LOG_DEBUG, "Abort policy for callId: %s", callId.logData() );

    if ( removeInvite( callId ) )
    {
        cpLog(LOG_DEBUG, "Removed invite from map, callId:%s",
              callId.logData());
    }
    else
    {
        cpLog(LOG_ERR, "Cannot find invite in map for callId:%s",
              callId.logData());
    }

    try
    {
        PolRequest& polReq = PepAgent::instance().getPolReq( callId );
        polReq.implAbort();
    }
    catch ( VMissingDataException& e )
    {
        cpLog( LOG_INFO, "Extra call to abort policy - ignored" );
    }

#endif
}


void
MrshlPolicy::policyStart( const Data &callId )
{
#ifndef NO_POLICY

    // start policy
    cpLog( LOG_DEBUG, "Start policy for callId: %s", callId.logData() );
    try
    {
        PolRequest& polReq = PepAgent::instance().getPolReq( callId );
        polReq.implStart();
    }
    catch ( VMissingDataException& e )
    {
        cpLog( LOG_ALERT, "Start policy failed for callId: %s, reason: %s",
               callId.logData(),
               e.getDescription().c_str() );
    }

#endif
}


void
MrshlPolicy::policyFinish( const Data &callId )
{
#ifndef NO_POLICY

    // stop policy, need to call for both incoming/outgoing since either
    // end can hang up
    cpLog( LOG_DEBUG, "Stop policy for callId: %s", callId.logData() );
    try
    {
        PolRequest& polReq = PepAgent::instance().getPolReq( callId );
        polReq.implFinish();
    }
    catch ( VMissingDataException& e )
    {
        // could be called twice since don't know which end is
        // going to hang up first
        cpLog( LOG_INFO, "Extra call to to finish policy - ignored" );
    }

#endif
}

bool
MrshlPolicy::policyAuthorize( const Sptr < SipEvent > sipEvent,
                              Sptr < InviteMsg > inviteMsg )
{
#ifndef NO_POLICY

    // check with policy server for authorization for outgoing call
    // note, the policy server responds via a callback function

    Data callId = inviteMsg->getCallId().getCallIdData();

    cpLog( LOG_DEBUG, "Policy Server authorize To:%s, From:%s, CallId:%s",
           inviteMsg->getTo().getUser().logData(),
           inviteMsg->getFrom().getUser().logData(),
           callId.logData() );

    try
    {
        LocalScopeAllocator lo;
        PepAgent::instance().auth( inviteMsg->getTo().getUser(),
                                   inviteMsg->getFrom().getUser(),
                                   callId.getData(lo) );

        addInvite( callId, sipEvent );
    }
    catch ( VDuplicateIdException& e )
    {
        // policy server authorization rejected
        return false;
    }

    return true;

#else
    ///NO policy is used so just proxy the message
    ///sipEvent->getSipStack()->sendAsync( inviteMsg );
    return true;

#endif
}

bool
MrshlPolicy::policyValidate( const Sptr < SipEvent > sipEvent,
                             Sptr < InviteMsg > inviteMsg )
    #ifndef NO_POLICY
    throw (VDuplicateIdException&)
    #endif    
{
#ifndef NO_POLICY

    // check with policy server for validation of incoming call
    // note, the policy server responds via a callback function
    //
    Data callId = inviteMsg->getCallId().getCallIdData();

    cpLog( LOG_INFO, "Policy Server validate To:%s, From:%s, CallId:%s",
           inviteMsg->getTo().getUser().logData(),
           inviteMsg->getFrom().getUser().logData(),
           callId.logData() );

    Data ospToken;
    int tokenLen = 0;

    if ( MrshlPolicy::myUseContentData )
    {
        // get the token from the sip message
        //
        Sptr<SipOsp> sipOsp;
        Sptr<SipContentData> contentDat;
        int contentCount = inviteMsg->getNumContentData();
        for ( int i = 0; i < contentCount; i++ )
        {
            contentDat = inviteMsg->getContentData( i );
            sipOsp.dynamicCast ( contentDat );
            if ( sipOsp != 0 )
            {
                Sptr<SipContentLength> scLen = sipOsp->getContentLength();
                tokenLen = int(scLen->getLength());

                ospToken = sipOsp->getData();

                // remove token from invite message
                inviteMsg->removeContentData(i);

                break;
            }
        }
    }
    else
    {
        SipOspAuthorization ospAuth = inviteMsg->getSipOspAuthorization();
        tokenLen = ospAuth.get().length();
        ospToken = ospAuth.get();

        // remove token from invite message
        SipOspAuthorization nullOspAuth;
        inviteMsg->setSipOspAuthorization( nullOspAuth );
    }

    try
    {
        // if no token exists, reject call
        if ( tokenLen == 0 )
            throw VDuplicateIdException( "No Token found in Invite Message",
                                         __FILE__, __LINE__ );
        else // found token
        {
            COPSData copsData( ospToken, tokenLen );
            LocalScopeAllocator lo;
            PepAgent::instance().validateToken( inviteMsg->getTo().getUser(),
                                                inviteMsg->getFrom().getUser(),
                                                callId.getData(lo),
                                                copsData );
        }

        addInvite(callId, sipEvent);
    }
    catch ( VDuplicateIdException& e )
    {
        // policy server authorization failed
        cpLog( LOG_INFO, "No Token found in Invite Message");
        return false;
    }

    return true;

#else
    ///NO policy is used so just proxy the message
    //Do it in the caller, as it is syncronous
    //sipEvent->getSipStack()->sendAsync( inviteMsg );
    return true;
#endif
}


bool
MrshlPolicy::enableQoS( const Sptr<SipEvent> sipEvent,
                        const Sptr<StatusMsg> statusMsg )
{
#ifndef NO_POLICY

#ifdef HAS_RSVPLIB

    cpLog( LOG_DEBUG, "Enabling QoS" );

    // retrieve Invite associated with sipEvent
    Sptr<InviteMsg> savedInvite = sipEvent->getInvite();
    if( savedInvite == 0 )
    {
        cpLog( LOG_DEBUG, "Enable QoS failed: No invite found for SIP Event" );
        return false;
    }

    //read remote SDP from 180 msg
    SipSdp* remoteSdp = dynamic_cast<SipSdp*>
        ( statusMsg->getContentData(0) );
    
    if( !remoteSdp )
    {
        cpLog( LOG_ERR,"Enable QoS failed: No remote SDP in 180");
        return false;
    }
    else if( !remoteSdp->isValidSdp() )
    {
        cpLog( LOG_ERR,"Enable QoS failed: Remote SDP in 180 is invalid");
        return false;
    }

    //read local SDP from matching invite
    SipSdp* localSdp = dynamic_cast<SipSdp*> ( savedInvite->getContentData(0) );
    if( !localSdp )
    {
        cpLog( LOG_ERR,"Enable QoS failed: No local SDP in matching invite");
        return false;
    }
    else if( !localSdp->isValidSdp() )
    {
        cpLog( LOG_ERR,"Enable QoS failed: Local SDP in invite is invalid");
        return false;
    }
    LocalScopeAllocator lo;
    string localIpAddress = localSdp->getConnAddress().getData(lo);
    string remoteIpAddress = remoteSdp->getConnAddress().getData(lo);
    int localPort = localSdp->getRtpPort();
    int remotePort = remoteSdp->getRtpPort();

    SipCallId callId = savedInvite->getCallId();

    //enable QoS in PepAgent
    cpLog( LOG_DEBUG, "Enable QoS for callId: %s [%s:%d -> %s:%d]",
                       callId.getCallIdData().logData(),
                       localIpAddress.c_str(), localPort,
                       remoteIpAddress.c_str(), remotePort );

    try
    {
        PepAgent::instance().enableQoS( callId.getCallIdData().getData(lo),
                                        localIpAddress, localPort,
                                        remoteIpAddress, remotePort );
    }
    catch ( VException &e )
    {
        cpLog( LOG_ERR,"Exception caught while enabling QoS:%s",
               e.getDescription().c_str());
        return false;
    }

#endif
    
    return true;
#else
    return false;
#endif
}


bool
MrshlPolicy::disableQoS( const SipCallId &callId )
{
#ifndef NO_POLICY

#ifdef HAS_RSVPLIB
    //enable QoS in PepAgent
    cpLog( LOG_DEBUG, "Disable QoS for callId: %s",
                      callId.getCallIdData().logData() );

    try
    {
        LocalScopeAllocator lo;
        PepAgent::instance().disableQoS( callId.getCallIdData().getData(lo) ); 
    }
    catch ( VException &e )
    {
        cpLog( LOG_ERR,"Exception caught while disabling QoS:%s",
               e.getDescription().c_str());
        return false;
    }
#endif
    
    return true;
#else
    return false;
#endif
}
