
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const DialPlanContainer_cxx_Version =
    "$Id: DialPlanContainer.cxx,v 1.26 2002/02/25 05:51:19 sprajpat Exp $";
#include "cpLog.h"
#include <stdio.h>
#include "DialPlanContainer.hxx"
#include "ProvisionInterface.hxx"


Sptr < DialPlanContainer > DialPlanContainer::theDPContainer = 0;

/**
 * Return the singleton Dial plan container or create it if
 * it does not already exist. 
 * @return Dial plan container singleton
 */
DialPlanContainer&
DialPlanContainer::instance()
{
    if ( theDPContainer == 0 )
    {
        theDPContainer = new DialPlanContainer;
    }

    return (*theDPContainer);
}

/**
 * Constructor
 */
DialPlanContainer::DialPlanContainer()
{};

/**
 * Destructor
 */
DialPlanContainer::~DialPlanContainer()
{};

/**
 * Add Dial plan elements one at a time
 * A new plan is created if the plan is not already in the container.
 * @param  planName  Unique name of plan to which the element is to be added.
 * @param  key  Regular expression
 * @param  contacts  List of contacts
 */
void
DialPlanContainer::add( const Data& planName, const Data& key, const  vector < Data > & contacts )
{
    /// find the plan in the container
    if (DPContainer.count(planName) == 0)
    {
        /// The plan does not exist, so create it
        Sptr < Plan > workingPlan = new Plan;
        workingPlan->addEntry(key, contacts);
        DPContainer[planName] = workingPlan;
    }
    else
    {
        DPContainer[planName]->addEntry(key, contacts);
    }
};

/**
 * Remove the specified plan
 * @param  planName  Identifier of plan to be removed.
 */
void
DialPlanContainer::remove( const Data& planName)
{
    /// search the container for the requested plan
    PlanIter i = DPContainer.find(planName);

    if (i != DPContainer.end() )
    {
        /// clear out the plan
        i->second->clear();
        /// remove the plan from the table
        DPContainer.erase(i);
    }

};

/**
 * Search plan using regular expression matching to find key.
 * Retrieve and return contact list.
 * Throw an exception if plan is not found.
 * @param  planName  Unique identifier of plan to be searched.
 * @param  key  String to be matched against regular expressions in plan.
 * @return  Contact list 
 */
bool
DialPlanContainer::search( const Data& planName, const Data& key, vector < Data > & planFound)
{
    cpLog( LOG_DEBUG, "DialPlanContainer::search");

    if (DPContainer.count(planName) == 0)
    {
        cpLog(LOG_DEBUG, " Plan is missing for %s", planName.logData());
        return false;
    }

    if (true == DPContainer[planName]->search(key, planFound))
    {
        return true;
    }
    else
    {
        return false;
    }
};

/**
 * Look for and obtain a specific plan from the container
 * @param  planName  Identifier of plan to search for.
 * @param  workingPlan  The obtained plan.
 * @return true if found, false if not
 */
bool
DialPlanContainer::getPlan( const Data& planName, Sptr < Plan > workingPlan)
{
    /// search the container for the requested plan
    PlanIter i = DPContainer.find(planName);

    if (i != DPContainer.end() )
    {
        workingPlan = i->second;
        return true;
    }
    else
        return false;
};

/**
 * create a new plan and store it against the given name
 * @param  planName  Unique identifier of new plan.
 * @param  workingPlan  The plan object created.
 */
void
DialPlanContainer::create( const Data& planName, Sptr < Plan > workingPlan )
{
    Sptr < Plan > newPlan = new Plan;
    DPContainer.insert(pair < Data, Sptr < Plan > > (planName, newPlan));
    workingPlan = newPlan;
};


/**
 * Display the specified plan.
 * @param  planName  Unique identifier of plan to be displayed.
 */
void
DialPlanContainer::display( const Data& planName)
{
    cpLog( LOG_DEBUG, "DialPlanContainer::display");
    if (DPContainer.count(planName) == 0)
    {
        cpLog(LOG_DEBUG, " Plan is missing for %s", planName.logData());
        return ;
    }

    DPContainer[planName]->display();
}

/**
 * The Provisioning Server is notifying us of a change in a Dial Plan. 
 * Obtain the plan from the PS and store it.
 * @param  data  Contents of xml file.
 * @param  fileName  Name of plan file.
 * @param  deleteFlag  Flag indicating if file was deleted.
 */
void
updateDialPlan(const string& data, const string& fileName, bool deleteFlag)
{

    cpLog( LOG_DEBUG, "updateDialPlan triggered");

    /// Strip off the '.xml' extension
    string withOutXml = fileName;
    string::size_type pos = fileName.find(".xml");
    if (pos != string::npos)
        withOutXml = fileName.substr(0, pos);

    Data identifier(withOutXml);

    vector < VPlanElement > planHolder;

    try
    {
        /// Retrieve the Dial Plan specific info from the PS
        ProvisionInterface::instance().getDialPlan(fileName, planHolder);
    }
    catch (VException& e)
    {
        /// failed to connect to the Provisioning Server
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
        return;
    }

    DialPlanContainer& dpContainer = DialPlanContainer::instance();

    /// Remove the current Dial Plan
    dpContainer.remove(identifier);

    /// Create the Dial Plan with the new data
    for ( vector < VPlanElement > ::iterator i = planHolder.begin() ; i != planHolder.end() ; ++i)
    {
        Data key(i->getKey());
        dpContainer.add(identifier, key, i->getContactList());
    }

    dpContainer.display(identifier);

} /// end of updateDialPlan

