#ifndef GroupContainer_HXX
#define GroupContainer_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

#include <map>
#include <vector>
#include "VMutex.h"
#include "Sptr.hxx"
#include "VServerGroupContainer.hxx"
#include "Group.hxx"

static const char* const GroupContainerVersion = "$Id: GroupContainer.hxx,v 1.20 2001/02/06 23:14:12 marcl Exp $";

/**
 *  This class defines a singleton container of Groups and
 *  the servers belonging to them.
 *  It also maintains a list of all of the valid group names.
 */

class GroupContainer
{
    public:
        /**
         *  Return the only Group container or create it
         *  if it doesn't already exist. 
         * @return  Reference to the singleton for this object
         */
        static GroupContainer& instance();

        /// dtor
        virtual ~GroupContainer();

        /**
         * Verify if a specific group is valid
         * @param  group  Group identifier
         * @param  type  Type of servers belonging to the given group
         * @return  True if the group is valid, false otherwise.
         */
        bool isValid( const Data& group, ServerType& type );

        /**
         * Process a notification of an update to group data in provisioning.
         */
        void handleUpdateFromPS();

        /**
         * Add a group to the container
         * @param  key  Group identifier
         * @param  listofServers  Servers belonging to the group to be added
         */
        void addGroup( const Data& key,
                       const Sptr < vector < Data > > listofServers);

        /**
         * Add a server to a group already in the container
         * @param  key  Group identifier
         * @param  server  Server to be added to the group's list of servers
         */
        void addServerToGroup( const Data& key,
                               const Data& server);

        /**
         * Remove the specified group from the map
         * @param  key  Group identifier
         */
        void removeGroup( const Data& key);

        /**
         * Remove server from group  
         * @param  key  Group identifier
         * @param  server  Server to be removed
         */
        void removeServerFromGroup( const Data& key,
                                    const Data& server);

        /**
         * Search for a given group and obtain its list of servers
         * @param  key  Group
         * @param  listofServers  The servers belonging to the specified group
         * @return 
         */
        bool search( const Data& key, Sptr < vector < Data > > & listofServers ) ;

        /**
         * Obtain a server from the specified group
         * @param  key  Group identifier
         * @param  server  The resulting server found within the given group
         * @return  True if a server was found, false if the group is not in the container
         */
        bool getServer( const Data& key, Data& server);

        /**
         * Display a specific group
         * @param  key  Name of group to be displayed
         */
        void display( const Data& key);

        /**
         * print list of servers belonging to this Group 
         * @param  listofServers  List of servers to print
         */
        void displayGroup( const vector < Data > & listofServers);

    protected:
        /// Constructor
        GroupContainer();

        /**
         *  Obtain the valid groups from provisioning
         */
        void getValidGroupsFromPS();

        /**
         * Obtain valid groups of a specific type from provisioning
         * @param  type  Type of server
         */
        void getValidGroupsOfType( const ServerType type);

        /// This is the single instance of GroupContainer
        static Sptr < GroupContainer > theGroupContainer;

    private:
        /// Defines a map of group names to lists of servers.
        typedef map < Data, Sptr < vector < Data > > > Table;
        /// Iterator for the map
        typedef Table::iterator TableIter;

        /// Data member instance of the map
        mutable Table groupList;

        /// mutex object
        VMutex lock;

        /// List of the all the valid group names.
        Sptr < vector < Group > > itsValidGroupList;
};

#endif
