
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Plan_cxx_Version =
    "$Id: Plan.cxx,v 1.11 2002/02/25 05:51:19 sprajpat Exp $";

#include"Plan.hxx"
#include "cpLog.h"

Plan::Plan()
{
};


Plan::~Plan()
{
    clear();
};


/**
 * Create a new entry by adding it to the bottom of the list
 * Order is important when it comes time to search through the list.
 * @param  key  Regular expression
 * @param  contacts  Contact list 
 */
void
Plan::addEntry( const Data& key, const  vector < Data > & contacts )
{
    // cpLog( LOG_DEBUG,"Plan::addEntry, adding: %s", key.logData());

    /// Create the element
    PlanElement element(key, contacts);

    PlanMutex.WriteLock();

    /// add it to the end of the vector
    itsPlan.push_back(element);

    PlanMutex.Unlock();
};

/**
 * Delete the specified key and its related contacts from the container.
 * @param  key  Regular expression to be deleted.
 */
void
Plan::deleteEntry( const Data& key )
{
    PlanMutex.WriteLock();

    for (vector < PlanElement > ::iterator i = itsPlan.begin();
            i != itsPlan.end();
            i++ )
    {
        if (key == i->getTextKey())
        {
            itsPlan.erase(i);
        }
    }

    PlanMutex.Unlock();
};

/**
 * Remove the whole  plan - will need to
 * be done when a change is made
 */
void
Plan::clear( )
{
    PlanMutex.WriteLock();

    itsPlan.clear();

    PlanMutex.Unlock();
};


/**
 * Search table using regexp matching. Retrieve and return a contact list
 * @param  key  String to match against the regular expressions in the table.
 * @param  elementFound List of contacts corresponding to expression matched.
 * @return True if found, false if not.
 * NOTE: If a key is found, and regular expression using grouping and 
 *       the contacts use $1, $2(replacement notations), then the
 *       contacts would be changed accordingly.
 */
bool
Plan::search( const Data& key , vector < Data > & elementFound)
{
    //  cpLog( LOG_DEBUG,"Plan::search");

    PlanMutex.ReadLock();
    LocalScopeAllocator lo;

    for (vector < PlanElement > ::iterator i = itsPlan.begin();
            i != itsPlan.end();
            ++i )
    {
        cpLog( LOG_DEBUG, "regexp in dialplan: %s", i->getTextKey().logData());

        /// attempt to match the passed in key with the expression and
		/// fill in the registers.
        if ( REG_NOMATCH != regexec( &(i->getKey()), key.getData(lo), 
							MAX_NUM_REGMATCH, 
							(regmatch_t *)i->getMatch(), 0) )
        {
            cpLog( LOG_DEBUG, "Key: %s matched RegExp: %s",
                   key.logData(), i->getTextKey().logData());

            PlanMutex.Unlock();

            /// Return the found contact list
            elementFound = i->getContacts();

			/// Check to see if the contacts needs to be replaced.(Regexec
			/// replacement may have been requested)
			if( i->isToBeReplaced() ) 	{
				cpLog( LOG_DEBUG, "Invoking regular expression replacement" );
				i->replace( key, elementFound );
			} 
            return true;
        }
    }

    PlanMutex.Unlock();

    cpLog(LOG_DEBUG, "Failed to find the dial element");
    return false;
};


/**
 * Display all the stored expressions and their related contacts.
 */
void
Plan::display()
{
    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        PlanMutex.ReadLock();

        cpLog( LOG_DEBUG, "-----------------------------");
        cpLog( LOG_DEBUG, " Plan ");
        for (vector < PlanElement > ::iterator dpIter = itsPlan.begin(); dpIter != itsPlan.end(); ++dpIter )
        {
            cpLog( LOG_DEBUG, " Expression: %s", dpIter->getTextKey().logData());

            /// print the data contained in the vector
            const vector < Data > & currentContacts = dpIter->getContacts();
            for (vector < Data > ::const_iterator i = currentContacts.begin(); i != currentContacts.end(); ++i)
            {
                cpLog( LOG_DEBUG, "    Contact: %s", i->logData() );
            }
        }
        cpLog( LOG_DEBUG, "-----------------------------");

        PlanMutex.Unlock();
    }
}
