
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const SipContactIf_cxx_Version =
    "$Id: SipContactIf.cxx,v 1.94 2002/12/12 01:19:38 chok Exp $";
#include <unistd.h>
#include <cassert>

#include "DialPlanContainer.hxx"
#include "GroupContainer.hxx"
#include "IPPlanContainer.hxx"
#include "ProvisionInterface.hxx"
#include "RSServerContainer.hxx"
#include "SipContactIf.hxx"
#include "SubscriberContainer.hxx"
#include "RSProxy.hxx"

#ifdef VOCAL_HAS_ENUM
#include "formatter.h"
#include "resolvers.h"
#define MAX_ENUM_RESULTS 10
#endif

using namespace Vocal;

const string WILDCARDUSER = "$USER";

/**
 * Constructor
 */
SipContactIf::SipContactIf()
{}

/**
 * Copy Constructor
 * @param  rhs  Object to copy from
 */
SipContactIf::SipContactIf(const SipContactIf& rhs)
{
    if (this != &rhs)
    {
        copyRhsToThis( rhs );
    }
}

/**
 * Set object equal to specified object.
 * @param  rhs  Object to copy from
 * @return  This object
 */
SipContactIf&
SipContactIf::operator=(const SipContactIf& rhs)
{
    if (this != &rhs)
    {
        copyRhsToThis(rhs);
    }

    return *this;
}

/**
 * Copy data from specified object into this object.
 * @param  rhs  Object from which to copy.
 */
void
SipContactIf::copyRhsToThis(const SipContactIf& rhs)
{}

/**
 * Destructor
 */
SipContactIf::~SipContactIf()
{}


/**
 * Update is triggered when a REGISTER message comes in to the RS.
 * The 'To' field is to be used as the key to search the Subscriber
 * container. The Term contacts are updated as well as the expire time.
 * Note: Update only deals with Subscribers
 * Note: The 'To' field only contains the user portion.
 * @param  to  Subscriber identifier to be updated.
 * @param  from  Identifier of individual who performed the registration.
 * @param  expires  Expire time in seconds until subscriber needs to be re-registered.
 * @param  newContacts  New terminating contact list to store
 * @param  existingContacts  Previously stored contact list 
 * @return Response code to be passed back in response to the Register.
 */
int
SipContactIf::update( const Data& to, const Data& from, Data& expires,
                      const Sptr < vector < Data > > & newContacts, vector < Data > & existingContacts )
{
    /// obtain the subscriber container
    SubscriberContainer& theContainer = SubscriberContainer::instance();

    /// transform the expires time to an integer
    unsigned long requestedExpiry = expires.convertInt();

    /// The expiry time given by the PS
    unsigned long expireTimeAllowed = theContainer.getMaxExpireTime();

    int responseCode = theContainer.updateOnRegister(to, from, newContacts, requestedExpiry, expireTimeAllowed, existingContacts);

    /// convert int to Data and put in expires
    Data exp(expireTimeAllowed);
    /// pass back the allowed time before registration must be performed again.
    expires = exp;

    //  theContainer.display(to);

    return responseCode;
}

/**
 * Generate a contact list for the given originator and destination.
 * The contact will be generated accordingly:
 * If the Req Uri is found in the Subscriber container, then
 *    From field's Calling Contacts
 *    Req Uri's Called Contacts
 *    Req Uri's Term Contacts
 *
 * Otherwise, if the Req Uri matches a key in one of the dial plans, then
 *    From field's Calling Contacts
 *    Dial Plan Entry's Contacts
 * @param  orig  Originator's user field
 * @param  dest  Destination Sip address
 * @param  destUserIsPhone  Flag to indicate if destinations user field is a phone number or ip address
 * @param  contacts  Holder for the contact list to be returned
 * @return  Response code to be passed back in the status response to the Invite.
 */
int
SipContactIf::generateContactList(const Data& orig, const Data& dest, bool destUserIsPhone, bool & destIsRegistered,  vector <Data> & contacts)
{

    int responseCode = 302;

    cpLog(LOG_DEBUG, "Orig:[%s], Dest:[%s]", orig.logData(), dest.logData());

    /// Create a copy of orig and strip off all but the user portion
    Data origUserOnly(orig);

    /// obtain the subscriber container
    SubscriberContainer& theContainer = SubscriberContainer::instance();

    /// Update originator's data from Provisioning if necessary.
    theContainer.checkProvisioned(origUserOnly);

    /// Attempt to find orig in the Subscriber container
    if (theContainer.searchFrom(origUserOnly, contacts))
    {
        cpLog( LOG_DEBUG, "Found 'FROM' in Subscriber");
    }
    else
    {
        cpLog( LOG_DEBUG, "Could not find 'FROM' in Subscriber");
    }

    /// Create a copy of dest and strip off all but the user portion
    Data destUserOnly(getUser(dest));

    bool destExpired = false;

    /// Update destination's data from Provisioning if necessary.
    theContainer.checkProvisioned(destUserOnly);

    /// SearchContainers on the 'uri' parameter
    if (theContainer.searchTo(destUserOnly, contacts, destExpired))
    {
        cpLog( LOG_DEBUG, "Found 'Uri' in Subscriber");

        if (destExpired)
        {
            destIsRegistered = false;
            /// Return 480 if subscriber has not registered,
            /// but allow the call to go through features first. 
            if (0 == contacts.size())
            {
                cpLog( LOG_DEBUG, "Subscriber is expired -- no contacts, send 480");
                responseCode = 480;
                return responseCode;
            }
            else
	    {
                cpLog( LOG_DEBUG, "Subscriber is expired, but has contacts");
	    }
        }
        else
        {
            destIsRegistered = true;
        }

    }
#ifdef VOCAL_HAS_ENUM
    else
    {
        /// Attempt to locate number using ENUM
        Data enumContact;
        if (true == queryEnum(destUserOnly, enumContact))
        {
            cpLog( LOG_DEBUG, "Found 'Uri' using NetNumber ENUM lookup");
            /// Need to inject a marshal gateway before this
            /// outside address.
            /// TODO Need to make the gateway group name configurable.
            Data gatewayMarshalGroup = "GatewayGroup";
            contacts.push_back(gatewayMarshalGroup);
            contacts.push_back(enumContact);
        }
#endif
        else
        {
            cpLog( LOG_DEBUG, "Did not find 'Uri' %s in Subscriber", destUserOnly.logData() );

            if (destUserIsPhone)
            {
                /// Search the Dial Plan container
    
                // We have the potential to handle multiple dial plans.
                // For now we will only use a default dial plan.
                Data planName("defaultPhonePlan");
                vector < Data > phoneContacts;
                if (true == DialPlanContainer::instance().search(planName, dest, phoneContacts))
                {
                    contacts.insert(contacts.end(), phoneContacts.begin(), phoneContacts.end());
                    cpLog( LOG_DEBUG, "Found 'Uri' in Dial Plan");
                }
            }
            else
            {
                /// Search the IP Plan container

                // We have the potential to handle multiple ip plans.
                // For now we will only use a default ip plan.
                Data planName("defaultIpPlan");
                vector < Data > ipContacts;
                if (true == IPPlanContainer::instance().search(planName, dest, ipContacts))
                {
                    contacts.insert(contacts.end(), ipContacts.begin(), ipContacts.end());

                    cpLog( LOG_DEBUG, "Found 'Uri' in IP Plan");
                }
            }
        }
#ifdef VOCAL_HAS_ENUM
    }
#endif

    if (cpLogGetPriority() >= LOG_DEBUG)
    {
        cpLog( LOG_DEBUG, "---------------------------------------------------------");
        cpLog( LOG_DEBUG, "Contact list generated from Caller [%s] and Called [%s]", origUserOnly.logData(), destUserOnly.logData());
        /// Display the contact list
        vector < Data > ::iterator contactIter = contacts.begin();
        for ( ; contactIter != contacts.end(); ++contactIter)
        {
            cpLog( LOG_DEBUG, "Contact [%s]", contactIter->logData());
        }
        cpLog( LOG_DEBUG, "Contact list size %d", contacts.size());
        cpLog( LOG_DEBUG, "---------------------------------------------------------");
    }

    if (contacts.size() == 0)
    {
        /// Empty contact list
        responseCode = 404;
    }

    return responseCode;
} // end of generateContactList


int
SipContactIf::generateTermOnlyContactList(const Data& dest, 
                                          bool & destIsRegistered, 
                                          vector <Data> & contacts)
{
   int responseCode = 302;

   /// Create a copy of dest and strip off all but the user portion
   Data destUserOnly(getUser(dest));

   bool destExpired = false;

   /// obtain the subscriber container
   SubscriberContainer& theContainer = SubscriberContainer::instance();

   /// Update destination's data from Provisioning if necessary.
   theContainer.checkProvisioned(destUserOnly);

   /// SearchContainers on the 'uri' parameter
   if (theContainer.searchTo(destUserOnly, contacts, destExpired, true))
   {
      cpLog( LOG_DEBUG, "Found 'Uri' in Subscriber");
        
      if (destExpired)
      {
         destIsRegistered = false;
         /// Return 480 if subscriber has not registered,
         /// but allow the call to go through features first. 
         if (0 == contacts.size())
         {
            cpLog( LOG_DEBUG, "Subscriber is expired -- no contacts, send 480");
            responseCode = 480;
            return responseCode;
         }
	 else
	 {
            cpLog( LOG_DEBUG, "Subscriber is expired but has contacts");
	 }
      }
      else
      {
         destIsRegistered = true;
      }
   }

   cpLog( LOG_DEBUG, "---------------------------------------------------------");
   cpLog( LOG_DEBUG, "Contact list generated from Called [%s]", destUserOnly.logData());

   /// Display the contact list
   vector < Data > ::iterator contactIter = contacts.begin();
   for ( ; contactIter != contacts.end(); ++contactIter)
   {
      cpLog( LOG_DEBUG, "Contact [%s]", contactIter->logData());
   }
   cpLog( LOG_DEBUG, "Contact list size %d", contacts.size());
   cpLog( LOG_DEBUG, "---------------------------------------------------------");

   if (contacts.size() == 0)
   {
      /// Empty contact list
      responseCode = 404;
   }

   return responseCode;
} // end of generateContactList



/**
 * Determine the location of the top Via in the contact list
 * @param  contacts  Contact list to be searched
 * @param  lastTwoVias  List of Via Sip addresses
 * @param  contactIter Location of the contact following the given Via
 * @return  True if Via was found, false otherwise
 */
bool 
SipContactIf::locateVia (vector <Data> & contacts,
                         const vector < Data > & lastTwoVias,
                         vector <Data>::iterator& contactIter)
{
    /// Create a copy of 'Via' and strip off the 'sip:'
    vector < Data > ::const_iterator viaIter = lastTwoVias.begin();
    Data viaEntry(*viaIter);
    cpLog( LOG_DEBUG, "top Via: %s", viaEntry.logData());
    viaEntry = stripOffSip(viaEntry);

    /// Set up the top Via by getting the server and group type
    /// Vias are already in the form host:port
    ServerType viaType = SERVER_Unknown;
    vector < Data > viaGroups;
    if (!RSServerContainer::instance().isValid(viaEntry, viaType, viaGroups))
    {
        cpLog(LOG_DEBUG, "*****Failed to get VIA group info");
    }

    //  cpLog(LOG_INFO, "*****VIA groups size:%d", viaGroups.size());

    bool viaFound = false;

    /// Determine if the top Via is a Marshal
    if ( SERVER_MS == viaType)
    {
        cpLog( LOG_DEBUG, "Top Via is a Marshal");
        if (++viaIter != lastTwoVias.end())
        {
            /// Create a copy of 'Via' and strip off the 'sip:'
            Data viaSecondEntry(*viaIter);
            cpLog( LOG_DEBUG, "2nd via: %s", viaSecondEntry.logData());
            viaSecondEntry = stripOffSip(viaSecondEntry);

            ServerType serverType = SERVER_Unknown;
            vector < Data > listofGroups;
            /// If the previous Via is not a server in our system then the
            /// call is entering our system. The Marshal will be at the
            /// top of the Via list, but will not be in the contact list.
            /// In this case the Via is considered to be found.
            if (RSServerContainer::instance().isValid(viaSecondEntry, serverType, listofGroups))
            {
                cpLog( LOG_DEBUG, "Previous Via is in our system, search for MS in contact list");
            }
            else
            {
                cpLog( LOG_DEBUG, "Previous Via is not in our system, take first contact");
                viaFound = true;
            }
        }
        else
        {
            cpLog( LOG_DEBUG, "Only one Via, take first contact");
            viaFound = true;
        }
    } /// end of top via is a Marshal

    contactIter = contacts.begin();

    if (true == viaFound)
    {
        return true;
    }
    else // Loop through the contact list in search of a match on the top VIA.
    {
        Data thisContact, thisContactHostPort;
        ServerType contactType = SERVER_Unknown;

        /// Locate the via in the contact list
        for ( ; contactIter != contacts.end(); ++contactIter)
        {
            cpLog( LOG_DEBUG, "Via not found yet");
            /// set up the current contact
            thisContact = *contactIter;
            //          cpLog(LOG_, "***** Contact Iter [%s]", (*contactIter).logData());
            cpLog(LOG_DEBUG, "this Contact [%s]", thisContact.logData());

            thisContactHostPort = getHostPort(thisContact);
            //          cpLog(LOG_INFO, "***** thiscontactHostPort [%s]", thisContactHostPort.logData());

            /// Check if contact matches the via
            if (thisContactHostPort == viaEntry)
            {
                cpLog( LOG_DEBUG, "Via matches contact's host:port  [%s]", thisContactHostPort.logData() );
                viaFound = true;
                ++contactIter;
                break;
            }

            /// Obtain the contact's groups
            vector < Data >  contactGroups;
            if (RSServerContainer::instance().isValid(thisContactHostPort, contactType, contactGroups))
            {
                /// needed the contactGroup info
            }

            /// Contact may be a group name or an address of the same group as the Via.
            /// Compare it to the Via's associated group or groups
            for (vector < Data > ::iterator groupIter = viaGroups.begin(); groupIter != viaGroups.end(); ++groupIter)
            {
                cpLog( LOG_DEBUG, "Comparing contact to via's group [%s]", (*groupIter).logData() );

                if ( thisContact == *groupIter )
                {
                    /// Via is of the same group type as this contact
                    cpLog( LOG_DEBUG, "Via's group matches this Contact [%s]", thisContact.logData() );
                    viaFound = true;
                    ++contactIter;
                    break;
                }

                /// Only checking the first group since a server should only belong to one group.
                if (contactGroups.begin() != contactGroups.end() )
                {
                    if ( *(contactGroups.begin()) == *groupIter )
                    {
                        cpLog( LOG_DEBUG, "Via and Contact are addresses belonging to the same group");
                        viaFound = true;
                        ++contactIter;
                        break;
                    }
                }
            }

            if (true == viaFound)
            {
                return true;
                break;
            }

        } /// done matching via to contact

        if (true == viaFound)
        {
            return true;
        }
    }

    return false;
} // end of locateVia

/**
 * Starting with the iterator location try to validate
 * a contact which will be used in a Status Response
 * to an INVITE message.
 * @param  contacts  Contact list to be searched
 * @param  contactIter Location of the first contact to be validated
 * @param  finalContact  Validated contact 
 * @return  Response code for Status Message
 */
int 
SipContactIf::obtainNextContact(const vector <Data> & contacts,
                                vector <Data>::iterator& contactIter,
                                bool destIsPhoneUser,
                                bool destIsRegistered,
                                Data & validContact)
{
    int responseCode = 302;

    GroupContainer& groupContainer = GroupContainer::instance();
    ServerType serverType = SERVER_Unknown;
    ServerType groupType = SERVER_Unknown;
    Data thisContact, thisContactHostPort;

    /// Via has been found, determine the next contact
    cpLog( LOG_DEBUG, "***** Via found, determine next contact");
    for ( ; contactIter != contacts.end(); ++contactIter)
    {
        thisContact = *contactIter;
        cpLog(LOG_DEBUG, "Validating Contact [%s]", thisContact.logData());
        //          cpLog(LOG_INFO, "***** Contact Iter [%s]", (*contactIter).logData());

        /// Contact can either be a server group name or server address.
        if (groupContainer.isValid(thisContact, groupType) )
        {
            cpLog( LOG_DEBUG, "Contact is a group name");
            /// contact is a group name, get an active address
            if (groupContainer.getServer(thisContact, validContact) )
            {
                /// Found a valid address for this contact
                cpLog( LOG_DEBUG, "Found a good Server");

                prepareContact(validContact, WILDCARDUSER, destIsPhoneUser);

                break;
            }
            else /// no active server available
            {
                /// No active servers are available for this group
                //cpLog( LOG_DEBUG,"Group's server type: %d", groupType );

                if ( SERVER_MS == groupType)
                {
                    /// fail the call - No server is available
                    responseCode = 503;
                    cpLog( LOG_NOTICE, 
			   "No active server for this Marshal. Failing call." );
                    return responseCode;
                    break;
                }
                /// A feature server can be skipped
                cpLog( LOG_DEBUG, 
		       "No active server available, skipping this contact");
                continue;
            }
        } /// end contact is a group name
        else /// contact is an address
        {
            thisContactHostPort = getHostPort(thisContact);
            //          cpLog(LOG_INFO, "***** thiscontactHostPort [%s]", thisContactHostPort.logData());

            cpLog( LOG_DEBUG, "Contact is address [%s]", thisContactHostPort.logData());

            /// Verify that contact is in our system and obtain the server type
            vector < Data > listOfGroups;
            if (RSServerContainer::instance().isValid(thisContactHostPort, serverType, listOfGroups))
            {
                /// Verify that contact is active
                if (Active == RSServerContainer::instance().getStatus(thisContactHostPort))
                {
                    /// Found a valid contact
                    validContact = thisContact;
                    break;
                }
                else
                {
                    cpLog( LOG_DEBUG, "Contact is NOT an active address");
                    /// If it is a marshal, look for another active marshal in the same group
                    if (SERVER_MS == serverType)
                    {
                        Data contactGroup(listOfGroups.front());
                        if (groupContainer.getServer(contactGroup, validContact) )
                        {
                            /// Found a valid contact
                            cpLog( LOG_DEBUG, "Found a good Server");
                            prepareContact(validContact, WILDCARDUSER, destIsPhoneUser);
                        }
                        else
                        {
                            cpLog( LOG_NOTICE, 
				   "Contact is an inactive Marshal with no alternates - fail call");
                            /// fail the call - No marshal is available
                            responseCode = 503;
                            return responseCode;
                        }
                        break;
                    } /// end MS == serverType
                    else
                    {
                        /// Otherwise feature servers can be skipped if
                        /// specified server is inactive.
                        /// Not going to look for another in the same group
                        continue;
                    }

                } /// contact is not active
            } /// contact is in our system

            else
            {
                cpLog( LOG_DEBUG, "Contact is an address, but doesn't match a server in our system. Use it as is.");
                /// Address doesn't match anything in our system
                /// Assume it is valid and send it as the next contact
                validContact = thisContact;
                break;
            }
        } /// end of contact is an address
    } /// end of contact for loop

    if (validContact == "")
    {   
        cpLog( LOG_DEBUG, "No valid contact found, failing call");
        if (destIsRegistered)
        {
            cpLog( LOG_DEBUG, "Subscriber is registered -- send 404");
            responseCode = 404;
        }
        else
        {
            responseCode = 480;
        }
    }

    return responseCode;
} // end of obtainNextContact

/**
 * Starting with the iterator location try to validate
 * each contact which will then be placed in a vector to be 
 * returned and used in a Status Response to an INVITE message.
 * @param  contacts  Contact list to be searched
 * @param  contactIter Location of the first contact to be validated
 * @param  destIsPhoneUser If true user parm = phone, else user parm = ip
 * @param  destIsRegistered Dialled number is a registered user.
 * @param  finalContacts  Validated contacts
 * @return  Response code for Status Message
 */
int
SipContactIf::validateRemainingContacts(const vector <Data> & contacts,
					vector <Data>::iterator& contactIter,
					bool destIsPhoneUser,
					bool destIsRegistered,
					vector <Data> & finalContacts)
{
    int responseCode = 302;
    
    GroupContainer& groupContainer = GroupContainer::instance();
    ServerType serverType = SERVER_Unknown;
    ServerType groupType = SERVER_Unknown;
    Data thisContact, thisContactHostPort, validContact;
    
    
    for ( ; contactIter != contacts.end(); ++contactIter)
    {
        thisContact = *contactIter;
        cpLog(LOG_DEBUG, "Validating Contact [%s]", thisContact.logData());
        //          cpLog(LOG_INFO, "***** Contact Iter [%s]", (*contactIter).logData());
	
        /// Contact can either be a server group name or server address.
	    if (groupContainer.isValid(thisContact, groupType) )
	    {
		cpLog( LOG_DEBUG, "Contact is a group name");
		/// contact is a group name, get an active address
		    if (groupContainer.getServer(thisContact, validContact) )
		    {
			/// Found a valid address for this contact
			cpLog( LOG_DEBUG, "Found a good Server");
			
			prepareContact(validContact, WILDCARDUSER, destIsPhoneUser);
			finalContacts.push_back(validContact);
			continue;
		    }
		    else /// no active server available
		    {
			/// No active servers are available for this group
			//cpLog( LOG_DEBUG,"Group's server type: %d", groupType );
			
			if ( SERVER_MS == groupType)
			{
			    if ( 0 == finalContacts.size())
			    {
				/// fail the call - No server is available
				responseCode = 503;
				cpLog( LOG_NOTICE, 
				       "No active server for this Marshal. Failing call." );
				return responseCode;
				break;
			    }
			    else
			    {
				/// Use contacts up until this point.
				cpLog( LOG_DEBUG, "No active server for this Marshal. Stopping validation." );
				break;
			    }
			}
			/// A feature server can be skipped
			    cpLog( LOG_DEBUG, "No active server available, skipping this contact");
			    continue;
		    }
	    } /// end contact is a group name
		else /// contact is an address
		{
		    thisContactHostPort = getHostPort(thisContact);
		    //          cpLog(LOG_INFO, "***** thiscontactHostPort [%s]", thisContactHostPort.logData());
		    
		    cpLog( LOG_DEBUG, "Contact is an address");
		    
		    /// Verify that contact is in our system and obtain the server type
			vector < Data > listOfGroups;
			if (RSServerContainer::instance().isValid(thisContactHostPort, serverType, listOfGroups))
			{
			    /// Verify that contact is active
			    if (Active == RSServerContainer::instance().getStatus(thisContactHostPort))
			    {
				/// Found a valid contact
				finalContacts.push_back(thisContact);
				continue;
			    }
			    else
			    {
				cpLog( LOG_DEBUG, "Contact is NOT an active address");
				/// If it is a marshal, look for another active marshal in the same group
				    if (SERVER_MS == serverType)
				    {
					Data contactGroup(listOfGroups.front());
					if (groupContainer.getServer(contactGroup, validContact) )
					{
					    /// Found a valid contact
					    cpLog( LOG_DEBUG, "Found a good Server");
					    prepareContact(validContact, WILDCARDUSER, destIsPhoneUser);
					    finalContacts.push_back(validContact);
					    continue;
					}
					else
					{
					    if ( 0 == finalContacts.size())
					    {
						cpLog( LOG_NOTICE, 
						       "Only contact is an inactive Marshal with no alternates - fail call");
						/// fail the call - No marshal is available
						    responseCode = 503;
						    return responseCode;
					    }
					    cpLog( LOG_DEBUG, "Contact is an inactive Marshal with no alternates. Stop validation.");
					    /// Use contacts up until this point.
						break;
					}
				    } /// end MS == serverType
					else
					{
					    /// Otherwise feature servers can be skipped if
					    /// specified server is inactive.
					    /// Not going to look for another in the same group
					    continue;
					}
				    
			    } /// contact is not active
				} /// contact is in our system
			    
			    else
			    {
				cpLog( LOG_DEBUG, "Contact is an address, but doesn't match a server in our system. Use it as is.");
				/// Address doesn't match anything in our system
				/// Assume it is valid and send it as the next contact
				    finalContacts.push_back(thisContact);
				    continue;
			    }
		} /// end of contact is an address
		} /// end of contact for loop
	
	
	if ( 0 == finalContacts.size())
	{
	    cpLog( LOG_DEBUG, "No valid contact found, failing call");
	    if (destIsRegistered)
	    {
		cpLog( LOG_DEBUG, "Subscriber is registered -- send 404");
		responseCode = 404;
	    }
	    else
	    {
		cpLog( LOG_DEBUG, "Subscriber is unregistered -- send 480");
		responseCode = 480;
	    }
	}
	
	return responseCode;
} // end of validateRemainingContacts


/**
 *  Look for a wildcard within each contact and replace it
 *  with the given user name.
 *  @param  userName  String which will replace wildcards
 *  @param  contacts  Contact list to be modified
 */
void 
SipContactIf::replaceWildcard(const Data & userName, vector <Data> & contacts)
{
    LocalScopeAllocator lo, lo2;
    for ( vector < Data > ::iterator i = contacts.begin(); i != contacts.end(); ++i )
    {
        string url( i->getData(lo) );

        // check for $USER string and replace with uriUser
        string::size_type pos = url.find(WILDCARDUSER);
        cpLog(LOG_DEBUG, " contact %s", url.c_str());
        if (string::npos != pos)
        {
            url.replace(pos, WILDCARDUSER.size(), userName.getData(lo2));
            cpLog(LOG_DEBUG, " changed contact %s", url.c_str());
        }

        *i = url;
    }
} // end of fillInWildcard


/**
 * strip off all but host:port of the contact
 * @param  contact  String to be manipulated
 * @return  Only the host:port portion of the passed in string.
 */
string
SipContactIf::getHostPort(const Data& contact )
{
    string workingContact = stripOffSip(contact);
    string::size_type pos = workingContact.find(";");
    if (pos != string::npos)
    {
        workingContact = workingContact.substr(0, pos);
        //      cpLog( LOG_DEBUG, "Stripped off ';user=ip/phone' [%s]", workingContact.c_str());
    }

    pos = workingContact.find("@");
    if (pos != string::npos)
    {
        workingContact = workingContact.substr(pos + 1, workingContact.size());
        //      cpLog( LOG_DEBUG, "Stripped off user@ [%s]", workingContact.c_str());
    }

    cpLog( LOG_DEBUG, "Stripped off all but contact's host:port [%s]", workingContact.c_str());
    return workingContact;

} /// end of getHostPort

/**
 * strip off all but user portion of the passed in SIP address
 * @param  sipAddress  String to be manipulated
 * @return  Only the user portion of the passed in string.
 */
string
SipContactIf::getUser(const Data& sipAddress)
{
    string userName = stripOffSip(sipAddress);
    string::size_type pos = userName.find("@");
    if (pos != string::npos)
    {
        userName = userName.substr(0, pos);
        cpLog( LOG_DEBUG, "Stripped off all but user: %s", userName.c_str());
    }
    return userName;
}

/**
 * Strip off 'sip:' from the given string
 * @param  withSip  String to be manipulated
 * @return  String with 'sip:' removed
 */
string
SipContactIf::stripOffSip(const Data& withSip)
{
    LocalScopeAllocator lo;
    string withoutSip = withSip.getData(lo);
    string::size_type pos = withoutSip.find("sip:");
    if (pos != string::npos)
    {
        withoutSip = withoutSip.substr(pos + 4, withoutSip.size());
        //      cpLog( LOG_DEBUG, "Stripped off sip: [%s]", withoutSip.c_str());
    }

    return withoutSip;

} /// end of stripOffSip


void
SipContactIf::prepareContact(Data & contact, const Data & userField, const bool isPhone)
{
    /// Add 'sip:' to the front of the identifier
    Data padding("sip:");
    /// Add user
    padding += userField;
    /// Add @
    padding += "@";
    contact = padding + contact;

    /// Add ;user=ip/phone
    if (true == isPhone)
    {
        padding = ";user=phone";
    }
    else
    {
        padding = ";user=ip";
    }
    contact += padding;
    cpLog( LOG_DEBUG, "Prepared contact: %s", contact.logData());
}

#ifdef VOCAL_HAS_ENUM
/*
 * Query ENUM (E164 to Url Mapping) to obtain a Sip url for the 
 * given E.164 number. 
 * @param  userField  E.164 number to be used in the query 
 * @param  resultingUrl  Corresponding Sip Url for given E.164 number.
 * @return  True if query returned a valid sip url, false otherwise
 */
bool
SipContactIf::queryEnum( Data & userField, Data & resultingUrl )
{
    bool activateEnum = false;
    /// Run time switch to turn ENUM service on and off.
    if (false == activateEnum)
        return false;

    ENUM_record *dns_results[MAX_ENUM_RESULTS];
    int ret;
    char* filter = "sip";
//  char *client_default_domains[]={"trial.e164.com.",NULL,NULL,NULL};
    char *client_default_domains[]={"trial.e164.com.","test.e164.enumworld.com.",NULL,NULL};
    char **root_domains = client_default_domains;

    /// Resolve the ENUM record
    LocalScopeAllocator lo;
    if((ret=enum_resolve((char*)(userField.getData(lo)), filter, root_domains, dns_results, MAX_ENUM_RESULTS)) < 1)
    {
        cpLog(LOG_DEBUG, "resolve_enum error: [%d] for user [%s]", ret, userField.logData());
        return false;
    }
    else
    {
        cpLog(LOG_DEBUG, "Results of resolve_enum [%s]", dns_results[0]->uri);
    }

    /// extract Uri from record
    if((resultingUrl = (dns_results[0]->uri)) )
    {
        cpLog(LOG_DEBUG, "No uri for [%s]", userField.logData());
        return false;
    }
    else
    {
        cpLog(LOG_DEBUG, "Valid uri: [%s]", resultingUrl.logData());
    }

    /// Clean up
    for(int i=0;i<ret;i++)
    {
        free_enum_record(dns_results[i]);
    }

    return true;
}
#endif
// End of File
