
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000-2003 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
static const char* const Subscriber_cxx_Version =
    "$Id: Subscriber.cxx,v 1.30.10.2 2003/03/13 17:03:01 bko Exp $";
#include <list>
#include "Subscriber.hxx"
#include "cpLog.h"
#include <algorithm>

const Data UNKNOWN = "";

/**
* Constructor
*/
Subscriber::Subscriber()
{

CallingContacts = new vector < Data > ;
CalledContacts = new vector < Data > ;
TermContacts = new vector < Data > ;
    secTilExpired = 0;
    RegTime = new VTime;
    StaticReg = false;
    ProvisionedState = MasterUnprovisioned;
    masterName = UNKNOWN;

}

/**
 * Copy Constructor 
 * @param  rhs  Subscriber object to copy
 */
Subscriber::Subscriber( const Subscriber& rhs )
{
    if ( this != &rhs )
    {
        CallingContacts = new vector < Data > ;
        CalledContacts = new vector < Data > ;
        TermContacts = new vector < Data > ;
        RegTime = new VTime;
        copyRhsToThis( rhs );
    }
}

/**
 * Constructor for master user with provisioning data.
 * @param  name  Master name
 * @param  calling  Contact list of subscriber's calling features.
 * @param  called  Contact list of subscriber's called features.
 * @param  staticFlag  Mark subscriber's static registration status.
 */
Subscriber::Subscriber( const Data& name,
            const Sptr < vector < Data > > calling,
            const Sptr < vector < Data > > called,
            const bool staticFlag) :
      StaticReg(staticFlag),
      ProvisionedState(Provisioned), 
      masterName(name)
{
    CallingContacts = new vector < Data > ;
    *CallingContacts = *calling;
    CalledContacts = new vector < Data > ;
    *CalledContacts = *called;
    TermContacts = new vector < Data > ;
    secTilExpired = 0;
    RegTime = new VTime;

}

/**
 * Set this subscriber equal to the given subscriber.
 * @param  rhs  Subscriber object to copy.
 * @return Reference to this subscriber
 */
Subscriber&
Subscriber::operator=( const Subscriber& rhs )
{
    if ( this != &rhs )
        copyRhsToThis( rhs );
    return *this;
}

/**
 * Copy the data in the given subscriber into this subscriber.
 * @param  rhs  Subscriber object to copy.
 */
void
Subscriber::copyRhsToThis( const Subscriber& rhs )
{
    *CallingContacts = *(rhs.CallingContacts);
    *CalledContacts = *(rhs.CalledContacts);
    *TermContacts = *(rhs.TermContacts);
    secTilExpired = rhs.secTilExpired;
    *RegTime = *(rhs.RegTime);
    StaticReg = rhs.StaticReg;
    ProvisionedState = rhs.ProvisionedState;
    masterName = rhs.masterName;
}


/**
 * Get the contact list used when this subscriber is making a call.
 * @param  contacts  Subscriber's calling contact list.
 */
Subscriber::~Subscriber()
{
    if (CallingContacts.getPtr() != 0) CallingContacts->clear();
    if (CalledContacts.getPtr() != 0) CalledContacts->clear();
    if (TermContacts.getPtr() != 0) TermContacts->clear();
}

/**
 * Get the contact list used when this subscriber is making a call.
 * @param  contacts  List to which subscriber's calling contacts will be appended.
 */
void
Subscriber::appendCalling( vector < Data > & contacts ) const
{
    for (vector < Data > ::iterator i = CallingContacts->begin(); i != CallingContacts->end(); ++i)
    {
        contacts.push_back(*i);
    }
}

/**
 * Get the contact list used when terminating on this subscriber.
 * @param  contacts  List to which subscriber's called contacts will be appended.
 */
void
Subscriber::appendCalled( vector < Data > & contacts ) const
{
    for (vector < Data > ::iterator i = CalledContacts->begin(); i != CalledContacts->end(); ++i)
    {
        // Avoid loop by keeping only the last one of the duplicated contacts
	// use an algorithm from the algorithm header (remove)
        contacts.erase(remove(contacts.begin(), contacts.end(), *i),
                       contacts.end());
	contacts.push_back(*i);
    }
}

/**
 * Get the contact list used when terminating on this subscriber.
 * @param  contacts  List to which subscriber's Term contacts will be appended.
 */
void
Subscriber::appendTerm( vector < Data > & contacts ) const
{
    for (vector < Data > ::iterator i = TermContacts->begin(); i != TermContacts->end(); ++i)
    {
        contacts.push_back(*i);
    }
}

/**
 * Get the amount of seconds remaining until registration becomes expired.
 * @return Seconds remaining before expiry.
 */
const unsigned long
Subscriber::getExpiry( ) const
{
    return secTilExpired;
}

/**
 * Erase the contacts used when this subscriber makes a call.
 */
void
Subscriber::clearCalling( )
{
    CallingContacts->clear();
}

/**
 * Erase the contacts used when a call is made to this subscriber.
 */
void
Subscriber::clearCalled( )
{
    CalledContacts->clear();
}

/**
 * Erase the contacts used when a call is made to this subscriber.
 */
void
Subscriber::clearTerm( )
{
    TermContacts->clear();
}

/**
 * Clear the expiry timer.
 */
void
Subscriber::clearTimer( )
{
    secTilExpired = 0;
}

/**
 * Store the feature contacts used when this subscriber makes a call.
 * @param  calling  List of contacts for calls made from this subscriber.
 */
void
Subscriber::setCalling (const Sptr < vector < Data > > calling)
{
    *CallingContacts = *calling;
}

/**
 * Store the feature contacts used when a call terminates on this subscriber.
 * @param  called  List of contacts for calls made to this subscriber.
 */
void
Subscriber::setCalled (const Sptr < vector < Data > > called)
{
    *CalledContacts = *called;
}

/**
 * Store the contacts used when a call terminates on this subscriber.
 * @param  term  List of contacts for calls made to this subscriber.
 */
void
Subscriber::setTerm (const Sptr < vector < Data > > term )
{

    /// A Register message could be used to de-register a user
    /// in this case the contact will be "*". Need to clear the contacts.
    if ((1 == term->size()) && ((*term)[0] == "*"))
    {
        TermContacts->clear();
        clearTimer();
    }
    else
    {
        //In future if RS supports returning mutiple-contacts in 302
        //Enable the following code
#if 0
  	list<Data> tData(TermContacts->begin(), TermContacts->end());
        for( vector < Data >::const_iterator itr = term->begin();
            itr != term->end(); itr++)
        {
                tData.push_back(*itr);
        }
	tData.unique();
	TermContacts->clear();
	for(list<Data>::iterator itr = tData.begin(); itr != tData.end(); itr++)
	{
	    TermContacts->push_back(*itr);
	}
#else
	if(term->size())
	{
           *TermContacts = *term;
	}
#endif
    }
}

/**
 * Store the expiry time remaining before re-registration.
 * @param  expire  Time in seconds before another register message is required.
 */
void
Subscriber::setExpiry( const unsigned long expire )
{
    secTilExpired = expire;
}

/**
 * Store the current time to mark when this Subscriber was last registered.
 */
void
Subscriber::setRegTime()
{
    VTime CurrentTime;
    *RegTime = CurrentTime;
    //  cpLog( LOG_DEBUG, "Subscriber::setRegTime -  seconds: %u fractions: %u", RegTime->getSeconds(), RegTime->getFractional());

}

/**
 * Check if subscriber's registration has expired and deregister if necessary.
 * @param  secondsRemaining  Amount of time in seconds before re-registration is required.
 * @return true if subscriber is expired, false otherwise.
 */
bool
Subscriber::expired(long & secondsRemaining)
{
    if (StaticReg)
    {
        secondsRemaining = INT_MAX;
        return false;
    }
    else
    {
        // cpLog( LOG_DEBUG, "Checking if registration has expired yet");
        /// Get the current time
        VTime CurrentTime;

        /// VTime calculations are done in msec. Convert delta
        /// of Current time and Registration time to seconds.
        unsigned long secondsSinceRegister = (CurrentTime - *RegTime) / 1000;

        if (( secondsSinceRegister > secTilExpired ) || (secTilExpired == 0 ))
        {
            /// Subscriber has expired. De-register the Subscriber
#if 1
            clearTerm();   /// in the future, clear out the terminating contacts.
#endif
            clearTimer();

            secondsRemaining = 0;
            return true;
        }
        else
        {
            secondsRemaining = secTilExpired - secondsSinceRegister;
            return false;
        }
    }
}

/**
 * Check to see if subscriber is an alias.
 * @param  name  Subscriber's identifier.
 * @return true if subscriber is an alias, false otherwise.
 */
bool
Subscriber::isAlias(const Data& name) const
{
    if ( AliasUnprovisioned == ProvisionedState )
    {
        return true;
    }
    else
    if ( MasterUnprovisioned == ProvisionedState )
    {
        return false;
    }
    else
    {
        if ( (masterName == name) || (UNKNOWN == masterName))
            return false;
    }
    return true;
}

/**
 *  Mark the subscriber as being for an unprovisioned Alias or Master
 *  Once provisioned, it could be representing both.
 *  @param  flag  Subscriber's alias status
 */
void
Subscriber::setAlias(const bool flag)
{
    if (true == flag)
    {
        ProvisionedState = AliasUnprovisioned;
    }
    else
    {
        ProvisionedState = MasterUnprovisioned;
    }
}

/**
 *  Mark the subscriber as having received data from provisioning
 *  @param  flag  Subscriber's provisioned status
 */
void 
Subscriber::setProvisioned(const bool flag)
{
    if (true == flag)
    {
        ProvisionedState = Provisioned;
    }
    else
    {
        ProvisionedState = MasterUnprovisioned;
    }
}

/**
 *  Check if Subscriber has gotten data from provisioning yet
 *  @return  true if provisioning data has been obtained for this subscriber
 */
bool 
Subscriber::isProvisioned() const
{
    if (Provisioned == ProvisionedState)
    {
        return true;
    }
    else
    {
        return false;
    }
}
