/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const DefaultAdmissionPolicy_cxx_Version = 
  "$Id: DefaultAdmissionPolicy.cxx,v 1.23 2002/05/15 11:35:43 icahoon Exp $";


#include "DefaultAdmissionPolicy.hxx"
#include "GatekeeperInfo.hxx"
#include "EndpointInfoMap.hxx"
#include "RASGKProtocol.hxx"
#include "Service.hxx"
#include "VLog.hxx"
#include <strstream>


using Vocal::H323GK::DefaultAdmissionPolicy;
using Vocal::H323GK::AdmissionPolicy;
using Vocal::H323GK::RASGKProtocol;
using Vocal::SipH323Info::GatekeeperInfo;
using Vocal::SipH323Info::EndpointInfoMap;
using Vocal::SipH323Info::EndpointInfo;
using Vocal::Services::Service;
using Vocal::Transport::IPAddress;
using Vocal::Logging::VLog;


DefaultAdmissionPolicy::DefaultAdmissionPolicy(
    Service  	    	&   service,
    RASGKProtocol   	&   rasGkProtocol,
    GatekeeperInfo  	&   gatekeeperInfo,
    EndpointInfoMap 	&   endpointInfoMap
)
    :	AdmissionPolicy(service, rasGkProtocol, gatekeeperInfo, endpointInfoMap)
{
}


DefaultAdmissionPolicy::~DefaultAdmissionPolicy()
{
}

			
void        
DefaultAdmissionPolicy::processAdmissionRequest(
    Sptr<H225_RasMessage>      	rasPDU, 
    Sptr<IPAddress>         	remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::processAdmissionRequest");
    VLog    	    log(fn);

    H225_AdmissionRequest	&   arq = *rasPDU;

    H225_AdmissionRejectReason      reason(H225_AdmissionRejectReason::e_undefinedReason);
    Sptr<EndpointInfo>      	    srcEndpointInfo = 0, 
	    	    	    	    destEndpointInfo = 0;
    
    // Let's first see if we can make some sense of it.
    //
    if ( !validateAdmissionRequest(arq, srcEndpointInfo, destEndpointInfo, reason, remoteAddr) )
    {
    	buildAdmissionReject(arq, reason, remoteAddr);
	return;
    }

    // Let the derived class decide how to process it.
    //
    Proceed proceed = onAdmission(rasPDU, remoteAddr, 
    	    	    	    	  srcEndpointInfo, destEndpointInfo,
				  arq, reason);
    switch ( proceed )
    {
	case wait:
	{
	    // The gatekeeper is probably waiting for an external 
	    // event to drive the admission forward. 
	    //
	    VDEBUG(log) << fn << ": waiting." << VDEBUG_END(log);
	    return;
	}

	case confirm:
	{
	    // Confirm
	    //
	    buildAdmissionConfirm(arq, srcEndpointInfo, destEndpointInfo, remoteAddr);

    	    break;
	}
	default:
	{
    	    buildAdmissionReject(arq, reason, remoteAddr);
	    break;
	}
    }
}


void   	    
DefaultAdmissionPolicy::processDisengageRequest(
    Sptr<H225_RasMessage>      	rasPDU,
    Sptr<IPAddress>         	remoteAddr
)
{
    const string    fn("DefaultRegistrationPolicy::processDisengageRequest");
    VLog    	    log(fn);

    // Used for per call memory analysis
    //
    #if ( 0 ) // defined(USE_LTHEAP)    
    {
	static  int     drq_count = 0;

	extern void HeapDump(FILE *, bool, bool, bool);
	strstream	    fileName;
	fileName << "callList" << drq_count << ".txt" << ends;
	FILE * filePtr = fopen(fileName.str(), "w");
	if ( filePtr )
	{
    	    VINFO(log) << fn << ": Logging memory to " << fileName.str() << VINFO_END(log);

	    HeapDump(filePtr, true, true, true);

	    fclose(filePtr);
	}
	else
	{
    	    VINFO(log) << fn << ": ARGH!! : " << fileName << VINFO_END(log);
	}
	fileName.freeze(false);
    }
    #endif // defined(USE_LT_HEAP)
    
    H225_DisengageRequest	&   drq = *rasPDU;

    H225_DisengageRejectReason      reason(H225_DisengageRejectReason::e_notRegistered);
    Sptr<EndpointInfo>      	    endpointInfo = 0;

    // Let's first see if we can make some sense of it.
    //
    if ( !validateDisengageRequest(drq, endpointInfo, reason) )
    {
    	buildDisengageReject(drq, reason, remoteAddr);
	return;
    }
    assert( endpointInfo != 0 );

    remoteAddr = responseAddress(endpointInfo, remoteAddr);

    // Let the derived class decide how to process it.
    //
    Proceed proceed = onDisengage(rasPDU, remoteAddr, endpointInfo, drq, reason);

    switch ( proceed )
    {
	case wait:
	{
	    // The gatekeeper is probably waiting for an external 
	    // event to drive the disengage forward. 
	    //
	    VDEBUG(log) << fn << ": waiting." << VDEBUG_END(log);
	    return;
	}

	case confirm:
	{
	    // Confirm
	    //
	    buildDisengageConfirm(drq, endpointInfo, remoteAddr);

	    break;
	}
	default:
	{
    	    buildDisengageReject(drq, reason, remoteAddr);
	    
	    break;
	}
    }
}			    


void   	    
DefaultAdmissionPolicy::processDisengageResponse(
    Sptr<H225_RasMessage>      	rasPDU,
    Sptr<IPAddress>         	remoteAddr
)
{
    const string fn("DefaultRegistrationPolicy::processDisengageConfirm");
    VLog    	 log(fn);

    onDisengageResponse(rasPDU);
}	    


void	    	    
DefaultAdmissionPolicy::disengage(
    H225_DisengageReason     	&   reason,
    Sptr<EndpointInfo>  	    endpointInfo,
    Sptr<IPAddress>     	    remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::disengage");
    VLog    	    log(fn);

    // Disengage
    // {
    //     requestSeqNum	:= 1
    //     endpointIdentifier	:= [from endpointInfo]
    //     conferenceID		:= [from finalize]
    //     callReferenceValue	:= [from finalize]
    //     disengageReason	:= [param]
    //     callIdentifier	:= [from finalize]
    // }

    H225_RasMessage    	    	    pdu;
    pdu.SetTag(H225_RasMessage::e_disengageRequest);

    H225_DisengageRequest	&   drq = pdu;

    drq.m_requestSeqNum = 1;

    drq.m_endpointIdentifier = endpointInfo->endpointIdentifier().c_str();
    
    drq.m_disengageReason = reason;
    
    finalizeDisengageRequest(drq, endpointInfo);

    // Ship it.
    //
    myRasGkProtocol.disengageRequest(pdu, remoteAddr);
}


ostream &           
DefaultAdmissionPolicy::writeTo(ostream & out) const
{
    return ( out << "DefaultAdmissionPolicy" );
}


Vocal::H323GK::RASPolicy::Proceed	    
DefaultAdmissionPolicy::onAdmission(
    Sptr<H225_RasMessage>      	    rasPDU, 
    Sptr<IPAddress> 	    	    remoteAddr,
    Sptr<EndpointInfo>  	    srcEptInfo,
    Sptr<EndpointInfo>  	    destEptInfo,
    H225_AdmissionRequest	&   arq,
    H225_AdmissionRejectReason  &   reason
)
{
    if ( srcEptInfo != 0 && arq.m_answerCall == false )
    {
        srcEptInfo->recordAdmission();
    }
    if ( destEptInfo != 0 && arq.m_answerCall == true )
    {
        destEptInfo->recordAdmission();
    }
    return ( confirm );
}

void	    DefaultAdmissionPolicy::finalizeAdmissionConfirm(
    H225_AdmissionRequest	&   arq,
    H225_AdmissionConfirm	&   acf,
    Sptr<EndpointInfo>  	    srcendpointInfo,
    Sptr<EndpointInfo>  	    destEndpointInfo
)
{
    // By default do nothing.
}


void	    
DefaultAdmissionPolicy::finalizeAdmissionReject(
    H225_AdmissionRequest	&   arq,
    H225_AdmissionReject    	&   arj
)
{
    // By default do nothing.
}


Vocal::H323GK::RASPolicy::Proceed     
DefaultAdmissionPolicy::onDisengage(
    Sptr<H225_RasMessage>    	    rasPDU, 
    Sptr<IPAddress> 	    	    remoteAddr,
    Sptr<EndpointInfo>  	    endpointInfo,
    H225_DisengageRequest 	&   drq,
    H225_DisengageRejectReason  &   reason
)
{
    endpointInfo->recordDisengage();
    
    return ( confirm );
}

	
void	    
DefaultAdmissionPolicy::finalizeDisengageRequest(
    H225_DisengageRequest   	&   drq,
    Sptr<EndpointInfo>  	    endpointInfo
)
{
    // By default do nothing.
}


void	    
DefaultAdmissionPolicy::finalizeDisengageConfirm(
    H225_DisengageRequest   	&   drq,
    H225_DisengageConfirm   	&   dcf,
    Sptr<EndpointInfo>  	    endpointInfo
)
{
    // By default do nothing.
}


void	    
DefaultAdmissionPolicy::finalizeDisengageReject(
    H225_DisengageRequest   	&   drq,
    H225_DisengageReject    	&   drj
)
{
    // By default do nothing.
}


void	    
DefaultAdmissionPolicy::onDisengageResponse(
    Sptr<H225_RasMessage>    	    rasPDU
)
{
    // By default do nothing.
}


Sptr<IPAddress>     
DefaultAdmissionPolicy::responseAddress(
    Sptr<EndpointInfo>  	    endpointInfo,
    Sptr<IPAddress>     	    remoteAddr
)
{
    // First try to get it from the endpointInfo
    //
    if  (   myGatekeeperInfo.returnToSender() == false
        &&  endpointInfo != 0 
        )
    {
    	remoteAddr = new IPAddress(endpointInfo->rasAddress());
    }

    // The last resort is to use the address obtained from the socket.
    //
    return ( remoteAddr );
}


bool
DefaultAdmissionPolicy::validateAdmissionRequest(
    H225_AdmissionRequest	    &   arq,
    Sptr<EndpointInfo>  	    &	srcEndpointInfo,
    Sptr<EndpointInfo>  	    &	destEndpointInfo,
    H225_AdmissionRejectReason      &	reason,
    Sptr<IPAddress>     	    &	remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::validateAdmissionRequest");
    VLog    	    log(fn);

    srcEndpointInfo = 0;
    destEndpointInfo = 0;

    // Pull the endpointIdentifier out of the arq.
    //
    string  endpointIdentifier;
    endpointIdentifier = arq.m_endpointIdentifier.GetValue();

    Sptr<EndpointInfo> endpointInfo 
    	= myEndpointInfoMap.findByEndpointId(endpointIdentifier);

    // No endpointIdentifier, no admission.
    //    
    if ( endpointInfo == 0 )
    {
    	reason.SetTag(H225_AdmissionRejectReason::e_invalidEndpointIdentifier);
    	return ( false );
    }

    // Update the remote address.
    //
    remoteAddr = responseAddress(endpointInfo, remoteAddr);
    
    // For gatekeeperRouted calls, the gatekeeper will be either the
    // source when receiving a call, or the destination when placing
    // a call.
    //
    Sptr<EndpointInfo>	gkEndpointInfo 
    	= myEndpointInfoMap.findByEndpointId(
	    	    	    	myGatekeeperInfo.gatekeeperIdentifier());

    // We are answering the call, so pull the src alias out of the arq.
    // If the alias isn't available, check the srcCallSignalAddress.
    // This is so that we can find the endpoint information for the
    // endpoint placing the call (the source). Note that we won't use the 
    // prefix to try to identify the source endpoint, since the prefix 
    // doesn't uniquely identify an endpoint.
    //
    if ( arq.m_answerCall )
    {
    	// Endpoint is answering the call, thus it is the destination.
	//
    	destEndpointInfo = endpointInfo;

	if  ( myGatekeeperInfo.gatekeeperRouted() )
	{
	    srcEndpointInfo = gkEndpointInfo;
	    return ( true );
	}

    	srcEndpointInfo = myEndpointInfoMap.findByAlias(arq.m_srcInfo);
	
	if ( srcEndpointInfo == 0 )
	{
	    // Check the srcCallSignalAddress
	    //
    	    if  (   arq.HasOptionalField(H225_AdmissionRequest::e_srcCallSignalAddress) 
    	    	&&  arq.m_srcCallSignalAddress.GetTag() == H225_TransportAddress::e_ipAddress 
		)
	    {
    	    	IPAddress	callSignalAddress;
		
		H225_TransportAddress_ipAddress & csAddr = arq.m_srcCallSignalAddress;
        	callSignalAddress.setIP(csAddr.m_ip.GetPointer());
		callSignalAddress.setPort(csAddr.m_port);

            	srcEndpointInfo = myEndpointInfoMap.findByCallSignalAddress(callSignalAddress);
	    }
	}
        
    	if ( srcEndpointInfo == 0 )
	{
    	    reason.SetTag(H225_AdmissionRejectReason::e_callerNotRegistered);
	    return ( false );
	}
		
    	// If we have gotten here, we are answering a call and we have
	// both the src and dest endpointInfo. So we can return successfully.
	//
	return ( true );
    }

    // We are placing the call, so we already have the src endpoint info
    // from the endpointIdentifier. We need to find the 
    //    
    srcEndpointInfo = endpointInfo;

    if ( myGatekeeperInfo.gatekeeperRouted() )
    {
	destEndpointInfo = gkEndpointInfo;
	return ( true );
    }
    
    // Try to find the destination endpoint by the destination alias
    //
    if  ( arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo) )
    {
        destEndpointInfo = myEndpointInfoMap.findByAlias(arq.m_destinationInfo);
    }

    // Try to match the destination endpoint by the call signal address
    //
    if ( destEndpointInfo == 0 )
    {
	IPAddress	callSignalAddress;

    	if  ( 	arq.HasOptionalField(H225_AdmissionRequest::e_destCallSignalAddress) 
    	    &&	arq.m_destCallSignalAddress.GetTag() == H225_TransportAddress::e_ipAddress )
	{
	    H225_TransportAddress_ipAddress & csAddr = arq.m_destCallSignalAddress;
            callSignalAddress.setIP(csAddr.m_ip.GetPointer());
	    callSignalAddress.setPort(csAddr.m_port);

            destEndpointInfo = myEndpointInfoMap.findByCallSignalAddress(callSignalAddress);
	}
    }

    bool outOfResources = false;
        
    // Try to match the destination endpoint to a gateway given the alias
    //
    if (    destEndpointInfo == 0 
        &&  arq.HasOptionalField(H225_AdmissionRequest::e_destinationInfo)
        ) 
    {
        destEndpointInfo = findGateway(arq.m_destinationInfo, outOfResources);
    }
    
    if ( destEndpointInfo == 0 )
    {
        if ( !outOfResources )
        {
    	    reason.SetTag(H225_AdmissionRejectReason::e_calledPartyNotRegistered);
        }
        else
        {
            reason.SetTag(H225_AdmissionRejectReason::e_resourceUnavailable);
        }
	return ( false );
    }

    // If we have gotten here, we are originating a call and we have
    // both the src and dest endpointInfo. So we can return successfully.
    //
    return ( true );
}


void	    
DefaultAdmissionPolicy::buildAdmissionConfirm(
    H225_AdmissionRequest   	&   arq, 
    Sptr<EndpointInfo>  	    srcEndpointInfo,
    Sptr<EndpointInfo>  	    destEndpointInfo,
    Sptr<IPAddress>     	    remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::buildAdmissionConfirm");
    VLog    	    log(fn);

    // AdmissionConfirm
    // {
    //     requestSeqNum    	    	:= [from arq]
    //     bandWidth			:= [from arq]
    //     callModel			:= direct
    //     destCallSignalAddress	:= [from destEndpointInfo]
    //     willRespondToIRR		:= false
    //     uuiesRequested
    // 	   {   
    // 	    	setup			:= false
    // 	    	callProceeding		:= false
    // 	    	connect			:= false
    // 	    	alerting		:= false
    // 	    	information		:= false
    // 	    	releaseComplete		:= false
    // 	    	facility		:= false
    // 	    	progress		:= false
    // 	    	empty			:= false
    // 	   }
    // }

    H225_RasMessage    	    	    pdu;
    pdu.SetTag(H225_RasMessage::e_admissionConfirm);

    H225_AdmissionConfirm	&   acf = pdu;

    acf.m_requestSeqNum = arq.m_requestSeqNum;

    acf.m_bandWidth = arq.m_bandWidth;
    
    if ( myGatekeeperInfo.gatekeeperRouted() )
    {
        acf.m_callModel.SetTag(H225_CallModel::e_gatekeeperRouted);
    }
    else
    {
        acf.m_callModel.SetTag(H225_CallModel::e_direct);
    }

    acf.m_destCallSignalAddress.SetTag(H225_TransportAddress::e_ipAddress);
    H225_TransportAddress_ipAddress & csAddr = acf.m_destCallSignalAddress;
    if ( myGatekeeperInfo.gatekeeperRouted() )
    {
	memcpy( &csAddr.m_ip[0], 
    		&(myGatekeeperInfo.callSignalAddress().getIPAddress().sin_addr.s_addr), 
		4);
	csAddr.m_port = myGatekeeperInfo.callSignalAddress().getPort();
    }
    else
    {
	memcpy( &csAddr.m_ip[0], 
    		&(destEndpointInfo->callSignalAddress().getIPAddress().sin_addr.s_addr), 
		4);
	csAddr.m_port = destEndpointInfo->callSignalAddress().getPort();
    }
    
    acf.m_willRespondToIRR = false;
    
    acf.m_uuiesRequested.m_setup = false;
    acf.m_uuiesRequested.m_callProceeding = false;
    acf.m_uuiesRequested.m_connect = false;
    acf.m_uuiesRequested.m_alerting = false;
    acf.m_uuiesRequested.m_information = false;
    acf.m_uuiesRequested.m_releaseComplete = false;
    acf.m_uuiesRequested.m_facility = false;
    acf.m_uuiesRequested.m_progress = false;
    acf.m_uuiesRequested.m_empty = false;


    finalizeAdmissionConfirm(arq, acf, srcEndpointInfo, destEndpointInfo);

    // Ship it.
    //

    ostrstream	confirmLog;

    // Source asking for admission.
    //
    if ( srcEndpointInfo != 0  && arq.m_answerCall == false )
    {
    	confirmLog  << "Admitted calling endpoint =\n{" 
		    << *srcEndpointInfo << '}';
    }

    // Destination asking for admission.
    //
    if ( destEndpointInfo != 0 && arq.m_answerCall == true )
    {
    	confirmLog  << "Admitted called endpoint =\n{" 
		    << *destEndpointInfo << '}';
    }
    
    confirmLog << ends;
    
    myRasGkProtocol.admissionConfirm(pdu, remoteAddr, confirmLog);

    confirmLog.freeze(false);
}


void	    
DefaultAdmissionPolicy::buildAdmissionReject(
    H225_AdmissionRequest       &   arq, 
    H225_AdmissionRejectReason  &   reason,
    Sptr<IPAddress>     	    remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::buildAdmissionReject");
    VLog    	    log(fn);

    // AdmissionReject
    // {
    //     requestSeqNum    	:= [from arq]
    //     rejectReason		:= [param]
    // }

    H225_RasMessage    	    	    pdu;
    pdu.SetTag(H225_RasMessage::e_admissionReject);

    H225_AdmissionReject	&   arj = pdu;

    arj.m_requestSeqNum = arq.m_requestSeqNum;
    
    arj.m_rejectReason = reason;

    finalizeAdmissionReject(arq, arj);

    // Ship it.
    //
    myRasGkProtocol.admissionReject(pdu, remoteAddr);
}


bool
DefaultAdmissionPolicy::validateDisengageRequest(
    H225_DisengageRequest	&   drq,
    Sptr<EndpointInfo>      	&   endpointInfo,
    H225_DisengageRejectReason  &   reason
)
{
    // Pull the endpointIdentifier out of the arq.
    //
    string  endpointIdentifier;
    endpointIdentifier = drq.m_endpointIdentifier.GetValue();

    endpointInfo = myEndpointInfoMap.findByEndpointId(endpointIdentifier);
    
    if ( endpointInfo == 0 )
    {
    	reason.SetTag(H225_DisengageRejectReason::e_notRegistered);
    	return ( false );
    }

    return ( true );
}


void	    	    
DefaultAdmissionPolicy::buildDisengageConfirm(
    H225_DisengageRequest	&   drq, 
    Sptr<EndpointInfo>  	    endpointInfo,
    Sptr<IPAddress>     	    remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::buildDisengageConfirm");
    VLog    	    log(fn);

    // DisengageConfirm
    // {
    //     requestSeqNum	:= [from drq]
    // }

    H225_RasMessage    	    	    pdu;
    pdu.SetTag(H225_RasMessage::e_disengageConfirm);

    H225_DisengageConfirm	&   dcf = pdu;

    dcf.m_requestSeqNum = drq.m_requestSeqNum;
    
    finalizeDisengageConfirm(drq, dcf, endpointInfo);

    // Ship it.
    //
    ostrstream	confirmLog;

    confirmLog << "Disengage endpoint =\n{" << *endpointInfo << '}' << ends;

    myRasGkProtocol.disengageConfirm(pdu, remoteAddr, confirmLog);
    
    confirmLog.freeze(false);
}


void	    	    
DefaultAdmissionPolicy::buildDisengageReject(
    H225_DisengageRequest       &   drq,
    H225_DisengageRejectReason  &   reason,
    Sptr<IPAddress>     	    remoteAddr
)
{
    const string    fn("DefaultAdmissionPolicy::buildDisengageConfirm");
    VLog    	    log(fn);

    // DisengageReject
    // {
    //     requestSeqNum	:= [from drq]
    // }

    H225_RasMessage    	    	    pdu;
    pdu.SetTag(H225_RasMessage::e_disengageReject);

    H225_DisengageReject	&   drj = pdu;

    drj.m_requestSeqNum = drq.m_requestSeqNum;

    drj.m_rejectReason = reason;
    
    finalizeDisengageReject(drq, drj);

    // Ship it.
    //
    myRasGkProtocol.disengageReject(pdu, remoteAddr);
}


Sptr<EndpointInfo>  
DefaultAdmissionPolicy::findGateway(
    H225_ArrayOf_AliasAddress   &   aliases,
    bool                        &   outOfResources
)
{
    const string    fn("DefaultAdmissionPolicy::findGateway");
    VLog            log(fn);
    
    Sptr<EndpointInfo>  endpointInfo = 0;
    
    string  alias("");
    size_t  size = aliases.GetSize();
    bool    found = false;

    // Look for the e.164 alias. No e.164 alias, no match.
    //
    for ( size_t i = 0; i < size; i++ )
    {
        if ( aliases[i].GetTag() == H225_AliasAddress::e_dialedDigits )
        {
            PASN_IA5String & e164 = aliases[i];
            alias = e164.GetValue();
            found = true;
            break;
        }
    }

    if ( found == true )
    {
        // Find the gateway that matches this alias.
        //
        string  matched_prefix("");

        EndpointInfoSet gateways 
            = myEndpointInfoMap.findByPrefix(alias, matched_prefix);

        endpointInfo = chooseGateway(gateways, outOfResources);

        if ( endpointInfo != 0 )
        {
            VDEBUG(log) << fn << ": alias = \"" << alias
                        << "\" matched prefix = \"" << matched_prefix 
                        << '"' << VDEBUG_END(log);

            // See if we need to strip the prefix
            //
            if ( endpointInfo->stripPrefix() )
            {
                assert( alias.size() >= matched_prefix.size() );

                if ( alias == matched_prefix )
                {
                    alias.erase();
                }
                else
                {
                    alias = alias.substr(matched_prefix.size());       
                }

                VDEBUG(log) << fn << ": stripped prefix, alias = \"" 
                            << alias 
                            << '"' << VDEBUG_END(log);
            }

            // Build a new alias list for...
            //
            H225_ArrayOf_AliasAddress       new_aliases;

            size = new_aliases.GetSize();
            new_aliases.SetSize(size + 1);

            new_aliases[size].SetTag(H225_AliasAddress::e_dialedDigits);

            PASN_IA5String  &   new_e164 = new_aliases[size];
            new_e164 = alias.c_str();

            // ... the new endpoint info. We do not want to overwrite
            // the alias of the gateway in the list, so build a new one.
            //
            endpointInfo = new EndpointInfo(*endpointInfo);
            endpointInfo->alias(new_aliases);
        }
    }
    
    return ( endpointInfo );
}


Sptr<EndpointInfo>  
DefaultAdmissionPolicy::chooseGateway(
    const EndpointInfoSet   &   gateways,
    bool                    &   outOfResources
)
{
    Sptr<EndpointInfo> endpointInfo = 0;
    
    outOfResources = false;
    
    for ( EndpointInfoSet::const_iterator i = gateways.begin();
          i != gateways.end();
          ++i
        )
    {
        if ( !(*i)->outOfResources() )
        {
            endpointInfo = *i;
            outOfResources = false;
            break;
        }
        else
        {
            outOfResources = true;
        }
    }

    return  ( endpointInfo );
}
