/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const RASGKProtocol_cxx_Version = 
    "$Id: RASGKProtocol.cxx,v 1.17 2002/05/15 11:35:43 icahoon Exp $";


#include "RASGKProtocol.hxx"
#include "PollFifo.hxx"
#include "DiscoveryPolicy.hxx"
#include "RegistrationPolicy.hxx"
#include "AdmissionPolicy.hxx"
#include "LocationPolicy.hxx"
#include "ResourcesAvailablePolicy.hxx"
#include "SocketOptions.hxx"
#include "PollFifoService.hxx"
#include "VLog.hxx"
#include "CallTrace.hxx"
#include <strstream>


using Vocal::H323GK::RASGKProtocol;
using Vocal::Transport::Protocol;
using Vocal::Services::PollFifoService;
using Vocal::H323GK::DiscoveryPolicy;
using Vocal::H323GK::RegistrationPolicy;
using Vocal::H323GK::AdmissionPolicy;
using Vocal::H323GK::LocationPolicy;
using Vocal::H323GK::ResourcesAvailablePolicy;
using Vocal::IO::FileDescriptor;
using Vocal::Transport::IPAddress;
using Vocal::Transport::UDPSocket;
using Vocal::Transport::SocketOptions;
using Vocal::Logging::VLog;


RASGKProtocol::RASGKProtocol(
    PollFifoService    	    &	pollFifoService,
    const IPAddress 	    &   rasAddress
)
    :	Protocol("RASGK"),
    	myPollFifoService(pollFifoService),
	myDiscoveryPolicy(0),
	myRegistrationPolicy(0),
	myAdmissionPolicy(0),
        myLocationPolicy(0),
        myResourcesAvailablePolicy(0),
	myRasAddress(rasAddress),
	myRasSocket(0),
	myInPerStreamSize(1024)
{
    // Create and bind the socket to the ras address specified.
    //
    myRasSocket = new UDPSocket(myRasAddress);
    myRasSocket->setNonblocking();

    // Make it bsd compatible so that errors show up when they happen,
    // not on the system call after. This may change with future versions
    // of the kernel.
    //
    SocketOptions   	rasSocketOptions(*myRasSocket);
    rasSocketOptions.bsdCompatible(true);

    awaitIncoming(true);
}
 

RASGKProtocol::~RASGKProtocol()
{
}


void	    	    	    
RASGKProtocol::setDiscoveryPolicy(DiscoveryPolicy & discoveryPolicy)
{
    myDiscoveryPolicy = &discoveryPolicy;
}


void	    	    	    
RASGKProtocol::setRegistrationPolicy(RegistrationPolicy & registrationPolicy)
{
    myRegistrationPolicy = &registrationPolicy;
}


void	    	    	    
RASGKProtocol::setAdmissionPolicy(AdmissionPolicy & admissionPolicy)
{
    myAdmissionPolicy = &admissionPolicy;
}
	

void	    	    	    
RASGKProtocol::setLocationPolicy(LocationPolicy & locationPolicy)
{
    myLocationPolicy = &locationPolicy;
}
	

void	    	    	    
RASGKProtocol::setResourcesAvailablePolicy(ResourcesAvailablePolicy & resourcesAvailablePolicy)
{
    myResourcesAvailablePolicy = &resourcesAvailablePolicy;
}
	

const FileDescriptor &	    
RASGKProtocol::getFileDescriptor() const
{
    assert( myRasSocket != 0 );
    return ( *myRasSocket );
}


void	    	    	    
RASGKProtocol::discoveryConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream 	    	&   confirmLog
)
{
    VLog    log;
    H225_GatekeeperConfirm	&   gcf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "GCF") << gcf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}
					

void	    	    	    
RASGKProtocol::discoveryReject(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_GatekeeperReject	&   grj = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "GRJ") << grj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}

					
void	    	    	    
RASGKProtocol::registrationConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream	    	&   confirmLog
)
{
    VLog    log;
    H225_RegistrationConfirm	&   rcf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "RCF") << rcf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::registrationReject(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_RegistrationReject	&   rrj= pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "RRJ") << rrj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::unregistrationRequest(	
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_UnregistrationRequest	&   urq = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "URQ") << urq << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::unregistrationConfirm(	
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream 	    	&   confirmLog
)
{
    VLog    log;
    H225_UnregistrationConfirm	&   ucf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "UCF") << ucf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::unregistrationReject(	
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_UnregistrationReject	&   urj = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "URJ") << urj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::admissionConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream      	&   confirmLog
)
{
    VLog    log;
    H225_AdmissionConfirm	&   acf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "ACF") << acf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}					


void	    	    	    
RASGKProtocol::admissionReject(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_AdmissionReject	&   arj = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "ARJ") << arj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::disengageRequest(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_DisengageRequest	&   drq = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "DRQ") << drq << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::disengageConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream    	&   confirmLog
)
{
    VLog    log;
    H225_DisengageConfirm	&   dcf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "DCF") << dcf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::disengageReject(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_DisengageReject	&   drj = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "DRJ") << drj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void                        
RASGKProtocol::locationRequest(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_LocationRequest	&   lrq = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "LRQ") << lrq << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}
                                        

void                       
RASGKProtocol::locationConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote,
    ostrstream    	&   confirmLog
)
{
    VLog    log;
    H225_LocationConfirm	&   lcf = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "LCF") << lcf 
    	<< '\n' << confirmLog.str() << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void                       
RASGKProtocol::locationReject(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_LocationReject 	&   lrj = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "LRJ") << lrj << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void                       
RASGKProtocol::resourcesAvailableConfirm(
    H225_RasMessage     &   pdu,
    Sptr<IPAddress>         remote
)
{
    VLog    log;
    H225_ResourcesAvailableConfirm	&   rac = pdu;
    
    CALLTRACE(log, CT_OUT, *remote, CT_RAS, "RAC") << rac << CALLTRACE_END(log);

    sendMessage(pdu, remote);
}


void	    	    	    
RASGKProtocol::onIncomingAvailable()
{
    const string    fn("RASGKProtocol::onIncomingAvailable");
    VLog	    log(fn);

    // Read the message from the socket.
    //
    Sptr<IPAddress> 	remoteAddr = new IPAddress;

    myInPerStream = new PPER_Stream;
    
    int bytesReceived = myRasSocket->receiveFrom
    	    	    	(
    	    	    	    myInPerStream->GetPointer(myInPerStreamSize),
			    myInPerStreamSize,
			    *remoteAddr
			);

    // If we receive an empty message, return.
    //
    if ( bytesReceived == 0 )
    {
    	return;
    }
    
    // See if we need to grow the buffer.
    //
    if ( bytesReceived == (int)myInPerStreamSize )
    {

	VDEBUG(log) << fn << ": resizing PER stream, old size = " 
	    	    << myInPerStreamSize
		    << ", new size = " << (myInPerStreamSize * 2)  
		    << VDEBUG_END(log);

    	// Double the size.
	//
	myInPerStreamSize *= 2;
    }

    VINFO(log)  << fn << ": packet arrived from address = { " << *remoteAddr 
                << " } packet dump, size = " << bytesReceived;

    u_int8_t    *   bytes = myInPerStream->GetPointer();
	
    for ( int i = 0; i < bytesReceived; i++ )
    {
    	if( i % 16 == 0 )
	{
	    VINFO_CONT(log) << "\n\t";
	}
	VINFO_CONT(log) << setfill('0') << setw(2) << hex 
	    	        << (unsigned)bytes[i] << ' ';
    }
    VINFO_CONT(log) << VINFO_END(log);
		
    // Decode the RAS message.
    //    
    Sptr<H225_RasMessage>   rasPDU = new H225_RasMessage;

    if ( !rasPDU->Decode(*myInPerStream) )
    {
    	VWARN(log)  << fn << ": Incoming RAS PDU decode failed." 
	    	    << VWARN_END(log);
	return;
    }

    // Forward the rasPDU and remoteAddress to the appropriate policy.
    //
    switch ( rasPDU->GetTag() )
    {

    	// Received GRQ
	//
    	case H225_RasMessage::e_gatekeeperRequest:
	{
	    if ( myDiscoveryPolicy )
	    {
		H225_GatekeeperRequest	&   grq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "GRQ Unicast") 
		    << grq << CALLTRACE_END(log);

    	    	myDiscoveryPolicy->processUnicastDiscoveryRequest(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding GRQ. DiscoveryPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
    	}
	
	// Received RRQ
	//
    	case H225_RasMessage::e_registrationRequest:
	{
	    if ( myRegistrationPolicy )
	    {
		H225_RegistrationRequest	&   rrq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "RRQ") 
		    << rrq << CALLTRACE_END(log);

    	    	myRegistrationPolicy->processRegistrationRequest(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding RRQ. RegistrationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}
	
	// Received URQ
	//
    	case H225_RasMessage::e_unregistrationRequest:
	{
	    if ( myRegistrationPolicy )
	    {
		H225_UnregistrationRequest	&   urq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "URQ") 
		    << urq << CALLTRACE_END(log);

    	    	myRegistrationPolicy->processUnregistrationRequest(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding URQ. RegistrationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

    	// Received UCF
	//
    	case H225_RasMessage::e_unregistrationConfirm:
	{
	    if ( myRegistrationPolicy )
	    {
		H225_UnregistrationConfirm	&   ucf = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "UCF") 
		    << ucf << CALLTRACE_END(log);

	    	myRegistrationPolicy->processUnregistrationResponse(rasPDU, remoteAddr);
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding UCF. RegistrationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

    	// Received URJ
	//
    	case H225_RasMessage::e_unregistrationReject:
	{
	    if ( myRegistrationPolicy )
	    {
		H225_UnregistrationReject	&   urj = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "URJ") 
		    << urj << CALLTRACE_END(log);

	    	myRegistrationPolicy->processUnregistrationResponse(rasPDU, remoteAddr);
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding UCF. RegistrationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	// Received ARQ
	//
    	case H225_RasMessage::e_admissionRequest:
	{
	    if ( myAdmissionPolicy )
	    {
		H225_AdmissionRequest	    	&   arq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "ARQ") 
		    << arq << CALLTRACE_END(log);

	    	myAdmissionPolicy->processAdmissionRequest(rasPDU, remoteAddr);
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding ARQ. AdmissionPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}
	
	// Received DRQ
	//
    	case H225_RasMessage::e_disengageRequest:
	{
	    if ( myAdmissionPolicy )
	    {
		H225_DisengageRequest	    	&   drq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "DRQ") 
		    << drq << CALLTRACE_END(log);

	    	myAdmissionPolicy->processDisengageRequest(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding DRQ. AdmissionPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	// Received DCF
	//
    	case H225_RasMessage::e_disengageConfirm:
	{
	    if ( myAdmissionPolicy )
	    {
		H225_DisengageConfirm	    	&   dcf = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "DCF") 
		    << dcf << CALLTRACE_END(log);

	    	myAdmissionPolicy->processDisengageResponse(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding DCF. AdmissionPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	// Received DRJ
	//
    	case H225_RasMessage::e_disengageReject:
	{
	    if ( myAdmissionPolicy )
	    {
		H225_DisengageReject	    	&   drj = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "DRJ") 
		    << drj << CALLTRACE_END(log);

	    	myAdmissionPolicy->processDisengageResponse(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding DRJ. AdmissionPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	// Received LRQ
	//
    	case H225_RasMessage::e_locationRequest:
	{
	    if ( myLocationPolicy )
	    {
		H225_LocationRequest    &   lrq = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "LRQ") 
		    << lrq << CALLTRACE_END(log);

	    	myLocationPolicy->processUnicastLocationRequest(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding LRQ. LocationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}


	// Received LCF
	//
    	case H225_RasMessage::e_locationConfirm:
	{
	    if ( myLocationPolicy )
	    {
		H225_LocationConfirm    &   lcf = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "LCF") 
		    << lcf << CALLTRACE_END(log);

	    	myLocationPolicy->processLocationResponse(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding LCF. LocationPolicy not set."
	    		    << VWARN_END(log);
	    }
	    break;
	}


	// Received LRJ
	//
    	case H225_RasMessage::e_locationReject:
	{
	    if ( myLocationPolicy )
	    {
		H225_LocationReject &   lrj = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "LRJ") 
		    << lrj << CALLTRACE_END(log);

	    	myLocationPolicy->processLocationResponse(rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding LRJ. LocationPolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	// Received RAI
	//
    	case H225_RasMessage::e_resourcesAvailableIndicate:
	{
	    if ( myResourcesAvailablePolicy )
	    {
		H225_ResourcesAvailableIndicate &   rai = *rasPDU;
		
		CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, "RAI") 
		    << rai << CALLTRACE_END(log);

	    	myResourcesAvailablePolicy->processResourcesAvailableIndication(
                                                rasPDU, remoteAddr); 
	    }
	    else
	    {
    		VWARN(log)  << fn << ": Discarding RAI. ResourcesAvailablePolicy not set." 
	    		    << VWARN_END(log);
	    }
	    break;
	}

	default:
	{
	    CALLTRACE(log, CT_IN, *remoteAddr, CT_RAS, string(rasPDU->GetTagName())) 
		<< "{\n  " << setprecision(2) << *rasPDU 
		<< "\n}" << CALLTRACE_END(log);

    	    VDEBUG(log) << fn << ": Unhandled RAS PDU"
		    	<< ", message type: " << rasPDU->GetTagName()
	    	    	<< VDEBUG_END(log);
    	    break;
	}
    }
}


void	    	    	    
RASGKProtocol::onOutgoingAvailable()
{
    const string    fn("RASGKProtocol::onOutgoingAvailable");
    VLog	    log(fn);

    while ( !myOutgoingMessages.empty() )
    {
    	Sptr<OutgoingMessage> outgoingMessage = myOutgoingMessages.front();
	
        myRasSocket->sendTo
	(
    	    outgoingMessage->outgoingStream->GetPointer(),
	    outgoingMessage->outgoingStream->GetSize(),
	    *outgoingMessage->remoteAddr
	);

    	myOutgoingMessages.pop();
    }
    
    awaitOutgoing(false);

    myPollFifoService.registerProtocol(*this);
}


void	    	    	    
RASGKProtocol::sendMessage(
    H225_RasMessage  	&   rasPDU,
    Sptr<IPAddress>         remote
)
{
    const string    fn("RASGKProtocol::sendMessage");
    VLog	    log(fn);

    if ( remote == 0 )
    {
    	return;
    }

    Sptr<PPER_Stream> outPerStream = new PPER_Stream;
        
    rasPDU.Encode(*outPerStream);
    outPerStream->CompleteEncoding();

    try
    {    
        int bytesSent = myRasSocket->sendTo
			(
			    outPerStream->GetPointer(),
			    outPerStream->GetSize(),
			    *remote
			);


	VINFO(log) << fn << ": sending packet to address = { " << *remote
        	   << " } packet dump, size = " << bytesSent;

        u_int8_t    *   bytes = outPerStream->GetPointer();

	for ( int i = 0; i < bytesSent; i++ )
	{

    	    if( i % 16 == 0 )
	    {
		VINFO_CONT(log) << "\n\t";
	    }
	    VINFO_CONT(log) << setfill('0') << setw(2) << hex 
	    	    	    << (unsigned)bytes[i] << ' ';
	}
	VINFO_CONT(log) << VINFO_END(log);
    }
    catch ( Vocal::SystemStatus & )
    {
    	Sptr<OutgoingMessage> outgoingMessage = new OutgoingMessage;
	outgoingMessage->outgoingStream = outPerStream;
	outgoingMessage->remoteAddr = remote;
	
	myOutgoingMessages.push(outgoingMessage);
	
	awaitOutgoing(true);
	
	myPollFifoService.registerProtocol(*this);
    }
}
