/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000, 2001 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const EndpointInfo_cxx_Version = 
    "$Id: EndpointInfo.cxx,v 1.20 2002/12/10 00:09:32 bko Exp $";


#include "EndpointInfo.hxx"
#include "VLog.hxx"

#ifdef __FreeBSD__
#define MAXINT INT_MAX
#else
#include <values.h>
#endif


using Vocal::SipH323Info::EndpointInfo;
using Vocal::Transport::IPAddress;
using Vocal::TimeAndDate::TimeVal;
using Vocal::Logging::VLog;
using Vocal::Threads::WriteLock;
using Vocal::Threads::ReadLock;
using Vocal::Services::Event;
using Vocal::FifoBase;
using Vocal::TimeVal;


EndpointInfo::EndpointInfo()
    :	myEndpointIdentifier(""),
        myEndpointType(UNDEFINED_TYPE),
        myStripPrefix(false),
        myStaticRegistration(false),
    	myTimeToLive(0),
	myTimeToLiveTimerEventId(0),
        myOutOfResources(false),
        myRegistrationCount(0),
        myAdmissionCount(0),
        myDisengageCount(0),
        myResourcesAvailableCount(0)
{
}


EndpointInfo::EndpointInfo(H225_RegistrationRequest & rrq)
    :	myEndpointIdentifier(""),
        myEndpointType(UNDEFINED_TYPE),
        myStripPrefix(false),
        myStaticRegistration(false),
    	myTimeToLive(0),
	myTimeToLiveTimerEventId(0),
        myOutOfResources(false),
        myRegistrationCount(0),
        myAdmissionCount(0),
        myDisengageCount(0),
        myResourcesAvailableCount(0)
{
    update(rrq);
}


EndpointInfo::EndpointInfo(
    const string        &   eptId,
    const IPAddress     &   csAddr,
    const list<string>  &   prefixes,
    bool                    strip_prefix
)
    :	myEndpointIdentifier(eptId),
        myEndpointType(GATEWAY_TYPE),
        myStripPrefix(strip_prefix),
        myStaticRegistration(true),
        myCallSignalAddress(csAddr),
    	myTimeToLive(0),
	myTimeToLiveTimerEventId(0),
        myOutOfResources(false),
        myRegistrationCount(0),
        myAdmissionCount(0),
        myDisengageCount(0),
        myResourcesAvailableCount(0)
{
    prefix(prefixes);
}


EndpointInfo::EndpointInfo(const EndpointInfo & src)
{
    ReadLock    lock(src.myMutex);
    
    myEndpointIdentifier        = src.myEndpointIdentifier;
    myEndpointType              = src.myEndpointType;
    myAlias                     = src.myAlias;
    myPrefix                    = src.myPrefix;
    myStripPrefix               = src.myStripPrefix;
    myStaticRegistration        = src.myStaticRegistration;
    myRasAddress                = src.myRasAddress;
    myCallSignalAddress         = src.myCallSignalAddress;
    myTimeToLive                = src.myTimeToLive;
    myTimeToLiveTimerEventId    = src.myTimeToLiveTimerEventId;
    myOutOfResources            = src.myOutOfResources;

    myRegistrationCount         = src.myRegistrationCount;
    myFirstRegistrationTime     = src.myFirstRegistrationTime;
    myLastRegistrationTime      = src.myLastRegistrationTime;

    myLastKeepAliveTime         = src.myLastKeepAliveTime;
    
    myAdmissionCount            = src.myAdmissionCount;
    myFirstAdmissionTime        = src.myFirstAdmissionTime;
    myLastAdmissionTime         = src.myLastAdmissionTime;

    myDisengageCount            = src.myDisengageCount;
    myFirstDisengageTime        = src.myFirstDisengageTime;
    myLastDisengageTime         = src.myLastDisengageTime;

    myResourcesAvailableCount    = src.myResourcesAvailableCount;
    myFirstResourcesAvailableTime = src.myFirstResourcesAvailableTime;
    myLastResourcesAvailableTime  = src.myLastResourcesAvailableTime;

    myLastActivityTime          = src.myLastActivityTime;
}


EndpointInfo::~EndpointInfo()
{
}


const EndpointInfo &
EndpointInfo::operator=(const EndpointInfo & src)
{
    WriteLock    lock(myMutex);

    if ( this != &src )
    {
    	ReadLock    lock(src.myMutex);
	
        myEndpointIdentifier        = src.myEndpointIdentifier;
        myEndpointType              = src.myEndpointType;
        myAlias                     = src.myAlias;

        // If the endpoint has statically registered, don't allow
        // the prefix to be clobbered.
        //
        if ( !myStaticRegistration )
        {
            myPrefix                    = src.myPrefix;
            myStripPrefix               = src.myStripPrefix;
            myStaticRegistration        = src.myStaticRegistration;
        }

        myRasAddress                = src.myRasAddress;
        myCallSignalAddress         = src.myCallSignalAddress;
        myTimeToLive                = src.myTimeToLive;
    }
    return ( *this );
}


const bool  
EndpointInfo::operator==(const EndpointInfo & src) const
{
    if ( this == &src )
    {
        return ( true );
    }
    
    ReadLock    lock(myMutex);
    ReadLock    srcLock(src.myMutex);
    
    return ( myEndpointIdentifier == src.myEndpointIdentifier );
}


const bool  
EndpointInfo::operator<(const EndpointInfo & src) const
{
    if ( this == &src )
    {
        return ( false );
    }
    
    ReadLock    lock(myMutex);
    ReadLock    srcLock(src.myMutex);
    
    return ( myEndpointIdentifier < src.myEndpointIdentifier );
}


bool	    	
EndpointInfo::update(H225_RegistrationRequest &	rrq)
{
    // Put the write lock in it's own scope. Protects against the case 
    // where someone in the future adds an early exit but forgets to unlock.
    //
    {
        WriteLock   lock(myMutex);

        // Pull the endpointIdentifier out of the rrq.
        //
        if ( rrq.HasOptionalField(H225_RegistrationRequest::e_endpointIdentifier) )
        {
    	    string eptId;
	    eptId = rrq.m_endpointIdentifier.GetValue();
    	    myEndpointIdentifier = eptId;
        }

        // Pull the endpointType out of the rrq.
        //
        myEndpointType = UNDEFINED_TYPE;

        if ( rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_terminal) )
        {
            myEndpointType = TERMINAL_TYPE;
        }
        if ( rrq.m_terminalType.HasOptionalField(H225_EndpointType::e_gateway) )
        {
            myEndpointType = GATEWAY_TYPE;
        }

        // Get the alias.
        //
        myAlias = rrq.m_terminalAlias;

        // If this is a gateway endpoint type and this endpoint wasn't created
        // statically, pull out the supported prefix.
        //
        if ( myEndpointType == GATEWAY_TYPE  && !myStaticRegistration )
        {
            updatePrefix(rrq.m_terminalType.m_gateway);
        }

        // Get the ras address. Look for the first IPAddress.
        //
        IPAddress rasAddr;

        for ( size_t i = 0; i < rrq.m_rasAddress.GetSize(); i++ )
        {
            if ( rrq.m_rasAddress[i].GetTag() == H225_TransportAddress::e_ipAddress )
            {
    	        H225_TransportAddress_ipAddress & rasIP = rrq.m_rasAddress[i];
                rasAddr.setIP(rasIP.m_ip.GetPointer());
	        rasAddr.setPort(rasIP.m_port);

	        myRasAddress = rasAddr;

	        break;
            }
        }

        // Get the callSignalAddress.
        //
        IPAddress csAddr;

        for ( size_t i = 0; i < rrq.m_callSignalAddress.GetSize(); i++ )
        {
            if ( rrq.m_callSignalAddress[i].GetTag() == H225_TransportAddress::e_ipAddress )
            {
    	        H225_TransportAddress_ipAddress & csIP = rrq.m_callSignalAddress[i];
                csAddr.setIP(csIP.m_ip.GetPointer());
	        csAddr.setPort(csIP.m_port);

	        myCallSignalAddress = csAddr;

	        break;
            }
        }

        // Get the timeToLive out of the rrq. If they do not want to
        // do keep alive, we will set this value to 0.
        //
        if ( rrq.HasOptionalField(H225_RegistrationRequest::e_timeToLive) )
        {
    	    myTimeToLive = rrq.m_timeToLive;
        }
    }
        
    return ( valid() );
}


bool	    	
EndpointInfo::valid() const
{
    ReadLock    lock(myMutex);

    // If endpoint type is a terminal or undefined, there must be 
    // a phone number to be valid.
    //
    if  ( (myEndpointType == TERMINAL_TYPE || myEndpointType == UNDEFINED_TYPE)
        &&  !myAlias.hasPhone() 
        )
    {
        return ( false );
    }
    
    // Must have an endpoint identifier to be valid.
    //
    if ( myEndpointIdentifier.size() == 0 )
    {
        return ( false );
    }
    
    // Must either be statically registered or have a RAS address
    // to be valid.
    //
    if( !myStaticRegistration && myRasAddress.isAny() )
    {
        return ( false );
    }
    
    // Must have a call signaling address to be valid.
    //
    if ( myCallSignalAddress.isAny() )
    {
        return ( false );
    }
    
    return ( true );
}


void	    	
EndpointInfo::endpointIdentifier(const string & endptId)
{
    WriteLock   lock(myMutex);
    myEndpointIdentifier = endptId;
}


string
EndpointInfo::endpointIdentifier() const
{
    ReadLock    lock(myMutex);
    return ( myEndpointIdentifier );
}


void            
EndpointInfo::endpointType(Type type)
{
    WriteLock   lock(myMutex);
    myEndpointType = type;
}


const char *    
EndpointInfo::typeName(Type type)
{
    static const char * typeName[TYPE_SIZE] =
    {
        "Undefined",
        "Terminal",
        "Gateway",
        "MCU"
    };
    return ( typeName[type >= TYPE_SIZE ? UNDEFINED_TYPE : type] );
};


EndpointInfo::Type            
EndpointInfo::endpointType() const
{
    ReadLock    lock(myMutex);
    return ( myEndpointType );
}


void	    	
EndpointInfo::alias(const H225_ArrayOf_AliasAddress & pAlias)
{
    WriteLock   lock(myMutex);
    myAlias.alias(pAlias);
}


H225_ArrayOf_AliasAddress
EndpointInfo::alias() const
{
    ReadLock    lock(myMutex);
    return ( myAlias.alias() );
}


void	    	
EndpointInfo::prefix(const H225_GatewayInfo & gwInfo)
{
    WriteLock   lock(myMutex);
    updatePrefix(gwInfo);
}


void	    	
EndpointInfo::prefix(const list<string> & prefixes)
{
    H225_ArrayOf_AliasAddress   aliases;
    
    for ( list<string>::const_iterator  i = prefixes.begin();
          i != prefixes.end();
          ++i
        )
    {
        PINDEX  size = aliases.GetSize();
        aliases.SetSize(size + 1);

        aliases[size].SetTag(H225_AliasAddress::e_dialedDigits);

        PASN_IA5String  &   alias = aliases[size];
        alias = i->c_str();
    }
    
    WriteLock    lock(myMutex);
    myPrefix = aliases;
}


H225_ArrayOf_AliasAddress
EndpointInfo::prefix() const
{
    ReadLock    lock(myMutex);
    return ( myPrefix );
}


void            
EndpointInfo::stripPrefix(bool strip_prefix)
{
    WriteLock   lock(myMutex);
    myStripPrefix = strip_prefix;
}


bool            
EndpointInfo::stripPrefix() const
{
    ReadLock    lock(myMutex);
    return ( myStripPrefix );
}


void            
EndpointInfo::staticRegistration(bool static_reg)
{
    WriteLock   lock(myMutex);
    myStaticRegistration = static_reg;
}


bool            
EndpointInfo::staticRegistration() const
{
    ReadLock    lock(myMutex);
    return ( myStaticRegistration );
}


void	    	
EndpointInfo::rasAddress(const IPAddress & rasAddr)
{
    WriteLock   lock(myMutex);
    myRasAddress = rasAddr;
}


IPAddress
EndpointInfo::rasAddress() const
{
    ReadLock    lock(myMutex);
    return ( myRasAddress );
}


void	    	
EndpointInfo::callSignalAddress(const IPAddress & csAddr)
{
    WriteLock   lock(myMutex);
    myCallSignalAddress = csAddr;
}


IPAddress
EndpointInfo::callSignalAddress() const
{
    ReadLock    lock(myMutex);
    return ( myCallSignalAddress );
}


void	    	
EndpointInfo::timeToLive(u_int32_t ttl)
{
    // We need to eventually convert this to milliseconds, which is signed,
    // so there is no use keeping a number greater than MAXINT/1000.
    //
    static u_int32_t	max_ttl = MAXINT / 1000;
    
    WriteLock   lock(myMutex);
    myTimeToLive = ( ttl > max_ttl ? max_ttl : ttl );
}


u_int32_t  	
EndpointInfo::timeToLive() const
{
    ReadLock    lock(myMutex);
    return ( myTimeToLive );
}


void	    	
EndpointInfo::timeToLiveTimerEventId(FifoBase< Sptr<Event> >::EventId eventId)
{
    WriteLock   lock(myMutex);
    myTimeToLiveTimerEventId = eventId;
}


FifoBase< Sptr<Event> >::EventId 
EndpointInfo::timeToLiveTimerEventId() const
{
    ReadLock    lock(myMutex);
    return ( myTimeToLiveTimerEventId );
}


void            
EndpointInfo::outOfResources(bool oor)
{
    WriteLock   lock(myMutex);
    myOutOfResources = oor;
}


bool            
EndpointInfo::outOfResources() const
{
    ReadLock    lock(myMutex);
    return ( myOutOfResources );
}
        

void        
EndpointInfo::recordRegistration()
{
    WriteLock   lock(myMutex);
    
    if ( myRegistrationCount == 0 )
    {
        myFirstRegistrationTime.now();
        myLastRegistrationTime = myFirstRegistrationTime;
        myLastKeepAliveTime = myLastRegistrationTime;
    }
    else
    {
        myLastRegistrationTime.now();
    }

    ++myRegistrationCount;
    
    // Just in case it rolls, don't reset first time.
    //
    if ( myRegistrationCount == 0 )
    {
        myRegistrationCount = 1;
    }

    myLastActivityTime = myLastRegistrationTime;
}
        

u_int32_t   
EndpointInfo::registrationCount() const
{
    ReadLock    lock(myMutex);
    return ( myRegistrationCount );
}
        
        
TimeVal     
EndpointInfo::firstRegistrationTime() const
{
    ReadLock    lock(myMutex);
    return ( myFirstRegistrationTime );
}


TimeVal     
EndpointInfo::lastRegistrationTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastRegistrationTime );
}


void        
EndpointInfo::recordKeepAlive()
{
    WriteLock   lock(myMutex);

    myLastKeepAliveTime.now();

    myLastActivityTime = myLastKeepAliveTime;
}
        

TimeVal     
EndpointInfo::lastKeepAliveTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastKeepAliveTime );
}


void        
EndpointInfo::recordAdmission()
{
    WriteLock   lock(myMutex);
    
    if ( myAdmissionCount == 0 )
    {
        myFirstAdmissionTime.now();
        myLastAdmissionTime = myFirstAdmissionTime;
    }
    else
    {
        myLastAdmissionTime.now();
    }

    ++myAdmissionCount;

    // Just in case it rolls, don't reset first time.
    //
    if ( myAdmissionCount == 0 )
    {
        myAdmissionCount = 1;
    }


    myLastActivityTime = myLastAdmissionTime;
}
                 

u_int32_t   
EndpointInfo::admissionCount() const
{
    ReadLock    lock(myMutex);
    return ( myAdmissionCount );
}


u_int32_t    
EndpointInfo::activeAdmissionCount() const
{
    ReadLock    lock(myMutex);
    
    return  ( myAdmissionCount > myDisengageCount 
                ? myAdmissionCount - myDisengageCount
                : 0
            );
}
          
TimeVal     
EndpointInfo::firstAdmissionTime() const
{
    ReadLock    lock(myMutex);
    return ( myFirstAdmissionTime );
}


TimeVal     
EndpointInfo::lastAdmissionTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastAdmissionTime );
}


void        
EndpointInfo::recordDisengage()
{
    WriteLock   lock(myMutex);
    
    if ( myDisengageCount == 0 )
    {
        myFirstDisengageTime.now();
        myLastDisengageTime = myFirstDisengageTime;
    }
    else
    {
        myLastDisengageTime.now();
    }

    ++myDisengageCount;

    // Just in case it rolls, don't reset first time.
    //
    if ( myDisengageCount == 0 )
    {
        myDisengageCount = 1;
    }


    myLastActivityTime = myLastDisengageTime;
}


u_int32_t   
EndpointInfo::disengageCount() const
{
    ReadLock    lock(myMutex);
    return ( myDisengageCount );
}


TimeVal     
EndpointInfo::firstDisengageTime() const
{
    ReadLock    lock(myMutex);
    return ( myFirstDisengageTime );
}


TimeVal     
EndpointInfo::lastDisengageTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastDisengageTime );
}


void        
EndpointInfo::recordResourcesAvailable()
{
    WriteLock   lock(myMutex);
    
    if ( myResourcesAvailableCount == 0 )
    {
        myFirstResourcesAvailableTime.now();
        myLastResourcesAvailableTime = myFirstResourcesAvailableTime;
    }
    else
    {
        myLastResourcesAvailableTime.now();
    }

    ++myResourcesAvailableCount;

    // Just in case it rolls, don't reset first time.
    //
    if ( myResourcesAvailableCount == 0 )
    {
        myResourcesAvailableCount = 1;
    }


    myLastActivityTime = myLastResourcesAvailableTime;
}


u_int32_t   
EndpointInfo::resourcesAvailableCount() const
{
    ReadLock    lock(myMutex);
    return ( myResourcesAvailableCount );
}


TimeVal     
EndpointInfo::firstResourcesAvailableTime() const
{
    ReadLock    lock(myMutex);
    return ( myFirstResourcesAvailableTime );
}


TimeVal     
EndpointInfo::lastResourcesAvailableTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastResourcesAvailableTime );
}


TimeVal     
EndpointInfo::lastActivityTime() const
{
    ReadLock    lock(myMutex);
    return ( myLastActivityTime );
}


void        
EndpointInfo::clear()
{
    WriteLock   lock(myMutex);

    if ( myStaticRegistration = false )
    {
        myEndpointIdentifier = "";
        myEndpointType = UNDEFINED_TYPE;
        myPrefix.RemoveAll();
        myStripPrefix = false;
        myCallSignalAddress.clear();
    }

    myAlias.clear();
    myRasAddress.clear();

    myTimeToLive = 0;
    myTimeToLiveTimerEventId = 0;
    myOutOfResources = false;

    myRegistrationCount = 0;
    myFirstRegistrationTime.clear();
    myLastRegistrationTime.clear();

    myLastKeepAliveTime.clear();

    myAdmissionCount = 0;
    myFirstAdmissionTime.clear();
    myLastAdmissionTime.clear();

    myDisengageCount = 0;
    myFirstDisengageTime.clear();
    myLastDisengageTime.clear();

    myResourcesAvailableCount = 0;
    myFirstResourcesAvailableTime.clear();
    myLastResourcesAvailableTime.clear();

    myLastActivityTime.clear();
}


static const char *
printDate(const TimeVal & tv)
{
    #define DATEBUF_SIZE 256
    static char     datebuf[DATEBUF_SIZE];
    *datebuf = '\0';

    if ( tv.tv_sec != 0 )
    {
        strftime(datebuf, DATEBUF_SIZE, 
                 "%Y%m%d-%H%M%S", localtime(&tv.tv_sec));

        char msbuf[5];
        sprintf(msbuf, ".%3.3ld", (tv.tv_usec/1000));

        size_t remain = DATEBUF_SIZE - (size_t) strlen(datebuf);

        strncat(datebuf, msbuf, remain - 1);

        datebuf[DATEBUF_SIZE - 1] = '\0';
    }
    
    return ( datebuf );
}
        

ostream &       
EndpointInfo::writeTo(ostream & out) const
{
    ReadLock    lock(myMutex);
    
    static const string     pre("\n\t");

    out << pre << "endpointIdentifier    = " << myEndpointIdentifier.c_str()
        << pre << "endpointType          = " << typeName(myEndpointType)
        << pre << "alias                 = " << myAlias;

    size_t  size = myPrefix.GetSize();

    if ( size > 0 )
    {
        for ( size_t i = 0; i < size; i++ )
        {
            assert(myPrefix[i].GetTag() == H225_AliasAddress::e_dialedDigits);

            PASN_IA5String  &   e164 = myPrefix[i];

            if ( i == 0 )
            {
                out << pre << "prefix                = " << e164.GetValue();
            }
            else
            {
                out << ", " << e164.GetValue();
            }
        }

        out << pre << "strip prefix          = " 
            << ( myStripPrefix ? "true" : "false" );
    }

    out << pre << "rasAddress            = " 
        << myRasAddress.getIPAsString() << ':' << myRasAddress.getPort()

	<< pre << "callSignalAddress     = " 
        << myCallSignalAddress.getIPAsString() 
        << ':' << myCallSignalAddress.getPort()

        << pre << "ttl                   = " << myTimeToLive

        << pre << "out of resources      = "
        << ( myOutOfResources ? "true" : "false" )

        << pre << "registrations         = " << myRegistrationCount

        << pre << "first registration    = "
        << printDate(myFirstRegistrationTime);

    out << pre << "last registration     = "
        << printDate(myLastRegistrationTime);
        
    out << pre << "last keep alive       = "
        << printDate(myLastKeepAliveTime);
        
    out << pre << "admissions            = " << myAdmissionCount
        
        << pre << "first admission       = " 
        << printDate(myFirstAdmissionTime);
        
    out << pre << "last admission        = "
        << printDate(myLastAdmissionTime);
        
    out << pre << "disengages            = " << myDisengageCount
        
        << pre << "first disengage       = " 
        << printDate(myFirstDisengageTime);
        
    out << pre << "last disengage        = "
        << printDate(myLastDisengageTime);
        
    out << pre << "resources avail       = " 
        << myResourcesAvailableCount
        
        << pre << "first resources avail = " 
        << printDate(myFirstResourcesAvailableTime);
        
    out << pre << "last resources avail  = " 
        << printDate(myLastResourcesAvailableTime);
        
    out << pre << "last activity         = " 
        << printDate(myLastActivityTime)

        << '\n';
        
    return ( out );
}


void
EndpointInfo::updatePrefix(const H225_GatewayInfo & gwInfo)
{
    const string    fn("EndpointInfo::updatePrefix");
    VLog            log(fn);
    
    // Gateway info is a sequence of SupportedProtocols.
    // Each SupportedProtocol is a choice of capabilities.
    // Each capability has a sequence of SupportedPrefix.
    // A SupportedPrefix has a single alias for the prefix.
    //
    // I wish I knew why they are trying to make this as hard
    // as TCS. I really should ask Glen F. about this. *grin*
    //
    // Right now we are only going to deal with the voice caps,
    // but ideally we'll accept any capability, since we don't
    // really care what they talk on the other side of the GW at 
    // this point. We just need to know where to route the call 
    // based on the alias. Also, we route based on e.164 only. 
    // Perhaps we could route based on other fields in the alias, 
    // but there are better solutions for that... i.e. DNS and
    // DNS SRV, etc.
    //
    for ( size_t i = 0; i < gwInfo.m_protocol.GetSize(); i++ )
    {
        const H225_SupportedProtocols   &   supportedProtocols 
                                                = gwInfo.m_protocol[i];

        switch ( supportedProtocols.GetTag() )
        {
            case H225_SupportedProtocols::e_voice:
            {
                const H225_VoiceCaps & voiceCaps = supportedProtocols;

                if ( voiceCaps.HasOptionalField(H225_VoiceCaps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < voiceCaps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = voiceCaps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }

            case H225_SupportedProtocols::e_h323:
            {
                const H225_H323Caps & h323Caps = supportedProtocols;

                if ( h323Caps.HasOptionalField(H225_H323Caps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < h323Caps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = h323Caps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }

            case H225_SupportedProtocols::e_h320:
            {
                const H225_H320Caps & h320Caps = supportedProtocols;

                if ( h320Caps.HasOptionalField(H225_H320Caps::e_supportedPrefixes) )
                {
                    for ( size_t i = 0; i < h320Caps.m_supportedPrefixes.GetSize(); i++ )
                    {
                        PINDEX  size = myPrefix.GetSize();
                        myPrefix.SetSize(size + 1);
                        
                        myPrefix[size] = h320Caps.m_supportedPrefixes[i].m_prefix;
                    }
                }
                break;
            }
            
            // Do these later... perhaps.
            //
            case H225_SupportedProtocols::e_h310:
            case H225_SupportedProtocols::e_h321:
            case H225_SupportedProtocols::e_h322:
            case H225_SupportedProtocols::e_h324:
            case H225_SupportedProtocols::e_t120_only:

            case H225_SupportedProtocols::e_nonStandardProtocol:
            case H225_SupportedProtocols::e_nonStandardData:
            default:
            {
                // Skip it by default
                //
                break;
            }
        }
    }

    VDEBUG(log) << fn << ": " << myPrefix << VDEBUG_END(log);
}


