#if !defined(VOCAL_GATEKEEPEPER_INFO_HXX)
#define VOCAL_GATEKEEPEPER_INFO_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const GatekeeperInfo_hxx_Version = 
    "$Id: GatekeeperInfo.hxx,v 1.15 2002/05/15 09:43:31 icahoon Exp $";


#include <string>
#include "IPAddress.hxx"
#include "VMutex.h"
#include "EndpointInfo.hxx"
#include "NonCopyable.hxx"


/** Vovida Open Communication Application Library<br><br>
 */
namespace Vocal
{


class SystemException;


/** SIP to H.323 Object Information<br><br>
 */
namespace SipH323Info
{


using Vocal::Transport::IPAddress;


/** Configuration information available for the H323 Gatekeeper.<br><br>
 *
 *  @see    Vocal::Transport::IPAddress
 */ 
class GatekeeperInfo : Vocal::NonCopyable
{

    public:

    	/** Construct given the gatekeeper identifier, the prefix for
	 *  the endpoint identifier, a flag indicating whether the 
	 *  calls will be routed through the gatekeeper, and a flag
	 *  whether the gatekeeper is a standalone.
    	 */
    	GatekeeperInfo(
            list< Sptr<EndpointInfo> >      staticGateways,
	    const string                &   gatekeeperId = "",
	    const string                &   endpointId = "",
	    bool    	    	            standAlone = false,
	    bool    	    	            gkRouted = true,
            bool                            discoveryEnabled = true,
            bool                            registrationEnabled = true,
            bool                            admissionEnabled = true,
            bool                            locationEnabled = false,
            bool                            resourcesAvailableEnabled = true
	)
    	throw ( Vocal::SystemException );
				    

    	/** Virtual destructor.
    	 */
	virtual     	    ~GatekeeperInfo();


        /** If set, the gatekeeper will handle discovery messages.
         *  The default is true.
         */
        bool                discoveryEnabled() const;


        /** If set true, the gatekeeper wil handle discovery messages.
         */
        void                discoveryEnabled(bool);
        
         
        /** If set, the gatekeeper will handle registration messages.
         *  The default is true.
         */
        bool                registrationEnabled() const;


        /** If set true, the gatekeeper wil handle registration messages.
         */
        void                registrationEnabled(bool);


        /** If set, the gatekeeper will handle admission messages.
         *  The default is true.
         */
        bool                admissionEnabled() const;


        /** If set true, the gatekeeper wil handle admission messages.
         */
        void                admissionEnabled(bool);


        /** If set, the gatekeeper will handle location messages.
         *  The default is false (for now).
         */
        bool                locationEnabled() const;


        /** If set true, the gatekeeper wil handle location messages.
         */
        void                locationEnabled(bool);


        /** If set, the gatekeeper will handle resources available messages.
         *  The default is true.
         */
        bool                resourcesAvailableEnabled() const;


        /** If set true, the gatekeeper wil handle resources available messages.
         */
        void                resourcesAvailableEnabled(bool);


    	/** Generate the identifier for the next endpoint.
    	 */
    	string	    	    generateEndpointIdentifier();


    	/** Set the hostname to be used by the gatekeeper.
    	 */
    	void	    	    hostName(const string &);


    	/** Access the hostname used by the gatekeeper.
    	 */
	string      	    hostName() const;


    	/** Set the hostAddress to be used by the gatekeeper.
    	 */
	void        	    hostAddress(const IPAddress &);


    	/** Access the hostAddress used by the gatekeeper.
    	 */
	IPAddress   	    hostAddress() const;


        /** Access the list of statically registered gateways.
         */
        list < Sptr<EndpointInfo> >     staticGateways() const;
        
        
    	/** Set the gatekeeper identifier.
    	 */
    	void	    	    gatekeeperIdentifier(const string &);


    	/** Access the gatekeeper identifier.
    	 */
    	string	    	    gatekeeperIdentifier() const;


    	/** Set the RAS address.
    	 */
	void	    	    rasAddress(const IPAddress &);


    	/** Access the RAS address.
    	 */
	IPAddress   	    rasAddress() const;


    	/** Set the RAS multicast address.
    	 */
	void	    	    rasMulticastAddress(const IPAddress &);


    	/** Access the RAS multicast address.
    	 */
	IPAddress   	    rasMulticastAddress() const;


    	/** Set the call signaling address.
    	 */
	void	    	    callSignalAddress(const IPAddress &);


    	/** Access the call signaling address.
    	 */
	IPAddress   	    callSignalAddress() const;


    	/** Set the endpoint identifier prefix.
    	 */
	void	    	    endpointIdentifier(const string &);


    	/** Access the endpoint identifier prefix.
    	 */
	string	    	    endpointIdentifier() const;


    	/** Set the gatekeeper routed property.
    	 */
    	void	    	    gatekeeperRouted(bool);


    	/** Access the gatekeeper routed property.
    	 */
    	bool	    	    gatekeeperRouted() const;


    	/** Set the stand alone property.
    	 */
	void	    	    standAlone(bool);


    	/** Access the stand alone property.
    	 */
	bool	    	    standAlone() const;


        /** If set, the responses should be sent to the ip address
         *  of the sender, rather than using the ras address. This
         *  can be useful for testing.
         */
        void                returnToSender(bool);
        
        
        /** If set, the responses should be sent to the ip address
         *  of the sender, rather than using the ras address. This
         *  can be useful for testing.
         */
        bool                returnToSender();
        

    	/** Access the h.225 protocol identifer.
    	 */
    	string	    	    h225ProtocolID() const;


    private:

        bool                        myDiscoveryEnabled;
        bool                        myRegistrationEnabled;
        bool                        myAdmissionEnabled;
        bool                        myLocationEnabled;
        bool                        myResourcesAvailableEnabled;
	string	    	            myHostName;
	IPAddress    	            myHostAddress;
        list< Sptr<EndpointInfo> >  myStaticGateways;
    	string	    	            myGatekeeperIdentifier;
	IPAddress   	            myRasAddress;
        IPAddress                   myRasMulticastAddress;
	IPAddress   	            myCallSignalAddress;
	string	    	            myEndpointIdentifier;
	bool	    	            myGatekeeperRouted;
	bool	    	            myStandAlone;
        bool                        myReturnToSender;
    	string	    	            myH225ProtocolID;

	static	u_int32_t           myEndpointNumber;
	
	mutable VMutex	            myMutex;
};


} // namespace SipH323Info
} // namespace Vocal


#endif // !defined(VOCAL_GATEKEEPEPER_INFO_HXX)
