#if !defined(VOCAL_SIP_TO_H323_REGISTRATION_POLICY_HXX)
#define VOCAL_SIP_TO_H323_REGISTRATION_POLICY_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const SipToH323RegistrationPolicy_hxx_Version = 
    "$Id: SipToH323RegistrationPolicy.hxx,v 1.4 2002/07/10 22:09:00 icahoon Exp $";


#include "DefaultRegistrationPolicy.hxx"
#include "H323RegistrationEvent.hxx"


/** Vovida Open Communication Application Library<br><br>
 */
namespace Vocal
{


/** SIP to H.323 Call Signaling Gateway<br><br>
 */
namespace SipH323CSGW
{


using Vocal::H323GK::RASGKProtocol;
using Vocal::SipH323Info::GatekeeperInfo;
using Vocal::SipH323Info::EndpointInfoMap;
using Vocal::SipH323Info::EndpointInfo;
using Vocal::H323Events::H323RegistrationEvent;
using Vocal::Services::Service;
using Vocal::Transport::IPAddress;
using Vocal::Services::Event;

/** Sip to H323 Call Signaling Gateway specific registration policy.<br><br>
 *
 *  This provides the link between the H323 Registration procedures and the
 *  SIP REGISTER procedures. Thus this policy must know the signaling Fifo
 *  of the SIP proxy.
 *
 *  @see    Vocal::H323GK::DefaultRegistrationPolicy
 *  @see    Vocal::Services::Service
 *  @see    Vocal::SipH323Info::GatekeeperInfo
 *  @see    Vocal::SipH323Info::EndpointInfoMap
 *  @see    Vocal::Transport::IPAddress
 *  @see    Vocal::H323Events::H323RegistrationEvent
 *  @see    FifoBase
 *  @see    Vocal::Services::Event
 *  @see    Sptr
 */
class SipToH323RegistrationPolicy : public Vocal::H323GK::DefaultRegistrationPolicy
{

    public:


    	/** Create the registration policy.
	 *
	 *  @param  service 	
	 *  	    a reference to the service controling this policy
	 *
	 *  @param  rasGKProtocol
	 *  	    the protocol responsible for receiving and parsing RAS PDUs
	 *
	 *  @param  gatekeeperInfo
	 *  	    configuration data about the gatekeeper
	 *
	 *  @param  endpointInfoMap
	 *  	    the database of endpoint information
	 *
	 *  @param  sipProxyFifo
	 *  	    the communication link between H.323 and SIP 
	 */
    	SipToH323RegistrationPolicy(
	    Service      	    	&   service,
	    RASGKProtocol   	    	&   rasGKProtocol,
	    GatekeeperInfo  	    	&   gatekeeperInfo,
	    EndpointInfoMap 	    	&   endpointInfoMap,
	    FifoBase< Sptr<Event> > 	&   sipProxyFifo
	);


    	/** Virtual destructor
	 */
	virtual		    ~SipToH323RegistrationPolicy();


    	/** EventObserver callback. We will be interested
	 *  in H323RegistrationEvents from the SipProxy.
	 */
    	virtual bool	    interestedIn(const Sptr<Event> &);


    	/** EventObserver callback to handle H323RegistrationEvents.
	 */
    	virtual void	    onEvent(Sptr<Event>);


    	/** Write the registration policy to an ostream.
	 */
    	ostream &           writeTo(ostream & out) const;


    protected:
	

	/** Called on receipt of a keep alive registration request. 
	 *
    	 *  On wait, the processing stops. It is the responsibility of
    	 *  the derived class to retain message state.
	 *
    	 *  On confirm, the processing continues processing the request.
	 *
    	 *  On reject, a registration reject is sent, and the user needs
    	 *  to supply a reason.
	 */
	virtual Proceed	    onKeepAliveRegistration(
		Sptr<H225_RasMessage>	    	    rasPDU, 
		Sptr<IPAddress> 	    	    remoteAddr,
	    	Sptr<EndpointInfo>  	    	    endpointInfo,
	    	H225_RegistrationRequest    	&   karrq,
		H225_RegistrationRejectReason	&   reason);


	/** Called on receipt of a registration request. 
	 *
    	 *  On wait, the processing stops. It is the responsibility of
    	 *  the derived class to retain message state.
	 *
    	 *  On confirm, the processing continues processing the request.
	 *
    	 *  On reject, a registration reject is sent, and the user needs
    	 *  to supply a reason.
	 */
    	virtual Proceed	    onRegistration(
	    Sptr<H225_RasMessage>	    	    rasPDU, 
	    Sptr<IPAddress> 	    	    	    remoteAddr,
	    Sptr<EndpointInfo>  	    	    endpointInfo,
	    H225_RegistrationRequest    	&   rrq,
	    H225_RegistrationRejectReason	&   reason);


	/** Called on receipt of a unregistration request. 
	 *
    	 *  On wait, the processing stops. It is the responsibility of
    	 *  the derived class to retain message state.
	 *
    	 *  On confirm, the processing continues processing the request.
	 *
    	 *  On reject, a unregistration reject is sent, and the user needs
	 *  to supply a reason.
	 */
    	virtual Proceed     onUnregistration(
	    Sptr<H225_RasMessage>	        rasPDU, 
	    Sptr<IPAddress> 	    	    	remoteAddr,
	    Sptr<EndpointInfo>    	    	endpointInfo,
	    H225_UnregistrationRequest  &	urq,
	    H225_UnregRejectReason      &   	reason);
	

    private:
    	
	FifoBase< Sptr<Event> >     	    	&   mySipProxyFifo;
	Sptr<H323RegistrationEvent> 	    	    myH323RegistrationEvent;
};


} // namespace SipH323CSGW
} // namespace Vocal


#endif // !defined(VOCAL_SIP_TO_H323_REGISTRATION_POLICY_HXX)
