/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const EvtSipStatus_cxx_Version = 
    "$Id: EvtSipStatus.cxx,v 1.25 2002/07/10 07:10:23 icahoon Exp $";


#include <assert.h>
#include <stdlib.h>

#include "VLog.hxx"

#include "AckMsg.hxx"
#include "SipContact.hxx"
#include "SipUrl.hxx"
#include "SipVia.hxx"
#include "SipRetryAfter.hxx"
#include "ValAssistant.h"

#include "EvtCallOffHook.hxx"
#include "Q931Defs.hxx"

#include "RouterInfo.hxx"
#include "Address.hxx"

#include "EvtSipStatus.hxx"
#include "FeatureTrCallSig.hxx"
#include "TranslatorConsts.hxx"
#include "Contacts.hxx"
#include "Consts.hxx"

using namespace translator;
using Vocal::Transport::IPAddress;
using Vocal::Logging::VLog;

EvtSipStatus::
EvtSipStatus(TrRouterSIP* sm)
    : EvtSIPTrCallSigMessage(sm)
{
}


void EvtSipStatus::
inConnectingH2S(FeatureTrCallSig& trcallsig)
{
    const string    fn("EvtSipStatus::inConnectingH2S");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    typedef void (EvtSipStatus::* Handler)(
        FeatureTrCallSig&, int, const Contact* contact);
    static Handler handlers[] =
    {
        &EvtSipStatus::handleResponse, // 000 - 099
	&EvtSipStatus::handleResponse, // 100 - 199
	&EvtSipStatus::handleConfirm,  // 200 - 299
	&EvtSipStatus::handleRedirect, // 300 - 399
	&EvtSipStatus::handleReject,   // 400 - 499
	&EvtSipStatus::handleReject,   // 500 - 599
	&EvtSipStatus::handleReject,   // 600 - 699
    };

    // (depends from H.323/SIP parser)

    int status = getMessage().getStatusLine().getStatusCode();
    int group = status / 100;
    int numGroups = sizeof(handlers) / sizeof(Handler);
    
    if (group < 0 || group >= numGroups)
    {
	// wrong status code
	VWARN(log) << fn << ": Got SIP status message with wrong code "
		   << status << " for call "
		   << trcallsig.getSM() << VWARN_END(log);
	group = 4;
	status = 400;
    }

    Contact* contact =
	trcallsig.getContacts().findByBranch(getMessage());

    if (contact)
    {
        VDEBUG(log) << fn << ": Entering sip status handler" << VDEBUG_END(log);
        (this->*handlers[group])(trcallsig, status, contact);
        VDEBUG(log) << fn << ": Leaving sip status handler" << VDEBUG_END(log);

	// Find contact once more because contact list may be moved
        Contact* contact =
	    trcallsig.getContacts().findByBranch(getMessage());
	assert(contact);

	trcallsig.getContacts().update(contact, getMessage());

        VDEBUG(log) << fn << ": Current stats = "
		    << trcallsig.getContacts().getStatus()
		    << VDEBUG_END(log);

        if (trcallsig.getContacts().getStatus() >= 300)
        {
            // Can not establish call: no contacts accepted it
            VDEBUG(log) << fn << ": Call is rejected" << VDEBUG_END(log);

            if (trcallsig.getContacts().size() -
                trcallsig.getContacts().redirects()
                > 1)
            {
                trcallsig.setReason(msip::respBusyEverywhere);
            }
            else
            {
                if (status >= 400)
                {
                    trcallsig.setReason(status);
                }
                else
                {
                    trcallsig.setReason(msip::respLoopDetected);
                }
            }

            trcallsig.getSM()->getSIPConnection()->disable();
            trcallsig.getSM()->complete();
        }
    }
    else
    {
        VWARN(log) << fn << ": Got SIP status message from unknown contact for call "
                   << trcallsig.getSM() << " (ignored)" << VWARN_END(log);
    }
}


void EvtSipStatus::
inWait200(FeatureTrCallSig& trcallsig)
{
    const string    fn("EvtSipStatus::inWait200");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    // (depends from H.323/SIP parser)

    int status = getMessage().getStatusLine().getStatusCode();

    trcallsig.setState(&PatternTrCallSig::inSending, "inSending");
    Sptr<AckMsg> sipAck = new AckMsg(
        const_cast<StatusMsg&>(getMessage()));
    trcallsig.getSM()->getSIPConnection()->send(sipAck);

    if (status >= 200 && status < 300)
    {
        // Confirm
        trcallsig.setState(&PatternTrCallSig::inConnected, "inConnected");
        trcallsig.getSM()->getCapabilityTranslator().confirmSDP();
    }
    else if (status == 500)
    {
        // Collision
        Vocal::SipRetryAfter ra = getMessage().getRetryAfter();
        trcallsig.waitTime_ = 1000 * ra.getDuration();

        // If no timeout specified, generate random value (absolute time
        // is not supported)
        if (trcallsig.waitTime_ == 0)
        {
            trcallsig.waitTime_ = 1000 * ((rand() * TMR_COLLISION) / RAND_MAX + 1);
        }

        trcallsig.setState(&PatternTrCallSig::inCollision, "inCollision");
    }
    else if (status >= 300)
    {
        // Reject or unexpected redirect: no change in media paramerers?
        trcallsig.setState(&PatternTrCallSig::inConnected, "inConnected");
        trcallsig.getSM()->getCapabilityTranslator().cancelSDP();
    }
    else
    {
        // Provisional reply: ignore
        trcallsig.setState(&PatternTrCallSig::inConnected, "inConnected");
    }
}


void EvtSipStatus::
inOther(FeatureTrCallSig& trcallsig)
{
    const string    fn("EvtSipStatus::inOther");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    // Ignore
    VWARN(log) << "Got unexpected SIP status message for call "
               << trcallsig.getSM() << VWARN_END(log);
}


void EvtSipStatus::
handleReject(FeatureTrCallSig& trcallsig,
	     int status,
	     const Contact* contact)
{
    const string    fn("EvtSipStatus::handleReject");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    assert(this);
    assert(status >= 400);

    Sptr<AckMsg> sipAck = new AckMsg(const_cast<StatusMsg&>(getMessage()));
    trcallsig.getSM()->getSIPConnection()->send(sipAck);
}


void EvtSipStatus::
handleRedirect(FeatureTrCallSig& trcallsig,
	       int status,
	       const Contact* contact)
{
    const string    fn("EvtSipStatus::handleRedirect");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    assert(this);
    assert(status >= 300);
    assert(status < 400);

    Sptr<AckMsg> sipAck = new AckMsg(const_cast<StatusMsg&>(getMessage()));
    // TODO: set request line?
    trcallsig.getSM()->getSIPConnection()->send(sipAck);

    for (int i=0; i < getMessage().getNumContact(); i ++)
    {
        SipContact sipContact = getMessage().getContact(i);
        //SipUrl newUrl = sipContact.getUrl();
	
       Sptr<BaseUrl> newUrl = sipContact.getUrl();
       Sptr<SipUrl> sipUrl;

       if(newUrl.getPtr()!=0)
       {
         if(newUrl->getType()==SIP_URL)
         {
               sipUrl.dynamicCast(newUrl);
         }
         else
         {
               VEMERG(log) << fn << ": baseUrl is not SipUrl, cannot proceed..." << VEMERG_END(log);
         }
       }
       else
       {
         VEMERG(log) << fn << ": baseUrl is a null pointer, cannot proceed..." << VEMERG_END(log);
       }
	

	// TODO: remove the following to work as a standalone gateway
	sipUrl->setHost("0.0.0.0");
	sipUrl->setPort(0);
	
	newUrl = sipUrl;

        Sptr<InviteMsg> inviteMsg =
	    new InviteMsg(*(trcallsig.invite_), newUrl);

	// Be sure we have the same call ID and CSeq
	inviteMsg->setCallId(getMessage().getCallId());
	inviteMsg->setCSeq(getMessage().getCSeq());
	
	// Patch used to increase CSeq for apps not understanding branch
	/*
	SipCSeq cs = getMessage().getCSeq();
	cs.incrCSeq();
	inviteMsg->setCSeq(cs);
	trcallsig.invite_->setCSeq(cs);
	*/

	// Set branch number to 'Via'
	inviteMsg->flushViaList();
	SipVia via;
	via.setHost("0.0.0.0");
	via.setPort(0);
	via.setBranch(Data(trcallsig.getContacts().size()));
	inviteMsg->setVia(via);

	// New Invite is ready: send it and add url to the contact list
	if (trcallsig.getContacts().findByUrl(*sipUrl))
	{
	    // Loop detected; ignore
	    VWARN(log) << fn << ": Circular forwarding for call "
		       << trcallsig.getSM() << VWARN_END(log);
	}
	else
	{
	    trcallsig.getContacts().add(*sipUrl);
	    trcallsig.getSM()->getSIPConnection()->send(inviteMsg);
	    VDEBUG(log) << fn << ": message sent" << VDEBUG_END(log);
	}
    }
}


void EvtSipStatus::
handleConfirm(FeatureTrCallSig& trcallsig,
	      int status,
	      const Contact* contact)
{
    const string    fn("EvtSipStatus::handleConfirm");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    assert(this);
    assert(status >= 200);
    assert(status < 300);

    // 200 OK

    trcallsig.status_ = new StatusMsg(const_cast<StatusMsg&>(getMessage()));
    SipRequestLine reqLine = trcallsig.invite_->getRequestLine();
    
    Sptr<BaseUrl> baseUrl = new SipUrl(*contact);
    reqLine.setUrl(baseUrl);
    
    trcallsig.invite_->setRequestLine( reqLine );
    // TODO: store via and branch?

    // Translate SIP address to "display"

     Sptr<BaseUrl> newUrl = getMessage().getTo().getUrl();
     Sptr<SipUrl> sipUrl;

       if(newUrl.getPtr()!=0)
       {
         if(newUrl->getType()==SIP_URL)
         {
               sipUrl.dynamicCast(newUrl);
         }
         else
         {
               VEMERG(log) << fn << ": baseUrl is not SipUrl, cannot proceed..." << VEMERG_END(log);
         }
       }
       else
       {
         VEMERG(log) << fn << ": baseUrl is a null pointer, cannot proceed..." << VEMERG_END(log);
       }                                                                                        


    Translator::Address addressTranslator = *sipUrl;

    if (addressTranslator.hasPhone())
    {
        trcallsig.getSM()->
            getH323Call()->
            getCallSigConnection()->setToDisplay(addressTranslator.getPhone());
    }
    else
    {
        trcallsig.getSM()->
            getH323Call()->
            getCallSigConnection()->setToDisplay("Unknown number");
    }

    trcallsig.setState(&PatternTrCallSig::inSDPH2S, "inSDPH2S");

    SipTo to = getMessage().getTo();

    if (!(trcallsig.getSM()->getH323Call()->getRouterInfo()->
          routerMediaControl()))
    {
        if (to.getTokenValue("h245Address").length() > 0)
        {
            VDEBUG(log) << fn << ": Got 'h245Address' token: "
                        << VDEBUG_END(log);
            LocalScopeAllocator lo;
            IPAddress address = (const char*)(to.getTokenValue(H245Address).getData(lo));
            VDEBUG(log) << fn << ": Address=  " << address
                        << VDEBUG_END(log);
            trcallsig.getSM()->getH323Call()->getCallSigConnection()->
                setH245(address);

            trcallsig.setState(&PatternTrCallSig::inConnected, "inConnected");

            Sptr<AckMsg> sipAck = new AckMsg(
                const_cast<StatusMsg&>(getMessage()));
            trcallsig.getSM()->getSIPConnection()->send(sipAck);
        }
    }

    assert(trcallsig.getSM());
    router::EvtCallOffHook offhook(trcallsig.getSM()->getH323Call());
    offhook.start();

    // TODO: cancel other contacts
}


void EvtSipStatus::
handleResponse(FeatureTrCallSig& trcallsig,
	       int status,
	       const Contact* contact)
{
    const string    fn("EvtSipStatus::handleResponse");
    VLog            log(fn);
    
    VDEBUG(log) << fn << VDEBUG_END(log);

    assert(this);
    assert(status >= 0);
    assert(status < 200);

    if (status == 180 && !(trcallsig.alerted_))
    {
	// Alerting
	assert(trcallsig.getSM());
	assert(trcallsig.getSM()->getH323Call());

	router::BH225Alerting alerting;
	trcallsig.getSM()->getH323Call()->getCallSigConnection()->
	    sendAlerting(alerting);
	trcallsig.alerted_ = true;
    }
    else
    {
	// Ignore
	VINFO(log) << fn << ": Got SIP status 1xx message for call "
		   << trcallsig.getSM() << VINFO_END(log);
    }
}


EvtSIPB* EvtSipStatus::
clone_impl()
{
    return new EvtSipStatus(*this);
}


EvtSipStatus::TagType EvtSipStatus::
getTag() const throw()
{
    // (depends from SIP parser)
    return SIP_STATUS;
}


void EvtSipStatus::
updateID()
{
    // Ignore
}


#ifndef DISABLE_WRITER

ostream& EvtSipStatus::
writeTo(ostream& out) const
{
    out << "EvtSipStatus";
    EvtSIPTrCallSigMessage::writeTo(out);
    return out;
}

#endif
