/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const caConfigure_cxx_Version = "$Id: caConfigure.cxx,v 1.34 2002/12/10 00:09:32 bko Exp $";

#include <stdio.h>
#include "caConfigure.hxx"
#include "global.h"
#include "endpoint.hxx"
#include "potStateEngine.hxx"
#include "callProcessing.hxx"
#include "callAgent.hxx"
#include "cpLog.h"
#include "parse3tuple.h"
#include "cpUtil.hxx"
#include "CommandLine.hxx"
#include "ProvisionInterface.hxx"
#include "VMarshalServerData.hxx"
#include "CAServerContainer.hxx"
#include "SystemInfo.hxx"

///
const char* RgwEndpointConfigTagString [] =
    {
        "directoryNumber",
        "inputQueue",
        "unknown"
    };

/// Gets Endpoint from the List
Endpoint*
getEndpoint (char* epId)
{
    Endpoint* ep = NULL;

    EndpointMap::iterator p = epMap.find (epId);
    if (p != epMap.end())
    {
        ep = p->second;
    }
    return ep;
}    // getEndpoint


///Adds a new Endpoint to the list
Endpoint*
newEndpoint (char* epId)
{
    Endpoint* ep = NULL;

    if (ca.endpointCount < ENDPOINT_MAX)
    {
        ep = &eps[ca.endpointCount];
        if ( ep == 0 )
        {
            assert( ep != 0 );
            return ( ep );

        }
        ep->id = ca.endpointCount++;
        ep->epId = epId;
        epMap.insert (pair < string, Endpoint* > (string(epId), ep));
    }
    return ep;
}    // newEndpoint


///Initializes Endpoint data Structure
void
epInit (Endpoint* ep)
{
    if ( ep == 0 )
    {
        assert( ep != 0 );
        return ;
    }
    ep->gw = &gws[ca.gatewayCount];
    ep->iq = gws[ca.gatewayCount].iq;

    ep->call = NULL;
    ep->heldCall = NULL;
    ep->requestIdentifier[0] = 0;

    ep->sipMsg = 0;
    ep->sipMsgHeld = 0;
    ep->sipMsgXfer = 0;
    ep->sipMsgBuf = 0;
    ep->callType = BasicCall;

    ep->sipMsgRouteList = new vector < SipRoute* > ;
    if ( ep->sipMsgRouteList == 0 )
    {
        cpLog( LOG_ERR, "Out of memory !!");
        assert( ep->sipMsgRouteList != 0 );
        return ;
    }
    ep->sipMsgHeldRouteList = new vector < SipRoute* > ;
    if ( ep->sipMsgHeldRouteList == 0 )
    {
        cpLog( LOG_ERR, "Out of memory !!");
        assert( ep->sipMsgHeldRouteList != 0 );
        return ;
    }

    ep->registered = false;

}    // epInit


///
void
rgwEpInit (Endpoint* ep)
{
    epInit (ep);
    ep->stateEngine = &potStateEngine;
    ep->state = CallStateInit;
    ep->dn[0] = 0;
    ep->cfnaDn[0] = 0;
    ep->huntDn[0] = 0;
    ep->echoCancellation = true;
}    // rgwEpInit


///Inserts a DN 
void
addDn (char* data, Endpoint* ep)
{
    if ( ep == 0 )
    {
        cpLog( LOG_DEBUG, "The endpoint pointer is NULL !!");
        return ;
    }
    map < string, Endpoint* > ::iterator p = dnMap.find(data);
    if (p == dnMap.end())
    {
       dnMap.insert(pair < string, Endpoint* > (data, ep));
    }
    else
    {
        cpLog (LOG_WARNING, "Multiple appearance is not supported yet");
        cpLog (LOG_WARNING, "Endpoint %s DN %s", ep->epId.logData(), data);
    }
}    // addDn


///
void
parseRgwEndpointConfig (char* tag, char* type, char* value)
{
    char data [TUPLE3_MAX_LENGTH];
    char epId [EP_ID_MAX_LENGTH];

    sscanf (value, "%s %s", epId, data);

    Endpoint* ep = getEndpoint (epId);
    if (ep == NULL)
    {
        ep = newEndpoint (epId);
        if (ep != NULL)
        {
            rgwEpInit (ep);
        }
        else
        {
            cpLog (LOG_ERR, "Cannot create endpoint: %s", epId);
            return ;
        }
    }

    switch (string2i (tag, RgwEndpointConfigTagString, RgwEndpointConfigTagMax))
    {
        case RgwEndpointDnTag:
        {
            strcpy (ep->dn, data);
            // tbd: Use address translation table instead
            addDn (data, ep);
        }
        break;
        case RgwEndpointInputQueueTag:
        {
            int i;
            sscanf (data, "%d", &i);
            if (i < ca.threadCount)
            {
                ep->iq = &iqs [i];     // input queue
            }

        }
        break;
        default:
        {
            cpLog (LOG_ERR, "Unknown RgwEndpointConfigTag: %s", tag);
        }
    }    // switch
}    // parseRgwEndpointConfig



///
void
rgwEndpointConfigure (char* configFile)
{

    cpLog (LOG_INFO, "Residential endpoint config file: %s", configFile);
    if (parse3tuple (configFile, parseRgwEndpointConfig) == false)
    {
        cpLog (LOG_ERR, "parse3tuple() Failed for residential endpoint config file");
    }
}    // rgwEndpointConfigure


///Types of Gateways
const char* GatewayTypeString [] =
    {
        "Residential",
        "unknown"
    };


///
const char* RgwConfigTagString [] =
    {
        "domainName",
        "gatewayName",
        "remotePort",
        "proxyAddress",
        "proxyPort",
        "inputQueue",
        "agentPort",
        "endpoint",
        "unknown"
    };


///
void
parseRgwConfig (char* tag, char* type, char* value)
{
    char data [TUPLE3_MAX_LENGTH];
    int gwIdx = ca.gatewayCount;

    switch (string2i (tag, RgwConfigTagString, RgwConfigTagMax))
    {
        case RgwDomainNameTag:
        {
            sscanf (value, "%s", data);
            gws[gwIdx].domainName = data;
            gwMap.insert (pair < string, Gateway* > (gws[gwIdx].domainName,
                          &gws[gwIdx]));
        }
        break;
        case RgwGatewayNameTag:
        {
            sscanf (value, "%s", data);
            gws[gwIdx].gatewayName = data;
        }
        break;
        case RgwRemotePortTag:
        {
            sscanf (value, "%d", &(gws[gwIdx].remotePort));
        }
        break;
        case RgwProxyAddressTag:
        {
            sscanf (value, "%s", data);
            gws[gwIdx].address = data;
        }
        break;
        case RgwProxyPortTag:
        {
            sscanf (value, "%d", &(gws[gwIdx].port));
        }
        break;
        case RgwInputQueueTag:
        {
            int i;
            sscanf (value, "%d", &i);
            if (i < ca.threadCount)
            {
                gws[gwIdx].iq = &iqs [i];
                gws[gwIdx].cpThread = i;
                gws[gwIdx].edPool = &eds [i];     // call processing event data pool
            }

        }
        break;
        case RgwAgentPort:
        {
            sscanf (value, "%d", &gws[gwIdx].agentPort);
        }
        break;
        case RgwEndpointTag:
        {
            sscanf (value, "%s", data);
            rgwEndpointConfigure (data);
        }
        break;
        default:
        {
            cpLog (LOG_ERR, "Unknown RgwConfigTag = %s", tag);
        }
    }    // switch
}    // parseRgwConfig



///
void
rgwGwInit (Gateway* gw, char* configFile)
{
    if ( gw == 0 )
    {
        cpLog( LOG_DEBUG, "The gw pointer is NULL !!");
        assert( gw != 0 );
        return ;
    }
    gw->configFile = configFile;
    gw->type = Residential;
    gw->iq = &iqs [0];
    gw->agentPort = ca.localPort[0];
    gw->cpThread = 0;
    gw->edPool = &eds [0];
}    // rgwGwInit


///
void
showGateway (Gateway* gw)
{
    if ( gw == 0 )
    {
        cpLog( LOG_DEBUG, "The gw pointer is NULL !!");
        assert( gw != 0 );
        return ;
    }
    cpLog (LOG_INFO, "%s gateway at %s:%d",
           GatewayTypeString [gw->type],
           gw->address.c_str(),
           gw->port);
    cpLog (LOG_INFO, "Config file: %s",
           gw->configFile.c_str());
    cpLog (LOG_INFO, "NotifyEntity port: %d",
           gw->agentPort);
    cpLog (LOG_INFO, "Call Processing thread: %d",
           gw->cpThread);

}    // showGateway


///
void
rgwConfigure (char* configFile)
{
    int i = ca.gatewayCount;

    if (i < GATEWAY_MAX)
    {
        rgwGwInit (&gws[i], configFile);

        if (parse3tuple (configFile, parseRgwConfig) == false)
        {
            cpLog (LOG_ERR, "parse3tuple() Failed for residential gateway config file");
        }
        showGateway (&gws[i]);
        ca.gatewayCount++;
    }
}    // rgwConfigure


///
const char* CaConfigTagString [] =
    {
        "localPort",
        "maxNumberOfEndpoints",
        "maxNumberOfCalls",
        "numberOfThreads",
        "residentialGateway",
        "sipMyHost",
        "sipMyPort",
        "sipProxyHost",
        "sipProxyPort",
        "otherCaHost",
        "otherCaPort",
        "provisioningHost1",
        "provisioningPort1",
        "provisioningHost2",
        "provisioningPort2",
        "startupDelay",
        "earlyRtp",
        "registerRetrans",
        "dialPlan",
        "conferenceServer",
        "unknown"
    };


///
void
parseCaConfig (char* tag, char* type, char* value)
{
    char data [TUPLE3_MAX_LENGTH];

    switch (string2i (tag, CaConfigTagString, CaConfigTagMax))
    {
        case CaLocalPortTag:
        {
            int i;
            sscanf (value, "%d", &i);
            if ((agentPortIndex (i)) == 0 && (i != ca.localPort[0]))
            {
                if (ca.localPortCount < RECEIVE_PORT_MAX)
                {
                    ca.localPort [ca.localPortCount++] = i;
                }
                else
                {
                    cpLog (LOG_ERR, "Too many localPort entries (%d max)",
                           RECEIVE_PORT_MAX);
                }
            }
        }
        break;
        case CaEndpointsMaxTag:
        {
            sscanf (value, "%d", &ca.endpointsMax);
        }
        break;
        case CaCallsMaxTag:
        {
            sscanf (value, "%d", &ca.callsMax);
            //tbd:            allocateCallData ();
        }

        break;
        case CaThreadsTag:
        {
            sscanf (value, "%d", &ca.threadCount);
        }
        break;
        case CaRgwFileTag:
        {
            sscanf (value, "%s", data);
            rgwConfigure (data);
        }
        break;
        case CaSipProxyHostTag:
        {
            sscanf (value, "%s", data);
            ca.sipProxyHost = new char [strlen (data) + 1];
            strcpy (ca.sipProxyHost, data);
        }
        break;
        case CaSipProxyPortTag:
        {
            int i;
            sscanf (value, "%d", &i);
            ca.sipProxyPort = i;
        }
        break;
        case CaOtherCaHostTag:
        {
            sscanf (value, "%s", data);
            ca.otherCaHost = new char [strlen (data) + 1];
            strcpy (ca.otherCaHost, data);
        }
        break;
        case CaOtherCaPortTag:
        {
            int i;
            sscanf (value, "%d", &i);
            ca.otherCaPort = i;
        }
        break;
        case CaStartupDelayTag:
        {
            int i;
            sscanf (value, "%d", &i);
            ca.startupDelay = i;
        }
        break;
        case CaEarlyRtpTag:
        {
            sscanf (value, "%s", data);
            ca.earlyRtp = new char [strlen (data) + 1];
            strcpy (ca.earlyRtp, data);
        }
        break;
        case CaRegisterRetransTag:
        {
            int i;
            sscanf (value, "%d", &i);
            ca.registerRetrans = i;
        }
        break;
        case CaDialPlanTag:
        {
            sscanf (value, "%s", data);
            ca.dialPlan = new char [strlen (data) + 1];
            strcpy (ca.dialPlan, data);
        }
        break;
        case CaConferenceServerTag:
        {
            sscanf (value, "%s", data);
            ca.conferenceServer = new char [strlen (data) + 1];
            strcpy (ca.conferenceServer, data);
        }
        break;
        default:
        {
            cpLog (LOG_ERR, "Unknown CaConfigTag = %s", tag);
        }
    }    // switch
}    // parseCaConfig



///
void
caInit (void)
{
    //    ca.localPort[0] = 2427;
    ca.localPort[0] = 2428;
    ca.localPortCount = 1;
    ca.threadCount = 1;
    ca.gatewayCount = 0;
    ca.endpointCount = 0;
}    // caInit


///
void
showAgent (void)
{
    cpLog (LOG_INFO, "Number of Message Receiving Threads = %d (%d max)",
           ca.localPortCount,
           RECEIVE_PORT_MAX);
    cpLog (LOG_INFO, "Default Call Agent Port: %d", ca.localPort[0]);

    for (int i = 1; i < ca.localPortCount; i++)
    {
        cpLog (LOG_INFO, "Additional Call Agent Port: %d", ca.localPort[i]);
    }
    cpLog (LOG_INFO, "Number of Call Processing Threads = %d (%d max)",
           ca.threadCount,
           CP_THREAD_MAX);
    cpLog (LOG_INFO, "Number of Gateways = %d (%d max)",
           ca.gatewayCount,
           GATEWAY_MAX);
    cpLog (LOG_INFO, "Number of Endpoints = %d (%d max)",
           ca.endpointCount,
           ENDPOINT_MAX);
}    // showAgent


///
void
getCaHost(int port)
{
    Data myHost = theSystem.gethostAddress();

    ca.sipMyHost = new char[myHost.length() + 1];
    memset(ca.sipMyHost, myHost.length() + 1, '\0');
    LocalScopeAllocator lo;
    strcpy(ca.sipMyHost, myHost.getData(lo));
    ca.sipMyPort = port;

    // redundant callAgent is read from ca.cfg

}


///Gets the SIP marshal from the Provisioning
void
getMarshalHost()
{
    // since loopback testing could be used, don't read marshal address from pserver
    if (ca.sipProxyHost != 0) return ;

    list < string > temp;
    try
    {
        ProvisionInterface& pif = ProvisionInterface::instance();
        pif.getMarshalServers(temp);

        for (list < string > ::iterator i = temp.begin(); i != temp.end(); i++)
        {
            VMarshalServerData msData;
            string msName = *i;
            pif.getMarshalData(msName.c_str(), msData);

            // it connects to user agent marshal only
            if (strcasecmp(msData.getType().c_str(), "useragent") == 0)
            {
                string ip, port;
                string::size_type colonPos = msName.find(":");
                if (colonPos != string::npos)
                {
                    ip = msName.substr(0, colonPos);
                    port = msName.substr(colonPos + 1);
                }
                ca.sipProxyHost = new char[ip.length() + 1];
                memset(ca.sipProxyHost, ip.length() + 1, '\0');
                strcpy(ca.sipProxyHost, ip.c_str());
                ca.sipProxyPort = atoi(port.c_str());

                Data msId(msName);
                CAServerContainer::instance().addServer(msId, SERVER_MS);
            }
        }
    }
    catch (VException& e)
    {
        cpLog (LOG_ERR, "Failed to connect to pServer = %s:%d, backup = %s:%d. STOP!!!",
               ca.provisioningHost1, ca.provisioningPort1,
               ca.provisioningHost2, ca.provisioningPort2);
        assert (0);
    }

}

///
void
getServers(const int argc, const char* argv[])
{
    CommandLine::instance(argc, argv);

    Data pServer;
    Data pServerPort;
    Data backupPServer;
    Data backupPServerPort;

    {
	Data psHostPort = CommandLine::instance()->getString( "PSERVER" );
	char matched;
	pServer = psHostPort.matchChar(":", &matched);
	pServerPort = "6005";
	if(matched == ':') 
	{ 
	    pServerPort = psHostPort;
	}
	else 
	{
	    // no match, so the remainder must be the psHost
	    pServer = psHostPort;
	}
    }

    {
	Data psHostPort = CommandLine::instance()->getString( "PSERVERBACKUP" );
	char matched;
	backupPServer = psHostPort.matchChar(":", &matched);
	backupPServerPort = "6005";
	if(matched == ':') 
	{ 
	    backupPServerPort = psHostPort;
	}
	else 
	{
	    // no match, so the remainder must be the psHost
	    backupPServer = psHostPort;
	}
    }

    Data sipPort = Data(CommandLine::instance()->getInt( "SIP_PORT" ));

    // set debug level
    Data debugLevel = CommandLine::instance()->getString("LOGLEVEL");
    if (debugLevel == "LOG_DEBUG")
        cpLogSetPriority (LOG_DEBUG);
    else if (debugLevel == "LOG_DEBUG_STACK")
        cpLogSetPriority (LOG_DEBUG_STACK);

    // provision host 1
    ca.provisioningHost1 = new char[pServer.length() + 1];
    memset(ca.provisioningHost1, pServer.length() + 1, '\0');
    LocalScopeAllocator lo;
    strcpy(ca.provisioningHost1, pServer.getData(lo));
    ca.provisioningPort1 = pServerPort.convertInt();

    // provision host 2
    ca.provisioningHost2 = new char[backupPServer.length() + 1];
    memset(ca.provisioningHost2, backupPServer.length() + 1, '\0');
    strcpy(ca.provisioningHost2, backupPServer.getData(lo));
    ca.provisioningPort2 = backupPServerPort.convertInt();

    // initial connection to provision host
    try
    {
        ProvisionInterface::instance(ca.provisioningHost1, ca.provisioningPort1,
                                     ca.provisioningHost2, ca.provisioningPort2);
    }
    catch (VException& e)
    {
        cpLog (LOG_ERR, "Failed to connect to pServer = %s:%d, backup = %s:%d. STOP!!!",
               ca.provisioningHost1, ca.provisioningPort1,
               ca.provisioningHost2, ca.provisioningPort2);
        assert (0);
    }

    // get the callAgent address and port
    getCaHost(sipPort.convertInt());

    // get marshal address
    getMarshalHost();

}

///
void
caConfigure (const int argc, const char* argv [])
{
    
    caInit ();

    string configFile = "ca.cfg";     // default configuration file will probably
    // be ../../config/ca/ca.cfg
    cpLog (LOG_INFO, "Call Agent configuration file: %s", configFile.c_str() );

    if (parse3tuple (configFile.c_str(), parseCaConfig) == false)
    {
        cpLog (LOG_ERR, "parse3tuple() Failed for root config file");
    }

    getServers(argc, argv);

    showAgent ();
}    // caConfigure
